#! /usr/bin/env python
# ChangeNIFLighting.py

# Script to change the NIF named material parameters so that FF1 NIFs
# don't appear as shadowed in FF2.  Can be run from game console.

import struct
import string
import sys
import os
import os.path
import re

# Change the NIFs in for all the character.nif files in the#
# subdirectories of the given path.  For example
# 
# ChangeAllNIFLightings('c:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\Custom\\Art\\library\\characters')
# 
# The numbered parameters are optional.  I took these default values
# from Renegade's posts in
# http://www.freedomreborn.net/phpbb/viewtopic.php?t=31173

def ChangeAllNIFLightings(path='',paramlist=[(1,1.0),(2,1.0),(3,1.0),(10,0.235),(11,0.235),(12,0.235)]):
   for NIF in FindCharacterNIFs(path):
       ChangeLighting(NIF,paramlist=paramlist)

def ChangeAllNIFLightingAndConvert(path='',paramlist=[(1,1.0),(2,1.0),(3,1.0),(10,0.235),(11,0.235),(12,0.235)]):
   SetConverterPath()
   for NIF in FindCharacterNIFs(path):
       ChangeLighting(NIF,paramlist=paramlist)
       KF = NIF[:-len('.nif')]+'.kf'
       execstr = '"%s" "%s" "%s" -a "%s" "%s"' % (nifconvertpath,NIF,NIF,KF,KF)
       status = os.system(execstr)
       

# The fname argument is the complete pathname of the NIF file.  For
# example
# 'c:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\Custom\\Art\\library\\characters\\vxmale_muscled\\character.nif'
# The numbered parameters are optional.  I took these default values
# from Renegade's posts in
# http://www.freedomreborn.net/phpbb/viewtopic.php?t=31173
def ChangeLighting(fname,paramlist=[(1,1.0),(2,1.0),(3,1.0),(10,0.235),(11,0.235),(12,0.235)]):
   f=open(fname,'rb')
   dat=f.read()
   f.close()

   print 'ChangeLighting: working on %s' % fname

   # convert parameter list to a dictionary
   params = {}
   for p in paramlist:
       params[p[0]] = p[1]
   # Now parse the parameter list and gen f/r regexp
   fstr = '(NiMaterialProperty.{4,10}?)(\\w\\w+.*?)(\xff{8}.{2})'
   sstr = '\\1\\2\\3'
   unchanged = 3
   for i in range(1,15):
       if i in params.keys():  # param was passed
           fstr = fstr + '.{4}'
           sstr = sstr + struct.pack('f',params[i])
       else:                   # keep existing value
           unchanged = unchanged + 1
           fstr = fstr + '(.{4})'
           sstr = sstr + '\\%d' % unchanged
   fstr = re.compile(fstr,re.DOTALL)
   dnew, count = re.subn(fstr,sstr,dat)

   if (dat == dnew) :
       print ' ... no change in NIF'
   else:
       print ' ... changed %d nodes' % (count,)
       try:
           i = 1
           backupname = '%s.bak' % (fname,)
           while os.path.isfile(backupname):
               i = i + 1   # first numbered .bak will be '.bak2'
               backupname = '%s.bak%d' % (fname,i)
           os.rename(fname,backupname)
       except OSError:
           print 'ChangeLighting: Unable to create backup of file <%s>, possible Read-Only problem.  Quitting' % (fname,)
           raise
       except:
           print 'ChangeLighting: Unable to create backup of file <%s>.  Quitting' % (fname,)
           raise
       
       f=open(fname,'wb')
       f.write('%s' % dnew)
       f.close()

def ChangeLightingAndConvert(fname,paramlist=[(1,1.0),(2,1.0),(3,1.0),(10,0.235),(11,0.235),(12,0.235)]):
   SetConverterPath()
   ChangeLighting(fname,paramlist=paramlist)
   KF = fname[:-len('.nif')]+'.kf'
   execstr = '"%s" "%s" "%s" -a "%s" "%s"' % (nifconvertpath,fname,fname,KF,KF)
   status = os.system(execstr)
   print '<%s> returned %s' % (execstr,str(status))
       
# For every character.nif file that also has a character.nif.bak file,
# delete the character.nif files and rename the character.nif.bak file
# to character.nif.  The effectively undoes the changes by the other
# functions above.
def RevertToBAKs(path=''):
   BAKs = FindCharacterBAKs(path)
   for b in BAKs:
       if os.path.isfile(b):
           corrupted = b[:-len('.bak')]
           if os.path.isfile(corrupted):
               os.remove(corrupted)
           os.rename(b,corrupted)

# Given a path name, return a list of full pathnames of all
# character.nif files in the subdirectories of that path.  For example
# FindCharacterNIFs('c:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich\\Custom\\Art\\library\\characters')

def FindCharacterNIFs(dirpath):
   if dirpath != '':
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif'),fulldirpaths)
           return filter(os.path.isfile,fullNIFpaths)
       else:
           return []
   else:   # no path given, try to guess
       # first from FF folder
       dirpath = os.path.join(GetFFPath(),'Custom\\Art\\library\\characters')
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif'),fulldirpaths)
           NIFs = filter(os.path.isfile,fullNIFpaths)
       # now add from Win temp dir
       dirpath = os.path.join(GetWinTempPath(),'Custom\\Art\\library\\characters')
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif'),fulldirpaths)
           NIFs.extend(filter(os.path.isfile,fullNIFpaths))
       return NIFs

def FindCharacterBAKs(dirpath):
   if dirpath != '':
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif.bak'),fulldirpaths)
           return filter(os.path.isfile,fullNIFpaths)
       else:
           return []
   else:   # no path given, try to guess
       # first from FF folder
       dirpath = os.path.join(GetFFPath(),'Custom\\Art\\library\\characters')
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif.bak'),fulldirpaths)
           NIFs = filter(os.path.isfile,fullNIFpaths)
       # now add from Win temp dir
       dirpath = os.path.join(GetWinTempPath(),'Custom\\Art\\library\\characters')
       dirpath = os.path.normpath(dirpath)
       if os.path.isdir(dirpath):
           dirpaths = os.listdir(dirpath)
           fulldirpaths = map(lambda p, dp=dirpath: os.path.join(dp,p),dirpaths)
           fulldirpaths = filter(os.path.isdir,fulldirpaths)
           fullNIFpaths = map(lambda dp: os.path.join(dp,'character.nif.bak'),fulldirpaths)
           NIFs.extend(filter(os.path.isfile,fullNIFpaths))
       return NIFs

# Set the path where resides the nifconvert program
nifconvertpath = ''
def SetConverterPath(trypath=''):
   global nifconvertpath
   if os.path.isfile(trypath): # a valid path was passed
       nifconvertpath = trypath
   if not os.path.isfile(nifconvertpath): # it's not already a valid path
       ffpath = GetFFPath()
       if os.path.isfile(os.path.join(ffpath,'nifconvert.exe')):
           nifconvertpath = os.path.join(ffpath,'nifconvert.exe')
       elif os.path.isfile(os.path.join(ffpath, 'Custom\\Art\\library\\characters', 'nifconvert.exe')):
           nifconvertpath = os.path.isfile(os.path.join(ffpath, 'Custom\\Art\\library\\characters', 'nifconvert.exe'))
       elif os.path.isfile(os.path.join(GetWinTempPath(), 'Custom\\Art\\library\\characters', 'nifconvert.exe')):
           nifconvertpath = os.path.isfile(os.path.join(GetWinTempPath(), 'Custom\\Art\\library\\characters', 'nifconvert.exe'))
       else:   # see if there's one laying around in a NIF folder
           NIFs = FindCharacterNIFs('')
           NIFs.append('')
           for NIF in NIFs:
               testpath = os.path.join(os.path.dirname(NIF),'nifconvert.exe')
               if os.path.isfile(testpath):
                   break
           nifconvertpath = testpath
   if not os.path.isfile(nifconvertpath): # it's still not a valid path
       print 'SetConverterPath: No valid path set for nifconvert.exe.'

# From datfiles.py

# Determine if the code is being run from within the game
def inGame():
   return 'ff' in sys.modules.keys()

# Determine which version of FF is running.  Returns default if not run from within the
# game.  Otherise returns 1 for Freedom Force and 2 for Freedom Force vs. The Third Reich.
def ffver(default = 0):
   if sys.argv[0] == 'FreedomForce':
       return 1
   elif sys.argv[0] == 'Freedom Force vs the 3rd Reich':
       return 2
   else:
       print 'ffver: unrecognized game version, returning %s' % str(default)
       return default

# Returns FF directory when run from within the game. If run from outside the game,
# gameVersion only works for FF2 and it will grab the FF directory
# from the Windows registry.
def GetFFPath():
   if ( inGame() ):  # from within game
       gameVersion = ffver()
       if gameVersion >  0:
           return os.path.dirname(sys.executable)
   else:
       import _winreg
       return _winreg.QueryValueEx(_winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                                   'SOFTWARE\\Irrational Games\\FFVTTR\\'),
                                   'InstallDir')[0]
   return ''

# Find the path where temporary files are saved.  This path is
# returned lower-case.  This is NOT TESTED for the first game!
# Intended for use with the second game where this stuff is stored in
# places like 'C:\Documents and Settings\vivalalibertad\Local
# Settings\Application Data\Irrational Games\Freedom Force vs the 3rd
# Reich\' *** Need to see if this works on pre-Win2000 machines.
# If AssumeVersion is not 0, we assume that to be the desired version
# and return a path accordingly.  This makes some assumptions about
# the location of the install (for version 2, for version 1 it just
# returns the mod path).
def GetWinTempPath():
   _WinTempPath = ''
   if inGame():
       syspath = map(string.lower,sys.path)
       headstr = string.lower(os.environ['USERPROFILE'])
       tailstring = 'temp'
       for p in syspath:
           if ( p[:len(headstr)]==headstr ) and ( p[-len(tailstring):]==tailstring ):
               _WinTempPath = os.path.normpath(p[:-len(tailstring)])
   else:   # we're runing outside the game environement
       _WinTempPath = os.path.join(os.path.normpath(os.environ['USERPROFILE']),
                                   'Local Settings','Application Data',
                                   'Irrational Games','Freedom Force vs the 3rd Reich')
   return _WinTempPath

# Run from command-line

if __name__ == '__main__':
   import sys

   params = []
   if len(sys.argv)>1:
       path = sys.argv[1]
       if len(sys.argv)>2:
           params = eval(sys.argv[2])
##            print 'params = %s' % params
   else:
       path = ''
##    print 'sys.version=%s  path=<%s> params=%s' % (sys.version,path,params)
##    print FindCharacterNIFs(path)
   if FindCharacterNIFs(path):
       if params:
##            print 'running for all with params = %s' % params
           ChangeAllNIFLightings(path,params)
       else:
##            print 'running for all with default params'
           ChangeAllNIFLightings(path)
   elif os.path.isfile(path):
       if params:
##            print 'running for %s with params = %s' % (path,params)
           ChangeLighting(path,params)
       else:
##            print 'running for %s with default params' % path
           ChangeLighting(path)
   else:
       try:
           ffpath = GetFFPath()
       except:
           ffpath = 'c:\\Program Files\\Irrational Games\\Freedom Force vs The 3rd Reich'
       basename = os.path.basename(sys.argv[0])
       print '%s can be called without any arguments, in which case it tries to find ' % (basename,)
       print 'and convert all character.nif files it can find in the FFvT3R directories.\n'
       print '%s can be called with a path argument.\n' % (basename,)
       print 'The path can be the folder containing your NIF folders.'
       print 'For example:'
       print '    %s "%s"\n'% (basename,os.path.join(ffpath,'Custom\\Art\\library\\characters'))
       print 'The path can also be a particular NIF file.  For example:'
       print '    %s "%s"\n' % \
             (basename,os.path.join(ffpath,'Custom\\Art\\library\\characters\\mymesh\\character.nif'))
       print 'If a second argument is present, it is taken to be a list of pairs of'
       print 'parameter numbers (from 1 to 14) and the values they are to be'
       print 'assigned.  For example:'
       print '    %s "%s" \'[(1,1.0),(2,0.98),(3,0.78),(10,0.235),(11,0.235),(12,0.235)]\'\n' % \
             (basename,os.path.join(ffpath,'Custom\\Art\\library\\characters\\mymesh\\character.nif'))
       print 'is the same as the previous example with parameter 1 set to 1.0, '
       print 'parameter 2 set to 0.98, and so on.'