###################################################################################################
#
# Keep Building Damage plug-in for Freeroam mods
# 
# by m25
#
# Save building damage whenever a player leaves a map
# and then restore that damage whenever the player
# re-enters the map.
#
# Probably more amusing than useful.
#
#
###################################################################################################

FREEROAMKBD_VERSION ={
'majorVersion': 1, # official major version, ex.: 3 for 3.0.4
'minorVersion': 0, # official update version, ex.: 0 for 3.0.4
'patchVersion': 0, # official patch version, ex.: 4 for 3.0.4
'buildVersion': 0, # for each new main branch, new build number; for someone's branch, name of the main build from which this was branched
'buildBranch': "main" # either 'main' (followed by some identification for public releases, such as 'beta 1' or 'final') for merged code or name of the author of the branch, optionally followed by the branch's numbering
}
print "Loading freeroam_keepbuildingdamage.py v.%i.%i.%i build %i; branch = %s" %  (FREEROAMKBD_VERSION['majorVersion'], FREEROAMKBD_VERSION['minorVersion'], FREEROAMKBD_VERSION['patchVersion'], FREEROAMKBD_VERSION['buildVersion'], FREEROAMKBD_VERSION['buildBranch'])


import js

import m25event
import m25gametype
import m25getset
import m25report

m25report.Report_SetChannel('FREEROAM_KBD', 1)


# Register the callbacks.  freeroam.py will trigger these when starting and ending missions
m25event.Event_RegisterEvent('freeroam_initmission', 'FreeroamKBD_StartMission')
m25event.Event_RegisterEvent('freeroam_exitmission', 'FreeroamKBD_EndMission')
m25report.Report('FREEROAM_KBD', 'initialized')

def OnPostInit():
	pass
	
	
# Read in the campaign varaible (from m25getset) and then set the health
# of all of the buildings to their previous values.
# Note that using js.Object_SetAttr(building,'health') doesn't give the desired effect
# (buildings show a damage health bar, but then reset as soon as they are touched)
# so I'm going with the messier Trigger_Damage
def FreeroamKBD_StartMission(event):
	current_map = event.object
	buildings = m25getset.GetMissionVar('freeroamkbd_%s'%(current_map),default=[])
	for b in buildings:
		if js.Object_Exists(b[0]): js.Trigger_Damage(b[0],js.Object_GetAttr(b[0],'health')-b[1])
		m25report.Report('FREEROAM_KBD', 'building %s health set to %i.'%(b,int(js.Object_GetAttr(b[0],'health'))))
	m25report.Report('FREEROAM_KBD', 'building health set.')
	
		
# Find all of the buildings on the map and set their health levels
# in a campaign variable so that they can be restored later.
def FreeroamKBD_EndMission(event):
	current_map = event.object
	buildinglist = []
	buildings = filter(FreeroamKBD_IsBuilding, js.Mission_GetObjects())
	for b in buildings:
		buildinglist.append([b,js.Object_GetAttr(b,'health')])
		m25report.Report('FREEROAM_KBD', 'building %s health %i stored.'%(b,int(js.Object_GetAttr(b,'health'))))
	m25getset.SetMissionVar('freeroamkbd_%s'%(current_map), buildinglist, campaign=1)
	m25report.Report('FREEROAM_KBD', 'building health stored.')
	
def FreeroamKBD_IsBuilding(object):
	return js.Object_GetClass(object) & js.OC_BUILDING
