#####################################################################
#
# Simple Choice plug-in for EZ Script
# 
# by m25
#
# The player talks to an allied character, and then is presented
# with a list of choices (custom commands on the allied character).
# The player selects a choice, then an appropriate cutscene is
# played and the appropriate end condition is returned.
#
# Based on the chat encounter from m25enc.py
# 
#
#####################################################################

OPENDOOR_VERSION ={
'majorVersion': 1, # official major version, ex.: 3 for 3.0.4
'minorVersion': 0, # official update version, ex.: 0 for 3.0.4
'patchVersion': 0, # official patch version, ex.: 4 for 3.0.4
'buildVersion': 0, # for each new main branch, new build number; for someone's branch, name of the main build from which this was branched
'buildBranch': "main" # either 'main' (followed by some identification for public releases, such as 'beta 1' or 'final') for merged code or name of the author of the branch, optionally followed by the branch's numbering
}
print "Loading     m25enc_OPENDOOR.py v.%i.%i.%i build %i; branch = %s" %  (OPENDOOR_VERSION['majorVersion'], OPENDOOR_VERSION['minorVersion'], OPENDOOR_VERSION['patchVersion'], OPENDOOR_VERSION['buildVersion'], OPENDOOR_VERSION['buildBranch'])

import ff
import js

#import cshelper

import m25enc
import m25report

# Set this to 0 to turn off
# to 1 to report to the console
# to 2 to report to the game screen and the console
m25report.Report_SetChannel('OPENDOOR', 1)

#This is the function that m25story will call
# It creates the encounter dictionary and then calls the setup function
def DoEncOpendoor(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], effect=None, opendoor=None, fog=None):
	m25report.Report('OPENDOOR', 'Initializing on marker %s'%(marker))
	encounter = EncOpenDoor(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives, effect, opendoor, fog)
	return m25enc.SetupEncounter(enc_id, encounter)

# Do the actual creation of the encounter dictionary and return it
# This is just working with data - no encounter code is actually run yet - that happens in m25enc.SetupEncounter
def EncOpenDoor(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[], effect='effect_door_single_opening', opendoor='nazihq_door_open', fog=None):
	#initialize the encounter dictionary
	encounter = {}
	#set up the variables the simple choice encounter uses
	encounter['state'] = [
		['door_effect', effect],
		['door_open', opendoor],
		['door_fog', fog],
		['door_ended', 0],
	]
	#setup the on_examined callback
	encounter['on_examined'] = ['OpenDoor_OpenDoor']
	#inherit the setup from the chat encounter (villains and minions will fight, the player will be able to talk to an ally)
	encounter = m25enc.CombineEncounters(encounter, m25enc.EncFight(enc_id, villains, minions, allies, objects, cutscenes, return_function, marker, objectives))
	#add the open command on the door
	encounter = m25enc.CombineEncounters(encounter, m25enc.EncExamine(objectgroup='object', command='OPEN', cutscene='opendoorcs', allow_enemies=1))
	#override the end function
	encounter['check_end'] = ['OpenDoor_CheckEnd']	
	return encounter	
	
def OpenDoor_OpenDoor(enc_id, object):
	door = m25enc.GetEncVar(enc_id, 'examine_object')
	standin = '%i_standin'%enc_id
	js.Object_SpawnAt('fx_standin',js.Get_ObjectPos(door), standin, js.Object_GetOrientation(door))
	m25enc.SetEncVar(enc_id, 'effect_standin', standin)
	js.Object_Destroy(door)
	effect = m25enc.GetEncVar(enc_id, 'door_effect')
	if not effect: effect, opendoor = OpenDoor_GetDoorDefaults(enc_id)
	if not m25enc.GetEncVar(enc_id,'door_open'): m25enc.SetEncVar(enc_id, 'door_open', opendoor)
	js.Object_PlayEffect(standin, effect, 'OpenDoor_Complete', js.FX_TRACK_OBJECT_ROT, enc_id)

def OpenDoor_Complete(event):
	enc_id = int(event.user)
	standin = m25enc.GetEncVar(enc_id, 'effect_standin')
	door = '%i_newdoor'%enc_id
	opendoor = m25enc.GetEncVar(enc_id,'door_open')
	js.Object_SpawnAt(opendoor,js.Get_ObjectPos(standin), door, js.Object_GetOrientation(standin))
	js.Object_Destroy(standin)
	fog = m25enc.GetEncVar(enc_id, 'door_fog')
	if fog: js.Object_Destroy(fog)
	m25enc.EncCheckEnd(enc_id)
		
# This function is called whenever:
#		a villain, minion or ally is ko'd
#		the player selects the open command on the door
def OpenDoor_CheckEnd(enc_id):
	if m25enc.GetEncVar(enc_id, 'examine_complete', default=0) and m25enc.Enc_IsAllGone(enc_id, ['villain','minion']) and (not m25enc.GetEncVar(enc_id,'door_ended')):
		m25enc.SetEncVar(enc_id, 'door_ended', 1)
		m25enc.Enc_CompleteObjective(enc_id)
		m25enc.PlayEncCutscene(enc_id, 'endcs', return_function='Enc_CleanUp(%i, ["door open", "end"])'%enc_id)

		
def OpenDoor_GetDoorDefaults(enc_id):
	door = m25enc.GetEncVar(enc_id, 'examine_object')
	door_types = {
		'nazihq_wide_door_closed' : ['effect_door_wide_opening', 'nazihq_wide_door_open'],
		'nazihq_door_closed' : ['effect_door_single_opening', 'nazihq_door_open'],
		'nazihq_wide_door_tall_closed' : ['effect_door_wide_tall_opening', 'nazihq_wide_door_tall_open'],
		'nazihq_door_tall_closed' : ['effect_door_tall_opening', 'nazihq_door_tall_open'],
	}
	template = js.Object_GetTemplate(door)
	if template in door_types.keys():
		return door_types[template]
	m25report.Report('ERROR', 'OpenDoor_GetDoorDefaults could not find an entry for %s'%repr(template))
	return ['effect_door_single_opening', 'nazihq_door_open']