#####################################################################
#
# Simple Choice plug-in for EZ Script
# 
# by m25
#
# The player talks to an allied character, and then is presented
# with a list of choices (custom commands on the allied character).
# The player selects a choice, then an appropriate cutscene is
# played and the appropriate end condition is returned.
#
# Based on the chat encounter from m25enc.py
# 
#
#####################################################################

SIMPLECHOICE_VERSION ={
'majorVersion': 1, # official major version, ex.: 3 for 3.0.4
'minorVersion': 0, # official update version, ex.: 0 for 3.0.4
'patchVersion': 0, # official patch version, ex.: 4 for 3.0.4
'buildVersion': 0, # for each new main branch, new build number; for someone's branch, name of the main build from which this was branched
'buildBranch': "main" # either 'main' (followed by some identification for public releases, such as 'beta 1' or 'final') for merged code or name of the author of the branch, optionally followed by the branch's numbering
}
print "Loading     m25enc_simplechoice.py v.%i.%i.%i build %i; branch = %s" %  (SIMPLECHOICE_VERSION['majorVersion'], SIMPLECHOICE_VERSION['minorVersion'], SIMPLECHOICE_VERSION['patchVersion'], SIMPLECHOICE_VERSION['buildVersion'], SIMPLECHOICE_VERSION['buildBranch'])

import ff
import js

import m25enc
import m25event
import m25report

# Set this to 0 to turn off
# to 1 to report to the console
# to 2 to report to the game screen and the console
m25report.Report_SetChannel('SIMPLECHOICE', 1)

#This is the function that m25story will call
# It creates the encounter dictionary and then calls the setup function
def DoEncSimplechoice(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], return_function='', marker='encounter', objectives=[], choice1='Yes', choice2='No', choice3='', choice4='', choice5=''):
	m25report.Report('SIMPLECHOICE', 'Initializing on marker %s'%(marker))
	encounter = EncSimpleChoice(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives, [choice1,choice2,choice3,choice4,choice5])
	return m25enc.SetupEncounter(enc_id, encounter)
	
# Do the actual creation of the encounter dictionary and return it
# This is just working with data - no encounter code is actually run yet - that happens in m25enc.SetupEncounter
def EncSimpleChoice(enc_id, villains=[], minions=[], allies=[], objects=[], cutscenes=[], marker='encounter', return_function='', objectives=[], choices=['Yes','No','','','']):
	#initialize the encounter dictionary
	encounter = {}
	#set up the variables the simple choice encounter uses
	encounter['state'] = [
		['choices', choices],
		['choice_made', None],
		['choices_set', 0]
	]
	#inherit the setup from the chat encounter (villains and minions will fight, the player will be able to talk to an ally)
	encounter = m25enc.CombineEncounters(encounter, m25enc.EncChat(enc_id, villains, minions, allies, objects, cutscenes, marker, return_function, objectives))
	#override the end function (chat sets it the Chat_CheckEnd - we want it to be SimpleChoice_CheckEnd)
	encounter['check_end'] = ['SimpleChoice_CheckEnd']
	return encounter	
	

# This function is called whenever:
#		a villain, minion or ally is ko'd
#		the player selects the talk to command on an ally
#		the player selects one of the choices from the encounter (after having talked to the ally)
def SimpleChoice_CheckEnd(enc_id):
	m25report.Report('SIMPLECHOICE', 'checking end conditions.')
	choices = m25enc.GetEncVar(enc_id, 'choices')
	talked_to = m25enc.GetEncVar(enc_id,'examine_object')	
	if m25enc.GetEncVar(enc_id, 'examine_complete', default=0):
		if m25enc.GetEncVar(enc_id, 'choice_made'):
			#remove choices from target
			for c in choices:
				try:
					js.Mission_RemoveCustomAction('CHOOSE:%s'%(c), '', talked_to)
				except:
					pass
			#complete the objective, if any, and play the cutscene corresponding to the choice made
			m25enc.Enc_CompleteObjective(enc_id)
			choice_code = m25enc.GetEncVar(enc_id, 'choice_made')
			m25report.Report('SIMPLECHOICE', 'Ending on choice %s'%(choice_code))
			m25enc.PlayEncCutscene(enc_id, choice_code+'cs', return_function='Enc_CleanUp(%i,%s)'%(enc_id, repr([choice_code,"end"])))
		else:
			if not m25enc.GetEncVar(enc_id, 'choices_set'):
				#set up the choices
				num = 0
				for c in choices:
					if c != '':
						js.Mission_CustomAction('CHOOSE:%s'%(c), '', talked_to, 'SimpleChoice_Choice%i'%(num), 10, 0)
						m25report.Report('SIMPLECHOICE', 'Setting choice %s on %s'%(c,talked_to))
					num = num + 1
				m25enc.SetEncVar(enc_id, 'choices_set', 1)
	else:
		if m25enc.Enc_IsAllKO(enc_id, 'ally'):
			m25enc.Enc_FailObjective(enc_id)
			#all of the allies were ko'd, so the choice cannot happen!
			m25enc.PlayEncCutscene(enc_id, 'allieslostcs', return_function = 'Enc_CleanUp(%i,%s)'%(enc_id,repr(["allies lost","end"])))


def SimpleChoice_Choice0(target, hero):
	enc_id = m25enc.Enc_GetCharEnc(target)
	m25enc.SetEncVar(enc_id, 'choice_made', m25enc.GetEncVar(enc_id,'choices')[0])
	m25report.Report('SIMPLECHOICE', 'Player has chosen %s on %s in encounter %i'%(m25enc.GetEncVar(enc_id,'choice_made'),target,enc_id))
	m25enc.EncCheckEnd(enc_id)
						
def SimpleChoice_Choice1(target, hero):
	enc_id = m25enc.Enc_GetCharEnc(target)
	m25enc.SetEncVar(enc_id, 'choice_made', m25enc.GetEncVar(enc_id,'choices')[1])
	m25report.Report('SIMPLECHOICE', 'Player has chosen %s on %s in encounter %i'%(m25enc.GetEncVar(enc_id,'choice_made'),target,enc_id))
	m25enc.EncCheckEnd(enc_id)
			
def SimpleChoice_Choice2(target, hero):
	enc_id = m25enc.Enc_GetCharEnc(target)
	m25enc.SetEncVar(enc_id, 'choice_made', m25enc.GetEncVar(enc_id,'choices')[2])
	m25report.Report('SIMPLECHOICE', 'Player has chosen %s on %s in encounter %i'%(m25enc.GetEncVar(enc_id,'choice_made'),target,enc_id))
	m25enc.EncCheckEnd(enc_id)

def SimpleChoice_Choice3(target, hero):
	enc_id = m25enc.Enc_GetCharEnc(target)
	m25enc.SetEncVar(enc_id, 'choice_made', m25enc.GetEncVar(enc_id,'choices')[3])
	m25report.Report('SIMPLECHOICE', 'Player has chosen %s on %s in encounter %i'%(m25enc.GetEncVar(enc_id,'choice_made'),target,enc_id))
	m25enc.EncCheckEnd(enc_id)
						
def SimpleChoice_Choice4(target, hero):
	enc_id = m25enc.Enc_GetCharEnc(target)
	m25enc.SetEncVar(enc_id, 'choice_made', m25enc.GetEncVar(enc_id,'choices')[4])
	m25report.Report('SIMPLECHOICE', 'Player has chosen %s on %s in encounter %i'%(m25enc.GetEncVar(enc_id,'choice_made'),target,enc_id))
	m25enc.EncCheckEnd(enc_id)
