from cshelper import *
from js import *
import event
import freeroam
from ffx import *

######################################################################
#
# freeroam_campaign.py
#  by Dr Mike v1.0 04/07/05
#
#  this is where campaign-specific data lives, describing the layout of your map,
#   cutscenes, stuff like that
#
######################################################################

#first entry is the object template that takes you to this map
#second entry is the name of the map
#third entry is the condition that must be evaluated for the entry to activate
#fourth entry is the colour of the arrow to use

doorways=[
['ffx_cityblock1','cityblock1','1',js.ARROW_YELLOW],
['ffx_cityblock2','cityblock2','1',js.ARROW_YELLOW],
['ffx_cityblock3','cityblock3','1',js.ARROW_YELLOW],
['ffx_cityblock4','cityblock4','1',js.ARROW_YELLOW],
['ffx_cityblock5','cityblock5',"Mission_GetAttr('microwave')",js.ARROW_YELLOW],
['ffx_apartment','apartment','1',js.ARROW_YELLOW],
['ffx_apartment2','apartment2',"Mission_GetAttr('bjfound')",js.ARROW_YELLOW],
['ffx_sewers','sewers',"(Mission_GetAttr('antsfound') and Mission_GetAttr('shadowend')==0) or (Mission_GetAttr('subsfound') and Mission_GetAttr('subsend')==0)",js.ARROW_YELLOW],
['ffx_sewers2','sewers2',"Mission_GetAttr('the_bard') and Mission_GetAttr('shadowend')==0",js.ARROW_RED],
['ffx_tophat','tophat',"Mission_GetAttr('tophatopen') and Mission_GetAttr('pinstripe')==0",js.ARROW_RED],
['ffx_fortress','fortress',"Mission_GetAttr('microwave')",js.ARROW_YELLOW],
['ffx_mechbase','',"Mission_GetAttr('mechbaseopen') and Mission_GetAttr('arc2')<5",js.ARROW_RED],
]

#the names of the campaign variables that are stored in your savegames and transmitted between levels when you
#go between maps etc

campVariables=[
'prestigeTotal', #this one must be in the list - its tracks the total value of prestige stored in game memory
'intro', #mandatory - has the intro c/s played yet?
'intro2', #mandatory - has the second intro text played yet?
'timeofday', #mandatory - used by day and night system
'antspawn',
'antsfound',
'subsfound',
'subsend',
'mechspawn',
'mechcount',
'prevmechcount',
'mrmech',
'mechbaseopen',
'bjfound',
'shadowend',
'raptorspawn',
'raptorcount',
'gangstercount',
'bulletfound',
'tophatopen',
'tophatthug_with_bat',
'tophatthug_with_gun',
'tophatmafioso',
'tophatcop_crooked',
'pinstripe',
'pdcount',
'shadowfound',
'shadowend',
'sewers2darkman_purple',
'sewers2darkman_blue',
'sewers2ant_soldier',
'sewers2ant_worker',
'bulletexplode',
'bulletappear',
'clonecount',
'dejacount',
]

#weapons that can be bought from the arms dealer
campWeapons=[
'ffx_handgun',
'ffx_grenade',
'ffx_zapgun',
'ffx_lure',
'ffx_scanner',
'ffx_jetpack',
'ffx_fawkes',
]

#recruitable heroes who can team up with your custom hero
campHeroes=[
['blackbird','TEAMUP:BLACKBIRD',],
['bullet','TEAMUP:BULLET',],
['microwave','TEAMUP:MICROWAVE',],
['the_bard','TEAMUP:THEBARD',],
['quetzalcoatl','TEAMUP:QUETZLCOATL',],
]

def teamup_bullet(target,char):
    name=getFormName()
    pos=Get_ObjectPos(char)
    spawn(name,'bullet',(pos[0]+200,pos[1]+200,pos[2]))
    RegTimer('bulletrun',5,0,name,char)
    freeroam.FreeRoam_OnTeamUp(char,'bullet')

def bulletrun(event):
    name=event.object
    char=event.string
    Trigger_Power(name,char,'ffx_sprint','')

def teamup_blackbird(target,char):
    b=getNearestBuilding(ff.theHero())
    if b!='':
        spawn(getFormName(),'blackbird',Get_ObjectPos(b))
    freeroam.FreeRoam_OnTeamUp(char,'blackbird')

def teamup_quetzalcoatl(target,char):
    freeroam.FreeRoam_OnTeamUp(char,'quetzalcoatl')
    Mission_SetSunlight(90,90,2,2,(1,1,1))

def teamup_the_bard(target,char):
    b=getNearestBuilding(ff.theHero())
    if b!='':
        spawn(getFormName(),'the_bard',Get_ObjectPos(b))
    freeroam.FreeRoam_OnTeamUp(char,'the_bard')

def teamup_microwave(target,char):
    freeroam.FreeRoam_OnTeamUp(char,'microwave')
    m=getAll('microwave')[0]
    Object_PlayEffect(m,'effect_microwave_teleport_in')

def getNearestBuilding(char):
    minDist=3000*3000
    target=''
    pos=Get_ObjectPos(char)
    for o in Mission_GetObjects():
        if Object_GetClass(o)&OC_BUILDING:
            bp=Get_ObjectPos(o)
            dx=pos[0]-bp[0]
            dy=pos[1]-bp[1]
            d2=dx*dx+dy*dy
            if d2<minDist:
                minDist=d2
                target=o
    return target


#story arcs - used to generate objectives
ARC_CRIME=1
ARC_MECHS=2
ARC_SHADOW=3
ARC_DEJAVU=4
ARC_QUETZ=5

arcs=[
#organised crime arc
(
"keep the streets safe from organised crime",
"confront the crime boss in the top hat nightclub",
"defeat pinstripe and all of his goons",
),
#blackjack and the robots
(
"defeat the mech-men and talk to your mysterious ally",
"keep defeating mechs as you find them, and report back to jack",
"confront mr. mechanical in his lair (talk to jack to get the location)",
"defeat mr. mechanical!",
),
#ants, shadow etc
(
"follow the ants down into the sewers",
"defeat the darkmen who live in the sewers",
"defeat and interrogate the bard",
"follow the bard into shadow's inner sanctum",
"defeat shadow and all her minions",
),
#deja vu and clones etc
(
"defeat and interrogate microwave",
"go to the freedom fortress to investigate",
"defeat and interrogate bullet and find the keys to the freedom fortress",
"enter the freedom fortress",
"defeat deja vu and his minions and destroy the cloning machine",
),
#quetz clone
(
"defeat any raptors or legionnaires you find near the museum",
"defeat the quetzalcoatl clone",
),
]

def updateArc(index,state):
    freeroam.FreeRoam_SetArcState(index,state)

#special overrides for junction behaviour - sometimes a junction doesn't lead to another map
def overrideDoor(target):
    t=Object_GetTemplate(target)
    if t=='ffx_mechbase':
        play(mrmech_cs)
        return 1
    if t=='ffx_fortress' and Mission_GetAttr('bulletexplode')==0:
        forceSpeak('microwave','MISSPCH_1_MW_04')
        return 1
    return 0

#functions that get called when an encounter starts or ends
#put all campaign specific cutscenes here

##############################################################
#
# arc 1 - Ants, shadow, etc
#
##############################################################

#when the first team of ants is destroyed, open up the sewer level
def ant_worker_end(obj,num):
    ant_soldier_end(obj,num)

def ant_soldier_end(obj,num):
    if num==0 and Mission_GetAttr('antsfound')==0:
        play(ant_cs)
        #activate the doorways
        Mission_SetAttr('antsfound',1)
        freeroam.FreeRoam_InitDoorType('ffx_sewers')
        updateArc(ARC_SHADOW,1)

ant_cs=[
(
        "speak(ff.theHero(),'MISSPCH_1_HR_01')",
),
(
        "speak('narrator','MISSPCH_1_AA_07')",
)
]

def darkman_blue_start(obj):
    updateArc(ARC_SHADOW,2)

def darkman_purple_start(obj):
    updateArc(ARC_SHADOW,2)

def darkman_purple_end(obj,num):
    count=Mission_GetAttr('pdcount')+1
    Mission_SetAttr('pdcount',count)
    pdcsplayed=0
    try:
        pdcsplayed=Mission_GetAttr('pdcsplayed')
    except:
        pass
    if num==0 and pdcsplayed==0 and count>5:
       turnTo(ff.theHero(),obj)
       RegTimer('pd2',3,0,obj)
       revive(obj)
       disable(obj)

def pd2(event):
    obj=event.object
    Object_SetAttr(obj,'health',1)
    turnTo(obj,ff.theHero())
    play(pd_cs)

def EnterBard(event):
    hpos=Get_ObjectPos(ff.theHero())
    marker=(hpos[0],hpos[1]+50,hpos[2])
    spawn('the_bard','the_bard',marker)
    FFX_DestroyPortrait('the_bard')
    SetTeam('the_bard',2)
    SetBrawler('the_bard',1)
    play(bard_cs)

#the fifth purple darkman you defeat talks
pd_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look(ff.theHero(),-200,0)",
    "speak(getAll('darkman_purple')[0],'MISSPCH_1_PD_01')",
),
(
    "Camera_LookAtObject(ff.theHero(),-200,0,360,10,CPM_SCROLLTO,0)",
	"speak(getAll('darkman_purple')[0],'MISSPCH_1_PD_02')",
),
(
    "Camera_LookAtObject(ff.theHero(),-200,0,0,10,CPM_SCROLLTO,0)",
	"speak(ff.theHero(),'MISSPCH_1_HR_05')",
),
(
    "for o in getAll('darkman_purple'):Trigger_Damage(o,2)",
    "Mission_SetAttr('pdcsplayed',1)",
    "regTimer('EnterBard',15)",
    "ff.War()",
	"endCS()",
),
]

#played when he gets close to a hero
bard_cs=[
(
    "startCS()",
	"ff.Peace()",
    "Object_SetSecondaryState('the_bard',SCSTATE_INVISIBLE,100,0)",
    "look(ff.theHero(),-200,0)",
	"speak(ff.theHero(),'MISSPCH_1_HR_06')",
),
(
    "Object_SetSecondaryState('the_bard',SCSTATE_INVISIBLE,0,REMOVE_STATE)",
    "Object_PlayEffect('the_bard','effect_ffx_pheremones')",
	"turnTo(ff.theHero(),'the_bard')",
	"turnTo('the_bard',ff.theHero())",
    "Camera_LookAtObject('the_bard',-200,0,0,1,CPM_SCROLLTO,0)",
    "speak('the_bard','MISSPCH_1_TB_01')",
),
(
    "Camera_LookAtObject('the_bard',-200,0,360,10,CPM_SCROLLTO,0)",
    "speak('the_bard','MISSPCH_1_TB_02')",
),
(
    "Camera_LookAtObject('the_bard',-200,0,0,10,CPM_SCROLLTO,0)",
    "speak('the_bard','MISSPCH_1_TB_03')",
),
(
    "regAlmostDead('the_bard','onHeroDown')",
    "ff.War()",
	"endCS()",
    "regTimer('bardhere',1)",
),
]

def bardhere(event):
    updateArc(ARC_SHADOW,3)

#played when he's defeated
the_bard_cs=[
(
    "speak('the_bard','MISSPCH_1_TB_04')",
	"regTimer('opensewers2',5)",
),
]

def opensewers2(event):
    freeroam.FreeRoam_InitDoorType('ffx_sewers2')
    updateArc(ARC_SHADOW,4)

#map specific startups - lighting, music etc.
def init_sewers():
    Mission_SetAttr('customlight',1)
    Mission_SetSunlight(-140,70,0.4,0.4,(0.5,0.8,0.6))
    Mission_SetShadows(-140,70,(0.2,0.2,0.4))
    Mission_SetProjShadowBackground((1,1,1))
    Mission_SetProjShadowColor((.6,.6,.6,1))
    Mission_SetAttr('customMusic',1)
    Sound_PlayMusic('music_11elem1')
    #spawn the bard if he's undefeated, and set up his defeat conditions etc...
    if Mission_GetAttr('arc3')==3:
        marker=freeroam.FreeRoam_GetEncounterMarker()
        spawn('the_bard','the_bard',marker)
        FFX_DestroyPortrait('the_bard')
        SetTeam('the_bard',2)
        SetBrawler('the_bard',1)
        regAlmostDead('the_bard','onHeroDown')

def init_sewers2():
    #set up custom lighting and music for this level
    Mission_SetAttr('customlight',1)
    Mission_SetAttr('customEncounters',1)
    Mission_SetSunlight(-140,70,0.4,0.4,(0.5,0.8,0.6))
    Mission_SetShadows(-140,70,(0.2,0.2,0.4))
    Mission_SetProjShadowBackground((1,1,1))
    Mission_SetProjShadowColor((.6,.6,.6,1))
    Mission_SetAttr('customMusic',1)
    Sound_PlayMusic('music_11elem1')
    #spawn customised encounters
    freeroam.FreeRoam_CustomEncounter('darkman_purple','pd',6)
    freeroam.FreeRoam_CustomEncounter('darkman_blue','bd',7)
    freeroam.FreeRoam_CustomEncounter('ant_soldier','sa',4)
    freeroam.FreeRoam_CustomEncounter('ant_worker','wa',2)
    spawn('shadow','shadow','pos_shadow')
    SetTeam('shadow',2)
    SetBrawler('shadow',1)
    if Mission_GetAttr('shadowfound')==0:
        Mission_SetAttr('shadowfound',1)
        play(shadow_cs)
    else:
        speak('shadow','MISSPCH_1_SH_02')
    updateArc(ARC_SHADOW,5)

def end_sewers2():
    print 'end_top hat:var=%d'%(Mission_GetAttr('shadowend'))
    if Mission_GetAttr('shadowend')==0:
        Mission_SetAttr('shadowend',1)
        print 'big bonus1'
        freeroam.FreeRoam_XPBonus(500)
        freeroam.FreeRoam_PrestigeBonus(2000)
        if len(getAll('the_bard'))==0:
            spawn('the_bard','the_bard','pos_fin_tb')
        play(shadow_end_cs)
        updateArc(ARC_SHADOW,6)

shadow_cs=[
(
    "startCS()",
	"ff.Peace()",
	"look('tosewers2',-200,-135)",
    "speak('shadow','MISSPCH_1_SH_01')",
),
(
    "forceSpeak('the_bard','MISSPCH_1_TB_05')",
),
(
    "ff.War()",
	"endCS()",
),
]

shadow_end_cs=[
(
    "startCS()",
	"ff.Peace()",
	"wait(3)",
),
(
    "teleport(getAll('the_bard')[0],'pos_fin_tb')",
    "revive(getAll('the_bard')[0])",
	"teleport(ff.theHero(),'pos_fin_hero')",
    "revive(ff.theHero())",
	"teleport('shadow','pos_fin_sh')",
	"playTransition()",
	"look('pos_fin_tb',-200,90)",
),
(
	"forceSpeak('the_bard','MISSPCH_1_TB_06')",
),
(
	"speak(ff.theHero(),'MISSPCH_1_HR_07')",
),
(
    "regTimer('shadowexit',3)",
),
]

def shadowexit(event):
    spawn('sewers1','ffx_cityblock2',ff.theHero())
    freeroam.FreeRoam_OnExit('sewers1',ff.theHero())

##############################################################
#
# arc 2 - Mechmen, blackjack
#
##############################################################

#when the first wave of robots appears - spawn an old guy nearby who helps out
def mech_man_flame_start(obj):
    if Mission_GetAttr('bjfound')==0 and Object_Exists('bj')==0:
        pos=Get_ObjectPos(obj)
        spawn('bj','jack',(pos[0]+randint(-50,50),pos[1]+randint(-50,50),pos[2]))
        regDetectHero(obj,'onSeeBJ')
    updateArc(ARC_MECHS,1)

def init_apartment2():
    #make blackjack say stuff when you return to his base
    mc=Mission_GetAttr('mechcount')
    pmc=Mission_GetAttr('prevmechcount')
    if mc>pmc:
        Mission_SetAttr('prevmechcount',mc)
        if pmc==0:
            play(arms0_cs)
        if mc>=3 and pmc<3:
            play(arms1_cs)
        if mc>=6 and pmc<=6:
            play(arms2_cs)
        if mc>=10 and pmc<=10:
            play(arms3_cs)
            Mission_SetAttr('mechbaseopen',1)

arms0_cs=[("speak('arms_dealer','MISSPCH_1_BW_01')",),]
arms1_cs=[("speak('arms_dealer','MISSPCH_1_BW_02')",),]
arms2_cs=[("speak('arms_dealer','MISSPCH_1_BW_03')",),]
arms3_cs=[("speak('arms_dealer','MISSPCH_1_BW_04')",),]

def mech_man_flame_end(obj,num):
    mech_man_end(obj,num)

def mech_man_end(obj,num):
    if num==0 and Mission_GetAttr('bjfound')==0 and Object_Exists('bj'):
        Mission_CustomAction('CUSTOM_TALKTO','','bj','OnTalkToBJ',5,0)
    if num==0:
        count=Mission_GetAttr('mechcount')+1
        Mission_SetAttr('mechcount',count)
# unlock new weapons here for buying at jack's
        if count==3:
            freeroam.FreeRoam_AddWeapon('ffx_lure')
            freeroam.FreeRoam_AddWeapon('ffx_fawkes')
            freeroam.FreeRoam_AddWeapon('ffx_jetpack')
        if count==6:
            freeroam.FreeRoam_AddWeapon('ffx_zapgun')
        if count==10:
           updateArc(ARC_MECHS,3)

def onSeeBJ(event):
    OnUseWatch('','bj')
    speak('bj','MISSPCH_1_BJ_01')
    addMoveGoal('bj',event.object)
    addArrow('abj','bj',delete=1)

def OnTalkToBJ(target,char):
    Mission_RemoveCustomAction('CUSTOM_TALKTO','','bj')
    Mission_SetAttr('bjfound',1)
    freeroam.FreeRoam_InitDoorType('ffx_apartment2')
    play(bj_cs)
    Mission_RemoveArrow('abj')
    #add some weapons
    freeroam.FreeRoam_AddWeapon('ffx_grenade')
    freeroam.FreeRoam_AddWeapon('ffx_handgun')
    freeroam.FreeRoam_AddWeapon('ffx_scanner')
    updateArc(ARC_MECHS,2)

def OnUseWatch(dummy,char):
    Object_PlayEffect(char,'effect_ffx_telepathy')
    for o in Mission_GetDynamicObjects():
        if Object_GetTemplate(o)[:8]=='mech_man':
            Object_SetPrimaryState(o,PCSTATE_BLANK,20,0)
    uses=FFX_ObjectGetAttr(char,'shots0')-1
    FFX_ObjectSetAttr(char,'shots0',uses)
    if uses<=0:
        Mission_RemoveCustomAction('CUSTOM_WATCH',char,char)
    else:
        Mission_StatusText('comm scrambler: %i charges left'%(uses))

def OnGiveWatch(event):
    char=ff.theHero()
    FFX_ObjectSetAttr(char,'shots0',5)
    Mission_CustomAction('CUSTOM_WATCH',char,char,'OnUseWatch',5,0)
    speak('narrator','MISSPCH_1_AA_06')

def initthehero(char):
    if FFX_ObjectGetAttr(char,'shots0')>0:
        Mission_CustomAction('CUSTOM_WATCH',char,char,'OnUseWatch',5,0)

bj_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look('bj',-200,0)",
	"wait(5)",
    "speak('bj','MISSPCH_1_BJ_02')",
	"turnTo('bj',ff.theHero())",
	"turnTo(ff.theHero(),'bj')",
),
(
    "AI_Animate('bj','melee')",
    "Camera_LookAtObject('bj',-200,0,360,20,CPM_SCROLLTO,CA_MOVE)",
),
(
    "AI_Animate(ff.theHero(),'drop_pole')",
    "speak(ff.theHero(),'MISSPCH_1_HR_02')",
),
(
    "Trigger_Move('bj','move',(0,0,0),'',0,1)",
    "speak('bj','MISSPCH_1_BJ_03')",
),
(
    "speak('narrator','MISSPCH_1_AA_05')",
),
(
    "regTimer('OnGiveWatch',2)",
    "ff.War()",
    "endCS()",
),
]

mrmech_cs=[
(
    "startCS()",
    "ff.Peace()",
    "turn(ff.theHero(),'pos_mr_mech')",
    "regTimer('buildingDown',1)",
    "look(ff.theHero(),0,180)",
    "Object_Destroy('mechbase')",
    "speak(ff.theHero(),'MISSPCH_1_HR_09')",
),
(
#once mr mech is spawned, onStepCB is called, causing this step to complete! tricky!
	"spawn('jack','jack','base_entry')",),
(
    "forceSpeak('mr_mechanical','MISSPCH_1_MR_01')",
),
(
    "forceSpeak('mr_mechanical','MISSPCH_1_MR_02')",
),
(
    "speak('jack','MISSPCH_1_BJ_04')",
),
(
    "Object_Destroy('jack')",
    "regTimer('notifymech',1)",
	"endCS()",
	"ff.War()",
),
]

def buildingDown(event):
    #just in case - the building could have been destroyed already!
    spmm=0
    if Object_Exists('mech_building')==0:
        spmm=1
    elif Object_IsAlive('mech_building')==0:
        spmm=1

    if spmm==0:
        Trigger_Damage('mech_building',20)
        regTimer('buildingDown',1)
        print 'damaging building: hp=%d'%(Object_GetAttr('mech_building','health'))
    else:
        spawn('mrmech','mr_mechanical','pos_mr_mech')
        SetTeam('mrmech',2)
        SetBrawler('mrmech')
        if isPlaying():
            disable('mrmech')
        Object_SetAttr('mrmech','noFade',0)
        regDeath('mrmech','OnMrMechDown')
        regTimer('onStepCB',0.1)

def init_cityblock4():
    #ifmr mech has appeared but isn't yet defeated, destroy his building and spawn him again
    if Mission_GetAttr('arc2')==4:
        RegTimer('buildingDown',0.1)

def notifymech(event):
    updateArc(ARC_MECHS,4)

def OnMrMechDown(event):
    updateArc(ARC_MECHS,5)
    freeroam.FreeRoam_XPBonus(300)
    freeroam.FreeRoam_PrestigeBonus(1000)
    forceSpeak('jack','MISSPCH_1_BW_05')


##############################################################
#
# arc 3 - Deja Vu & clones of FF members
#
##############################################################

def duplicate_cop_end(obj,num):
    duplicate_fem_end(obj,num)

def duplicate_male_end(obj,num):
    duplicate_fem_end(obj,num)

def duplicate_fem_end(obj,num):
    if num==0:
        Mission_SetAttr('clonecount',Mission_GetAttr('clonecount')+1)

def dejavu_clone_end(obj,num):
    if num==0:
        Mission_SetAttr('dejacount',Mission_GetAttr('dejacount')+1)
    if Mission_GetAttr('dejacount')==3:
        freeroam.FreeRoam_InitDoorType('fortress')
        updateArc(ARC_DEJAVU,4)
        speak(ff.theHero(),'MISSPCH_1_MW_05')

def microwave_start(obj):
    speak(obj,'MISSPCH_1_MW_01')
    addKillGoal(obj,ff.theHero(),PRI_HI)
    regAlmostDead(obj,'onHeroDown')
    updateArc(ARC_DEJAVU,1)

def onHeroDown(event):
    char=event.object
    print '%s(%s) is down!'%(char,Object_GetTemplate(char))
    Object_SetPrimaryState(char,PCSTATE_STUNNED,100000,PERMANENT_STATE)
    addAttention(char)
    addInterrogation(char,'onInterrogateHero')
    #hack to recalculate his target 'gender' ie EVIL females from now on
    if Object_GetTemplate(char)=='the_bard':
        initpheremones(char,0)
    if Object_GetTemplate(char)=='microwave':
        print 'removing all microwave clones!'
        for o in getAll('microwave'):
            if o!=char:
                 kill(o)
                 print 'killing %s'%(o)

def onInterrogateHero(target,char):
    hero=Object_GetTemplate(target)
    exec('play(%s_cs)'%(hero))
    kill(target)
    revive(target)
    Object_SetAttr(target,'health',Object_GetAttr(target,'maxHealth'))
    if hero!='bullet':
        Mission_SetAttr(hero,1)
        FFX_AddPortrait(target)
        FFX_ObjectSetAttr(target,'encounter',0)
        SetTeam(target,1)
    clearGoals(target)
    removeAttention(target)
    normalize(target)

microwave_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look(ff.theHero(),-200,0)",
    "wait(5)",
    "forceSpeak('microwave','MISSPCH_1_MW_02')",
	"turnTo(getAll('microwave')[0],ff.theHero())",
	"turnTo(ff.theHero(),getAll('microwave')[0])",
),
(
    "Camera_LookAtObject(ff.theHero(),-200,0,360,20,CPM_SCROLLTO,CA_MOVE)",
),
(
    "speak(ff.theHero(),'MISSPCH_1_HR_03')",
),
(
    "forceSpeak('microwave','MISSPCH_1_MW_03')",
	"regTimer('openblock5',1)",
	"endCS()",
	"ff.War()",
),
]

def openblock5(event):
    freeroam.FreeRoam_InitDoorType('ffx_cityblock5')
    updateArc(ARC_DEJAVU,2)

def bullet_start(obj):
    if Mission_GetAttr('bulletappear')==0:
        Mission_SetAttr('bulletappear',1)
        speak(obj,'MISSPCH_1_BL_01')
    if Mission_GetAttr('bulletexplode')==0:
        regAlmostDead(obj,'onHeroDown')
    addKillGoal(obj,ff.theHero(),PRI_HI)
    updateArc(ARC_DEJAVU,3)

def bullet_end(obj):
    if Mission_GetAttr('bulletexplode'):
        Trigger_Explosion(obj,'evil male Annihilate')
        Object_SetAttr(obj,'noFade',0)

bullet_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look(ff.theHero(),-200,0)",
    "forceSpeak('bullet','MISSPCH_1_BL_02')",
	"turnTo(getAll('bullet')[0],ff.theHero())",
	"turnTo(ff.theHero(),getAll('bullet')[0])",
),
(
	"endCS()",
	"ff.War()",
	"regTimer('bullet_explode',5)",
),
]

def bullet_explode(event):
    Mission_SetAttr('bulletexplode',1)
    char=getAll('bullet')[0]
    Trigger_Explosion(char,'evil male Annihilate')
    kill(char)
    Object_SetAttr(char,'noFade',0)

def init_fortress():
    Mission_SetAttr('customlight',1)
    Mission_SetAttr('customEncounters',1)
    Mission_SetAttr('customMusic',1)
    if Mission_GetAttr('bullet')==0:
        updateArc(ARC_DEJAVU,5)
        spawn('dj','dejavu','pos_dejavu')
        SetTeam('dj',2)
        SetBrawler('dj',1)
        Object_SetAttr('clone_machine','maxHealth',80)
        Object_SetAttr('clone_machine','minHealth',0)
        Object_SetAttr('clone_machine','health',80)
        Sound_PlayMusic('music_02cuba1')
        regDeaths('dejavu','onFortressKO')
        play(fortress_cs)

def fortressspawn(event):
    list=('duplicate_fem','duplicate_male','duplicate_cop','dejavu_clone')
    if len(getAllBaddies())<10 and Object_Exists('clone_machine'):
        if Object_IsAlive('clone_machine'):
            index=int(randint(0,360)/100)
            template=list[index]
            name=getDummyName()
            freeroam.FreeRoam_Spawn(name,template,Get_MarkerPos('pos_clone_machine'))
            Object_PlayEffect('clone_machine','effect_deja_clone')
            RegDeath(name,'onFortressKO')
            Trigger_Move(name,ff.theHero())
    RegTimer('fortressspawn',10)

def onFortressKO(event):
    print 'onFortressKO'
    num=len(getAllBaddies())
    if len(getAllBaddies())==0:
        play(fortress_win_cs)
        Mission_SetAttr('bullet',1)
        updateArc(ARC_DEJAVU,6)

fortress_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look('bullet_machine',-200,180)",
	"wait(1)",
),
(
    "AI_Animate('dj','ranged')",
    "speak('dj','MISSPCH_1_DJ_01')",
),
(
    "for o in getAllHeroes():turnTo(o,'_impobj_9')",
    "forceSpeak('bullet','MISSPCH_1_BL_03')",
),
(
    "look(ff.theHero(),-200,145)",
	"wait(2)",
	"regTimer('fortressspawn',5,0)",
),
(
    "ff.War()",
    "endCS()",
),
]

fortress_win_cs=[
(
    "startCS()",
	"wait(3)",
),
(
    "playTransition()",
	"spawn('bullet','bullet','winpos_bullet')",
	"revive(ff.theHero())",
	"teleport(ff.theHero(),'winpos_hero')",
),
(
    "look('winpos_hero',-200,0)",
    "turnTo('bullet',ff.theHero())",
	"turnTo(ff.theHero(),'bullet')",
    "speak('bullet','MISSPCH_1_BL_04')",
	"wait(3)",
),
(
    "Camera_LookAtMarker('winpos_hero',-200,0,180,10,CPM_SCROLLTO,0)",
),
(
    "speak('bullet','MISSPCH_1_BL_05')",
),
(
    "endCS()",
    "look('winpos_hero',-200,180)",
    #line up the special briefing
    "Mission_SetAttr('briefing',1)",
    "UI_Base_Enable(1)",
    "UI_Enable(0)",
	"ff.RecruitHeroes()",
    "spawn('i_am_a_base','ffx_cityblock5','pos_dejavu')",
),
]

briefing1_cs=[
(
    "startCS()",
	"UI_Base_Enable(0)",
	"UI_Enable(0)",
	"look('winpos_hero',-200,180)",
	"speak('bullet','MISSPCH_1_BL_06')",
),
(
    "forceSpeak('liberty_lad','MISSPCH_1_LL_01')",
),
(
    "forceSpeak('mentor','MISSPCH_1_ME_01')",
),
(
    "forceSpeak('man_o_war','MISSPCH_1_MO_01')",
),
(
    "endBriefingCS()",
),
]

##############################################################
#
# arc 4 - raptors & quetz
#
##############################################################

def legionnaire_start(obj):
    raptor_start(obj)

def legionnaire_end(obj,num):
    raptor_end(obj,num)

def centurion_start(obj):
    raptor_start(obj)

def centurion_end(obj,num):
    raptor_end(obj,num)

def raptor_start(obj):
    if Mission_GetAttr('raptorspawn')==0:
        updateArc(ARC_QUETZ,1)

#the 15th raptor/roman spawns quetzalcoatl's clone nearby
def raptor_end(obj,num):
    count=Mission_GetAttr('raptorcount')+1
    Mission_SetAttr('raptorcount',count)
    if count==15:
        spawn('quetzalcoatl','quetzalcoatl',obj)
        FFX_DestroyPortrait('quetzalcoatl')
        SetupCharacter('quetzalcoatl')
        SetTeam('quetzalcoatl',2)
        SetBrawler('quetzalcoatl',1)
        play(quetz_cs)
        updateArc(ARC_QUETZ,2)
        regAlmostDead('quetzalcoatl','onHeroDown')

#called when the clone first appears
quetz_cs=[
(
    "startCS()",
    "ff.Peace()",
    "look('quetzalcoatl',-200,0)",
	"wait(5)",
    "speak('quetzalcoatl','MISSPCH_1_QZ_01')",
),
(
    "Camera_LookAtObject('quetzalcoatl',-200,0,360,20,CPM_SCROLLTO,CA_MOVE)",
),
(
    "ff.War()",
    "endCS()"
),
]

#called when hes interrogated - the real one returns and joins you
quetzalcoatl_cs=[
(
    "startCS()",
	"ff.Peace()",
	"RPG_SetActive(0)",
    "look('quetzalcoatl',-200,0)",
    "ff.spawnAwayFrom('quetz','quetzalcoatl','quetzalcoatl',200)",
    "Mission_SetSunlight(90,90,2,2,(1,1,1))",
	"Trigger_Move('quetz','fly',Get_ObjectPos(ff.theHero()))",
	"Trigger_Explosion('quetzalcoatl','generic electrical explosion')",
	"Object_PlayEffect('quetzalcoatl','effect_ffx_thunder')",
	"turnTo(ff.theHero(),'quetz')",
	"wait(3)",
	"wait(1)",
),
(
    "Camera_LookAtObject(ff.theHero(),-100,0,0,5,CPM_SCROLLTO,0)",
    "kill('quetzalcoatl')",
),
(
    "Object_Destroy('quetzalcoatl')",
    "speak('quetz','MISSPCH_1_QZ_02')",
),
(
    "speak('quetz','MISSPCH_1_QZ_03')",
),
(
    "turnTo('quetz',ff.theHero())",
    "speak(ff.theHero(),'MISSPCH_1_HR_04')",
),
(
    "AI_Animate('quetz','area')",
    "speak('quetz','MISSPCH_1_QZ_04')",
),
(
    "regTimer('quetznormal',1)",
    "RPG_SetActive(1)",
    "ff.War()",
    "endCS()",
),
]

def queztnormal(event):
    updateArc(ARC_QUETZ,3)
    SetTeam('quetz',1)

##############################################################
#
# arc 5 - organised crime!
#
##############################################################

def init_cityblock2():
    updateArc(ARC_CRIME,1)

def mafioso_start(obj):
    #set them up to talk if enough are KO'd and the top hat isn't open yet
    if Mission_GetAttr('gangstercount')>=9 and Mission_GetAttr('tophatopen')==0:
        regAlmostDead(obj,'gangstertalk')

def mafioso_end(obj,num):
    #add to the gangster count
    count=Mission_GetAttr('gangstercount')+1
    Mission_SetAttr('gangstercount',count)

def gangstertalk(event):
    obj=event.object
    if Mission_GetAttr('tophatopen')==0:
        Object_SetPrimaryState(obj,PCSTATE_STUNNED,20,0)
        speakNoCB(obj,'MISSPCH_1_MF_01')
        RegTimer('gangsterDie',10,0,obj)
        freeroam.FreeRoam_InitDoorType('ffx_tophat')
        Mission_SetAttr('tophatopen',1)
        updateArc(ARC_CRIME,2)
    else:
        kill(obj)

def gangsterDie(event):
    kill(event.object)

def init_tophat():
    #set up custom lighting and music for this level
    Mission_SetAttr('customlight',1)
    Mission_SetAttr('customEncounters',1)
    Mission_SetAttr('customMusic',1)
    Sound_PlayMusic('music_02cuba1')
    RegTimer('tophatlights',0.1,0)
    #spawn customised encounters
    freeroam.FreeRoam_CustomEncounter('thug_with_bat','group1marker',6)
    freeroam.FreeRoam_CustomEncounter('mafioso','group2marker',6)
    freeroam.FreeRoam_CustomEncounter('cop_crooked','group3marker',6)
    freeroam.FreeRoam_CustomEncounter('thug_with_gun','group4marker',6)
    spawn('pinstripe','pin_stripe','pos_pinstripe')
    SetTeam('pinstripe',2)
    SetBrawler('pinstripe',1)
    if Mission_GetAttr('blackbird')==0:
        play(blackbird_cs)
        updateArc(ARC_CRIME,3)
    else:
        speak('pinstripe','MISSPCH_1_PS_03')
    Object_SetAttr('dancefloor','minHealth',Object_GetAttr('dancefloor','maxHealth'))

blackbird_cs=[
(
    "startCS()",
	"ff.Peace()",
	"look('tophat',-200,0)",
	"speak('pinstripe','MISSPCH_1_PS_01')",
	"wait(5)",
),
(
    "turnTo(ff.theHero(),'pinstripe')",
    "Camera_LookAtObject('tophat',-200,0,90,10,CPM_SCROLLTO,0)",
),
(
    "Mission_SetAttr('blackbird',1)",
    "spawn('blackbird','blackbird','pos_blackbird')",
    "speak('pinstripe','MISSPCH_1_PS_02')",
),
(
    "Camera_LookAtObject('blackbird',-200,0,90,2,CPM_SCROLLTO,0)",
	"speak('blackbird','MISSPCH_1_BB_01')",
),
(
    "Camera_LookAtObject('tophat',-200,0,90,0,1,0)",
	"wait(1)",
),
(
    "ff.War()",
	"endCS()",
),
]

def end_tophat():
    print 'end_top hat:var=%d'%(Mission_GetAttr('pinstripe'))
    if Mission_GetAttr('pinstripe')==0:
        Mission_SetAttr('pinstripe',1)
        print 'big bonus1'
        freeroam.FreeRoam_XPBonus(300)
        freeroam.FreeRoam_PrestigeBonus(1000)
        for bb in getAll('blackbird'):
            revive('blackbird')
        if len(getAll('blackbird'))==0:
            spawn('blackbird','blackbird','pos_fin_bb')
            Object_SetAttr('blackbird','health',0)
        play(tophat_end_cs)
        updateArc(ARC_CRIME,4)

tophat_end_cs=[
(
    "startCS()",
	"ff.Peace()",
	"wait(3)",
),
(
    "teleport(getAll('blackbird')[0],'pos_fin_bb')",
	"teleport(ff.theHero(),'pos_fin_hero')",
	"teleport('pinstripe','pos_fin_ps')",
	"playTransition()",
	"look('pos_fin_bb',-200,90)",
),
(
	"speak(getAll('blackbird')[0],'MISSPCH_1_BB_02')",
),
(
    "regTimer('tophatexit',3)",
),
]

def tophatexit(event):
    freeroam.FreeRoam_NextLevel('cityblock1')

def tophatlights(event):
    index=event.user
    if index==0:
        FFX_Sun(50,40,0.2,1,(1,0.2,0.2))
    elif index==1:
        FFX_Sun(50,100,0.2,1,(0.2,1,1))
    elif index==2:
        FFX_Sun(50,160,0.2,1,(0.2,1,0.2))
    elif index==3:
        FFX_Sun(50,220,0.2,1,(1,0.2,1))
    elif index==4:
        FFX_Sun(50,280,0.2,1,(0.2,0.2,1))
    elif index==5:
        FFX_Sun(50,340,0.2,1,(1,1,0.2))
    index=index+1
    if index==6:
        index=0
    RegTimer('tophatlights',0.7,index)

##############################################################
#
# arc 6 - subterrestrials
#
##############################################################

def subterrestrial_end(obj,num):
    if num==0 and Mission_GetAttr('subsfound')==0:
        speak(getAllHeroes()[0],'MISSPCH_1_HR_08')
        #activate the doorways
        Mission_SetAttr('subsfound',1)
        freeroam.FreeRoam_InitDoorType('ffx_sewers')