######################################################################
## 
## Built-In Function Wrapper
## By pimthe
## 
######################################################################
print "Starting Built-In Function Wrapper 1.5"

# DESCRIPTION
# The wrappers allow you to add behaviours to existing functions and change variables without having to change other existing scripts
# (the various mods and danger rooms, for example) as you would have were you to use a normal function (i.e., with a unique name)
# as wrapper. This is mostly useful for FFvsTTR's built-in functions, since the source code can't be changed directly.

# Feel free to change or add to the code here. Just make sure you maintain compatibility,
# so external modules which could be using these functions keep working. :)

# IMPORTANT
# - Be careful, since adding a wrapper function will change every single use of the built-in function!
# - Less is more... wrapping lots of Python code around a frequently used built-in function will probably slow down the game a lot!
# - IMPORTANT: If a module importing js is loaded after the mission starts, calls to the wrapped functions done within up to 10 seconds
#   after might use the build-in version instead.

# DEPENDENCIES
# - FFX 3.2+
# - DAT Files

# VERSION HISTORY
# 0.1 2004?:
# - First tests within ffx.py; added a wrapper for Object_CalcPrestige().
# 0.5 2006-06-27:
# - Changed the method; now it also works when using the js. prefix (from anywhere).
# - Removed testBuiltIn() function to wrap the wrapper, as it doesn't work with the new method. Use module prefixing instead.
# 1.0 2006-08-02:
# - Changed the method again; now it also works when using no prefix (from anywhere).
# - now also works with global variables, unless the variable needs to be accessed at startup.
# - To use the default non-wrapped function or variable, prefix its name with "WrapCode_", i.e., WrapCode_Object_CalcPrestige().
# - First release as a module.
# 1.1 2006-08-22:
# - Added Mission_GetObjects().
# 1.2 2006-08-25:
# - Updated Object_CalcPrestige() to FFX3.2-style character attributes.
# 1.3 2006-09-11:
# - Removed Mission_GetObjects(). Please check manually for Object_GetTerrain instead if needed.
# - Added Mission_CustomAction(), Mission_RemoveCustomAction() and Mission_ClearCustomActions() to prevent duplicate entries.
# 1.4 2006-09-18:
# - The module will now attempt more aggressively to replace the .js versions in case other modules import that
#   after functionwrapper is loaded.
# 1.5 2006-09-20:
# - Removed the checking loop altogether and moved the functionwrapper import after the plugin imports in ffx.py.
# - Removed the speechIDs variable manipulation example, as it seemed rather pointless.
# - Improved speed of Mission_CustomAction()-related wrappers.
# - Made Mission_CustomAction() wrapper compatible with the oneShot=1 parameter; this has limitations, however - see function for details.
# 1.6 2006-09-22:
# - Removed Mission_CustomAction() and variants, as the additional processing caused problems with custom commands intensive attributes
#   such as Power Absorber.


######################################################################
## 
## low-level stuff header
## please check next section for the wrappers themselves
## 
######################################################################

import ff
import js
import ffx
import datfiles
import missionobjvar
import sys

WRAP_CODE = []
def wrapCode(name, module='js'):
    global WRAP_CODE
    WRAP_CODE.append((name, module))




######################################################################
## 
## Place your wrappers here
## 
######################################################################

############ Object_CalcPrestige() ############
# A wrapper for the built-in (.js module) function
# TO DO
# - Update Object_CalcPrestige() to take into account FFX 3.2 dynamic attribute disabling/enabling

def Object_CalcPrestige(object):
    """Allows to modify the character prestige value formula; mostly useful for in-game changes.
    
    This could be used, for example, to correctly calculate the value of characters with attributes added 
    only via script (such as Latent Mutant), since those attributes aren't visible to the normal system."""
    
    #print "Object_CalcPrestige wrapper used"
    value = WrapCode_Object_CalcPrestige(object) #*needs* to be written as WrapCode_functionName; not a function, but a variable pointing to the real function - see below
    value = value + ffx.Object_CalcFFXForCharPrestige(object) #duplicate this line and change the function to add others modifiers
    if value < 1: # characters should have a cost
        value = 1
    return value

#add Object_CalcPrestige() to the code wrapper
# (module parameter isn't mentioned since Object_CalcPrestige() is built-in (i.e., uses .js))
wrapCode('Object_CalcPrestige')

def Object_CalcFFXForCharPrestige(char):
    myAttr = ffx.getCharacterSpecificAttributes(char)
    attrCost = 0
    if len(myAttr) > 0:
        if ffx.FFX_ObjectGetAttr(char, 'latentMutantTriggered'): #Activated Mutants' cost shouldn't calculate both this attribute and the new one(s)
            attrCost = -300
        attrDAT = datfiles.Campaign_ReadAttributes()
        for attrName in myAttr:
            if attrDAT.has_key(attrName):
                attrCost = attrCost + attrDAT[attrName]
        #print "    value modified by %i"%attrCost
    return attrCost




######################################################################
## 
## low-level stuff footer
## 
######################################################################

for pair in WRAP_CODE:
    print pair
    name = pair[0]
    module = pair[1]
    if not globals().has_key('WrapCode_%s' % name):
        exec "globals()['WrapCode_%s'] = %s.%s" % (name, module, name)
    exec '%s.%s = %s' % (module, name, name)
    exec 'ff.%s = %s' % (name, name)
    exec "globals()['%s'] = %s" % (name, name)
