#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
######################################################################
## 
## General Utilities
## By pimthe
## 
######################################################################
print "Starting General Utilities 1.0.1"

# DESCRIPTION
# A small collection of general-purpose helper functions which aren't really specific 
# to Freedom Force.

# Feel free to change or add to the code here. Just make sure you maintain compatibility,
# so external modules which could be using these functions keep working. :)

# DEPENDENCIES
# - Imports missionobjvar (included with FFX 3+) and string (included with FFvsTTR) modules
# - This module is used by heightcheck.py, mlogreader.py and firehydrants1_0.py.

# VERSION HISTORY
# 0.1 2006-06-22:
# - Added makeDict(), formerly in heightcheck.py

# 1.0 2006-07-15:
# - Added listAdd(), listRemove() and invertDict(), formerly in mlogreader.py
# - Added isDifferenceAcceptable(), stringToList(), stringToTuple(), sequenceToString(),
#   formerly in ffx.py
# - Added getUniquesElements()
# - First release.

# 1.0.1 2007-10-15:
# - Added file encoding and made missionobjvar conditional for compatibility when used externally with Python 2.5+.
#   Note that this still doesn't prevent you from calling generalutils functions which require missionobjvar.
#   such as listAdd() and listRemove().


try:
    import missionobjvar
except:
    print "generalutils warning: cannot import module 'missionobjvar'"
import string


######################################################################
## 
## Generic manipulation
## 
######################################################################

# Returns a string, list or tuple of elements present only in the first of two strings, lists, tuples or dictionary keys
# Added FFX 3.2. Replaces older inA_notInB()
def getUniquesElements(listA,listB):
    return filter(lambda a: a not in listB, listA)


######################################################################
## 
## Dictionaries manipulations
## 
######################################################################

# Returns a dictionary from two lists
def makeDict(k, v):
    if len(k) != len(v):
        print "makeDict() fatal error:"
        print "    %s"%k
        print "    isn't the same length as"
        print "    %s"%v
        return {}
    d = {}
    for i in k:
        d.update({i: v[k.index(i)]})
    return d

# Returns a copy of a dictionary with keys as values and vice-versa
def invertDict(myDict):
    k = myDict.keys()
    v = myDict.values()
    d = {}
    for i in v:
        d.update({i: k[v.index(i)]})
    return d


######################################################################
## 
## Lists and tuples manipulations
## 
######################################################################
   
# Adds an element to a persistent list using the missionobjvar module.
# Also adds the element to a global variable list for faster access.
def listAdd(listVar, listValue):
    if not missionobjvar.Mission_VarExists(listVar):
        print "listAdd(): Creating %s"%listVar
        missionobjvar.Mission_SetVar(listVar, [listValue])
        globals()["%s"%listVar] = [listValue]
        return 1
    else:
        li = missionobjvar.Mission_GetVar(listVar)
        if li.count(listValue) == 0:
            print "listAdd(): adding %s to %s"%(listValue, listVar)
            li.append(listValue)
            missionobjvar.Mission_SetVar(listVar, li)
            globals()["%s"%listVar] = li #synchronises the global variable list with the persistent variable list
            return 1
        else:
            print "listAdd(): %s is already in %s"%(listValue, listVar)
            return 0

# Removes an element from a persistent list using the missionobjvar module.
# Also removes the element from its global variable list.
def listRemove(listVar, listValue):
    if not missionobjvar.Mission_VarExists(listVar):
        print "listRemove() fatal error: %s not found"%listVar
        return 0
    else:
        li = missionobjvar.Mission_GetVar(listVar)
        if li.count(listValue) == 0:
            print "listRemove(): %s wasn't found in %s"%(listValue, listVar)
            return 0
        else:
            print "listRemove(): %s removed from %s"%(listValue, listVar)
            li.remove(listValue)
            missionobjvar.Mission_SetVar(listVar, li)
            globals()["%s"%listVar] = li #synchronises the global variable list with the persistent variable list
            return 1

# Convert a tuple or a list to a string. Arrays and dictionaries are not supported.
# Added FFX 3.2. Replaces tupleToString().
def sequenceToString(theVar, delimiter = ' '):
    theString = ""
    varType = str(type(theVar))
    if (varType == "<type 'tuple'>") or (varType == "<type 'list'>"):
        for v in theVar:
            theString = theString + str(v) + str(delimiter)
        return theString[0:-1]
    elif str(type(theVar)) == "<type 'int'>": # 1-item tuples are considered integers instead
        return str(theVar)
    else:
        print "sequenceToString() fatal error:"
        print "    variable is not a valid list or tuple"
        return ""


######################################################################
## 
## strings manipulations
## 
######################################################################

# Returns a version of a string containing only lowercase a-z and 0-1 characters. Could be useful for identifying that "Spider-Man",
# "spider_man" and "spiderman" refer to the same basic character.
# Added FFX 3.2
def cleanUpString(originalString):
    originalString = string.lower(originalString)
    newString = ""
    for c in originalString:
        if ((ord(c) > 47) and (ord(c) < 58)) or ((ord(c) > 96) and (ord(c) < 123)):
            newString = newString + c
    return newString

# Convert a string to a tuple 
# Added in FFX 2.5
def stringToTuple(theVar, delimiter = ' '):
    theTuple = tuple(stringToList(theVar, delimiter))
    return theTuple

# Converts string back to a list of strings, integers and float.
# IMPORTANT: Other data types (variables, long integers, etc.) are not supported.; they' ll be converted to one of the above three types.
def stringToList(theVar, delimiter = ' '):
    theListOfStrings = string.split(theVar, delimiter) #WARNING! syntax obsolete in more recent Python versions?
    # if any of the list's string elements are integers, convert them
    theList = []
    for item in theListOfStrings:
        try:
            i = float(item)
            j = int(i)
            if i == j:
                theList.append(j)
            else:
                theList.append(i)
        except:
            theList.append(item)
    return theList


######################################################################
## 
## numbers manipulations
## 
######################################################################

# Is number x's value close enough to number y's?
# Added FFX 3.0
def isDifferenceAcceptable(x, y, maxDifference = 1):
    difference = abs(x - y)
    if difference - maxDifference < 0:
        return 0
    else:
        return 1

