######################################################################
## 
## Height Check module for FFvsTTR
## 
######################################################################
print "Starting Height Check module 1.4"

# Template_GetHeight(template)
# returns the height of a template object if known. Otherwise, if the datfiles module is loaded, tries to
# return the height of another template using the same mesh.
#
# checkHeight(templateList, precise = 'no', repeat = 0, replace = 0, giants = 1, campaign = 0)
# calculates the approximate height of multiple template objects and stores the results in heightcheckdata.py.
#
# checkHeightErrors()
# prints names and values of problematic objects.dat entries using the meshes list
# and returns a list of these entries.
#
# Other possibly useful functions can be found under "Mesh/template management" and "Helpers".

# Based on Dr Mike's brilliant flightcheck detection concept, as developped by M25.
#
# A big thank too to Stumpy for his falling object slow down code, which made this up to twenty times
# more precise (!) and for his MissionObjVar module.
#
# Height Check also uses Stumpy's upcoming DatFiles module (if available) to prevent character templates
# in objects.dat without the corresponding entry in characters.dat from crashing the game 'custom_colossus'
# is an example of such problematic entries. If you want to try Height Check but don't have Datfiles yet,
# just make certain the template list you provide as argument to checkHeight() doesn't contain characters
# without both entries.

# IMPORTANT:
# - Please read the comments just above the definitions of Template_GetHeight() and checkHeight() below.
# - Don't forget to change the MODFOLDER variable below.
# - Add the included heightcheckdata.py in the same folder as heightcheck.py
# - For character templates: while results are usually precise enough, sometimes they're completely off.
#	I haven't found any pattern to these aberrations.

# v. 1.4 2006-07-09 (pimthe): 
# - Merged Kuertee's 1.3 with mine.

# v. 1.3 2006-07-09 (Kuertee):
# - Added custom character support. (Now that the complex attribute is used to retrieve a custom character's actual
#	template, this is now possible. All due to FFX3 - thanks Dr.Mike.)
# - tweak: Minimised output to the script.log.
# - bug-fix: Data type returned by BuildingTemplate_GetDimensions is now alwasy a Dictionary. Previously, it returned
#	None for buildings not found in the library.
# - bug-fix: Re-worked the code that kept track of which buildings still need to be measured.

# v. 1.3 2006-06-22 (pimthe):
# - Moved helpers to sysutils.py and generalutils.py.
#   (By the way, it's too bad there's no script access to the "m" key AI debugging menu, as toggling through 
#   the display withthe "," key gives access to the character's position, with the z-axis point starting on 
#   top of him rather than under him. :S )

# v. 1.2 2006-02-18: (additions by kuertee)
# - Added retrieval of building dimensions data. Because buildings cannot be destroyed (via Object_Destroy),
#	a special empty city terrain is required. This is included in this module's zip-file that can be
#	downloaded from http://www.frpmods.com/files/kuertee/heightcheckV1.2.zip. Detailed information
#	on how to retrieve building data is located closer to the bottom of this module.
# - BuildingTemplate_GetDimensions (string template) returns a dictionary of
#	{"nw":tuple position, "se": tuple position, "w": float width, "d": float depth, "h": float height}
# - rumble room game-mode: get building dimensions is required to measure the buildings in objects.dat.
#	use this tool to save the data in heightcheckdata.py prior to delivery of the mod.

# v. 1.1 2005-04-30:
# - Changed some erroneous height values in heightcheckdata.py, thanks to Kuertee.
# - Added limited and experimental mesh path detection using Stumpy's DatFiles. The results are stored
#	in heightcheckdata.py. HERO FILES AND SKIRMISH MODE CHARACTERS ARE CURRENTLY NOT SUPPORTED.
# - Template_GetHeight(): if the template hasn't been detected and if the datfiles module is loaded,
#	the function now tries to check if another template using the same mesh has an height value.
# - You can now use checkHeightErrors() to verify your heightcheckdata.py OBJECTS_HEIGHT entries for potential errors.
# - Updated updateFile() to work with multiple variables.
# - The checkHeight() test is now usually faster at the lowest speeds as it concludes after three passes
#	if all three results have been identical and if there not suspect values (0, 9, 19, 29, 57).
# - Replaced typographically correct symbols with the common simulacrii in these notes, as Python throwed
#	a warning. :/

# v. 1.0 2005-04-25:
# - Initial release.

#TO DO?:
# - Possibly add template mesh path check. This would allow returning the value of an existing template
#	using the same mesh with Template_GetHeight() instead of having to use checkHeight()
#	EDIT: v. 1.1: Done partially.
# - Find a way to get rid of the occasional crappy results. :(
# - Replace writing of custom template's height according to campaign with writing of custom template's
#	height with its true name?
# - Optimize speed. While it could be really useful - obviously -, I don't intend to. If you feel like
#	doing it or any other change, be my guest.

######################################################################
## 
## Initialisation
## 
######################################################################

import os
import cshelper
import missionobjvar
from js import *
import ff
import copy

import sysutils
import generalutils

try:
	import datfiles
	DATFILESLOADED = 1
except:
	DATFILESLOADED = 0
try:
	#kuertee: changed details file-name to heightcheckdata
#	  import details
#	  OBJECTS_HEIGHT = details.OBJECTS_HEIGHT
	import heightcheckdata
	OBJECTS_HEIGHT = heightcheckdata.OBJECTS_HEIGHT
except:
	OBJECTS_HEIGHT = {}
print "OBJECTS_HEIGHT: %i entries"%len(OBJECTS_HEIGHT)
try:
	#kuertee: changed details file-name to heightcheckdata
#	  NIF_OBJECTS = details.NIF_OBJECTS
	NIF_OBJECTS = heightcheckdata.NIF_OBJECTS
except:
	NIF_OBJECTS = {}
print "NIF_OBJECTS: %i entries"%len(NIF_OBJECTS)
CURRENT_SUBJECT_HEIGHT = []
SKIPPED_SUBJECTS_HEIGHT = []


#kuertee removed
# Since FFX isn't available yet for FFvsTTR, let's call this hopeful forward compatibility.
# As FFX apparently won't be installed in the main data folder, the code below is *very* unlikely to be 100% valid
# Anyway, just change the value of the last MODFOLDER for now, OK?
# try: 
#	  import ffxextrasS
#	  ffxextras = ffxextrasS #"import module as name" doesn't work in ye olde Python 1.5
#	  MODFOLDER = ffxextras.MODFOLDER
#	  print "importing mod's folder name from mod's ffxextrasS.py: '%s'"%MODFOLDER
# except:
#	  try:
#		  import ffxextras
#		  print "importing mod's folder name from general ffxextras.py: '%s'"%MODFOLDER
#		  MODFOLDER = ffxextras.MODFOLDER
#	  except:
#		  MODFOLDER = "data" #change this to your mod's root folder
#		  print "mod's folder name set directly in HeightCheck to: '%s'"%MODFOLDER
#kuertee removal end
#kuertee added
if DATFILESLOADED:
	MODFOLDER = datfiles.GetModPath()
else:
	MODFOLDER = sysutils.getModFolder() #Ep. 2006-06-22; of course, if SysUtils is available, DATFiles will be too anyway
#kuertee addition end
missionobjvar.Mission_SetVar('heightcheck_sampling_default', (1, 8))
Mission_SetAttr('heightcheck_startingpos', 2300)
Mission_SetAttr('heightcheck_campaign', 0)

######################################################################
## 
## Height detection
## 
######################################################################

# Template_GetHeight()
# Returns a tuple where the first value is the height. The second value is 1 if the template was found
# in the OBJECTS_HEIGHT dictionnary and 0 if not. If there is a result for other templates using this
# mesh path, the first value will return that the first of those results and the second value, zero;
# otherwise, both values will be zeroes.
def Template_GetHeight(template):
	#kuertee additions
	global gaMissingTemplates
	try: skXTest = gaMissingTemplates
	except: gaMissingTemplates = []
	#kuertee additions end
	if OBJECTS_HEIGHT.has_key(template):
		return (OBJECTS_HEIGHT[template], 1)
	elif DATFILESLOADED:
		meshTemplates = Template_GetMeshTemplates(template)
# 		if len(meshTemplates) != 0:
		#kuertee additions
		if len(meshTemplates) != 0 and template not in gaMissingTemplates:
			gaMissingTemplates.append (template)
		#kuertee additions end
			print "Template '%s' not found in list of template heights."%template
			print "	 but template '%s', using the same mesh, has been."%meshTemplates[0]
			return (OBJECTS_HEIGHT[meshTemplates[0]], 0)
	#kuertee additions
	if template in gaMissingTemplates: return (0, 0)
	gaMissingTemplates.append (template)
	#kuertee additions end
	print "Template_GetHeight() non-fatal error:"
	print "    template '%s' not found in list of template heights."%template
	print "    To detect approximate height, use"
	print "    checkHeight(templateList, precise = 'no', repeat = 0, replace = 0, giants = 1, campaign = 0)"
	return (0, 0)

# checkHeight()
# IMPORTANT:
# - Results for characters will be approximative. Those for generic objects should be far more exact.
# - Doesn't work on buildings, as they cannot be spawned correctly and can't be destroyed afterward.
#	 (To detect an existing building's height, just spawn someone above it.)
# - Doesn't work either on tanks (the 'crusher' attribute seems to be built-in for tanks and messes 
#	 with detection?)
# - Doesn't work with power proxies or powerups.
# - Not every template object will give reliable results:
#	 - Object whose apex isn't at the center (the T-Rex, for example) will give underestimated results.
#	 - Objects with the attribute 'physical' set to 0 won't be processed, as they can't be hit. They will
#	   return an height of 0. Idem for those with 'avoid'=0 or generic objects with 'standon'-0.
#	 - When using giants=0, objects taller than 60-110 units won't be detected correctly (for example,
#	   depending on gamespeed, the trex mesh (107 units) may or may not be detected correctly. Uncorrectly
#	   detected templates either return zero or a low number; for example, time_master may report as
#	   being 24 unit high, while its actual height is 170, as the mesh seems to have two hit targets or
#	   something...
#	   Even when using giants=1, objects taller than 255-260 units won't work (time_master is 170). 
#	   Uncorrectly detected giants will return an height of 0.
#	 - The smaller characters have a chance of returning an height of 0 if precise='no' and gamespeed 
#	   is high enough, especially if repeat is set to a low value.
#	 - A mesh's physical height for the game engine isn't necessarily exactly the same as it's visual height.
#	 - While results for character height generally have a small enough error margin, there are sometimes 
#	   aberrations with no known cause. Hence the use of multiple passes on characters (the repeat= argument)
#	   to reduce the problem. Occasionally, though, every result is worthless... I haven't found a pattern
#	   to this problem. :(
#	 - Objects with invalid mesh path in templates.dat will return 0.
# - Using the plus/minus Easter egg will not influence results. ;)
#
# Detecting height can be quite slow, especially when using precise='yes'. OK, extremely slow.
# Example, using the el_diablo template (height of 34):
# precise = 'no', repeat = 0, giants = 1, game speed of 1.0X:
#	  delay: 35 secs., normal error margin: 0-4, usually 1
# precise = 'no', repeat = 0, giants = 0, game speed of 1.0X:
#	  delay: 22 secs., normal error margin: 0-4, usually 1
# precise = 'no', repeat = 0, giants = 1, game speed of 0.4X:
#	  delay: 190 secs., normal error margin: 0-1
# precise = 'yes', repeat = 0, giants = 1, game speed of 1.0X:
#	  delay:  240 secs., normal error margin: 0-1
# precise = 'no', repeat = 1, giants = 0, game speed of 1.0X:
#	  delay:  4 secs., normal error margin: 0-4, but could be much higher
#
# Note that occasional errors due to the object passing through the subject template instead of bouncing off it will
# lenghten the process, as the next pass will start from the maximum possible height.
#	  
# A generic object takes about 8 seconds to be detected at 1X speed.
#
#
# The best solution for modmakers is to use checkHeight() themselves beforehand with 
#	  checkHeight(newTemplatesList)
# running at 0.4X game speed if you've got only a few new templates and no existing templates with
# a new mesh.
# If your list is quite long (running every built-in templates took me a few hours), use instead
#	  checkHeight(modifiedCharacterTemplatesList, replace = 1)
# running at 0.4X game speed (or higher if you don't need as much precision) and
#	  checkHeight(modifiedGenericTemplatesList, replace = 1)
# running at 2X game speed.
#
# Then, reserve the end-user in-game detection for custom characters added during the campaign, etc.
# Even then, you might wish to use it during cutscenes and/or in bases. Depending on how precise you
# need the detection to be, you could either use precise='yes' to compensate for the fact that
# user-selected game speed might be at up to 2X or you might set repeat = 10 to somewhat reduce the
# impact of erroneous results. Ex:
#	  height = Template_GetHeight(template)
#	  if height[1] == 0:
#		  checkHeight([template], precise = 'no', repeat = 10)
#
# Setting precise = 'yes' will temporarily reduce the time scale (game speed) to 3/10th of normal,
# as EVENT_TIMER can't occur often enough in normal time or faster, which may not be precise enough,
# giving an usual error margin of 0-10 at 2X speed. Most of the time, "precise = 'yes'" should give 
# an error margin of 0-2 at 2X speed. Computer performance will also influence results. To put thing 
# in perspective, el_diablo has an real height of 34 and a skull, of 11. Since game speed is irrelevant 
# to generic object detection, precise = 'yes' will be ignored for them.
#
# If repeat = 0, default value (8) will be used instead. Since the precision margin for generic objects
# is inferior to 0.1, repeat= will be ignored for them.
#
# If replace = 1, values for templates already in heightcheckdata.py will be overwritten.
#
# If giants = 0, giant-sized characters won't be detected correctly, but the first pass of the repeating
# detection will be speeded up considerably. This is currently ignored for generic templates.
#
# If campaign = 1, custom heroes will be saved to heightcheckdata.py. Otherwise they won't, since their template
# name is generated on the fly, making the entry useless in skirmish.
# 
# To cancel checkHeight() after the current iteration of the current template, use checkHeightCancel().
def checkHeight(templateList, precise = 'no', repeat = 0, replace = 0, giants = 1, campaign = 0):
	if repeat != 0:
		missionobjvar.Mission_SetVar('heightcheck_sampling', (1, int(repeat)))
	else:
		missionobjvar.Mission_SetVar('heightcheck_sampling', missionobjvar.Mission_GetVar('heightcheck_sampling_default'))
	if campaign != 0:
		Mission_SetAttr('heightcheck_campaign', 1)
	Mission_SetAttr('heightcheck_giants',giants)
	templateListCleaned = copy.copy(templateList)
	global SKIPPED_SUBJECTS_HEIGHT
	for template in templateList:
		if (replace == 0) and (OBJECTS_HEIGHT.has_key(template)):
			print "template '%s' is already in details.OBJECTS_HEIGHT. Skipped!"%template
			templateListCleaned.remove(template)
			SKIPPED_SUBJECTS_HEIGHT.append((template, 'replace=0 and already in heightcheckdata.py'))
		else:
			if template[0:16] == 'custom_template_':
				print "Template '%s' is a custom character."%template
			elif not Template_Exists(template):
				print "Template '%s' not found in objects.dat. Skipped!"%template
				templateListCleaned.remove(template)
				SKIPPED_SUBJECTS_HEIGHT.append((template, 'not found'))
			else:
				templateClass = Template_GetClassString(template)
				if (templateClass == 'GAME_OBJ_BUILDING') or (templateClass == 'GAME_OBJ_TANK') or (templateClass == 'GAME_OBJ_POWER_PROXY') or (templateClass == 'GAME_OBJ_POWERUP'):
					print "Template '%s' is a %s and can't be detected correctly. Skipped!"%(template, templateClass)
					templateListCleaned.remove(template)
					averageHeight(template, [])
					SKIPPED_SUBJECTS_HEIGHT.append((template, templateClass))
				elif DATFILESLOADED:
					charDB = datfiles.Campaign_ReadCharacters()
					if (templateClass == 'GAME_OBJ_HERO') or (templateClass == 'GAME_OBJ_VILLIAN') or (templateClass == 'GAME_OBJ_CIVILIAN') or (templateClass == 'GAME_OBJ_MINION') or (templateClass == 'GAME_OBJ_POLICE'):
						if not charDB.has_key(template):
							print "Character template '%s' entry missing from characters.dat. Skipped!"%template
							templateListCleaned.remove(template)
							SKIPPED_SUBJECTS_HEIGHT.append((template, 'entry missing from characters.dat'))
	if len(templateListCleaned) == 0:
		print "List of templates to check for height is empty."
		return
	checkHeightLoop(templateListCleaned, precise)

def checkHeightCancel():
	missionobjvar.Mission_SetVar('heightcheck_queue', [])

def checkHeightLoop(templateList, precise):
	if templateList == '':
		templateList = missionobjvar.Mission_GetVar('heightcheck_queue')
	if len(templateList) == 0:
		print "Checking height for selected templates complete."
		global SKIPPED_SUBJECTS_HEIGHT
		for skippedItem in SKIPPED_SUBJECTS_HEIGHT:
			print "Skipped: %s (%s)"%(skippedItem[0], skippedItem[1])
		SKIPPED_SUBJECTS_HEIGHT = []
		if Object_Exists('heighttest_platform'):
			Object_Destroy('heighttest_platform')
		if Object_Exists('heighttest_dummy'):
			Object_Destroy('heighttest_dummy')
		if Object_Exists('heighttest_subject'):
			Object_Destroy('heighttest_subject')
		return
	template = templateList[0]
	print "Checking height of %s: initialising"%template
	if template[0:16] == 'custom_template_':
		templateClass = 'GAME_OBJ_HERO'
	else:
		templateClass = Template_GetClassString(template)
	if templateClass == 'GAME_OBJ_GENERIC':
		startHeightTestGeneric(template, precise, templateList)
	else:
		startHeightTestCharacter(template, precise, templateList)

def OnCheckHeight(event):
	precise = event.string
	checkHeightLoop('', precise)

def startHeightTestCharacter(template, precise, templateList):
	sampling = missionobjvar.Mission_GetVar('heightcheck_sampling')
	if sampling[0] == 1:
		print "startHeightTestCharacter(): %i templates left to check"%len(templateList)
	print "sample %i of %i"%(sampling[0], sampling[1])
	if sampling[0] == sampling[1]:
		templateList.pop(0)
	missionobjvar.Mission_SetVar('heightcheck_queue', templateList)
	# put the char on a platform first, because objects tend not to fall lower than 83 units...
	if Object_Exists('heighttest_platform'):
		Object_Destroy('heighttest_platform')
	Event_RegisterSink(EVENT_TIMER, 'startHeightTestCharacter2', template, precise, 0.1, 0, 0) #delay needed between destruction and respawning

def startHeightTestCharacter2(event):
	template = event.object
	precise = event.string
	Object_SpawnAt('skull', (0, 0, 1989), 'heighttest_platform') #skull has an height of 11
	Object_SetAttr('heighttest_platform','mass', 1000000)
	Object_SetAttr('heighttest_platform','minForce', 1000000)
	Object_SetAttr('heighttest_platform','minHealth', 1)
	sampling = missionobjvar.Mission_GetVar('heightcheck_sampling')
	print "sample %i of %i"%(sampling[0], sampling[1])
	if sampling[0] == 1:	
		startingHeight = (Mission_GetAttr('heightcheck_giants') * 200) + 2070
	else:
		startingHeight = max(Mission_GetAttr('heightcheck_startingpos') + 30, 2070)
	print "startingHeight=%i"%startingHeight
	if Object_Exists('heighttest_dummy'):
		Object_Destroy('heighttest_dummy')
	Object_SpawnAt('trainer_hack', (0, 0, startingHeight), 'heighttest_dummy')
	if not Object_Exists('heighttest_subject'):
		dummies=['','','','']
		if template[0:16] == 'custom_template_':
			templateClass = 'GAME_OBJ_HERO'
		else:
			templateClass = Template_GetClassString(template)
		if templateClass == 'GAME_OBJ_HERO':
			#ripped from FFX NPCMimic() code
			heroCount=len(cshelper.getAllHeroes())
			index=0
			while heroCount<4:
				dummies[index] = "mobj_dummy_hc%i"%index
				cshelper.spawn(dummies[index], template, (99999,99999,0))
				heroCount=heroCount+1
				index=index+1
		Object_SpawnAt(template, (0, 0, 2000), 'heighttest_subject') #chars would move up to stand on platform, but not static objects
		for dummy in dummies:
			if dummy!='':
				Object_Destroy(dummy)
		if (Object_GetClass('heighttest_subject') & OC_CHARACTER):
			AI_SetEnabled('heighttest_subject', 0)
	Event_RegisterSink(EVENT_TIMER, 'startHeightTestCharacter3', template, precise, 0.2, startingHeight, 0)

def startHeightTestCharacter3(event):
	template = event.object
	precise = event.string
	startingHeight = event.user
	previousHeight = startingHeight
	if Object_GetAttr('heighttest_subject', 'physical') == 0:
		cancelHeightTest(template, precise)
		return
	print "precise = %s"%precise
	if precise == 'yes':
		Mission_SetAttr('heightcheck_timescale', ff.APP_TIME_SCALE)
		ff.APP_TIME_SCALE = 0.3
	Object_SetAttr('heighttest_subject','mass', 1)
	Object_SetAttr('heighttest_dummy', 'mass', 0.1)
	Object_SetAttr('heighttest_dummy', 'minForce', 1)
	Trigger_Force('heighttest_dummy', 0, 0, -1, 2, TF_ABSOLUTE)
	Event_RegisterSink(EVENT_TIMER, 'OnHeightTestCharacter', '%i'%startingHeight, precise, 0.02, previousHeight, 0) # can't use EVENT_DAMAGE because of passive defences

def startHeightTestGeneric(template, precise, templateList):
	# the exact sequence of events must be:
	# 1-spawn platform
	# 2-spawn flyer on platform
	# 3-trigger flight
	# 4-spawn subject
	# 5-destroy platform
	# 6-trigger land
	print "startHeightTestGeneric(): %i templates left to check"%len(templateList)
	template = templateList[0]
	missionobjvar.Mission_SetVar('heightcheck_queue', templateList)
	# put the char on a platform first, because otherwise they will pass through the subject template instead of landing on them
	if Object_Exists('heighttest_platform'):
		Object_Destroy('heighttest_platform')
	Event_RegisterSink(EVENT_TIMER, 'startHeightTestGeneric2', template, precise, 0.1, 0, 0) #delay needed between destruction and respawning

def startHeightTestGeneric2(event):
	template = event.object
	precise = event.string
	Object_SpawnAt('skull', (0, 0, 2300), 'heighttest_platform')
	#had to move everything within map limits, as falling characters don't stay outside. Interesting, no?
	Object_SpawnAt('alien_sergeant', (0, 0, 2311), 'heighttest_dummy')
	AI_SetEnabled('heighttest_dummy', 0)
	Trigger_Move('heighttest_dummy', 'fly', (0,0,2300))
	Event_RegisterSink(EVENT_TIMER, 'startHeightTestGeneric3', template, precise, 0.2, 2250, 0)

def startHeightTestGeneric3(event):
	template = event.object
	precise = event.string
	if not Object_Exists('heighttest_subject'):
		Object_SpawnAt(template, (0, 0, 2000), 'heighttest_subject') 
	if (Object_GetAttr('heighttest_subject', 'physical') == 0) or (Object_GetAttr('heighttest_subject', 'avoid') == 0) or (Object_GetAttr('heighttest_subject', 'standon') == 0):
		cancelHeightTest(template, precise)
		return
	templateList = missionobjvar.Mission_GetVar('heightcheck_queue')
	del templateList[0]
	missionobjvar.Mission_SetVar('heightcheck_queue', templateList)
	if Object_Exists('heighttest_platform'):
		Object_Destroy('heighttest_platform')
	Trigger_Move('heighttest_dummy', 'land', (0,0,2300))	  
	Event_RegisterSink(EVENT_TIMER, 'OnHeightTestGeneric', template, precise, 7.5, 0, 0)#the flying/landing makes it much longer than otherwise necessary :(

#don't test non-physical objects, since all results will be invalid
def cancelHeightTest(template, precise):
	print "template %s is non-physical or non-avoid or non-standon generic"%Object_GetTemplate('heighttest_subject')
	global SKIPPED_SUBJECTS_HEIGHT
	SKIPPED_SUBJECTS_HEIGHT.append((template, 'non-physical or non-avoid or non-standon generic'))
	Object_Destroy('heighttest_dummy')
	print "Character subject sampling cancelled."
	Object_Destroy('heighttest_subject')
	global CURRENT_SUBJECT_HEIGHT
	CURRENT_SUBJECT_HEIGHT = []
	averageHeight(template, CURRENT_SUBJECT_HEIGHT)
	templateList = missionobjvar.Mission_GetVar('heightcheck_queue')
	del templateList[0]
	missionobjvar.Mission_SetVar('heightcheck_queue', templateList)
	Event_RegisterSink(EVENT_TIMER, 'OnCheckHeight', '', precise, 0.2, 0, 0)

def OnHeightTestCharacter(event):
	startingHeight = eval(event.object)
	previousHeight = event.user
	precise = event.string
	pos = Get_ObjectPos('heighttest_dummy')
	print pos
	if (((pos[2] < previousHeight) and (pos[2] > 1990)) or ((previousHeight == pos[2]) and (pos[2] > startingHeight - 10))) and (pos[0] == 0)  and (pos[1] == 0):
		if (previousHeight - pos[2]) > 12:
			Trigger_Force('heighttest_dummy', 0, 0, 1, 12, TF_ABSOLUTE)###!
		elif (previousHeight - pos[2]) > 6:
			Trigger_Force('heighttest_dummy', 0, 0, 1, 9, TF_ABSOLUTE)###!
		elif (previousHeight - pos[2]) > 2:
			Trigger_Force('heighttest_dummy', 0, 0, 1, 6, TF_ABSOLUTE)###!
		else:
			Trigger_Force('heighttest_dummy', 0, 0, 1, 3, TF_ABSOLUTE)###!
		Event_RegisterSink(EVENT_TIMER, 'OnHeightTestCharacter', event.object, precise, 0.02, pos[2], 0)
		return
	startingHeight = int(pos[2])
	template = Object_GetTemplate('heighttest_subject')
	Object_Destroy('heighttest_dummy')
	objectHeight = int(previousHeight) - 2000
	global CURRENT_SUBJECT_HEIGHT
	if objectHeight < 0:
		print "Height of template %s error: negative number %i, result ignored"%(template, objectHeight)
		objectHeight = 0
		startingHeight = 2270
	else:
		CURRENT_SUBJECT_HEIGHT.append(objectHeight)
	sampling = missionobjvar.Mission_GetVar('heightcheck_sampling')
	print "sample %i of %i"%(sampling[0], sampling[1])
	if (sampling[0] + 1 > sampling[1]) or quickValidateCharacter():
		missionobjvar.Mission_SetVar('heightcheck_sampling', (1, sampling[1]))
		print "Character subject sampling complete."
		Object_Destroy('heighttest_subject')
		averageHeight(template, CURRENT_SUBJECT_HEIGHT)
		CURRENT_SUBJECT_HEIGHT = []
	else:
		missionobjvar.Mission_SetVar('heightcheck_sampling', (sampling[0] + 1, sampling[1]))
		Mission_SetAttr('heightcheck_startingpos', startingHeight)
	if precise == 'yes':
		ff.APP_TIME_SCALE = Mission_GetAttr('heightcheck_timescale')
		print "Height of template %s estimated to %i using precise method"%(template, objectHeight)
	else:
		print "Height of template %s roughly estimated to %i"%(template, objectHeight)
	Event_RegisterSink(EVENT_TIMER, 'OnCheckHeight', '', precise, 0.2, 0, 0)

#conclude the test after three passes if all three results have been identical and there not suspect values.
def quickValidateCharacter():
	if len(CURRENT_SUBJECT_HEIGHT) == 3:
		if CURRENT_SUBJECT_HEIGHT[0] == CURRENT_SUBJECT_HEIGHT[1] == CURRENT_SUBJECT_HEIGHT[2]:
			if (CURRENT_SUBJECT_HEIGHT[0] != 0) and (CURRENT_SUBJECT_HEIGHT[0] != 9) and (CURRENT_SUBJECT_HEIGHT[0] != 19) and (CURRENT_SUBJECT_HEIGHT[0] != 29) and (CURRENT_SUBJECT_HEIGHT[0] != 57):
				templateList = missionobjvar.Mission_GetVar('heightcheck_queue')
				if len(templateList) > 0:
					del templateList[0]
					missionobjvar.Mission_SetVar('heightcheck_queue', templateList)
				return 1
	return 0

def OnHeightTestGeneric(event):
	template = event.object
	precise = event.string
	pos = Get_ObjectPos('heighttest_dummy')
	print pos
	Object_Destroy('heighttest_dummy')
	objectHeight = int(pos[2]) - 2000
	global CURRENT_SUBJECT_HEIGHT
	if objectHeight < 0:
		print "Height of template %s error: negative number %i, result ignored"%(template, objectHeight)
		objectHeight = 0
	else:
		CURRENT_SUBJECT_HEIGHT.append(objectHeight)
	#ignore sampling for generic
	print "Generic subject sampling complete."
	Object_Destroy('heighttest_subject')
	averageHeight(template, CURRENT_SUBJECT_HEIGHT)
	CURRENT_SUBJECT_HEIGHT = [] 	   
	print "Height of template %s is %i"%(template, objectHeight)
	Event_RegisterSink(EVENT_TIMER, 'OnCheckHeight', '', precise, 0.2, 0, 0)

def averageHeight(template, sampling):
	print sampling
	result = 0
	if len(sampling) == 0:
		result = 0
	elif len(sampling) == 1:
		result = sampling[0]
	elif len(sampling) == 2:
		result = int((sampling[0] + sampling[1]) / 2)
	else:
		safeSampling = []
		for i in sampling:
			if (i != 9) and (i != 19) and (i != 29) and (i != 57): #the values 9, 19, 29 and 57 are quite often erroneous for some reason (rounded multiple of ~9.5?)
				safeSampling.append(i)
		if len(safeSampling) > 0:
			sampling = safeSampling
		else:
			print "%s: height estimate unreliable: only suspect results %s found"%(template, sampling)
		sampling.sort()
		sampling = sampling[1:-1] #remove extreme values, as they're probably erroneous
		if len(sampling) > 1:
			if sampling[-1] - sampling[0] > 12:
				# Remove the lower results. Small errors usually give overestimated results,
				# while big ones mostly give underestimated results.
				print "%s: height estimate unreliable: results ranging from %i to %i"%(template, sampling[0], sampling[-1])
				samplingGuess = []
				for i in sampling:
					if i >= sampling[-1] - 12:
						samplingGuess.append(i)
				sampling = samplingGuess
				print "    Guessing more probable results: %s"%sampling
		for i in sampling:
			result = result + i
		result = result / len(sampling)
	global OBJECTS_HEIGHT
	if OBJECTS_HEIGHT.has_key(template):
		if OBJECTS_HEIGHT[template] == result:
			print "%s: Previous height estimate was already at %i"%(template, result)
			return
		else:
			print "%s: Replacing previous height estimate of %i to %i"%(template, OBJECTS_HEIGHT[template], result)
			OBJECTS_HEIGHT[template] = result
	else:
		print "%s: Adding height estimate of %i"%(template, result)
		#print OBJECTS_HEIGHT
	#not very efficient if updating a single value or checking multiple templates...
	objects_height_keys = OBJECTS_HEIGHT.keys()
	objects_height_values = OBJECTS_HEIGHT.values()
	objects_height_keys.append(template)
	objects_height_values.append(result)
	OBJECTS_HEIGHT = generalutils.makeDict(objects_height_keys, objects_height_values)
	if not ((template[0:16] == 'custom_template_') and (Mission_GetAttr('heightcheck_campaign') == 0)):
		sysutils.updateFile('heightcheckdata.py', 'OBJECTS_HEIGHT', OBJECTS_HEIGHT)

######################################################################
## 
## Mesh/template management
## 
######################################################################

# Returns 1 if two templates are using the same mesh path.
def Templates_AreSameMesh(template1, template2):
	templateMesh1 = Template_GetMesh(template1)
	if templateMesh1 == '':
		return 0
	templateMesh2 = Template_GetMesh(template2)
	return templateMesh1 == templateMesh2

# For a given template, returns a list of the different templates based on the same mesh
# founds in objects.dat
def Template_GetMeshTemplates(template):
	templateMesh = Template_GetMesh(template)
	if templateMesh == '':
		return []
	return Mesh_GetTemplates(templateMesh)

# For a given template, returns the mesh path as found in objects.dat
def Template_GetMesh(template):
	if not DATFILESLOADED:
		print "Template_GetMesh() fatal error: datfiles module not loaded"
		return ''
	objectsDat = datfiles.Campaign_ReadObjects()
	if not objectsDat.has_key(template):
		#kuertee additions
		try: mbCustomExists = ff.Custom_Exists (template)
		except: mbCustomExists = 0
		if mbCustomExists:
			return datfiles.Campaign_ReadHeroFile (template + ".hero")["NIF"]
		#kuertee additions end
		print "Template_GetMesh() fatal error: template '%s' not found"%template
		return ''
	templateDat = objectsDat[template]
	return templateDat['NIF']

# For a given mesh, returns 1 if its list of templates include the templat4e passed as parameter
# verbose = 1 will activate the print comments
def Mesh_HasTemplates(mesh, template, verbose = 0):
	templatesList = Mesh_GetTemplates(mesh)
	if len(templatesList) == 0:
		if verbose:
			print "Mesh '%s' not found"%mesh
		return 0
	if templatesList.count(template) == 0:
		if verbose:
			print "Template '%s' not found in '%s'"%(template, mesh)
		return 0
	else:
		if verbose:
			print "Template '%s' found in '%s'"%(template, mesh)
		return 1		
	
# For a given mesh, returns a list of the different templates founds in objects.dat
def Mesh_GetTemplates(mesh):
	if len(NIF_OBJECTS) == 0:
		print "NIF_OBJECTS not found... reading from objects.dat"
		checkMeshes()
	if NIF_OBJECTS.has_key(mesh):
		return NIF_OBJECTS[mesh]
	else:
		return []

# Directly adds the template to the NIF_OBJECTS mesh path key
# and optionally to the heightcheckdata.py NIF_OBJECTS entry.
# verbose = 1 will activate the print comments
def Mesh_AddTemplate(mesh, template, verbose = 0, save = 0):
	global NIF_OBJECTS
	if not NIF_OBJECTS.has_key(mesh):
		print "Mesh_AddTemplate() fatal error: Mesh '%s' not found"%mesh
		return
	meshTemplates = NIF_OBJECTS[mesh]
	if meshTemplates.count(template) != 0:
		if verbose:
			print "Template '%s' already in mesh templates list"%template
	else:
		print "Adding template '%s' to mesh templates list"%template
		meshTemplates.append(template)
	if verbose:
		print mesh
		print meshTemplates
	if save:
		sysutils.updateFile('heightcheckdata.py', 'NIF_OBJECTS', NIF_OBJECTS)

# Removes the template from the NIF_OBJECTS mesh path key
# and optionally from the heightcheckdata.py NIF_OBJECTS entry.
# verbose = 1 will activate the print comments
def Mesh_RemoveTemplate(mesh, template, verbose = 0, save = 0):
	global NIF_OBJECTS
	if not NIF_OBJECTS.has_key(mesh):
		print "Mesh_RemoveTemplate() fatal error: Mesh '%s' not found"%mesh
		return
	meshTemplates = NIF_OBJECTS[mesh]
	if meshTemplates.count(template) == 0:
		if verbose:
			print "Template '%s' not found in mesh templates list"%template
	else:
		print "Removing template '%s' from mesh templates list"%template
		meshTemplates.remove(template)
	if verbose:
		print mesh
		print meshTemplates
	if save:
		sysutils.updateFile('heightcheckdata.py', 'NIF_OBJECTS', NIF_OBJECTS)

# Gets the mesh for every entry in objects.dat and creates a dictionary with the mesh path as key
# and the list of templates using this mesh as value.
# mode = 'list': (default) simply prints out the results
# mode = 'add': adds the results to the NIF_OBJECTS mesh path key and to the heightcheckdata.py NIF_OBJECTS entry
# mode = 'replace': replaces the content of the NIF_OBJECTS mesh path key and the heightcheckdata.py NIF_OBJECTS entry with the results
# verbose = 1 will activate the print comments
def checkMeshes(mode = 'list', verbose = 0):
	if verbose:
		print "checkMeshes(mode = '%s', verbose = %i)"%(mode, verbose)
	if not DATFILESLOADED:
		print "checkMeshes() fatal error: datfiles module not loaded"
		return
	objectsDat = datfiles.Campaign_ReadObjects()
	pairs = []
	for obj in objectsDat.keys():
		pairs.append([objectsDat[obj]['NIF'], obj])
	nifGroups = []
	for pair in pairs:
		added=0
		for group in nifGroups:
			if pair[0] == group[0]:
				group.append(pair[1])
				added=1
				break
		if added == 0:
			nifGroups.append([pair[0],pair[1]])
	nifs = []
	for group in nifGroups:
		nifs.append(group.pop(0))
	nifsDict = generalutils.makeDict(nifs, nifGroups)
	print "Meshes and found templates associated:"
	print nifsDict
	global NIF_OBJECTS
	if mode == 'replace':
		NIF_OBJECTS = nifsDict
		sysutils.updateFile('heightcheckdata.py', 'NIF_OBJECTS', NIF_OBJECTS)
	elif mode == 'add':
		newMeshes = []
		newTemplatesLists = []
		for mesh in nifsDict.keys():
			if NIF_OBJECTS.has_key(mesh):
				meshTemplates = NIF_OBJECTS[mesh]
				for template in nifsDict[mesh]:
					if meshTemplates.count(template) == 0:
						if verbose:
							print "Adding template '%s' to mesh templates list"%template
						meshTemplates.append(template)
			else:
				newMeshes.append(mesh)
				newTemplatesLists.append(nifsDict[mesh])
				if verbose:
					print "Adding new mesh '%s' with templates %s to NIF_OBJECTS"%(mesh, nifsDict[mesh])
		if newMeshes != []:
			detailsPyMeshes = NIF_OBJECTS.keys()
			detailsPyTemplatesLists = NIF_OBJECTS.values()
			for newMesh in newMeshes:
				detailsPyMeshes.append(newMesh)
			for newTemplatesList in newTemplatesLists:
				detailsPyTemplatesLists.append(newTemplatesList)
			NIF_OBJECTS = generalutils.makeDict(detailsPyMeshes, detailsPyTemplatesLists)
		sysutils.updateFile('heightcheckdata.py', 'NIF_OBJECTS', NIF_OBJECTS)

# Prints names and values of problematic objects.dat entries using the meshes list
# and returns a list of these entries
def checkHeightErrors():
	errorsFound = []
	if len(NIF_OBJECTS) == 0:
		print "NIF_OBJECTS not found."
		return ["NIF_OBJECTS not found."]
	for nif in NIF_OBJECTS.keys():
		heights = []
		objects = NIF_OBJECTS[nif]
		for obj in objects:
			if not OBJECTS_HEIGHT.has_key(obj):
				print "'%s' object template has no height defined."%obj
				print "    Possibly a character template without the corresponding characters.dat entry.\n"
				errorsFound.append(obj)
			else:
				heights.append(OBJECTS_HEIGHT[obj])
		for height in heights:
			if height != heights[0]: #if some templates are physical/avoid/standon and other not, they will appear here
				for obj in objects:
					if OBJECTS_HEIGHT.has_key(obj):
						print "%s: %i"%(obj, OBJECTS_HEIGHT[obj])
						errorsFound.append(obj)
				print ""
			elif ((height == 9) or (height == 19) or (height == 29) or (height == 57)): #those values are known to be sometimes erroneous; zero isn't considered here
				print "Suspect values found:"
				for obj in objects:
					if OBJECTS_HEIGHT.has_key(obj):
						print "%s: %i"%(obj, OBJECTS_HEIGHT[obj])
						errorsFound.append(obj)
				print ""
	return errorsFound


######################################################################
## 
## building dimensions
## 
######################################################################
from skXAngles import *

try:
	BUILDING_DIMENSIONS = heightcheckdata.BUILDING_DIMENSIONS
	BUILDINGS_IN_OBJECTS_DAT = heightcheckdata.BUILDINGS_IN_OBJECTS_DAT
	NEXT_BUILDING = heightcheckdata.NEXT_BUILDING
except:
	BUILDING_DIMENSIONS = {}
	BUILDINGS_IN_OBJECTS_DAT = []
	NEXT_BUILDING = 0
print "BUILDINGS_IN_OBJECTS_DAT: %i entries"%len(BUILDINGS_IN_OBJECTS_DAT)
print "BUILDING_DIMENSIONS: %i entries"%len(BUILDING_DIMENSIONS.keys ())

MEASUREMENT_SNAP = 25

# BuildingTemplate_GetDimensions (string: template)
# returns a dictionary with these keys and values:
# {	"nw": north-west offset from centre,
#	"se": south-east offset from centre,
#	"w": width,
#	"h": height,
#	"d": depth
# }
def BuildingTemplate_GetDimensions (template):
	if Object_Exists (template):
		template = Object_GetTemplate (template)
	if template in BUILDING_DIMENSIONS.keys ():
		return BUILDING_DIMENSIONS [template]
	print "Building_GetDimensions() non-fatal error:"
	print "    template '%s' not found in list of building dimensions."%template
	print "    To detect approximate dimensions, use"
	print "    getBuildingDimensions ()"
	return {}

def isBuildingTemplate (ts):
	return Template_GetClassString (ts) == "GAME_OBJ_BUILDING"

# getBuildingDimensions ()
# the buildings' dimensions cannot be retrieved "on-the-fly". their dimensions
#	need to be stored in heightcheckdata.py by using the "get building
#	dimensions" skirmish game-mode-tool.
# the process of measuring all buildings from the objects.dat file is quite slow.
#	and the tool only measures six buildings at a time. restarts of the
#	game-mode is required until all buildings have been measured.
# the actual measurement routine is "hacky":
#	1. spawn 4 thugs at the top of the building.
#	2. trigger them to move outwards
#	3. if they get stuck before hitting the ground, teleport them a short
#		distance forward
#	4. once they hit the ground, trigger them to move back towards the
#		building
#	5. their position from the building's origin (i.e. the building's local
#		(0,0,0)) is then used to generate the building's dimensions
# and note that the measurement occurs only on one building at a time. the time
#	to take to measure the 6 buildings can be reduced if 24 "caliper" thugs
#	are spawned at once. however, this is not really necessary.
# these buildings don't spawn properlyy, and therefore have erroneous dimensions:
#	cub_bldg_church, cub_bldg_white and ger_bldg_cathedral.
# luckily, ffvsttr has only 30 buildings.
def getBuildingsToGetDimensionsOf (tevent = "", tbReplace = 0):
	Mission_StatusText ("getting list of buildings...")
	see ()
	global gaBuildings
	global gaBuildingTemplates
	gaBuildings = []
	gaBuildingTemplates = []
	msFileName = os.path.join (MODFOLDER, "objects.dat")
	if not os.path.isfile (msFileName):
		msFileName = os.path.join ("Data", "objects.dat")
	maBuildings = datfiles.Campaign_ReadObjects (msFileName).keys ()
	maBuildings = filter (isBuildingTemplate, maBuildings)
	mi = 0
	while (mi < len (maBuildings)):
		if maBuildings [mi] in BUILDINGS_IN_OBJECTS_DAT:
			maBuildings.remove (maBuildings [mi])
		else: mi = mi + 1
	#if not len (BUILDINGS_IN_OBJECTS_DAT):
	#	miBuilding = 0
	#	sysutils.updateFile ('heightcheckdata.py', 'BUILDINGS_IN_OBJECTS_DAT', maBuildings)
	#else:
	#	miBuilding = NEXT_BUILDING
	#if miBuilding < len (maBuildings):
	if len (maBuildings):
		gaBuildingTemplates = maBuildings [:6]
		print `gaBuildingTemplates `
		if len (gaBuildingTemplates):
			for ms in gaBuildingTemplates:
				BUILDINGS_IN_OBJECTS_DAT.append (ms)
				mi = gaBuildingTemplates.index (ms)
				Object_Spawn (ms, ms, "building" + `mi+1`)
				gaBuildings.append (ms)
			sysutils.updateFile ('heightcheckdata.py', 'BUILDINGS_IN_OBJECTS_DAT', BUILDINGS_IN_OBJECTS_DAT)
			getNextBuildingDimensions ()
	else:
		Mission_StatusText ("all buildings already measured.")
getBuildingDimensions = getBuildingsToGetDimensionsOf

def getNextBuildingDimensions (tevent = ""):
	global gaBuildings
	global gsBuilding
	if len (gaBuildings):
		gsBuilding = gaBuildings [0]
		Mission_StatusText ("get dimensions of " + Object_GetTemplate (gsBuilding))
		gaBuildings.remove (gsBuilding)
		mtPos = Get_ObjectPos (gsBuilding)
		Object_SpawnAt ("thug_with_bat", mtPos, "checker1")#going n
		Object_SpawnAt ("thug_with_bat", mtPos, "checker2")#going e
		Object_SpawnAt ("thug_with_bat", mtPos, "checker3")#going s
		Object_SpawnAt ("thug_with_bat", mtPos, "checker4")#going w
		replaceArrow ("arrowToChecker", "checker1")
		disable ("checker1")
		disable ("checker2")
		disable ("checker3")
		disable ("checker4")
		regTimer ("startFromBuildingPos", 1)
	else:
		saveBuildingDimensionsToFile ()
ff.getNextBuildingDimensions = getNextBuildingDimensions

def startFromBuildingPos (tevent = ""):
	global gsBuilding
	global gaCheckerPos
	global gaCheckerTargets
	global gaCheckerMovements
	gaCheckerMovements = [0,0,0,0]
	global giHeight
	giHeight = Get_ObjectPos ("checker1")[2] - Get_ObjectPos (gsBuilding)[2]
	if giHeight > MEASUREMENT_SNAP:
		mtPos = Get_ObjectPos (gsBuilding)
		miO = fSkXResolveAngle (fSkXGetDegreesFromRadians (Object_GetOrientation (gsBuilding) [1]))
		gaCheckerPos = [mtPos, mtPos, mtPos, mtPos]
		global gtN
		global gtE
		global gtS
		global gtW
		gtN = fSkXGetPointFromVector (mtPos, miO, 500)
		gtE = fSkXGetPointFromVector (mtPos, fSkXResolveAngle (miO + 90), 1000)
		gtS = fSkXGetPointFromVector (mtPos, fSkXResolveAngle (miO + 180), 1000)
		gtW = fSkXGetPointFromVector (mtPos, fSkXResolveAngle (miO + 270), 1000)
		gaCheckerTargets = [gtN, gtE, gtS, gtW]
		Trigger_Move ("checker1", "walk", gtN)
		Trigger_Move ("checker2", "walk", gtE)
		Trigger_Move ("checker3", "walk", gtS)
		Trigger_Move ("checker4", "walk", gtW)
		regTimer ("loopUntilAllStopped", 1)
	else:#no height? could be rubble
		retrieveBuildingDimensions ()
ff.startFromBuildingPos = startFromBuildingPos

def loopUntilAllStopped (tevent = ""):
	Mission_StatusText ("measuring...")
	global gsBuilding
	global gaCheckerPos
	global gaCheckerTargets
	global gaCheckerMovements
	mtBuilding = Get_ObjectPos (gsBuilding)
	for mi in range (0, 4):
		ms = "checker" + `(mi + 1)`
		mtPos = Get_ObjectPos (ms)
		if fSkXGetDistance (mtPos, gaCheckerPos [mi]) > MEASUREMENT_SNAP:
			gaCheckerMovements [mi] = 1
		else:
			gaCheckerMovements [mi] = 0
		gaCheckerPos [mi] = mtPos
	if 1 in gaCheckerMovements:
		regTimer ("loopUntilAllStopped", 1)
	else:
		for mi in range (0, 4):
			ms = "checker" + `(mi + 1)`
			msMarker = ms + "marker"
# 			miO = Object_GetOrientation (ms) [1]
			miH = fSkXGetAngle (mtBuilding, gaCheckerTargets [mi])
			mtPos = Get_ObjectPos (ms)
			miD = fSkXGetDistance (mtBuilding, mtPos)
			if abs (mtPos [2] - mtBuilding [2]) > MEASUREMENT_SNAP:
				mtPos = fSkXGetPointFromVector (mtBuilding, miH, miD + MEASUREMENT_SNAP)
			if Marker_Exists (msMarker):
				Marker_SetPos (msMarker, mtPos)
			else:
				Marker_Spawn (msMarker, mtPos)
		regTimer ("teleportToMarker", 1)
ff.loopUntilAllStopped = loopUntilAllStopped

def teleportToMarker (tevent = ""):
	for mi in range (0,4):
		ms = "checker" + `mi + 1`
		msMarker = ms + "marker"
		teleport (ms, msMarker)
	regTimer ("startFromTeleport", 3)
ff.teleportToMarker = teleportToMarker

def startFromTeleport (tevent = ""):
	global gsBuilding
	global gaCheckerTargets
	global gtN
	global gtE
	global gtS
	global gtW
	maTargets = [gtN, gtE, gtS, gtW]
	mtBuilding = Get_ObjectPos (gsBuilding)
	mbStillMeasuring = 0
	for mi in range (0,4):
		ms = "checker" + `mi + 1`
		mtPos = Get_ObjectPos (ms)
		if abs (mtPos [2] - mtBuilding [2]) > MEASUREMENT_SNAP:
			gaCheckerTargets [mi] = maTargets [mi]
			Trigger_Move (ms, "walk", gaCheckerTargets [mi])
			mbStillMeasuring = 1
		elif gaCheckerTargets [mi] == maTargets [mi]:
			miH = fSkXGetAngle (mtBuilding, gaCheckerTargets [mi])
			miD = fSkXGetDistance (mtBuilding, mtPos)
			mtTarget = fSkXGetPointFromVector (mtBuilding, miH, miD * 0.75)
			gaCheckerTargets [mi] = mtTarget
			Trigger_Move (ms, "walk", gaCheckerTargets [mi])
			mbStillMeasuring = 1
	if mbStillMeasuring:
		regTimer ("loopUntilAllStopped", 1)
	else:
		retrieveBuildingDimensions ()
ff.startFromTeleport = startFromTeleport

def walkBack (tevent = ""):
	Mission_Status ("measuring...")
	global gsBuilding
	global gaCheckerTargets

def retrieveBuildingDimensions (tevent = ""):
	global gsBuilding
	global giHeight
	Mission_StatusText ("done for " + Object_GetTemplate (gsBuilding) + ".")
	mtBuilding = Get_ObjectPos (gsBuilding)
	mt = Get_ObjectPos ("checker1")
	mt2 = Get_ObjectPos ("checker2")
	mt3 = Get_ObjectPos ("checker3")
	mt4 = Get_ObjectPos ("checker4")
# 	print gsBuilding + ": " + `mtBuilding`
# 	print "nw:" + `mt4` + ", " + `mt`
# 	print "se:" + `mt2` + ", " + `mt3`
	miWidth = abs (mt2 [0] - mt4 [0])
	miDepth = abs (mt [1] - mt3 [1])
	miHeight = giHeight - mtBuilding [2]
	md = {}
	md ["w"] = int (miWidth)
	md ["d"] = int (miDepth)
	md ["h"] = int (miHeight)
	md ["nw"] = [int (mt4 [0] - mtBuilding [0]), int (mt [1] - mtBuilding [1]), int (mtBuilding [2])]
	md ["se"] = [int (mt2 [0] - mtBuilding [0]), int (mt3 [1] - mtBuilding [1]), int (mtBuilding [2])]
# 	print to_break
	BUILDING_DIMENSIONS [Object_GetTemplate (gsBuilding)] = md
	Object_Destroy ("checker1")
	Object_Destroy ("checker2")
	Object_Destroy ("checker3")
	Object_Destroy ("checker4")
	regTimer ("getNextBuildingDimensions", 1)

def saveBuildingDimensionsToFile (tevent = ""):
	Mission_StatusText ("saving building dimensions to file.")
	removeArrow ("arrowToChecker")
	sysutils.updateFile('heightcheckdata.py', 'BUILDING_DIMENSIONS', BUILDING_DIMENSIONS)
	sysutils.updateFile ("heightcheckdata.py", "NEXT_BUILDING", NEXT_BUILDING + 6)
	Mission_Win ()
