#####################################################################################
#
# Module: Custom Heads
# Author: m25
# 
# Description: Will move a character to a custom template, so that 
#  custom heads will work
#
# Dependencies: 
#
# Version 3.052
#
#####################################################################################

#print "Loading m25customheads.py v.3.2.0 final; branch = m25 branch 1"

import shutil
import os
import chardata
import datfiles
import js
import m25template
import time
import string
import m25report

import m25customheadsdata
from m25customheadsdata import *

m25report.Report_SetChannel('CUSTOMHEADS', 1)

#printing the current time so that it sits in the ff.log file
#this will show when the user started the game (not the mission)
#and will allow us to clear out custom heads that were created before the current start time
#Note that custom heads get cached BY THE GAME, not by me, so this is a workaround
print 'CustomHeadCurrentTime %f' % time.time()

def CustomHeads_GetTemplate(char):
	if 'customheads_ignore' not in globals().keys():
		customheads_ignore = []
	template = m25template.GetComplexTemplate(char)
	m25report.Report('CUSTOMHEADS', 'trying to resolve %s'%(template))
	#see if we are ignoring this template, since it didn't qualify previously
	if template in customheads_ignore:
		m25report.Report('CUSTOMHEADS', '%s is being ignored.'%(template))
		return None
	#check to see if this is a hero file
	complex_val = js.Object_GetAttr(char, 'complex')
	if complex_val < 1000: 
		m25report.Report('CUSTOMHEADS', '%s not a hero file.'%(template))
		customheads_ignore.append(template)
		return None
	#check to see if the template still has the custom prefix, meaning it isn't identified properly
	if template[0:len('custom')] == 'custom':
		m25report.Report('CUSTOMHEADS', '%s is not identified correctly.'%(template))
		print 'Customheads ' + template + ' is not identified correctly'
		customheads_ignore.append(template)
		return None
		
	#if the head list is out of date, clear it and start anew
	if m25customheadsdata.M25_HEADUPDATETIME < CustomHeads_GetGameLoadTime():
		m25report.Report('CUSTOMHEADS', 'Clearing custom heads list.')
		m25customheadsdata.M25_HEADLIST = []
		
	#check to see if we've already done a substitution for this template
	for m in m25customheadsdata.M25_HEADLIST:
		try:
			if m[1] == template:
				m25report.Report('CUSTOMHEADS', '%s found substitute %s'%(template,repr(m[0])))
				return m[0]
		except:
			pass
	#find the mesh location
	location = GetMeshLocation(char)
	print location
	if location == None:
		m25report.Report('CUSTOMHEADS', '%s cannot find mesh.'%(template))
		customheads_ignore.append(template)
		return None
	#see if there is a head.nif
	head_location = os.path.join(location,'head.nif')
	if not os.path.isfile(head_location):
		m25report.Report('CUSTOMHEADS', '%s has no head.nif - using default.'%(template))
		head_location = os.path.join(datfiles.GetModPath(),'defaulthead','head.nif')
		#customheads_ignore.append(template)
		#return None
	#find an open substitution slot
	slot = CustomHeads_GetOpenSlot()
	if slot == None:
		m25report.Report('CUSTOMHEADS', 'unable to find an open slot!')
		customheads_ignore.append(template)
		return None
	#try to remove the destination if it exists
	destination = os.path.join(datfiles.GetModPath(),'art','library','characters',slot)
	try: shutil.rmtree(destination)
	except: 
		m25report.Report('CUSTOMHEADS', 'could not remove destination directory for %s'%(template))
		pass
	#now create a clean directory for the destination
	try: os.mkdir(destination)
	except:
		m25report.Report('CUSTOMHEADS', 'cannot create destination directory for %s'%(template))
		return None
	#copy the head.nif to the destination
	try:
		shutil.copyfile(head_location,os.path.join(destination,'head.nif'))
	except:
		m25report.Report('CUSTOMHEADS', 'copy of head.nif failed for %s'%(template))
		return None
	#find the skin
	cdata = chardata.GetCharacterData(char)
	skin = cdata['skin']
	if skin in ['', None]: skin = 'standard'
	m25report.Report('CUSTOMHEADS', '%s has skin %s'%(template,skin))
	#copy the skin into the standard skin directory at the destination
	try:
		os.mkdir(os.path.join(destination,'skins'))
		#os.mkdir(os.path.join(destination,'skins','standard'))
		shutil.copytree(os.path.join(location,'skins',skin), os.path.join(destination,'skins','standard'))
	except:
		m25report.Report('CUSTOMHEADS', 'copy of skin failed for %s'%(template))
		print 'Customheads ' + template + ' copy of skin failed'
		return None
	#add the template to the substitution list
	m25customheadsdata.M25_HEADLIST.append([slot,template])
	m25customheadsdata.M25_HEADUPDATETIME = time.time()
	CustomHeads_WriteHeadSlots()
	#return the successful substitution
	m25report.Report('CUSTOMHEADS', '%s has new substitution %s'%(template,repr(slot)))
	return slot
	

def CustomHeads_GetOpenSlot():
	for val in range(1,m25customheadsdata.M25_HEADSLOTS):
		found = 0
		for m in m25customheadsdata.M25_HEADLIST:
			try:
				if m[0] == ('m25head' + `val`):
					found = 1
			except:
				pass
		if not found:
			return 'm25head' + `val`
	return None
	

def CustomHeads_WriteHeadSlots():
	filename = os.path.join(datfiles.GetModPath(),'missions','scripts','m25customheadsdata.py')
	f = open(filename,'w')
	f.write('M25_HEADSLOTS = ' + `m25customheadsdata.M25_HEADSLOTS` + '\n')
	f.write('M25_HEADUPDATETIME = ' + `m25customheadsdata.M25_HEADUPDATETIME` + '\n')
	f.write('M25_HEADLIST = [\n')
	for m in m25customheadsdata.M25_HEADLIST:
		f.write('\t' + `m` + ',\n')
	f.write(']\n')
	f.close()
	
	
def GetMeshLocation(object):
	cdata = chardata.GetCharacterData(object)
	charNIF = cdata['NIF']
	possibles = [
		os.path.join(datfiles.GetFFPath(),'custom','art',charNIF),
		os.path.join(datfiles.GetWinTempPath(),'custom','art',charNIF),
		os.path.join(datfiles.GetModPath(),'art',charNIF),
		os.path.join(datfiles.GetFFPath(),'Data','art',charNIF),
	]
	for p in possibles:
		if os.path.isfile(p):
			return os.path.dirname(p)
	return None

	
def CustomHeads_GetGameLoadTime():
	if 'customheads_gametime' not in globals().keys():
		customheads_gametime = 0
    #check to see if we've already processed the game start time
	if customheads_gametime != 0:
		return customheads_gametime
	#go through the ff.log file and find the first occurance of Current Time
	fp = open('ff.log', 'r')
	line = fp.readline()
	found = 0
	while line != '':
		words = string.split(line)
		try:
			if words[0] == 'CustomHeadCurrentTime':
				fp.close()
				customheads_gametime = float(words[1])
				print 'Game load time is ' + `customheads_gametime`
				return float(words[1])
		except:
			pass
		line = fp.readline()
	fp.close()
	return 0