#####################################################################################
#
# Module: m25event
# Author: m25
# 
# Description: set up events within the ff system
#
#
#
#####################################################################################

#print "Loading m25event.py v.3.2.0 final; branch = m25 branch 1"

import ff
from ff import *
import js
import missionobjvar
import m25report
import types

EVENT_PERSISTENT = 1

m25report.Report_SetChannel('EVENT', 0)

#adapted from stumpy's code in mlogreader
class UserEvent:
    def __init__(self, eventType=None, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0, custom=None):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal
        self.custom = custom

def Event_RegisterEvent(eventType, function, persistent=0):
	if eventType == None:
		m25report.Report('ERROR', 'm25event.Event_RegisterEvent - %s tried to register with non-existent event type'%(function))
	else:
		if type(eventType) == types.StringType:
			name = '_event_%s'%(eventType)
		else:
			name = '_event_%s'%(repr(eventType))
		callbacks = missionobjvar.Mission_GetVar(name, None)
		if callbacks: callbacks.append([function,persistent])
		else: callbacks = [[function,persistent]]
		missionobjvar.Mission_SetVar(name, callbacks)
		m25report.Report('EVENT', 'm25event.Event_RegisterEvent - %s registered with event type %s'%(function,repr(eventType)))

#event should be of the UserEvent type
def Event_TriggerEvent(eventType, event=None):
	m25report.Report('EVENT', 'm25event.Event_TriggerEvent for event type %s.'%(repr(eventType)))
	if eventType == None:
		m25report.Report('ERROR', 'm25event.Event_TriggerEvent - tried to trigger null event type')
	else:
		if type(eventType) == types.StringType:
			name = '_event_%s'%(eventType)
		else:
			name = '_event_%s'%(repr(eventType))
		callbacks = missionobjvar.Mission_GetVar(name, None)
		newcallbacks = []
		change = 0
		if callbacks:
			for c in callbacks:
				#call the function with the event as a parameter
				m25report.Report('EVENT', 'm25event.Event_TriggerEvent - calling %s for event type %s.'%(c[0],repr(eventType)))
				try:
					eval('ff.%s(event)'%(c[0]))
				except:
					m25report.Report('ERROR', 'in Event_TriggerEvent - bad function %s'%(c[0]))
					raise
				#keep the function callback if it is persistent
				if c[1] == EVENT_PERSISTENT: 
					newcallbacks.append(c)
				else:
					change = 1
		if change:
			missionobjvar.Mission_SetVar(name, newcallbacks)

def Event_TriggerFirst(eventType, event=None):
	m25report.Report('EVENT', 'm25event.Event_TriggerFirst for event type %s.'%(repr(eventType)))
	if eventType == None:
		m25report.Report('ERROR', 'm25event.Event_TriggerEvent - tried to trigger null event type')
		return None
	else:
		if type(eventType) == types.StringType:
			name = '_event_%s'%(eventType)
		else:
			name = '_event_%s'%(repr(eventType))
		callbacks = missionobjvar.Mission_GetVar(name, None)
		newcallbacks = []
		value = None
		change = 0
		if callbacks:
			for c in callbacks:
				#call the function with the event as a parameter
				if not value:
					m25report.Report('EVENT', 'm25event.Event_TriggerEvent - calling %s for event type %s.'%(c[0],repr(eventType)))
					try:
						value = eval('ff.%s(event)'%(c[0]))
					except:
						m25report.Report('ERROR', 'in Event_TriggerFirst - bad function %s'%(c[0]))
						raise
					#value = eval('ff.%s(event)'%(c[0]))
				#keep the function callback if it is persistent
				if c[1] == EVENT_PERSISTENT: 
					newcallbacks.append(c)
				else: 
					change = 1
		if change:
			missionobjvar.Mission_SetVar(name, newcallbacks)
		return value
		
def Event_UnRegisterEvent(eventType, function):
	if eventType == None:
		m25report.Report('ERROR', 'm25event.Event_TriggerEvent - tried to trigger null event type')
	else:
		name = '_event_%s'%(repr(eventType))
		callbacks = missionobjvar.Mission_GetVar(name, None)
		newcallbacks = []
		if callbacks:
			for c in callbacks:
				if c[0] != function:
					newcallbacks.append(c)
				else:
					m25report.Report('EVENT', 'm25event.Event_UnRegisterEvent - removing %s from event type %s.'%(function,repr(eventType)))
		missionobjvar.Mission_SetVar(name, newcallbacks)
        
		

def SetTimer(time, function, string1='', string2='', user=0):
	js.Event_RegisterSink(js.EVENT_TIMER, function, string1, string2, time, user, 0)
	
def SetProximity(function, char, object='GAME_OBJ_HERO', distance=30, user=0):
	js.Event_RegisterSink(js.EVENT_OBJ_PROXIMITY_ENTER, function, char, object, distance, user, 0)
