#####################################################################
#
# Utility for writing to the dat files
# m25
#
# Thanks to stumpy for figuring out the dat file structures
# and for discovering how to convert from dat files to gdbm files
# and back again.
#
# 
#
#####################################################################

print 'm25gdbm version 1.4'

#1.4
# character names capped at 15 characters
# power names capped at 29 characters
# added Dat_CopyTree (can copy over an existing tree)
# melee area fix applied to transfered characters
# powers default to 'not for custom'
# fixed bug in writing ffx files
#1.3
# DAT_Write_Word now writes in the correct order
# Dat_Insert_Data now works correctly with variable sized records
# Dat_Read_String now reads all characters (must be deliminated by '\000' now)
# Added unknown byte to powers definition
# Added movement constant to characters definition
# Added unknown bytes (movement grid) to characters definition
#1.2
# now transfers custom AI to the built-in character if defined
# added dat2gdbm and gdbm2dat so that this file is compatible with older versions of ffx3
# only one backup is kept, to minimize the cluttering of directories with numerous backup files
#1.1
# added support for captions.dat and captions.txt - used by m25pstory.py

import gdbm
import string
import os
import datfiles
from datfiles import *

FFX_AVAILABLE = 0
try:
	import ffxcustom
	import ffxcustom2
	import ffxattribs
	FFX_AVAILABLE = 1
	print 'ffx found'
except:
	FFX_AVAILABLE = 0
	print 'no ffx found'

AI_AVAILABLE = 0
try:
	import m25aidata
	AI_AVAILABLE = 1
	print 'custom aidata found'
except:
	AI_AVAILABLE = 0
	print 'no custom aidata found'
	
#-------------------------------------------------------------------------
#
# Definitions
#
#-------------------------------------------------------------------------	

# there is a lot of 'junk' in the attributes.dat file
# I'm not sure if said junk has any use.
# clean attributes seem to work just fine.
attribute_definition = {
'length': 68,
'parameters': {
	'name': {
		'type': 'string',
		'length': 31,
		'location': 0,
		},
	'cost': {
		'type': 'int',
		'length': 3,
		'location': 32,
		},
	},
}

captions_definition = {
'length': 4,
'parameters': {
	'emotion': {
		'type': 'byte',     #0=neutral, 1=angry, 2=sad, 3=surprised, 4=shifty, 5=wisecracking, 6=fanatical, 7=pained, 8=happy
		'length': 1,
		'location': 0,
		},
	'caption': {
		'type': 'string',
		'length': 0,
		'location': 4,
		},
	},
}

captions_strings = ['neutral', 'angry', 'sad', 'surprised', 'shifty', 'wisecracking', 'fanatical', 'pained', 'happy']

char_definition = {
'length': 1078,
'parameters': {
	'header': {
		'type': 'constant',
		'value': 'HDR4\005\000\000\000sCDef\016\000\000\000!\004\000\000',
		'length': 21,
		'location': 0,
		},
	'charName': {
		'type': 'string',
		'length': 30,
		'location': 21,
		},
	'CSBase': {
		'type': 'string',
		'length': 255,
		'location': 51,
		},
	'strength': {
		'type': 'byte',
		'length': 1,
		'location': 306,
		},
	'speed': {
		'type': 'byte',
		'length': 1,
		'location': 307,
		},
	'endurance': {
		'type': 'byte',
		'length': 1,
		'location': 308,
		},
	'energy': {
		'type': 'byte',
		'length': 1,
		'location': 310,
		},
	'agility': {
		'type': 'byte',
		'length': 1,
		'location': 311,
		},
	'characterAttributes': {
		'type': ['string', 'string', 'string', 'string', 'string'],
		'length': 30,
		'location': 312,
		},
	'tier_a': {
		'type': ['string', 'string', 'string', 'string', 'string'],
		'length': 30,
		'location': 462,
		},
	'tier_b': {
		'type': ['string', 'string', 'string', 'string', 'string'],
		'length': 30,
		'location': 612,
		},
	'unknown1': {             #the real movement grid?
		'type': 'byte',
		'length': 1,
		'location': 762,
		},
	'unknown2': {
		'type': 'byte',
		'length': 1,
		'location': 763,
		},
	'unknown3': {
		'type': 'byte',
		'length': 1,
		'location': 764,
		},
	'unknown4': {
		'type': 'byte',
		'length': 1,
		'location': 765,
		},
	'unknown5': {
		'type': 'byte',
		'length': 1,
		'location': 766,
		},
	'unknown6': {
		'type': 'byte',
		'length': 1,
		'location': 767,
		},
	'unknown7': {
		'type': 'byte',
		'length': 1,
		'location': 768,
		},
	'unknown8': {
		'type': 'byte',
		'length': 1,
		'location': 769,
		},
	'unknown9': {
		'type': 'byte',
		'length': 1,
		'location': 770,
		},
	'unknown10': {
		'type': 'byte',
		'length': 1,
		'location': 771,
		},
	'VID': {
		'type': 'string',
		'length': 2,
		'location': 772,
		},
	'AI': {
		'type': 'string',
		'length': 30,
		'location': 775,
		},
	'tier_a_start': {
		'type': 'byte',
		'length': 1,
		'location': 807,
		},
	'tier_b_start': {
		'type': 'byte',
		'length': 1,
		'location': 808,
		},
	'attrib_start': {
		'type': 'byte',
		'length': 1,
		'location': 811,
		},
	'camp_only': {
		'type': 'byte',
		'length': 1,
		'location': 812,
		},
	'movegrid': {
		'type': 'constant',
		'length': 4,
		'location': 814,
		'value': '\377\177\377\177'   #cop_crooked has 11 and 0 as last two bytes (others are all the same)
		},
	'movementRadius': {
		'type': 'float',
		'length': 4,
		'location': 818,
		},
	'alterEgo': {
		'type': 'string',
		'length': 30,
		'location': 822,
		},
	},
}

fx_definition = {
'length': 439,
'parameters': {
	'header': {
		'type': 'constant',
		'value': 'HDR4\006\000\000\x00sFXDef\001\000\000\000\241\001\000\000',
		'length': 21,
		'location': 0,
		},
	'FXName': {
		'type': 'string',
		'length': 30,
		'location': 22,
		},
	'NIF': {
		'type': 'string',
		'length': 256,
		'location': 52,
		},
	'NodeStart': {
		'type': 'string',
		'length': 30,
		'location': 307,
		},
	'NodeCore': {
		'type': 'string',
		'length': 30,
		'location': 337,
		},	
	'NodeEnd': {
		'type': 'string',
		'length': 30,
		'location': 367,
		},
	'PowerType': {
		'type': 'byte',
		'length': 1,
		'location': 397,
		},
	'SubType': {
		'type': 'byte',
		'length': 1,
		'location': 401,
		},
	'NodesTrackRot': {
		'type': 'byte',
		'length': 1,
		'location': 405,
		},
	'NodeLaunch': {
		'type': 'string',
		'length': 30,
		'location': 409,
		},
	},
}

obj_definition = {
'length': 2284,
'parameters': {
	'templateName': {
		'type': 'string',
		'length': 30,
		'location': 0,
		},
	'NIF': {
		'type': 'string',
		'length': 256,
		'location': 30,
		},
	'class': {
		'type': 'string',
		'length': 2,
		'location': 285,
		},
	'attributes': {
		'type': {
			'length': 36,
			'parameters': {
				'name': {
					'type': 'string',
					'length': 32,
					'location': 0,
					},
				'value': {
					'type': 'float',
					'length': 4,
					'location': 32,
					},
				},
			},
		'entries': 15,
		'length':15*36,
		'location': 316,
		},
	'skin': {
		'type': 'string',
		'length': 256,
		'location': 886,
		},
	'explosion': {
		'type': 'string',
		'length': 256,
		'location': 1171,
		},
	'arg2': {
		'type': 'string',
		'length': 256,
		'location': 1456,
		},
	'arg3': {
		'type': 'string',
		'length': 256,
		'location': 1741,
		},
	'arg4': {
		'type': 'string',
		'length': 256,
		'location': 2026,
		},
	}
}

pow_definition = {
'length': 184,
'parameters': {
	'PowerName': {
		'type': 'string',
		'length': 30,
		'location': 0,
		},
	'PowerType': {
		'type': 'byte',
		'length': 1,
		'location': 132,
		},
	'animation': {
		'type': 'string',
		'length': 30,
		'location': 60,
		},
	'FX': {
		'type': 'string',
		'length': 30,
		'location': 90,
		},
	'EPCost': {
		'type': 'byte',
		'length': 1,
		'location': 121,
		},
	'Speed': {
		'type': 'byte',
		'length': 1,
		'location': 122,
		},
	'Accuracy': {
		'type': 'byte',
		'length': 1,
		'location': 123,
		},
	'Stun': {
		'type': 'byte',
		'length': 1,
		'location': 124,
		},
	'Knockback': {
		'type': 'byte',
		'length': 1,
		'location': 125,
		},
	'RangeMin': {
		'type': 'byte',
		'length': 1,
		'location': 126,
		},
	'RangeMax': {
		'type': 'byte',
		'length': 1,
		'location': 127,
		},
	'Radius': {
		'type': 'byte',
		'length': 1,
		'location': 128,
		},
	'SubType': {
		'type': 'byte',
		'length': 1,
		'location': 136,
		},
	'Magnitude': {
		'type': 'byte',
		'length': 1,
		'location': 140,
		},
	'DamageType': {
		'type': 'byte',
		'length': 1,
		'location': 144,
		},	
	'MaxInstances': {
		'type': 'byte',
		'length': 1,
		'location': 152,
		},
	'DefenceFlags': {
		'type': 'byte',
		'length': 1,
		'location': 156,
		},
	'AttackFlags': {
		'type': 'word',
		'length': 2,
		'location': 156,
		},
	'notForCustom': {
		'type': 'byte',
		'length': 1,
		'location': 158,
		},
	'SpecialType': {
		'type': 'byte',
		'length': 1,
		'location': 160,
		},
	'powerID': {
		'type': 'long',
		'length': 4,
		'location': 168,
		},
	'AttackModesBlocked': {
		'type': 'byte',
		'length': 2,
		'location': 172,
		},
	'DamageTypesBlocked': {
		'type': 'word',
		'length': 2,
		'location': 176,
		},
	'unknown1': {         #shows up in turret_raygun_shield only
		'type': 'byte',
		'length': 1,
		'location': 178,
	},
	'Success': {
		'type': 'byte',
		'length': 1,
		'location': 180,
		},
	'Duration': {
		'type': 'byte',
		'length': 1,
		'location': 181,
		},
	'BlockType': {
		'type': 'byte',
		'length': 1,
		'location': 182,
		},
	}
}

res_definition = {
'length': 296,
'parameters': {
	'name': {
		'type': 'string',
		'length': 30,
		'location': 0,
		},
	'dir': {
		'type': 'string',
		'length': 256,
		'location': 30,
		},
	},
}

string_definition = {
'length': 0,
'parameters': {
	'string': {
		'type': 'string',
		'length': 0,
		'location': 0,
		}
	}
}

dat_definitions = {
	'attributes.dat': attribute_definition,
	'captions.dat': captions_definition,
	'characters.dat': char_definition,
	'fx.dat': fx_definition,
	'objects.dat': obj_definition,
	'powers.dat': pow_definition,
	'res.dat': res_definition,
	'strings.dat': string_definition,
}


#-------------------------------------------------------------------------
#
# Convert from gdbm block to dictionary
#
#-------------------------------------------------------------------------	

def Dat_Read_Definition(data, definition):
	result = {}
	param = definition['parameters']
	for p in param.keys():
		tp = param[p]['type']
		if type(tp) == types.ListType: result[p] = Dat_Read_List(data, param[p]['type'], param[p]['location'], param[p]['length'])
		elif type(tp) == types.DictType: 
			temp = []
			loc = param[p]['location']
			length = param[p]['type']['length']
			for e in range(0,param[p]['entries']):
				temp.append(Dat_Read_Definition(data[loc:loc+length], param[p]['type']))
				loc = loc + length
			#result[p] = Dat_Read_Definition(data[param[p]['location']:param[p]['location']+param[p]['length']], param[p]['type'])
			result[p] = temp
		else:
			if tp == 'string': result[p] = Dat_Read_String(data, param[p]['location'])
			elif tp == 'byte': result[p] = Dat_Read_Byte(data, param[p]['location'])
			elif tp == 'float': result[p] = Dat_Read_Float(data, param[p]['location'])
			elif tp == 'int': result[p] = Dat_Read_Int(data, param[p]['location'])
			elif tp == 'word': result[p] = Dat_Read_Word(data, param[p]['location'])
			elif tp == 'long': result[p] = Dat_Read_Long(data, param[p]['location'])
			elif tp == 'constant': result[p] = definition['parameters'][p]['value']
			else: pass
	return result
	
def Dat_Read_String(data, loc):
	word, index = '', loc
	#while (index < len(data)) and (data[index] > '\x00') and (data[index] <= '\xC8'):
	while (index < len(data)) and (data[index] > '\x00'):
		word = word + data[index]
		index = index + 1
	return word

def Dat_Read_Byte(data, loc):
	return ord(data[loc])
	
def Dat_Read_Float(data, loc):
	return struct.unpack('f',data[loc:loc+4])[0]
	
def Dat_Read_Int(data, loc):
	return struct.unpack('h',data[loc:loc+2])[0]
	
def Dat_Read_Word(data, loc):
	return ( ord(data[loc+1]) << 8 ) + ord(data[loc])
	
def Dat_Read_Long(data, loc):
	return struct.unpack('l',data[loc:loc+4])[0]
	
def Dat_Read_List(data, param_list, location, length):
	result_list = []
	for index in range(0,len(param_list)):
		tp = param_list[index]
		loc = location + (index*length)
		result = None
		if tp == 'string': result = Dat_Read_String(data, loc)
		elif tp == 'byte': result = Dat_Read_Byte(data, loc)
		elif tp == 'float': result = Dat_Read_Float(data, loc)
		elif tp == 'int': result = Dat_Read_Int(data, loc)
		elif tp == 'word': result = Dat_Read_Word(data, loc)
		elif tp == 'long': result = Dat_Read_Lone(data, loc)
		else: pass
		if result: result_list.append(result)
	return result_list


#-------------------------------------------------------------------------
#
# Convert from dictionary to gdbm block
#
#-------------------------------------------------------------------------		
	
def Dat_Write_Definition(source, definition, data=''):
	if not data: data = Dat_Fill_Data(definition['length'])
	param = definition['parameters']
	for p in param.keys():
		tp = param[p]['type']
		if (tp != 'constant') and (not source.has_key(p)): continue  #make sure that we have an entry from the source
		
		if type(tp) == types.ListType: data = Dat_Write_List(data, source[p], param[p]['type'], param[p]['location'], param[p]['length'])
		elif type(tp) == types.DictType: 
			loc = param[p]['location']
			length = param[p]['type']['length']
			for e in range(0,param[p]['entries']):
				data = Dat_Insert_Data(data, loc, Dat_Write_Definition(source[p][e], param[p]['type']))
				loc = loc + length
			#data = Dat_Insert_Data(data, param[p]['location'], Dat_Write_Definition(source[p], param[p]['type']))
		else:
			if tp == 'string': data = Dat_Write_String(data, source[p], param[p]['location'])
			elif tp == 'byte': data = Dat_Write_Byte(data, source[p], param[p]['location'])
			elif tp == 'float': data = Dat_Write_Float(data, source[p], param[p]['location'])
			elif tp == 'int': data = Dat_Write_Int(data, source[p], param[p]['location'])
			elif tp == 'word': data = Dat_Write_Word(data, source[p], param[p]['location'])
			elif tp == 'long': data = Dat_Write_Long(data, source[p], param[p]['location'])
			elif tp == 'constant': data = Dat_Insert_Data(data, param[p]['location'], param[p]['value'])
			else: pass
	return data

def Dat_Fill_Data(size=1024):
	newline = ''
	for n in range(0,size):
		newline = newline + '\000'
	return newline
	
def Dat_Insert_Data(data, position, value):
	#certain dats are variable in length (captions.dat)
	if len(value)+position >= len(data):
		return data[0:position] + value + '\000'
	#others have a fixed length
	else:
		return data[0:position] + value + data[position+len(value):]
	
def Dat_Write_String(data, value, position):
	return Dat_Insert_Data(data, position, value)
	
def Dat_Write_Byte(data, value, position):
	return Dat_Insert_Data(data, position, chr(value))
	
def Dat_Write_Float(data, value, position):
	return Dat_Insert_Data(data, position, struct.pack('f', value))
	
def Dat_Write_Int(data, value, position):
	return Dat_Insert_Data(data, position, struct.pack('h', value))
	
def Dat_Write_Word(data, value, position):
	part2 = chr(value >> 8)
	part1 = chr(value - ((value>>8) << 8))
	return Dat_Insert_Data(data, position, '%s%s'%(part1,part2))
	#return Dat_Insert_Data(temp, position+1, part2)
	
def Dat_Write_Long(data, value, position):
	return Dat_Insert_Data(data, position, struct.pack('l',value))
	
def Dat_Write_List(data, value, param_list, location, length):
	for index in range(0,len(param_list)):
		if index < len(value):
			tp = param_list[index]
			loc = location + (index*length)
			if tp == 'string': data = Dat_Write_String(data, value[index], loc)
			elif tp == 'byte': data = Dat_Write_Byte(data, value[index], loc)
			elif tp == 'float': data = Dat_Write_Float(data, value[index], loc)
			elif tp == 'int': data = Dat_Write_Int(data, value[index], loc)
			elif tp == 'word': data = Dat_Write_Word(data, value[index], loc)
			elif tp == 'long': data = Dat_Write_Long(data, value[index], loc)
			else: pass
	return data
	
#-------------------------------------------------------------------------
#
# Setup GDBM
#
#-------------------------------------------------------------------------	

def Dat_SetupGDBM(filename):
	Dat_DAT2GDBM(filename)
	
def Dat_FinalizeDat(filename):
	Dat_Backup_File(filename)
	Dat_GDBM2DAT(filename)
	#os.remove('%s.gdbm'%filename)

def Dat_Read_Dat(filename, definition=None, raw = 0):
	if not definition and not raw:
		try:
			definition = dat_definitions[filename]
		except:
			print 'ERROR: (Dat_Read_Dat) %s has no associated definition'%filename
			return {}
	filename = Dat_Find_Dat(filename)
	if not filename:
		print 'ERROR: unable to find %s'%filename
		return {}
	Dat_SetupGDBM(filename)
	filedict = {}
	db = gdbm.open('%s.gdbm'%filename, 'r') 
	e = db.firstkey()
	while e:
		if raw: filedict[e[0:-1]] = db[e]
		else: filedict[e[0:-1]] = Dat_Read_Definition(db[e], definition)
		e = db.nextkey(e)
	db.close()
	return filedict	
	
def Dat_Write_Dat(filename, entry_list):
	filename = Dat_Find_Dat(filename)
	if not filename:
		print 'ERROR: unable to find %s'%filename
		return
	Dat_SetupGDBM(filename)
	db = gdbm.open('%s.gdbm'%filename, 'ws')
	db.firstkey()
	for e in entry_list:
		db[e[0]] = e[1]
	db.close()
	Dat_FinalizeDat(filename)
	
	
def Dat_Find_Dat(filename):
	f = os.path.join(datfiles.GetModPath(), filename)
	if os.path.isfile(f): return f
	f = os.path.join(datfiles.GetModPath(),'Lang','English',filename)
	if os.path.isfile(f): return f
	return None	
		
def Dat_Backup_File(filename):
	global dat_backups_made
	if 'dat_backups_made' not in globals().keys(): dat_backups_made = []
	#don't back up the same file twice in one session - reduce the number of spurious backups
	if filename in dat_backups_made: return  
	i = 1
	newname = '%s.bak'%filename
	#while os.path.isfile(newname):
	#	i = i + 1
	#	newname = '%s.bak%d'%(filename,i)
	Dat_CopyFileFullPath(filename, newname)
	dat_backups_made.append(filename)

def Dat_CopyFileFullPath(source, target):
	source = os.path.normpath(source)
	target = os.path.normpath(target)
	if os.path.isfile(source):
		try:
			os.makedirs(os.path.dirname(target))
		except:
			pass
		f = open(source, 'rb')
		data = f.read()
		f.close()
		f = open(target, 'wb')
		f.write(data)
		f.close()	
		return 1
	else:
		print 'Dat_CopyFileFullPath: source file %s does not exist.'%source
		return 0

def Dat_CopyTree(source, target):
	files = []
	try:
		files = os.listdir(source)
	except (IOError, os.error), why:
		print "Can't copy %s to %s: %s" % (source, target, str(why))
	for f in files:
		if os.path.isdir(os.path.join(source,f)): Dat_CopyTree(os.path.join(source,f),os.path.join(target,f))
		if os.path.isfile(os.path.join(source,f)): Dat_CopyFileFullPath(os.path.join(source,f),os.path.join(target,f))			
		
#-------------------------------------------------------------------------
#
# Convert h to dictionary
#
#-------------------------------------------------------------------------		

def Dat_Write_Template(template, directory):
	#create name
	name = Dat_Create_Name(template)
	#get ch
	cdata = Campaign_ReadHeroFile('%s.hero'%template)

	
def Dat_Create_Name(template):
	'''Take a template and convert it into a name for the dat files.  It is cut off at 15 characters for the animportrait.'''
	name = ''
	for t in string.lower(template):
		if t in [' ','\t']: name = name + '_'
		else: name = name + t
	return string.strip(name[0:15])
			
def Dat_Create_Power_Name(template, powername):
	name = '%s %s'%(Dat_Create_Name(template),powername)
	return string.strip(name[0:29])
	
def Dat_Convert_Pow(cdata):
	#prefix = Dat_Create_Name(cdata['charName'])
	powgdbm = []
	for p in cdata['powers'].keys():
		temp = cdata['powers'][p]
		temp['PowerName'] = Dat_Create_Power_Name(cdata['charName'],p)
		temp['notForCustom'] = 1
		#apply range fix to area melee powers
		if (temp.get('PowerType') == datfiles.PT_MELEE) and (temp.get('AttackFlags') & datfiles.PT_ATTACK_MELEE_AREA):
			temp['RangeMin'], temp['RangeMax'] = 1, 5
		powgdbm.append(['%s\000'%temp['PowerName'], Dat_Write_Definition(temp, pow_definition)])
	return powgdbm	
	
def Dat_Convert_Char(cdata):
	cdef = {}
	for c in char_definition['parameters'].keys():
		if cdata.has_key(c): cdef[c] = cdata[c]
	cdef['header'] = char_definition['parameters']['header']['value']
	cdef['attrib_start'] = cdata['activeAttributes']
	cdef['tier_a'], cdef['tier_b'] = Dat_Split_Pow(cdata)
	cdef['tier_a_start'] = len(cdef['tier_a'])
	cdef['tier_b_start'] = len(cdef['tier_b'])
	cdef['charName'] = Dat_Create_Name(cdata['charName'])
	cdef['camp_only'] = 1
	cdef['movementRadius'] = 1.0
	cdef['movegrid'] = char_definition['parameters']['movegrid']['value']
	return [['%s\000'%cdef['charName'], Dat_Write_Definition(cdef, char_definition)]]
	

def Dat_Split_Pow(cdata):
	#prefix = Dat_Create_Name(cdata['charName'])
	positions= cdata['powerPositions']
	#get max position
	maxpos = 0
	for p in positions.keys():
		if positions[p] > maxpos: maxpos = positions[p]
	split = maxpos/2
	#split them among two tiers
	a, b = [], []
	for p in positions.keys():
		if positions[p] <= split: a.append(Dat_Create_Power_Name(cdata['charName'],p))
		else: b.append(Dat_Create_Power_Name(cdata['charName'],p))
	return a,b

def Dat_Convert_Obj(cdata):
	odef = {}
	for c in obj_definition['parameters'].keys():
		if cdata.has_key(c): odef[c] = cdata[c]
	odef['templateName'] = Dat_Create_Name(cdata['charName'])
	#attributes
	#copy all attributes over to a format the writer understands
	# be sure to assign a new complex attribute to differentiate it from the hero file
	attr = cdata['objectAttributes']
	alist = []
	complex_value = Dat_Find_Object_Attr(odef['templateName'], 'complex')
	if not complex_value:
		complex_value = Dat_Find_Highest_Complex()+1
	complex_found = 0
	for a in attr:
		temp = {}
		temp['name'] = a
		if a == 'complex': 
			temp['value'] = complex_value
			complex_found = 1
		else: temp['value'] = cdata[a]
		alist.append(temp)
	if not complex_found:
		alist.append({'name': 'complex', 'value': complex_value,})
	temp = {}
	temp['name'], temp['value'] = '', 0.0
	while len(alist) < 15:
		alist.append(temp)
	odef['attributes'] = alist
	return [['%s\000'%odef['templateName'], Dat_Write_Definition(odef, obj_definition)]]	
	
def Dat_Find_Object_Attr(name, attr):
	d = Dat_Read_Dat('objects.dat')
	if d.has_key(name):
		for a in d[name]['attributes']:
			if a['name'] == 'complex': return a['value']
	return None	
	
def Dat_Convert_Res(cdata):
	rdef = {}
	rdef['name'] = 'animportrait_%s'%Dat_Create_Name(cdata['charName'])
	rdef['dir'] = os.path.join(os.path.dirname(cdata['NIF']),'head.nif')
	return [['%s\000'%rdef['name'], Dat_Write_Definition(rdef, res_definition)]]
	
def Dat_Convert_String(cdata):
	existing_strings = Dat_Read_Dat('strings.dat')
	name = Dat_Create_Name(cdata['charName'])
	stringlist = []
	stringlist.append(['%s_01\000'%name, Dat_Write_Definition({'string': cdata['charName'],}, string_definition)])
	stringlist.append(['%s_desc_01\000'%name, Dat_Write_Definition({'string': '%s short description'%cdata['charName'],}, string_definition)])
	stringlist.append(['%s_desc2_01\000'%name, Dat_Write_Definition({'string': '%s bio'%cdata['charName'],}, string_definition)])
	for p in cdata['powers'].keys():
		stringlist.append(['%s_01\000'%Dat_Create_Power_Name(cdata['charName'],p), Dat_Write_Definition({'string': p,}, string_definition)])
		stringlist.append(['%s_desc_01\000'%Dat_Create_Power_Name(cdata['charName'],p), Dat_Write_Definition({'string': '%s description'%p,}, string_definition)])
	revisedlist = []
	for s in stringlist:
		if existing_strings.has_key(s[0][0:-1]): 
			print 'String: %s already in strings.dat - skipping.'%s[0][0:-1]
			pass  #don't overwrite existing strings
		else: revisedlist.append(s)
	return stringlist

def Dat_Convert_All(cdata):
	Dat_Write_Dat('powers.dat', Dat_Convert_Pow(cdata))
	Dat_Write_Dat('characters.dat', Dat_Convert_Char(cdata))
	Dat_Write_Dat('objects.dat', Dat_Convert_Obj(cdata))
	Dat_Write_Dat('res.dat', Dat_Convert_Res(cdata))
	Dat_Write_Dat('strings.dat', Dat_Convert_String(cdata))
	Dat_Strings_To_TextFile('strings')
	if inGame():
		datfiles.Campaign_MakeHeroIndex(ModPath='', indexFileName='', ForceRead=1)
		Dat_MoveMesh(cdata)
		if FFX_AVAILABLE:
			Dat_Convert_Table(cdata)
		if AI_AVAILABLE:
			Dat_TransferAI(cdata)

def Dat_Find_Highest_Complex():
	highest = 0
	objects = Dat_Read_Dat('objects.dat')
	for c in objects.keys():
		for a in objects[c]['attributes']:
			if a['name'] == 'complex':
				if a['value'] > highest: highest = a['value']
	return highest
		
	
#-------------------------------------------------------------------------
#
# Move Directory
#
#-------------------------------------------------------------------------	
	
def Dat_GetMeshLocation(cdata):
	#cdata = chardata.GetCharacterData(object)
	charNIF = cdata['NIF']
	possibles = [
		os.path.join(datfiles.GetFFPath(),'custom','art',charNIF),
		os.path.join(datfiles.GetWinTempPath(),'custom','art',charNIF),
		os.path.join(datfiles.GetModPath(),'art',charNIF),
		os.path.join(datfiles.GetFFPath(),'Data','art',charNIF),
	]
	for p in possibles:
		if os.path.isfile(p):
			return os.path.dirname(p)
	return None	
	
def Dat_MoveMesh(cdata):
	start = Dat_GetMeshLocation(cdata)
	if not start: 
		print 'ERROR: cannot find mesh at %s - aborting mesh transfer.'%cdata['NIF']
		return
	target = os.path.dirname(os.path.join(datfiles.GetModPath(),'art',cdata['NIF']))
	if start == target:
		print 'Source Mesh %s is already in the correct directory -  aborting mesh transfer.'%cdata['NIF']
		return
	head_location = os.path.join(start,'head.nif')
	if not os.path.isfile(head_location):
		head_location = os.path.join(datfiles.GetModPath(),'defaulthead','head.nif')
	#try: shutil.rmtree(target)
	#except: pass 
	#shutil.copytree(start, target)
	#shutil.copyfile(head_location,os.path.join(target,'head.nif'))
	Dat_CopyTree(start,target)
	Dat_CopyFileFullPath(head_location, os.path.join(target,'head.nif'))
	

#-------------------------------------------------------------------------
#
# Copy Table Entries
#
#-------------------------------------------------------------------------	

def Dat_Convert_Table(cdata):
	#import ffxcustom
	#import ffxcustom2
	name, newname = cdata['charName'], Dat_Create_Name(cdata['charName'])
	#don't need to convert the table entries if the name remains the same
	if name == newname: return
	#otherwise...
	Dat_Add_FFX_Entries(name, newname, 'ffxcustom')
	Dat_Add_FFX_Entries(name, newname, 'ffxcustom2')
	Dat_Save_FFXCustom('ffxcustom', Dat_Get_FFXCustomOrder())
	Dat_Save_FFXCustom('ffxcustom2')

#the ffx control centre doesn't work unless ffxcustom has a very specific look to it
#this function tries to replicate that
def Dat_Get_FFXCustomOrder():
	order = ['FFX_RULES_STASIS', 'FFX_RULES_FROZENDAMAGE', 'FFX_RULES_CLONECOMMANDS', 'FFX_RULES_NOPCAI', 'FFX_RULES_EXTRASTUN', 'FFX_RULES_PUSHEP', '', 'FFX_RULES_NOMINRANGE', '']
	for a in ffxattribs.FFX_ATTRIBUTES:
		order.append(a[1])
	order.append('FFX_CARRIERS')
	return order

#FFX_RULES_NOMINRANGE=1	
	
def Dat_Add_FFX_Entries(name, newname, file):
	entries = dir(eval(file))
	for e in entries:
		if e[0] != '_':
			table = eval('%s.%s'%(file,e))
			if type(table) == types.ListType:
				#clean existing entries first
				temp = table[0:]
				for t in temp:
					if t[0] == newname:
						table.remove(t)
				for t in table:
					if t[0] == name:
						#simply copy all info, but use the new name as the first element of the list
						table.append([newname] + t[1:])

def Dat_Save_FFXCustom(file, order=None):
	newlist = []
	#get all of the tables and their values
	if not order:
		entries = dir(eval(file))
	else: entries = order
	for e in entries:
		if e == '': newlist.append(['',''])
		else:
			if e[0] != '_': 
				try:
					table = eval('%s.%s'%(file,e))
					newlist.append([e,table])
				except:
					print '%s does not exist in %s'%(e,file)
		
	#back up the file in case something goes wrong
	filename = os.path.join(datfiles.GetModPath(),'Missions','Scripts',file+'.py')
	Dat_Backup_File(filename)
	#write out all of the tables
	f = open(filename,'w')
	for n in newlist:
		if type(n[1]) == types.ListType:
			f.write('%s=[\n'%n[0])
			for entry in n[1]:
				#we have to write this in a specific way for the ffx control centre to function
				#f.write('%s,\n'%repr(e))  #doesn't work with ffxcc
				f.write('[')
				comma = ''
				for p in entry:
					if type(p) == types.StringType: f.write('%s"%s"'%(comma,p))
					elif type(p) == types.IntType: f.write('%s%d'%(comma,p))
					else: f.write('%s%s'%(comma,repr(p)))
					comma = ','
				f.write('],\n')
			f.write(']\n\n')
		elif n[1] == '':
			f.write('\n')
		else:
			f.write('%s=%s\n'%(n[0],repr(n[1])))
	f.close()	
	
def Dat_Create_Carriers(cdata):
	prefix = Dat_Create_Name(cdata['charName'])
	carriers = []
	table = ffxcustom.FFX_CARRIERS
	for p in cdata['powers'].keys():
		temp = cdata['powers'][p]
		for t in table:
			#if p has type t[3] add a new entry to the table
			if success:
				table.append(t[0:2]+[p])
			

	
#-------------------------------------------------------------------------
#
# Write to comma delimited files
#
#-------------------------------------------------------------------------		
	
def Dat_Strings_To_TextFile(prefix):
	predata = Dat_Read_Dat(prefix+'.dat')
	data = {}
	for key in predata.keys():
		data[key] = predata[key]['string']
	if data:
		Dat_Write_StringsTxt(prefix+'.txt', data)
	
def Dat_Write_StringsTxt(file, data):
	filename = Dat_Find_Dat(file)
	Dat_Backup_File(filename)
	f = open(filename,'w')
	keys = data.keys()
	keys.sort()
	for d in keys:
		f.write('%s, %s\n'%(d,data[d]))		
	f.close()
	
def Dat_Write_CaptionsTxt():
	data = Dat_Read_Dat('captions.dat')
	filename = Dat_Find_Dat('captions.txt')
	Dat_Backup_File(filename)
	f = open(filename, 'w')
	keys = data.keys()
	keys.sort()
	for d in keys:
		f.write('%s, %s, %s\n'%(d,captions_strings[data[d]['emotion']],data[d]['caption']))
	f.close()

	
# def Dat_ReadCommaFile(filename):
# 	stringlist = {}
# 	f = open(filename, 'r')
# 	line = f.readline()
# 	while line:
# 		success, tag, description = m25string.SplitAroundPhrase(line,',')
# 		if success:
# 			stringlist[tag] = description
# 		line = f.readline()
# 	f.close()
# 	return stringlist

	
#-------------------------------------------------------------------------
#
# Copy AI
#
#-------------------------------------------------------------------------	

def Dat_TransferAI(cdata):
	name = cdata['charName']
	ai = Dat_FindAI(name)
	if ai:
		newai = Dat_TransferPowerList(cdata, ai)
		all_ai = Dat_AddAI(Dat_Create_Name(cdata['charName']), newai)
		Dat_WriteAIFile(all_ai)
	else:
		print 'Dat_TransferAI: %s has no custom AI.'%name
	
def Dat_AddAI(name, ai):
	all_ai = m25aidata.AIExtensions
	for a in all_ai:
		if a[0] == ai[0]:
			all_ai.remove(a)
	all_ai.append(ai)
	all_ai.sort()
	return all_ai
			
		
def Dat_FindAI(template):	
	try:
		filename = os.path.join(datfiles.GetHeroFilePath(),template+'.m25ai')
		f = open(filename,'r')
		cust = f.read()
		f.close()
		ai = eval(cust)
		return ai
	except:
		return None

	
def Dat_TransferPowerList(cdata, translist):
	newlist = []
	powlist = cdata['powers'].keys()
	for t in translist:
		if type(t) == types.ListType:
			newlist.append(Dat_TransferPowerList(cdata, t))
		elif type(t) == types.StringType:
			newlist.append(Dat_TransferPowerString(cdata, t))
		else:
			newlist.append(t)
	return newlist
		
def Dat_TransferPowerString(cdata, transstring):
	newstring = transstring
	#prefix = Dat_Create_Name(cdata['charName'])
	for p in cdata['powers'].keys():
		newstring = Dat_ReplacePowerLine(newstring, p, Dat_Create_Power_Name(cdata['charName'],p))
	return newstring

def Dat_ReplacePowerLine(line, powername, replacement):
	start = -1
	current = 0
	for i in line:
		if line[current:current+len(powername)] == powername:
			if (line[current-1] in ['"',"'"]) and (line[current+len(powername)] in ['"',"'",'\\']) and (not (line[current-2] in ['='])):
				start = current
		current = current + 1
	if start >= 0:
		newline = line[0:start-1] + `replacement` + line[start+len(powername)+1:]
		return newline
	return line

def Dat_WriteAIFile(all_ai):
	#all_ai = m25aidata.AIExtensions
	scriptpath = os.path.join(datfiles.GetModPath(),'Missions','Scripts')
	print scriptpath
	fname = os.path.join(scriptpath,'m25aidata.py')
	Dat_Backup_File(fname)
	aif = open(fname,'w')
	aif.write('AIExtensions=[\n')
	for a in all_ai:
		Dat_WriteAI(aif,a)
	aif.write('\n\n#end of file - there should be a close bracket here\n')
	aif.write(']\n')
	aif.close()

def Dat_WriteAI(file, ai):
	file.write('\n\n[' + `ai[0]` + ',\n')
	for section in ai[1:]:
		try:
			if section[0] == 'timer':
				file.write('\t["timer", ' + `section[1]` + ',\n')
				for tac in section[2:]:
					file.write('\t\t' + `tac` + ',\n')
				file.write('\t],\n')
			else:
				file.write('\t[' + `section[0]` + ',\n')
				for targ in section[1:]:
					file.write('\t\t' + `targ` + ',\n')
				file.write('\t],\n')
		except:
			pass
	file.write('],\n')
	
	
	
#from stumpy's datfiles.py (for backwards compatibility)	
	
# change the magic number so that a DAT file can be read in as a GDBM file
def Dat_DAT2GDBM(FileName):
	GDBM_MAGIC_NUM = '\xCE\x9A\x57\x13'
	gdbmFileName = FileName+'.gdbm'
	f = open(FileName,'rb')
	dat = f.read()
	f.close()
	f = open(gdbmFileName,'wb')
	gdbmdat = GDBM_MAGIC_NUM+dat[4:]
	f.write(gdbmdat)
	f.close()

# change the magic number so that a GDBM file can be read in as a DAT file
# NB the FileName is of the original DAT file.  E.g. 'power.dat'
def Dat_GDBM2DAT(FileName):
	DAT_MAGIC_NUM = '\xAD\xDB\xDE\xFA'
	gdbmFileName = FileName+'.gdbm'
	f = open(gdbmFileName,'rb')
	dat = f.read()
	f.close()
	f = open(FileName,'wb')
	datdat = DAT_MAGIC_NUM+dat[4:]
	f.write(datdat)
	f.close()
	
	
#-------------------------------------------------------------------------
#
# Replicate
#
#-------------------------------------------------------------------------

def Dat_Replicate(template, num=24):
	""" Take a character template and produce num-1 copies of it in the dat files.
		Used with m25spawn.py, it provides a space expensive workaround for ghosting meshes.
	"""
	objects = Dat_Read_Dat('objects.dat')
	chars = Dat_Read_Dat('characters.dat')
	res = Dat_Read_Dat('res.dat')
	objdata = objects[template]
	chardata = chars[template]
	resdata = res['animportrait_%s'%template]
	directory, nif = os.path.split(objdata['NIF'])
	objadd, charadd, resadd = [], [], []
	for i in range(2,num):
		newname = '%s%d'%(template,i)
		newdir = '%s%d'%(directory,i)
		chardata['charName'] = newname
		chardata['AI'] = '* None'
		charadd.append(['%s\000'%newname, Dat_Write_Definition(chardata, char_definition)])
		objdata['NIF'] = os.path.join(newdir, nif)
		objdata['templateName'] = newname
		objadd.append(['%s\000'%newname, Dat_Write_Definition(objdata, obj_definition)])
		resdata['name'] = 'animportrait_%s'%newname
		resdata['dir'] = os.path.join(newdir,'head.nif')
		resadd.append(['%s\000'%resdata['name'], Dat_Write_Definition(resdata, res_definition)])

	Dat_Write_Dat('characters.dat', charadd)
	Dat_Write_Dat('objects.dat', objadd)
	Dat_Write_Dat('res.dat', resadd)
	

	
#-------------------------------------------------------------------------
#
# Test Harness
#
#-------------------------------------------------------------------------

def Dat_Test():
	for file in dat_definitions.keys():
		Dat_Test_Dat(file)


def Dat_Test_Dat(dat_name):
	converted = Dat_Read_Dat(dat_name)
	raw = Dat_Read_Dat(dat_name, raw=1)
	check_powers = converted.keys()
	print '--- %s ---'%dat_name
	for cp in check_powers:
		rewrite = Dat_Write_Definition(converted[cp],dat_definitions[dat_name])
		if rewrite == raw[cp]:
			#print '%s passed'%(cp)
			pass
		else:
			print '%s failed'%(cp)
			Dat_CompareStrings(rewrite, raw[cp])
			
def Dat_CompareStrings(s1, s2):
	if len(s1) != len(s2):
		print 's1 has length %d, while s2 has length %d'%(len(s1), len(s2))
		print s1
		print s2
		return
	for i in range(0,len(s1)):
		if s1[i] != s2[i]:
			print '%d\t%d\t%d'%(i,ord(s1[i]),ord(s2[i]))
			
# import string
# def Dat_G(s):
# 	data = Dat_Read_Dat('objects.dat')
# 	for d in data.keys():
# 		if string.find(data[d]['NIF'],s) != -1:
# 			print d 
# 			print data[d]['NIF']

