#####################################################################################
#
# Module: m25generateai
# Author: m25
# 
# Description: automatically generate a custom ai for a character
#
#
# Version 3.052
#
#####################################################################################

#print "Loading m25generateai.py v.3.2.0 final; branch = m25 branch 1"

import ff

import datfiles
import chardata
import m25aidata
import js

import m25report
import m25template
import types
import m25getset
import m25event
import m25energy

import os
import struct

from m25aigeneratedata import *

import ffxcustom
import ffxmulti

GENAI_CHANNEL='GENERATE_AI'

genlist = []
char_attr = []

def GenerateAI(char):
	if m25template.GetComplexTemplate(char)[:7] == 'custom_':
		js.Mission_StatusText('WARNING: ' + char + ' has not been branded. Quit and run FFX control centre.')

	global genlist
	global char_attr
	char_attr = []
	if (not js.Object_GetAttr(char,'physical')) or (js.Object_GetClass(char)&js.OC_CIVILIAN):
		return None
	cdata = chardata.GetCharacterData(char)
	if cdata == {}: return None
	m25report.Report_SetChannel(GENAI_CHANNEL,1)
	ai = None
	timer = []
	targetenemy = []
	targets = GenerateAI_TargetEnemy(char)
	if targets != []:
		targetenemy = CombineLists(['target enemy'], targets)
	tactics = GenerateAI_TimerTactics(char)
	if tactics != []:
		timer = CombineLists(['timer', 'auto'], tactics)
	ai = [m25template.GetComplexTemplate(char),targetenemy,timer]
	if 'absorber' in char_attr:
		ai.append(['target any', 'object'])
	er = GetReserveLevel(char)
	if er > 0:
		ai.append(['energy reserve', datfiles.PT_EP_COST_NAMES[er]])
	genlist.append(ai[0])
	#PrintAI(ai)
	return ai

def AddCharAttr(attr):
	global char_attr
	char_attr.append(attr)
		
def SetReserveLevel(char, val):
	old = m25getset.GetObjectVar(char, 'reserve_level', default=0)
	if val > old:
		m25getset.SetObjectVar(char, 'reserve_level', val)	
	
def GetReserveLevel(char):		
	return m25getset.GetObjectVar(char, 'reserve_level', default=0, integer=1)
	
targetenemytype = {datfiles.PT_SPECIAL_AFTERBURNER : [],
	datfiles.PT_SPECIAL_ALTERATION : ['not blind','not panicked','not hypnotized','not staticked'],
	datfiles.PT_SPECIAL_ALTER_GENETIC_STRUCTURE : [],
	datfiles.PT_SPECIAL_ANIMATE_DEAD : ['dead','koed','not alive','lowerhealthpct 15'],
	datfiles.PT_SPECIAL_ANIMATE_DEAD_MINIONS :['dead','koed','not alive','lowerhealthpct 15'],
	datfiles.PT_SPECIAL_BEZERKER :[],
	datfiles.PT_SPECIAL_BUOYANCY : [],
	datfiles.PT_SPECIAL_BURROW : [],
	datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION : [],
	datfiles.PT_SPECIAL_CLOAK : [],
	datfiles.PT_SPECIAL_CLONE_SELF : [],
	datfiles.PT_SPECIAL_CLONE_TARGET : ['strongest'],
	datfiles.PT_SPECIAL_DEEP_SENSOR : [],
	datfiles.PT_SPECIAL_DENSITY_MAXIMIZE : [],
	datfiles.PT_SPECIAL_DIMENSION_SHIFT : ['not exiled'],
	datfiles.PT_SPECIAL_DISPLACE_IMAGE : [],
	datfiles.PT_SPECIAL_DISRUPT_OBJECT : [],
	datfiles.PT_SPECIAL_DOPPELGANGER : [],
	datfiles.PT_SPECIAL_ELECTRIFY_SELF : ['metal'],
	datfiles.PT_SPECIAL_EMPATHY : [],
	datfiles.PT_SPECIAL_ENERGIZE : [],
	datfiles.PT_SPECIAL_ENERGIZE_AREA : [],
	datfiles.PT_SPECIAL_EP_DRAW : ['higherenergy 50'],
	datfiles.PT_SPECIAL_EP_TRANSFER : [],
	datfiles.PT_SPECIAL_HP_BALANCE : ['higherhealththan'],
	datfiles.PT_SPECIAL_IGNITE : [],
	datfiles.PT_SPECIAL_IMAGE_PROJECTOR : [],
	datfiles.PT_SPECIAL_LAWORDER : [],
	datfiles.PT_SPECIAL_LIFE_GIFT : [],
	datfiles.PT_SPECIAL_LIVING_FIRES : [],
	datfiles.PT_SPECIAL_MOLECULAR_EXCITE : [],
	datfiles.PT_SPECIAL_MORPH : [],
	datfiles.PT_SPECIAL_NONE : [],
	datfiles.PT_SPECIAL_NORMALISE : [],
	datfiles.PT_SPECIAL_PATRIOT_CHARGE : ['highesthealth'],
	datfiles.PT_SPECIAL_POSSESSION : [],
	datfiles.PT_SPECIAL_POWER_NULLIFICATION : [],
	datfiles.PT_SPECIAL_RALLY : [],
	datfiles.PT_SPECIAL_REMOVE_DEFENCE : [],
	datfiles.PT_SPECIAL_SACRIFICE : [],
	datfiles.PT_SPECIAL_SONIC_BOOM : [],
	datfiles.PT_SPECIAL_SPEEDING_BULLET : [],
	datfiles.PT_SPECIAL_SPRINT : ['fastest'],
	datfiles.PT_SPECIAL_TELEPORT_SELF : [],
	datfiles.PT_SPECIAL_TELEPORT_SWAP : [],
	datfiles.PT_SPECIAL_TUMBLE : [],
	datfiles.PT_SPECIAL_TUNNEL : [],
	datfiles.PT_SPECIAL_UNCLOAK : []}	
	
targetenemydamagetype = {datfiles.PT_DAMAGE_ACID : ['weakness'],
	datfiles.PT_DAMAGE_ACID_BURN : ['not acid_burnt'],
	datfiles.PT_DAMAGE_BLIND : ['not blind'],
	datfiles.PT_DAMAGE_COLD : ['weakness'],
	datfiles.PT_DAMAGE_CRUSH : ['weakness'],
	datfiles.PT_DAMAGE_ELECTRICAL : ['weakness'],
	datfiles.PT_DAMAGE_ENERGY : ['weakness'],
	datfiles.PT_DAMAGE_FIRE : ['weakness'],
	datfiles.PT_DAMAGE_FREEZE : ['not frozen'],
	datfiles.PT_DAMAGE_HEX : ['not hexed'],
	datfiles.PT_DAMAGE_HYPNOSIS : ['not hypnotized'],
	datfiles.PT_DAMAGE_IRRADIATE : ['not irradiated'],
	datfiles.PT_DAMAGE_MENTAL_BLANK : ['not blanked'],
	datfiles.PT_DAMAGE_OBJECTIFY : ['not objectified'],
	datfiles.PT_DAMAGE_PANIC : ['not panicked'],
	datfiles.PT_DAMAGE_PIERCE : ['weakness'],
	datfiles.PT_DAMAGE_POLYMORPH : [],
	datfiles.PT_DAMAGE_RADIATION : ['weakness'],
	datfiles.PT_DAMAGE_RAGE : ['not enraged'],
	datfiles.PT_DAMAGE_SPECIAL : [],
	datfiles.PT_DAMAGE_SPEED_DECREASE : [],
	datfiles.PT_DAMAGE_SPEED_INCREASE : [],
	datfiles.PT_DAMAGE_STASIS : ['not staticked'],
	datfiles.PT_DAMAGE_TOADIFY : ['not toadified'],
	datfiles.PT_DAMAGE_WATER : [],
	datfiles.PT_DAMAGE_WEAKEN : []}	
	
targetenemyffxtype = {'absorb': ['metal or stone'],
'castle': [],
'cloak' : [],
'crystallise': ['not caged'],
'firecage': ['not caged'],
'gravityincrease': ['greaterstrength 0'],
'gravityreverse': [],
'heal': ['lowerhealthpct 50'],
'icepatch': ['not ice_control', 'not acrobatic'],
'leadership': [], 
'lifedrain': [],
'magnetic pull': ['metal'],
'magnetic throw': ['metal'],
'magnetise': [],
'mental damage': [],
'mimic': [],
'staticcage': ['not caged'],
'strengthdrain': ['greaterstrength 0'],
'phase': ['not displaced'], 
'pull': [],
'selfheal': [],
'shrink': ['not shrunk'],
'speed boost': [],
'strobe': [],
'summon': [],
'taunt': [],
'teleport': [],
'throw': [],
'transmute': [],
'vertigo': ['not vertigo'],
'weaken': ['not ffx_weakened'],
'zerog': ['not flying'],
}		
	
gendmgconversion = {datfiles.PT_DAMAGE_ACID : 'acid',
		datfiles.PT_DAMAGE_ACID_BURN : 'acid',
		datfiles.PT_DAMAGE_BLIND : 'mental',
		datfiles.PT_DAMAGE_COLD : 'cold',
		datfiles.PT_DAMAGE_CRUSH : 'crushing',
		datfiles.PT_DAMAGE_ELECTRICAL : 'electrical',
		datfiles.PT_DAMAGE_ENERGY : 'energy',
		datfiles.PT_DAMAGE_FIRE : 'fire',
		datfiles.PT_DAMAGE_FREEZE : 'cold',
		datfiles.PT_DAMAGE_HEX : 'mystical',
		datfiles.PT_DAMAGE_HYPNOSIS : 'mental',
		datfiles.PT_DAMAGE_IRRADIATE : 'radiation',
		datfiles.PT_DAMAGE_MENTAL_BLANK : 'mental',
		datfiles.PT_DAMAGE_OBJECTIFY : 'mystical',
		datfiles.PT_DAMAGE_PANIC : 'mental',
		datfiles.PT_DAMAGE_PIERCE : 'piercing',
		datfiles.PT_DAMAGE_POLYMORPH : 'mystical',
		datfiles.PT_DAMAGE_RADIATION : 'radiation',
		datfiles.PT_DAMAGE_RAGE : 'mental',
		datfiles.PT_DAMAGE_SPECIAL : 'none',
		datfiles.PT_DAMAGE_SPEED_DECREASE : 'none',
		datfiles.PT_DAMAGE_SPEED_INCREASE : 'none',
		datfiles.PT_DAMAGE_STASIS : 'mystical',
		datfiles.PT_DAMAGE_TOADIFY : 'mystical',
		datfiles.PT_DAMAGE_WATER : 'water',
		datfiles.PT_DAMAGE_WEAKEN : 'none'}
		
gendmgclassconversion = {datfiles.PT_MELEE : 'melee',
		datfiles.PT_RANGED : 'ranged',
		datfiles.PT_AREA : 'area',
		datfiles.PT_DIRECT : 'direct',
		datfiles.PT_ACTIVE_DEFENCE : 'none',
		datfiles.PT_PASSIVE_DEFENCE : 'none',
		datfiles.PT_SPECIAL : 'special'}			
	
def GenerateAI_TargetEnemy(char):
	target = []
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power != {} and not (power['PowerType'] in [datfiles.PT_PASSIVE_DEFENCE,datfiles.PT_ACTIVE_DEFENCE]):
			sp = targetenemytype[power['SpecialType']]
			if power['SpecialType'] == datfiles.PT_SPECIAL_NONE:
				sp = targetenemydamagetype[power['DamageType']]
			if sp != []:
				for s in sp:
					if s == 'weakness':
						s = 'weakness_' + gendmgclassconversion[power['PowerType']] + '_' + gendmgconversion[power['DamageType']]
					if not (s in target):
						target.append(s)
	if not IsRanged(char):
		target.append('not flying')
		#target.append('not flier')
	return target
	
def GenerateAI_TimerTactics(char):
	DeterminePositioning(char)
	tactics = []
	tactics = CombineLists(tactics,GenerateAI_Remedies(char))
	tactics = CombineLists(tactics,GenerateAI_Flee(char))
	tactics = CombineLists(tactics,GenerateAI_Powers(char))
	tactics = CombineLists(tactics,GenerateAI_Throw(char))
	tactics = CombineLists(tactics,GenerateAI_Movement(char))
	return tactics
	
def CombineLists(list1, list2):
	if (list1 == []) or (list1 == None):
		return list2
	if (list2 == []) or (list2 == None):
		return list1
	newlist = list1[0:]
	for t in list2:
		newlist.append(t)
	return newlist

def AddOred(string, addendum):
	if string == '':
		return addendum
	if addendum == '':
		return string
	return string + ' or ' + addendum

def AddToList(vlist, value):
	newlist = vlist
	if type(vlist)  == types.ListType:
		newlist = vlist[0:]
		del vlist
		if type(value) == types.ListType:
			for v in value:
				newlist.append(v)
		else:
			newlist.append(value)
	elif vlist == 'none' or vlist == None:
		newlist = value
	else:
		if type(value) == types.ListType:
			newlist = value[0:]
			del value
			newlist.append(vlist)
		else:
			newlist = [vlist,value]
	return newlist	
	
# flee melee	
	
def GenerateAI_Flee(char):
	flee = []
	global scannedkf
	time = GetAnimationTime(char,'run')
	dist = DetermineRangedDist(char,flee=1)
	if IsRanged(char):
		m25report.Report(GENAI_CHANNEL,char + ' is ranged.')
		movetype = 'move'
		if CanFly(char):
			movetype = 'fly'
		flee= ['TFleeMelee(mindist=' + `dist` + ',movetype=' + `movetype` + ',time=' + `time` + ')',]
	flee = CombineLists(flee,['TFleeMelee(type="any",subtype="emitting_damage"'+ ',time=' + `time` + ')'])
	# commented out until I can determine the performance cost of this tactic
	#if IsFast(char):
	#	flee = CombineLists(flee, ['TDodgeMissiles(time=' + `time` + ')'])
	return flee


# movement
		
def GenerateAI_Movement(char):
	movement = []
	global scannedkf
	time = GetAnimationTime(char,'run')
	dist = DetermineRangedDist(char,flee=0)
	if dist == 'short':
		dist == 'medium'
	if dist == 'ranged_min':
		dist == 'short'
	if CanFly(char) and GoodMelee(char):
		movement.append('TLand()')
	if IsRanged(char):
		movement.append('TMoveTo(mindist='+`dist` +' ,subtype="not emitting_damage"'+',time=' + `time` + ')')
	else:
		if HasRanged(char):
			movement.append('TMoveTo(mindist='+`dist` + ',time=' + `time` + ',subtype=["not emitting_damage","flying"])')
			movement.append('TMoveTo(mindist="melee",subtype=["not emitting_damage","not flying"]'+',time=' + `time` + ')')
		else:
			movement.append('TMoveTo(mindist="melee",subtype="not emitting_damage"'+',time=' + `time` + ')')
	return movement


# hero points	
	
def GenerateAI_Remedies(char):
	template = js.Object_GetTemplate(char)
	if template[:6] != 'custom':
		if js.Object_GetClass(char) & js.OC_MINION:
			return []
	else:
		m25report.Report(GENAI_CHANNEL,char + ' uses hero points.')
		return ["TUseRemedy()",]
	

# powers and attributes	
		
# add to the list of tactics, but sort them by weights (contained in the second field of the list)
def AppendByWeight(tactics, tac):
	if (tac == None) or (tac == []):
		return tactics[0:]
	num = 0
	newtactics = tactics
	#m25report.Report(GENAI_CHANNEL,`tactics` + ' + ' + `tac`)
	if type(tac[0]) == types.ListType:
		for tc in tac:
			newtactics = AppendByWeight(newtactics,tc)
			#print newtactics
		return newtactics
	if (len(tactics) > 0) and (type(tactics[0]) == types.ListType):
		for t in tactics:
			if t[1] <= tac[1]:
				newtactics = tactics[:num]
				newtactics.append(tac)
				for t in tactics[num:]:
					newtactics.append(t)
				return newtactics
			num = num + 1
	newtactics = tactics[0:]
	newtactics.append(tac)
	return newtactics

def StripWeights(tactics):
	nt = []
	for t in tactics:	
		nt.append(t[0])
	return nt
		
def GetEnergyWeight(power):
	epcost = datfiles.PT_EP_COST_NAMES[power['EPCost']]
	if epcost == 'high':
		return 3
	if epcost == 'medium':
		return 2
	if epcost == 'low':
		return 1
	return 0	

	
def GenerateAI_GetTactic(char,identifier,prefix=''):
	try:
		print "ff.AIGen_%s('%s')"%(identifier,char)
		return eval("ff.AIGen_%s('%s')"%(identifier,char))
	except:
		if AIGenerateTemplate.has_key(prefix+identifier):
			return AIGenerateTemplate[prefix+identifier]
		else:
			return []
		
def GenerateAI_Powers(char):
	tactics = []
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power == {}: continue
		tac = []
		# first see if there is a state swap on the power that needs a special tactic
		tac = GetFFXCarrier(char, power)
		if tac == []:
			# if a defensive power, simply generate the power tactic
			if (power['PowerType'] in [datfiles.PT_PASSIVE_DEFENCE,datfiles.PT_ACTIVE_DEFENCE]):
				#tac = AIGenerateTemplate[datfiles.PT_POWER_TYPES[power['PowerType']]]
				tac = GenerateAI_GetTactic(char,datfiles.PT_POWER_TYPES[power['PowerType']])
			# check if there is a special damage type that needs a special tactic (in the 'special type' field)
			elif AIGenerateTemplate.has_key(datfiles.PT_SPECIAL_TYPES[power['SpecialType']]):
				#tac = AIGenerateTemplate[datfiles.PT_SPECIAL_TYPES[power['SpecialType']]]
				tac = GenerateAI_GetTactic(char, datfiles.PT_SPECIAL_TYPES[power['SpecialType']])
			# check if there is a special damage type in the 'damage type' field
			elif AIGenerateTemplate.has_key(datfiles.PT_DAMAGE_TYPES[power['DamageType']]):
				tac = GenerateAI_GetTactic(char, datfiles.PT_DAMAGE_TYPES[power['DamageType']])
				#tac = AIGenerateTemplate[datfiles.PT_DAMAGE_TYPES[power['DamageType']]]
			# regular attack/aid power - generate the power tactic
			elif AIGenerateTemplate.has_key(datfiles.PT_POWER_TYPES[power['PowerType']]):
				#tac = AIGenerateTemplate[datfiles.PT_POWER_TYPES[power['PowerType']]]
				tac = GenerateAI_GetTactic(char, datfiles.PT_POWER_TYPES[power['PowerType']])
		weighted_tactic = GenerateTac(char, power, tac)
		if weighted_tactic != []:
			print weighted_tactic
			tactics = AppendByWeight(tactics, weighted_tactic)
	# now generate tactics for special FFX attributes
	attributes = cdata['characterAttributes']
	print attributes
	tac = []
	for a in attributes:
		extra = GetMultiAttribute(a)
		if extra != []:
			for e in extra:
				attributes.append(e)
		else:
			weighted_tactic = []
			#ADDED to allow custom functions outside of m25generateai.py
			tac = GenerateAI_GetTactic(char,a,prefix='FFX_')
			weighted_tactic = GenerateTac(char, a, tac)		
			if (weighted_tactic != []) and (weighted_tactic!=None):
				print weighted_tactic
				tactics = AppendByWeight(tactics, weighted_tactic)
	# remove the weights (they've been sorted already) so that they can be added to the AI definition upon return
	tactics = StripWeights(tactics)
	return tactics
	
def GetMultiAttribute(attr):
	extra_attributes = []
	for m in ffxmulti.ffxMultis:
		if attr == m[0]:
			for n in m[1:]:
				extra_attributes.append(n)
	return extra_attributes
					
	
def GenerateTac(char, power, tac):
	if len(tac) == 0: return []
	if type(tac[0]) == types.ListType:
		tactics = []
		for t in tac:
			tactics.append(GenerateTac(char, power, t))
		return tactics
	if len(tac) == 1:
		tactic = ''
		exec 'tactic=' + tac[0]
		return tactic
	if len(tac) == 2:
		tactic = ''
		weight = 0
		if (tac[0][0] == 'T') or (tac[0][:2] == 'AI'):
			tactic = tac[0]
		else:
			exec 'tactic=' + tac[0]
		if type(tac[1]) == types.IntType:
			weight = tac[1]
		else:
			exec 'weight=' + tac[1]
		return [tactic, weight]
	return []
	

specialtypestring = {datfiles.PT_SPECIAL_AFTERBURNER : 'afterburner',
	datfiles.PT_SPECIAL_ALTERATION : 'alteration',
	datfiles.PT_SPECIAL_ALTER_GENETIC_STRUCTURE : '',
	datfiles.PT_SPECIAL_ANIMATE_DEAD : '',
	datfiles.PT_SPECIAL_ANIMATE_DEAD_MINIONS :'',
	datfiles.PT_SPECIAL_BEZERKER : '',
	datfiles.PT_SPECIAL_BUOYANCY : 'buoyancy',
	datfiles.PT_SPECIAL_BURROW : '',
	datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION : '',
	datfiles.PT_SPECIAL_CLOAK : 'cloak',
	datfiles.PT_SPECIAL_CLONE_SELF : '',
	datfiles.PT_SPECIAL_CLONE_TARGET : '',
	datfiles.PT_SPECIAL_DEEP_SENSOR : '',
	datfiles.PT_SPECIAL_DENSITY_MAXIMIZE : 'density',
	datfiles.PT_SPECIAL_DIMENSION_SHIFT : 'exile',
	datfiles.PT_SPECIAL_DISPLACE_IMAGE : 'displace',
	datfiles.PT_SPECIAL_DISRUPT_OBJECT : '',
	datfiles.PT_SPECIAL_DOPPELGANGER : '',
	datfiles.PT_SPECIAL_ELECTRIFY_SELF : '',
	datfiles.PT_SPECIAL_EMPATHY : 'empathy',
	datfiles.PT_SPECIAL_ENERGIZE : 'energize',
	datfiles.PT_SPECIAL_ENERGIZE_AREA : '',
	datfiles.PT_SPECIAL_EP_DRAW : '',
	datfiles.PT_SPECIAL_EP_TRANSFER : 'energysurge',
	datfiles.PT_SPECIAL_HP_BALANCE : '',
	datfiles.PT_SPECIAL_IGNITE : '',
	datfiles.PT_SPECIAL_IMAGE_PROJECTOR : 'limeylure',
	datfiles.PT_SPECIAL_LAWORDER : '',
	datfiles.PT_SPECIAL_LIFE_GIFT : '',
	datfiles.PT_SPECIAL_LIVING_FIRES : '',
	datfiles.PT_SPECIAL_MOLECULAR_EXCITE : '',
	datfiles.PT_SPECIAL_MORPH : '',
	datfiles.PT_SPECIAL_NONE : '',
	datfiles.PT_SPECIAL_NORMALISE : '',
	datfiles.PT_SPECIAL_PATRIOT_CHARGE : '',
	datfiles.PT_SPECIAL_POSSESSION : '',
	datfiles.PT_SPECIAL_POWER_NULLIFICATION : 'powernull',
	datfiles.PT_SPECIAL_RALLY : 'rally',
	datfiles.PT_SPECIAL_REMOVE_DEFENCE : '',
	datfiles.PT_SPECIAL_SACRIFICE : '',
	datfiles.PT_SPECIAL_SONIC_BOOM : '',
	datfiles.PT_SPECIAL_SPEEDING_BULLET : '',
	datfiles.PT_SPECIAL_SPRINT : '',
	datfiles.PT_SPECIAL_TELEPORT_SELF : '',
	datfiles.PT_SPECIAL_TELEPORT_SWAP : '',
	datfiles.PT_SPECIAL_TUMBLE : '',
	datfiles.PT_SPECIAL_TUNNEL : '',
	datfiles.PT_SPECIAL_UNCLOAK : ''}	
	
damagetypestring = {datfiles.PT_DAMAGE_ACID : '',
	datfiles.PT_DAMAGE_ACID_BURN : 'acidburn',
	datfiles.PT_DAMAGE_BLIND : 'blind',
	datfiles.PT_DAMAGE_COLD : '',
	datfiles.PT_DAMAGE_CRUSH : '',
	datfiles.PT_DAMAGE_ELECTRICAL : '',
	datfiles.PT_DAMAGE_ENERGY : '',
	datfiles.PT_DAMAGE_FIRE : '',
	datfiles.PT_DAMAGE_FREEZE : 'frozen',
	datfiles.PT_DAMAGE_HEX : 'hex',
	datfiles.PT_DAMAGE_HYPNOSIS : 'hypnosis',
	datfiles.PT_DAMAGE_IRRADIATE : 'irradiate',
	datfiles.PT_DAMAGE_MENTAL_BLANK : 'blank',
	datfiles.PT_DAMAGE_OBJECTIFY : '',
	datfiles.PT_DAMAGE_PANIC : '',
	datfiles.PT_DAMAGE_PIERCE : '',
	datfiles.PT_DAMAGE_POLYMORPH : 'polymorph',
	datfiles.PT_DAMAGE_RADIATION : '',
	datfiles.PT_DAMAGE_RAGE : 'rage',
	datfiles.PT_DAMAGE_SPECIAL : '',
	datfiles.PT_DAMAGE_SPEED_DECREASE : 'speeddecrease',
	datfiles.PT_DAMAGE_SPEED_INCREASE : 'speedincrease',
	datfiles.PT_DAMAGE_STASIS : 'stasis',
	datfiles.PT_DAMAGE_TOADIFY : 'toadify',
	datfiles.PT_DAMAGE_WATER : '',
	datfiles.PT_DAMAGE_WEAKEN : 'weaken'}		
		
def GetFFXCarrier(char, power):
	if power == {}: return []
	template = m25template.GetComplexTemplate(char)
	powername = power['PowerName']
	if power.has_key('SpecialType'): specialtype = power['SpecialType']
	else:
		specialtype = datfiles.PT_SPECIAL_NONE
	if power.has_key('DamageType'): damagetype = power['DamageType']
	else: 
		damagetype = datfiles.PT_DAMAGE_WATER
	for e in ffxcustom.FFX_CARRIERS:
		if e[0] == template:
			if e[3] == powername:
				print e[1], e[2]
				return DetermineFFXTactic(char,power,e[1],e[2])
			else:
				if e[3] == 'all':
					if (e[1] == specialtypestring[specialtype]) or (e[1] == damagetypestring[damagetype]):
						print `e[1]` + ' swap to ' + `e[2]`
						return DetermineFFXTactic(char,power,e[1],e[2])
	return []	
	
def DetermineFFXTactic(char, power, swapfrom, swapto):
	if swapto == 'absorb': AddCharAttr('absorber')
	if swapfrom in ['empathy', 'cloak', 'displace', 'limeylure', 'rally']:
		key = 'FFXswap_'+swapfrom+'_'+swapto
		tac = GenerateAI_GetTactic(char,key)
		if tac == []:
			if swapto in ['phase', 'shield', 'cloak', 'accelerate', 'normalise', 'heal', 'leadership', 'remoteempathy']:
				key = 'FFXswap_'+swapfrom+'_ally'
			key = 'FFXswap_'+swapfrom+'_any'
		else: return tac
	else:
		key = 'FFXswap_'+swapto
	return GenerateAI_GetTactic(char,key)

# power generators


def Gen_FFX_livinglaser(char,power):
	if IsRanged(char):
		return ['TLivingLaserAway()',5]
	return ['TLivingLaser()',0]
	
def Gen_FFX_superleaper(char,power):
	if IsRanged(char):
		return ['TSuperLeapAway()',5]
	return ['TSuperLeap()',0]	

def Gen_PT_ACTIVE_DEFENCE(char,power):
	mindist = 0
	weight = 10
	rt = ''
	if power['AttackModesBlocked'] == datfiles.PT_RANGED_BLOCKED:
		mindist = 'melee'
	if power['DefenceFlags'] & datfiles.PT_DEFENCE_FLAG_REMOTE:
		rt = 'Remote'
	SetReserveLevel(char, power['EPCost'])
	return [CreateTactic(char,power,'TActiveDefense'+rt,mindist=mindist),weight]
	#return ['TActiveDefense' + rt + '(' + `power['PowerName']` + ', mindist=' + `mindist` +  ')',weight]

def Gen_PT_AREA(char,power):
	weight = GetEnergyWeight(power)+1
	return [CreateTactic(char,power,'TArea'),weight]	

def Gen_PT_DIRECT(char,power):
	weight = GetEnergyWeight(power)
	if power['Radius'] > 0:
		return [CreateTactic(char,power,tacticname='TDirectExplosive'),weight]
	return [CreateTactic(char,power,tacticname='TDirect'),weight]
	
def Gen_PT_MELEE(char,power):
	weight = GetEnergyWeight(power)
	mt = ''
	if power['AttackFlags'] & datfiles.PT_ATTACK_MELEE_AREA:
		mt = 'Area'
		weight = weight + 1
	if power['AttackFlags'] & datfiles.PT_ATTACK_AIRBORN_TARG_ONLY:
		mt = 'Air'
		weight = weight + 1
	return [CreateTactic(char,power,'TMelee'+mt),weight]

def Gen_PT_RANGED(char,power):
	weight = GetEnergyWeight(power)
	rt = ''
	if power['SubType'] == datfiles.PT_ATTACK_SUBTYPE_EXPLOSIVE:
		rt = 'Explosive'
		weight = weight + 1
	if power['SubType'] == datfiles.PT_ATTACK_SUBTYPE_CONE:
		rt = 'Cone'
		weight = weight + 1
	return [CreateTactic(char,power,'TRanged'+rt),weight]

	

		
def Gen_PT_SPECIAL_CLOAK(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TCloak',maxdist='short'),4]
		#return ['TCloak(' + `power['PowerName']` + ',maxdist="medium")',4]
	else:
		return [CreateTactic(char,power,'TCloak',mindist='short'),4]
		#return ['TCloak(' + `power['PowerName']` + ',mindist="short")',4]	
	
def Gen_PT_SPECIAL_IMAGE_PROJECTOR(char,power):
	SetReserveLevel(char, power['EPCost'])
	return [[CreateTactic(char,power,'TDecoy'), 10], ['TFleeMelee(situation="neardecoy short")',5]]
		
		
def Gen_PT_SPECIAL_SPRINT(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TSprintAway'),5]
	return [CreateTactic(char,power,'TSprint'),0]

def Gen_PT_SPECIAL_TELEPORT_SELF(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TTeleportAway'),5]
	return [CreateTactic(char,power,'TTeleport'),0]
		
def Gen_PT_SPECIAL_TUMBLE(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TTumbleAway'),5]
	return [CreateTactic(char,power,'TTumble'),0]	

def Gen_PT_SPECIAL_TUNNEL(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TTunnelTravelAway',time=10),5]
	return [CreateTactic(char,power,'TTunnelTravel',time=10),0]

def Gen_PT_SPECIAL_UNCLOAK(char,power):
	if IsRanged(char):
		return [CreateTactic(char,power,'TUncloak',mindist='short'),4]
		#return ['TUncloak(' + `power['PowerName']` + ',mindist="short")',4]
	else:
		return [CreateTactic(char,power,'TUncloak',maxdist='short'),4]
		#return ['TUncloak(' + `power['PowerName']` + ',maxdist="short")',4]



# throwing
	
def GenerateAI_Throw(char):
#	global scannedkf
	if CanThrow(char):
		pickup_time = GetAnimationTime(char,'pick_up_object')
		throw_time = GetAnimationTime(char,'throw_object')
		return ["TThrow(pickup_time="+`pickup_time` + ",throw_time=" + `throw_time` + ")",]
	return []


# helper functions

def CanThrow(char):
	if ((js.Object_GetAttr(char,'strength') >= 4) or chardata.Character_HasAttribute(char,'heavy lifter')) and (not chardata.Character_HasAttribute(char,'clumsy')):
		return 1
	return 0

def CountFarPowers(char):
	count = 0
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power != {} and power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT] and (not FriendPower(power)):
			count = count + 1
	return count
	
def CountNearPowers(char):
	count = 0
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if power != {} and power['PowerType'] in [datfiles.PT_MELEE,datfiles.PT_AREA] and (not FriendPower(power)):
			count = count + 1
	return count
	
def FriendPower(power):
	friendspecial = [datfiles.PT_SPECIAL_BUOYANCY,
		datfiles.PT_SPECIAL_CEREBRAL_NORMALISATION,
		datfiles.PT_SPECIAL_ENERGIZE,
		datfiles.PT_SPECIAL_ENERGIZE_AREA,
		datfiles.PT_SPECIAL_EP_TRANSFER,
		datfiles.PT_SPECIAL_LIFE_GIFT,
		datfiles.PT_SPECIAL_NORMALISE,
		datfiles.PT_SPECIAL_RALLY]
	if power['SpecialType'] in friendspecial:
		return 1
	return 0
	
def CanFly(char):
	flight_attributes = ['flier', 'levitate', 'ffqglider', 'ffqlowgravflier', 'fastflier', 'ffqropeswinger']
	for f in flight_attributes:
		if chardata.Character_HasAttribute(char,f):
			m25report.Report(GENAI_CHANNEL,char + ' can fly.')
			return 1
	return 0	
	
def IsMelee(char):
	return not IsRanged(char)	
	
def IsRanged(char):
	if (chardata.GetCharacterData(char)['strength'] >= 5) and HasMeleePower(char):
		return 0
	return CountFarPowers(char) >= CountNearPowers(char)

def GoodMelee(char):
	if GoodMeleeDamage(char) and GoodEndurance(char):
		return 1
	return 0
	
def GoodMeleeDamage(char):
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if (power['PowerType'] == datfiles.PT_MELEE) and (power['Magnitude'] >= 2):
			m25report.Report(GENAI_CHANNEL,char + ' has good melee damage.')
			return 1
	return 0
	
def GoodEndurance(char):
	if (chardata.GetCharacterData(char)['endurance']) >= 4:
		m25report.Report(GENAI_CHANNEL,char + ' has good endurance.')
		return 1
	return 0

def HasMeleePower(char):
	cdata = chardata.GetCharacterData(char)
	for powdata in cdata['powers'].keys():
		power = cdata['powers'][powdata]
		if (power['PowerType'] == datfiles.PT_MELEE):
			m25report.Report(GENAI_CHANNEL,char + ' has melee powers.')
			return 1
	return 0	
		
def IsFast(char):
	if chardata.GetCharacterData(char)['speed'] >=4:
		m25report.Report(GENAI_CHANNEL,char + ' has good speed.')
		return 1
	return 0	
	
def HasRanged(char):
	if CanThrow(char):
		return 1
	cdata = chardata.GetCharacterData(char)
	powers = []
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT]):
			return 1
	return 0
	
def CanDisplaceImage(char):
	cdata = chardata.GetCharacterData(char)
	powers = []
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'SpecialType' in power.keys() ) and (power['SpecialType'] == datfiles.PT_SPECIAL_DISPLACE_IMAGE):
			return 1
	return 0	
	
def CanCloak(char):
	cdata = chardata.GetCharacterData(char)
	powers = []
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'SpecialType' in power.keys() ) and (power['SpecialType'] == datfiles.PT_SPECIAL_CLOAK):
			return 1
	return 0		
	
def GetNumberOfAttacks(char):
	cdata = chardata.GetCharacterData(char)
	num = 0
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT,datfiles.PT_MELEE,datfiles.PT_AREA]):
			if not FriendPower(power):
				num = num + 1
	return num
	
def HasOneRangedAttack(char):
	cdata = chardata.GetCharacterData(char)
	powers = []
	num = 0
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT]) and not FriendPower(power):
			num = num + 1
	if num == 1: return 1
	return 0
	
def HasOneMeleeAttack(char):
	cdata = chardata.GetCharacterData(char)
	powers = []
	num = 0
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_MELEE,datfiles.PT_AREA]) and not FriendPower(power):
			num = num + 1
	if num == 1: return 1
	return 0
	
def HasManyDamageTypes(char):
	cdata = chardata.GetCharacterData(char)
	powers = []
	type1 = None
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT,datfiles.PT_AREA,datfiles.PT_MELEE]) and not FriendPower(power):
			if 'DamageType' in power.keys():
				if type1 != None:
					if type1 != power['DamageType']:
						return 1
				type1 = power['DamageType']
	return 0	
	
#Attack percentages
# - melee characters with ranged attacks should have a lower chance of using ranged attacks ??
# - ranged characters should have a higher chance of using ranged attacks ??
# - characters with a single attack should have a 100% chance of using it
def AdjustAttackPct(char,power):
	if power.has_key('PowerType') and power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT]:
		if HasOneRangedAttack(char) and IsRanged(char):
			return 100
	if power.has_key('PowerType') and power['PowerType'] in [datfiles.PT_MELEE,datfiles.PT_AREA]:
		if HasOneMeleeAttack(char) and IsMelee(char):
			return 100	

# - characters with more than one damage type should not use an attack if the enemy resists its damage type		
def AdjustSubtype(char,power,subtype):
	if HasManyDamageTypes(char) and ('DamageType' in power.keys()):
		if gendmgconversion[power['DamageType']] != 'none':
			s = 'not resistant_' + gendmgclassconversion[power['PowerType']] + '_' + gendmgconversion[power['DamageType']]
			return AddToList(subtype,s)
	return subtype

#-------------------------------------------------------------------------------------
#
# Positioning
#
#-------------------------------------------------------------------------------------	
	
def SetRangedDist(char, ranged_dist):
	m25getset.SetObjectVar(char,'_gen_ranged_dist',ranged_dist)
	
def GetRangedDist(char):
	return m25getset.GetObjectVar(char,'_gen_ranged_dist',default='short')

def SetMoveToRangeTarget(char, flee_type):
	m25getset.SetObjectVar(char,'_gen_fleetarget',flee_type)
	
def GetMoveToRangeTarget(char):
	return m25getset.SetObjectVar(char,'_gen_fleetarget',default=None)	
		
def SetMoveToRangeSituation(char, flee_type):
	m25getset.SetObjectVar(char,'_gen_fleesit',flee_type)
	
def GetMoveToRangeSituation(char):
	return m25getset.SetObjectVar(char,'_gen_fleesit',default=None)	
	
def SetMoveToMeleeSituation(char, flee_type):
	m25getset.SetObjectVar(char,'_gen_movetosit',flee_type)
	
def GetMoveToMeleeSituation(char):
	return m25getset.SetObjectVar(char,'_gen_movetosit',default=None)	
	
def SetMoveToMeleeTarget(char, flee_type):
	m25getset.SetObjectVar(char,'_gen_movetotarget',flee_type)
	
def GetMoveToMeleeTarget(char):
	return m25getset.SetObjectVar(char,'_gen_movetotarget',default=None)	
	
	
def DeterminePositioning(char):
	# determine ranged dist
	ranged_dist = DetermineRangedDist(char)
	#SetRangedDist(char,ranged_dist)
	m25report.Report(GENAI_CHANNEL,char + ' ranged distance = ' + `ranged_dist`)
	# move to melee when...
	melee_factors = DetermineMoveToMelee(char)
	#SetMoveToMeleeSituation(melee_factors[1])
	#SetMoveToMeleeTarget(melee_factors[0])
	# move to ranged when...
	range_factors = DetermineMoveToRange(char)
	#SetMoveToRangeSituation(range_factors[1])
	#SetMoveToRangeTarget(range_factors[0])

def DetermineRangedDist(char, flee=0):
	ranged_dist = 1000
	cdata = chardata.GetCharacterData(char)
	powers = []
	for pname in cdata['powers'].keys():
		power = cdata['powers'][pname]
		if ( 'PowerType' in power.keys() ) and (power['PowerType'] in [datfiles.PT_RANGED,datfiles.PT_DIRECT]):
			if power['RangeMax'] < ranged_dist:
				ranged_dist = power['RangeMax']
	if (ranged_dist == 1000):
		ranged_dist = 'short'
	else:
		ranged_dist = datfiles.PT_RANGE_NAMES[ranged_dist]
	# if we have a power with short range, make the character stay at FF.RANGED_MIN (otherwise they'll waffle)
	if flee:
		 if ranged_dist == 'short':
		 	ranged_dist = 'ranged_min'
		 elif ranged_dist == 'medium':
		 	ranged_dist = 'short'
		 elif ranged_dist == 'long':
		 	ranged_dist = 'medium'
		 else:
		 	ranged_dist = 'short'
	return ranged_dist
	
def DetermineMoveToMelee(char):
	target_factors = ['not emittingdamage']
	situation_factors = []
	if CanDisplaceImage(char) and IsMelee(char):
		situation_factors.append('displaced')
	if CanCloak(char) and IsMelee(char):
		situation_factors.append('cloaked')
	if IsRanged(char):
		target_factors.append('not canhurtrange')
	return [target_factors,situation_factors]

def DetermineMoveToRange(char):
	target_factors = ['emittingdamage']
	situation_factors = []
	if IsMelee(char):
		target_factors.append('not canhurtmelee')
	if (CanDisplaceImage(char) or CanCloak(char)) and IsRanged(char):
		situation_factors.append('displaced or cloaked')
	return [target_factors,situation_factors]


tactictypename = {
	datfiles.PT_RANGED:'TRanged',
	datfiles.PT_MELEE:'TMelee',
	datfiles.PT_AREA:'TArea',
	datfiles.PT_DIRECT:'TDirect',
	datfiles.PT_SPECIAL:'TSpecial',
}	
	
def CreateTactic(char, power, tacticname=None, pct=None, priority=None, type=None, subtype=None, situation=None, mindist=None, maxdist=None, time=None):
	if tacticname == None:
		tacticname = tactictypename[power['PowerType']]
	tactic = tacticname + '(' + repr(power['PowerName'])
	param_sep = ','
	if pct != None: 
		tactic = tactic + param_sep + 'pct=' + `pct`
		param_sep = ','
	else:
		if type == None or type == 'enemy':
			pct = AdjustAttackPct(char,power)
		if pct != None:
			tactic = tactic + param_sep + 'pct=' + `pct`
			param_sep = ','
	if priority != None: 
		tactic = tactic + param_sep + 'priority=' + `priority`
		param_sep = ','
	if type != None: 
		tactic = tactic + param_sep + 'type=' + `type`
		param_sep = ','
	#if type == None or type == 'enemy':
	#	subtype = AdjustSubtype(char,power,subtype)
	if subtype != None:
		tactic = tactic +param_sep + 'subtype=' + `subtype`
		param_sep = ','
	if situation != None: 
		tactic = tactic + param_sep + 'situation=' + `situation`
		param_sep = ','
	if mindist != None: 
		tactic = tactic + param_sep + 'mindist=' + `mindist`
		param_sep = ','
	if maxdist != None: 
		tactic = tactic + param_sep + 'maxdist=' + `maxdist`
		param_sep = ','
	if time == None:
		time = GetAnimationTime(char,power['animation'])
	if time != None: 
		tactic = tactic + param_sep + 'time=' + `time`
		param_sep = ','
	tactic = tactic + ')'
	return tactic
	
	
#-------------------------------------------------------------------------------------
#
# Output - both to the log and to the m25aidata file
# will also write to .m25ai files
#
#-------------------------------------------------------------------------------------	

	
def PrintAI(ai):
	print '[' + `ai[0]` + ','
	for section in ai[1:]:
		if section[0] == 'timer':
			print '\t["timer", ' + `section[1]` + ','
			for tac in section[2:]:
				print '\t\t' + `tac` + ','
			print '\t],'
		elif section[0] == 'target enemy':
			print'\t["target enemy",'
			for targ in section[1:]:
				print '\t\t' + `targ` + ','
			print '\t],'
		else:
			print '\t[' + `section[0]` + ','
			for targ in section[1:]:
				print '\t\t' + `targ` + ','
			print '\t],'
	print '],'

	
def GetAIFile():
	return m25aidata.AIExtensions
	
def AddAI(ai):
	if type(ai) != types.ListType:
		print 'bad add ai', ai
		return
	all_ai = GetAIFile()
	num = 0
	for a in all_ai:
		if type(a) != types.ListType:
			print 'bad ai', a
			return
		if a[0] == ai[0]:
			ai = MergeAI(a,ai)
			all_ai.remove(a)
			#all_ai.append(ai)
		num = num + 1
	oldai = FindHeroAI(ai[0])
	if oldai != None:
		ai = MergeAI(oldai,ai)
	all_ai.append(ai)
	all_ai.sort()
	
def MergeAI(oldai, newai):
	mergedai = newai[0:]
	reserved = []
	try:
		for t in newai[1:]:
			reserved.append(t[0])
	except:
		pass
	try:
		for element in oldai[1:]:
			if element[0] not in reserved:
			#if element[0] not in ['timer','target enemy','target any']:
				mergedai.append(element)
	except:
		pass
	return mergedai
	
def FindHeroAI(template):	
	try:
		filename = os.path.join(datfiles.GetHeroFilePath(),template+'.m25ai')
		f = open(filename,'r')
		cust = f.read()
		f.close()
		ai = eval(cust)
		return ai
	except:
		return None
	
	
def WriteAIFile():
	global genlist
	all_ai = GetAIFile()
	#FileName = os.path.abspath(FileName)
	scriptpath = os.path.join(datfiles.GetModPath(),'Missions','Scripts')
	print scriptpath
	fname = os.path.join(scriptpath,'m25aidata.py')
	aif = open(fname,'w')
	aif.write('AIExtensions=[\n')
	for a in all_ai:
		if a[0][:7] == 'custom_':
			print 'WriteAIFile: ' + a[0] + ' is unbranded and will not be written.'
		else:
		#try to write as .hero file ai first
			if CharIsHeroFile(a[0]) and (a[0] in genlist):
				print 'WriteAIFile: '+ a[0] + ' is a .hero file!'
				try:
					filename = os.path.join(datfiles.GetHeroFilePath(),a[0]+'.m25ai')
					print filename
					f = open(filename,'w')
					WriteAI(f,a,single=1)
					f.close()
				except:
					print 'WriteAI unable to write hero file ai ' + a[0]
			else:
				#print 'WriteAIFile: '+ a[0] + ' is NOT a .hero file!'
				WriteAI(aif,a)
	aif.write('\n\n#end of file - there should be a close bracket here\n')
	aif.write(']\n')
	aif.close()

def WriteAI(file, ai,single=0):
	if single:
		file.write('[' + `ai[0]` + ',\n')
	else:
		file.write('\n\n[' + `ai[0]` + ',\n')
	for section in ai[1:]:
		try:
			if section[0] == 'timer':
				file.write('\t["timer", ' + `section[1]` + ',\n')
				for tac in section[2:]:
					file.write('\t\t' + `tac` + ',\n')
				file.write('\t],\n')
			elif section[0] == 'target enemy':
				file.write('\t["target enemy",\n')
				for targ in section[1:]:
					file.write('\t\t' + `targ` + ',\n')
				file.write('\t],\n')
			else:
				file.write('\t[' + `section[0]` + ',\n')
				for targ in section[1:]:
					file.write('\t\t' + `targ` + ',\n')
				file.write('\t],\n')
		except:
			pass
	if single:
		file.write(']\n')
	else:			
		file.write('],\n')

genailist = []	
	
def GenerateAIFile(charlist):
	global genailist
	genailist = charlist[0:]
	js.Mission_StatusText('Creating AI for ... ' + `m25template.GetComplexTemplate(genailist[0])`)
	m25event.SetTimer(0.2,'GenerateNextAI')
	return
	
	for c in charlist:
		print 'Creating AI for ... ' + c
		js.Mission_StatusText('Creating AI for ... ' + c)
		ai = GenerateAI(c)
		if ai != None:
			AddAI(ai)
	WriteAIFile()
	
	
def GenerateNextAI(event):
	global genailist
	c = genailist[0]
	
	ai = GenerateAI(c)
	if ai != None:
		AddAI(ai)
	genailist=genailist[1:]
	if genailist == []:
		m25event.SetTimer(1,'GenerateAIEnd')
	else:
		js.Mission_StatusText('Creating AI for ... ' + `m25template.GetComplexTemplate(genailist[0])`)
		m25event.SetTimer(0.2,'GenerateNextAI')

def GenerateAIEnd(event):
	WriteAIFile()
	js.Mission_StatusText('Finished creating custom AIs.  Have fun with them.')
	m25event.SetTimer(3,'End_Create')
	
def End_Create(event):
	Mission_Win()	
	
#-------------------------------------------------------------------------------------
#
# keyframe reader
#
#-------------------------------------------------------------------------------------

scannedkf = {}

def GetAnimationTime(char,animation):
	global scannedkf
	filename = chardata.GetKeyframeFile(char)
	keyframes = {}
	try:
		keyframes = datfiles.Keyframe_ReadAnimationTimes(filename)
	except:
		print 'unable to read keyframes'
	if keyframes == {}:
		if scannedkf == {}:
			keyframes = scankf(char, filename)
		else:
			keyframes = scannedkf
	#print keyframes
	val = GetAnimVal(keyframes,animation)
	for letter in ['a','b','c','d','e','f','g','h','i']:
		newval = GetAnimVal(keyframes,animation+'_'+letter)
		if newval > val:
			val = newval
	if val == -1:
		val = 3
	return val

def GetAnimVal(keyframes, animation):
	if keyframes.has_key(animation):
		anim = keyframes[animation]
		if type(anim) == types.ListType:
			val =  anim[len(anim)-1][1]
		else:
			val = anim
		return ((int(val*100.0))/100.0)  #strip down to two decimal places
	return -1
	
def CharIsHeroFile(char):
	filename = os.path.join(datfiles.GetHeroFilePath(),char+'.hero')
	if not os.path.isfile(filename):
		return 0
	return 1

	
	
def scankf(char, filename):
	global scannedkf
	try:
		f = open(filename,'rb')
		dat = f.read()
		f.close()
	except:
		return {}
	offset = 0
	total = len(dat)
	while offset < total:
		t = None
		try:
			t = struct.unpack('i',dat[offset:offset+4])[0]
		except:
			t = None
		if (t != None) and (t < 30) and (t > 0):
			word = dat[offset+4:offset+4+t]
			if IsWord(word):
				#print word
				if IsAnim(dat, offset+4+t):
					endloc = scanforend(dat[offset+4+t:])
					#print 'endloc = ' + `endloc`
					t2 = struct.unpack('f',dat[offset+4+t+endloc-8:offset+4+t+endloc-4])[0]
					time = t2
					#print 'time = ' + `time`
					scannedkf[word] = time
					offset = offset + endloc + 4 + t
					#print 'new offset = ' + `offset`
					lastoffset = offset
		offset = offset + 1
	return scannedkf

def IsWord(word):
	for w in word:
		if w not in ['a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z','_','1','2','3','4','5','6','7','8','9','0']:	
			return 0
	return 1
		
def IsAnim(dat, offset):
	if dat[offset+4:offset+8] == 'Root':
		return 1
	if dat[offset+16:offset+16+len('Bip')] == 'Bip':
		return 1
	#print `dat[offset+12:offset+12+len('NiText')]`
	if dat[offset+12:offset+12+len('NiText')] == 'NiText':
		return 1
	return 0

def scanforend(dat):
	num = 0
	for d in dat:
		if dat[num:num+len('end')] == 'end':
			return num
		num = num + 1
	return 0		
