#####################################################################################
#
# Module: Mission and Object Variables
# Author: m25  (credit for the idea goes to DrMike)
# 
# Description: set all kinds of mission and object variables without generating errors
#
# Dependencies: stumpy's incredible missionobjvar.py, datfiles and m25string
#
# Version 
#
#####################################################################################

#print "Loading m25getset.py v.3.2.0 final; branch = m25 branch 1"

from ff import *
from js import *
from missionobjvar import *

import datfiles
import m25string
import os
import missionobjvar
import types
import m25report

#from time import *
import time

#3.041
# added support for DrMikes freeroam - ability to set variables to campaign (they can be saved and reloaded across map transitions)



GETSET_SAVEVARS = '_getset_save'
NOT_FOUND = '__NOT_FOUND'

#-------------------------------------------------------------------------------------
#
# Object Variables
#
#-------------------------------------------------------------------------------------	

def SetObjectVar(object, name, value,nosave=0,campaign=0):
	start = time.clock()
	WriteToObjectVarDictionary(object, name, value, campaign)
	if IsSavedVariablesOn() and (not nosave):
		if (not campaign) and (type(value) is types.IntType) or (type(value) is types.FloatType):
  			Mission_SetAttr(object+'_'+name, value)
		else:
			#print 'Saving object var ' + object + ' : ' + name + ' -> ' + `value`
			#missionobjvar.Object_SetVar(object, name, value)
			StoreVar('%s_%s'%(object,name),value)
	m25report.ReportTime('SetObjectVar', time.clock()-start)


def GetObjectVar(object, name, default=None, integer=0, nosave=0, number=0):
	start = time.clock()
	value = GetFromObjectVarDictionary(object,name)
	if value != NOT_FOUND: return value
	if IsSavedVariablesOn() and not nosave:
		try:
			value = Mission_GetAttr(object+'_'+name)
			if integer:
 				WriteToObjectVarDictionary(object,name,int(value))
 				return int(value)
 			WriteToObjectVarDictionary(object,name,value)
 			m25report.ReportTime('getObjectVar', time.clock()-start)
 			return value
 		except:
 			pass
		#value = missionobjvar.Object_GetVar(object, name, defaultVal=default, setToDefault=1)
		value = RestoreVar('%s_%s'%(object,name), default=default)
		#if value == None: value=default
		WriteToObjectVarDictionary(object,name,value)
		m25report.ReportTime('getObjectVar', time.clock()-start)
		return value
	if value == NOT_FOUND:
		WriteToObjectVarDictionary(object,name,default)
		m25report.ReportTime('getObjectVar', time.clock()-start)
		return default
	m25report.ReportTime('getObjectVar', time.clock()-start)
	return value


#-------------------------------------------------------------------------------------
#
# Mission Variables
#
#-------------------------------------------------------------------------------------				
	
def StoreMissionData():
	return
	StoreMissionSaveDictionary()
		
def SetMissionVar(name, value, nosave=0, campaign=0):
	start = time.clock()
	WriteToMissionVarDictionary(name, value, campaign)
	if IsSavedVariablesOn() and (not nosave):
 		if (not campaign) and (type(value) is types.IntType) or (type(value) is types.FloatType):
 			Mission_SetAttr(name, value)
 		else:
 			#print 'Saving mission var ' + name + ' -> ' + `value`
 			#missionobjvar.Mission_SetVar(name, value)
 			StoreVar(name,value)
 			#WriteToMissionSaveDictionary(name,value)
 			#StoreMissionSaveDictionary()
 	m25report.ReportTime('SetMissionVar', time.clock()-start)

def GetMissionVar(name, default=None, integer=0, nosave=0, number=0):
	start = time.clock()
	value = GetFromMissionVarDictionary(name)
	if value != NOT_FOUND:
		m25report.ReportTime('GetMissionVar', time.clock()-start)
		return value
	if IsSavedVariablesOn() and not nosave:
		try:
			value = Mission_GetAttr(name)
			if integer:
				WriteToMissionVarDictionary(name, int(value))
				m25report.ReportTime('GetMissionVar', time.clock()-start)
				return int(value)
			WriteToMissionVarDictionary(name, value)
			m25report.ReportTime('GetMissionVar', time.clock()-start)
			return value
		except:
			pass
		#value = GetFromMissionSaveDictionary(name)
		#value = missionobjvar.Mission_GetVar(name)
		value = RestoreVar(name, default=default)
		if value == None:
			WriteToMissionVarDictionary(name, default)
			m25report.ReportTime('GetMissionVar', time.clock()-start)
			return default
		WriteToMissionVarDictionary(name, value)
		m25report.ReportTime('GetMissionVar', time.clock()-start)
		return value
	if value == NOT_FOUND:
		WriteToMissionVarDictionary(name, default)
		m25report.ReportTime('GetMissionVar', time.clock()-start)
		return default
	m25report.ReportTime('GetMissionVar', time.clock()-start)
	return value


#-------------------------------------------------------------------------------------
#
# Checking for write-through on variables
#
#-------------------------------------------------------------------------------------				
			
def TurnOnSavedVariables():
	Mission_SetAttr(GETSET_SAVEVARS,1)
	
def TurnOffSavedVariables():
	Mission_SetAttr(GETSET_SAVEVARS,0)

def IsSavedVariablesOn():
	#return 0
	try:
		val = Mission_GetAttr(GETSET_SAVEVARS)
		return val
	except:
		if GetSet_InCampaign():
			TurnOnSavedVariables()
			return 1
		else:
			TurnOffSavedVariables()
			return 0


def GetSet_InCampaign():
    return 'OnGameTypeCheck' not in dir(ff)		
		
#-------------------------------------------------------------------------------------
#
# Variable Dictionary
#
#-------------------------------------------------------------------------------------					
						
vardictionary = {}
#savedict = {}



# def StartMVDUpdate():
# 	global vardictionary
# 	start = time.clock()
# 	if not js.Mission_TestAttr('_mvd'):
# 		print 'STARTMVD has value %f'%(js.Mission_TestAttr('_mvd'))
# 		js.Mission_SetAttr('_mvd',1)
# 		js.Event_RegisterSink(js.EVENT_TIMER, 'UpdateMissionVarDictionary', repr(vardictionary), '', 0.1, 0, 0)
# 		print 'UMVD took: %f'%(time.clock()-start)

# def UpdateMissionVarDictionary(event):
# 	global vardictionary
# 	start = time.clock()
# 	#if vardictionary == {}: vardictionary = eval('_destring(%s)'%(event.object))
# 	print repr(vardictionary)
# 	js.Event_RegisterSink(js.EVENT_TIMER, 'UpdateMissionVarDictionary', repr(vardictionary), '', 1, 0, 0)
# 	print 'UMVD took: %f'%(time.clock()-start)

# def _destring(string):
# 	return string
	
		
def GetFromMissionVarDictionary(name):
	if vardictionary.has_key(name):
		return vardictionary[name]['value']
	return NOT_FOUND

def WriteToMissionVarDictionary(name, value, campaign=0):
	if vardictionary.has_key(name):
		del vardictionary[name]
	vardictionary[name] = {}
	vardictionary[name]['value'] = value
	vardictionary[name]['campaign'] = campaign
	if campaign:
		AddCampaignVar(name)
			
# def WriteToMissionSaveDictionary(name, value):
# 	savedict[name] = value
# 	
# def GetFromMissionSaveDictionary(name):
# 	if savedict == {}: LoadMissionSaveDictionary()
# 	if savedict.has_key(name):
# 		return savedict[name]
# 	return NOT_FOUND
# 	
# def StoreMissionSaveDictionary():
# 	if IsSavedVariablesOn():
# 		missionobjvar.Mission_SetVar('_ai_mission_vars',savedict)
# 	
# def LoadMissionSaveDictionary():
# 	savedict = missionobjvar.Mission_GetVar('_ai_mission_vars',defaultVal={'__t':0})
# 	
# def RemoveMissionVar(var):
# 	if savedict.has_key(var):
# 		del savedict[var]
	
def GetFromObjectVarDictionary(object, name):
	if vardictionary.has_key(object+':'+name):
		return vardictionary[object + ':' + name]['value']
	return NOT_FOUND
			
def WriteToObjectVarDictionary(object, name, value, campaign=0):
	if vardictionary.has_key(object + ':' + name):
		del vardictionary[object + ':' + name]
	vardictionary[object + ':' + name] = {}
	vardictionary[object + ':' + name]['value'] = value
	vardictionary[object + ':' + name]['campaign'] = campaign
	if campaign:
		AddCampaignVar(object+':'+name)

def ClearVarDictionary():
	global vardictionary
	del vardictionary
	

#-------------------------------------------------------------------------------------
#
# Read/Write Variable Dictionary for free roaming campaigns
#
#-------------------------------------------------------------------------------------	

def AddCampaignVar(var):
	campvars = GetCampaignVars()
	if not var in campvars:
		campvars.append(var)
		SetCampaignVars(campvars)

def RemoveCampaignVar(var):
	campvars = GetCampaignVars()
	if var in campvars:
		campvars.remove(var)
		SetCampaignVars(campvars)
	
def SetCampaignVars(var):
	StoreVar('__campaign_vars', var)
	
def GetCampaignVars():
	return RestoreVar('__campaign_vars',default=[])

	
def WriteVarDictionary(clearfile=0):
	print 'WRITING VAR DICTIONARY'
	scriptpath = os.path.join(datfiles.GetModPath(),'Missions','Scripts')
	fname = os.path.join(scriptpath,'m25campaigntemp.py')
	file = open(fname,'w')
	#WARNING: I'm assuming that no values need to be stored at the start of a freeroam campaign
	#so the file will be blank
	if not clearfile:
		keys = GetCampaignVars()
		for k in keys:
			value = RestoreVar(k,NOT_FOUND)
			if value != NOT_FOUND:
				file.write('%s=%s\n'%(k,repr(value))) 	
# 		keys = vardictionary.keys()
# 		for k in keys:
# 			if vardictionary[k].has_key('campaign'):
# 				if vardictionary[k]['campaign']:
# 					file.write(k + '=' + `vardictionary[k]['value']` + '\n')
	else:
		file.write('started_campaign = ' + `time.clock` + '\n')
	file.close()
	
def GetSet_GetObjectAndVar(name):
	num = 0
	for n in name:
		if n == ':':
			return [name[:num],name[num:]]
		num = num + 1
	return [None,None]
	
def ReadVarDictionary():
	scriptpath = os.path.join(datfiles.GetModPath(),'Missions','Scripts')
	fname = os.path.join(scriptpath,'m25campaigntemp.py')
	all = []
	try:
		#print 'trying read of var dictionary'
		file = open(fname,'r')
		all = file.readlines()
		file.close()
	except:
		print 'm25getset.ReadVarDictionary(): file m25campaigntemp.py does not exist'
		return
	for expr in all:
		var = m25string.GetLeftString(expr,'=')
		value = m25string.GetRightString(expr,'=')
		#vardictionary[var] = {}
		exec 'val = getset_destring('+value+')'
		object, objectvar = GetSet_GetObjectAndVar(var)
		if object == None:
			SetMissionVar(var, val, campaign=1)
		else:
			SetObjectVar(object,newvar,val,campaign=1)
			
# 		if var[:9] == '_mission_':
# 			SetMissionVar(var[9:], val, campaign=1)
# 		else:
# 			object, newvar = GetSet_GetObjectAndVar(var)
# 			SetObjectVar(object,newvar,val,campaign=1)
		
		#print var, val
		#vardictionary[var]['value'] = val
		#vardictionary[var]['campaign'] = 1

def getset_destring(arg):
	return arg


#-------------------------------------------------------------------------------------
#
# Read/Write Variables to save game
#
#-------------------------------------------------------------------------------------	
	
import chunkstring

def StoreVar(name, var):
	missionobjvar.Mission_SetVar(name,var)
	return

def RestoreVar(name, default=None):
	if not missionobjvar.Mission_VarExists(name): return default
	return missionobjvar.Mission_GetVar(name)
	
	
def listtotuple(value):
	pass	
	

def StoreVarDictionary():
	start = time.clock()
	global vardictionary
	value = chunkstring.string2numtuple(repr(vardictionary))
	num = 1
	for v in value:
		js.Mission_SetAttr('_mvd%i'%num, v)
		num = num + 1
	js.Mission_SetAttr('_mvd_len',len(value))
	print 'Store took: %f'%(time.clock()-start)
	
def RestoreVarDictionary():
	start = time.clock()
	global vardictionary
	length = js.Mission_TestAttr('_mvd_len')
	value = []
	if length:
		for i in range(1,length+1):
			value.append(js.Mission_GetAttr('_mvd%i'%i))
		exec 'vardictionary = getset_destring(%s)'%(chunkstring.numtuple2string(tuple(value)))
	else:
		vardictionary = {}
	#print vardictionary
	print 'Restore took: %f'%(time.clock()-start)

	
