#####################################################################################
#
# Module: m25randomenc
# Author: m25
# 
# Description: Generate random lists of villains for encounters
#
# Version 3.053
#
#
#####################################################################################

#print "Loading m25randomenc.py v.3.2.0 final; branch = m25 branch 1"

import ff
import js
import random
import types
import string

import freeroam_native
import freeroam_customs
import freeroam_level

import m25ai
import m25name		#generate unique names
import m25getset	#store variables
import m25gametype	#detect which type of game we are playing
from m25gametype import *
import m25string
import m25report

from ffxtraits import *

m25report.Report_SetChannel('RANDOMENC',1)

def RandomEnc_GetAllHeroes():
	objects = js.Mission_GetDynamicObjects()
	heroes = []
	for obj in objects:
		if js.Object_GetClass(obj)&js.OC_CONTROLLABLE:
			try:
				if not js.Mission_GetAttr(obj+'_fakehero'):
					heroes.append(obj)
			except:
				heroes.append(obj)
	return heroes
        
# short circuiting these functions to improve speed	
	
def RandomEnc_CalcHeroSquadPrestige():
	return len(RandomEnc_GetAllHeroes()) * ff.RPG_GAME_DIFFICULTY
	
	prestige = m25getset.GetMissionVar('mission_hero_prestige',default=None)
	if prestige == None:
		heroes = RandomEnc_GetAllHeroes()
		prestige = 0
		for h in heroes:
			prestige = prestige + js.Object_CalcPrestige(h)
		m25getset.SetMissionVar('mission_hero_prestige',prestige)
	print 'Calcherosquadprestige has prestige ' + `prestige`
	return prestige
	
def RandomEnc_CalcMinionPrestige(minion):
	return 1
	
	prestige = m25getset.GetMissionVar('mission_'+minion+'_prestige',default=None)
	if prestige == None:
		name = m25name.GetNextObjectName()
		#spawn them out of the way and do calculations
		js.Object_SpawnAt(minion,(-10000,-10000,0),name)
		#cshelper.spawn(name,minion,(-10000,-10000,0))
		prestige = js.Object_CalcPrestige(name)
		js.Object_Destroy(name)
		m25getset.SetMissionVar('mission_'+minion+'_prestige',prestige)
	print 'Calcminionprestige ' + minion + ' has prestige ' + `prestige`
	return prestige
	
	
def RandomEnc_MakeMinionList(minions):
	if minions == []: return []
	minion_list = []
	hero_prestige = float(RandomEnc_CalcHeroSquadPrestige())
	total_prestige = 0
	while (total_prestige < hero_prestige):
		i = random.randint(0,len(minions)-1)
		selected = minions[i]
		if type(minions[i]) == types.ListType:
			selected = minions[i][1]
		#note that the checksubtype doesn't really work here, 
		# instead it filters out any non-existent templates
		# e.g. it doesn't make sense to try to spawn a 'wise' minion when the set of all minions is undefined
		#if not RandomEnc_CheckSubtype(selected, selected): selected = 'thug_with_bat'
		prestige = RandomEnc_CalcMinionPrestige(selected)
		minion_list.append(selected)
		total_prestige = total_prestige + prestige
		m25report.Report('RANDOMENC', 'selected %s : total prestige %i'%(selected, total_prestige))
		#print 'RandomEnc_MakeMinionList selected ' + selected + ' : total_prestige ' + `total_prestige`
	return minion_list

	
def RandomEnc_GetRandomEnemy(subtype=None):
	if m25gametype.GameType_IsFreeRoam():
		return RandomEnc_GetVillains(subtype)
	if m25gametype.GameType_IsCampaign():
		return RandomEnc_GetVillains(subtype)
	#in a skirmish game
	return RandomEnc_GetRandomUserEnemy(subtype)
	
def RandomEnc_CheckSubtype(template, subtype):
	if not subtype: return 1
	test_template = m25string.String_ConvertToTemplate(subtype)
	if template == test_template: 
		return 1
	if m25ai.AI_TemplateHasUserSubtype(template, subtype): 
		return 1
	for traitlist in ffxTraits:
		if subtype == string.lower(traitlist[0]):
			try:
				if js.Template_GetSpeechID(template) in traitlist[1:]:
					return 1
			except:
				pass
	return 0
	
def RandomEnc_GetRandomUserEnemy(subtype=None):
	storedtemplates = m25getset.GetMissionVar('story_enemytemplates',default=[])
	enemy = 'thug_with_bat'
	found, num = 0, 0
	if storedtemplates != []:
		while not found and (num < 50):
			random_pick = random.randint(0,len(storedtemplates)-1)
			enemy = storedtemplates[random_pick]
			if RandomEnc_CheckSubtype(enemy, subtype): 
				storedtemplates.remove(enemy)
				m25getset.SetMissionVar('story_enemytemplates',storedtemplates)
				found = 1
			num = num + 1
	if not found: enemy = 'thug_with_bat'
	m25report.Report('RANDOMENC',' Random user enemy selected %s'%(enemy))
	#print 'GetNextUserEnemy: ' + `enemy`
	return enemy
	
def RandomEnc_GetVillains(subtype=None):
	prestige_value = RandomEnc_CalcHeroSquadPrestige()
	possible_encounters = []
	for set in freeroam_native.encounterlist: possible_encounters.append(set)
	for set in freeroam_customs.customVillains: possible_encounters.append(set)
	found, num = 0, 0
	while not found and (num <50):
		found = random.randint(0,len(possible_encounters)-1)
		if not RandomEnc_ValidVillains(possible_encounters[found],prestige_value):
			found = 0
		if not RandomEnc_CheckSubtype(possible_encounters[found][0], subtype):
			found = 0
		num = num + 1
	if not found:
		return 'thug_with_bat'
	#create the list
	set = possible_encounters[found]
	villains = ['_set']
	#villain is first, villain cost second, condition is at the end, minion/minion cost in the middle
	total_prestige = 0
	villains.append(set[0])
	#find the minions
	minion_list = []
	index = 2
	while index < len(set)-1:
		minion_list.append(set[index])
		#store the prestige cost
		m25getset.SetMissionVar('mission_'+set[index]+'_prestige',set[index+1])
		index = index+2
	#randomly choose minions until our cost is used up
	mlist = RandomEnc_MakeMinionList(minion_list)
	for m in mlist: villains.append(m)
	return villains

def RandomEnc_ValidVillains(set, prestige_value):
	#make sure the villain isn't too powerful
	cost = set[1]
	#print cost, prestige_value*2
	#if cost > prestige_value*1.5:
	#	return 0		
	#now make sure the spawn condition is satisfied		
	condition = set[-1]
	valid = 1
	try:
		exec 'valid = '+ condition
	except:
		#print 'error in condition ' + `condition`
		valid = 0
	return valid
