#####################################################################
#
# Storylines
# 
# Represent missions as state machines.
# This file keeps track of what stage the mission is at and
# where to go next.
#
# Added support for free roaming campaigns
#
# Allows such wonders as auto-generated missions and the 'EZ mission scripter'.
# 
# Version 3.052
#
#
#####################################################################

#print "Loading m25story.py v.3.2.0 final; branch = m25 branch 1"

from random import randint

import ff
from ff import *
import js

import m25ai
import m25enc
from m25enc import *
import m25string
import m25report
import m25getset

import m25event

import m25gametype
import m25randomenc
import m25missionover

import m25gametype
import m25cutscene

import string

STORY_CHANNEL = 'STORY'
m25report.Report_SetChannel(STORY_CHANNEL,0)


#-------------------------------------------------------------------
#
# Story Utility functions
#
#-------------------------------------------------------------------

def GetStoryline():
	s = m25getset.GetMissionVar('_story',default=None)
	if s != None: return s
	try: 
		s = ff.Story
		m25getset.SetMissionVar('_story', s, campaign=0)
		return s
	except:
		return {}
	
def SetStoryline(story, nosave=0):
	m25getset.SetMissionVar('_story',story,campaign=0,nosave=nosave)		
		
def SaveStoryline():
	s = GetStoryline()
	SetStoryline(s)	

def isEncounterFunction(enc):
	if m25string.IsFunction(enc):
		if enc[0:5] == 'DoEnc':
			return 1
	return 0
		
def GenerateStorylineEncounterID(storyline, block):
	return m25enc.EncGetNextEncID()
	
#-------------------------------------------------------------------
#
# Story Variables
#
#-------------------------------------------------------------------				

def GetStoryVar(story,var,default=None):
	st = GetStoryline()[story]['variables']
	if st.has_key(var):
		return st[var]
	else:
		return default
		
def SetStoryVar(story,var,value):
	GetStoryline()[story]['variables'][var] = value
	SaveStoryline()

def SetGlobalStoryVar(var, value, default=None):
	if not GetStoryline().has_key('global'):
		GetStoryline()['global'] = {}
	GetStoryline()['global'][var] == value
	SaveStoryline()
	
def GetGlobalStoryVar(var, default=None):
	if not GetStoryline().has_key('global'):
		return default
	if GetStoryline()['global'].has_key(var):
		return GetStoryline()['global'][var]
	return default	
	
	
#-------------------------------------------------------------------
#
# Tracking active encounters within a story
#
#-------------------------------------------------------------------	
	
def GetActiveEncounters(story):
	return m25getset.GetMissionVar('_story_'+story+'_active_list', default=None)

def SetEncounterActive(story, encounter):
	if IsStoryTemporary(story): return
	active = m25getset.GetMissionVar('_story_'+story+'_active_list', default=None)
	if active == None:
		active = [encounter]
	else:
		active.append(encounter)
	camp = 1
	if IsStoryTemporary(story):
		camp = 0
	m25getset.SetMissionVar('_story_'+story+'_active_list', active, campaign=camp)
	m25report.Report(STORY_CHANNEL, 'Starting story ' + `active`)

def IsEncounterActive(story, encounter):
	active = m25getset.GetMissionVar('_story_'+story+'_active_list', default=[])
	return encounter in active

def StopEncounterActive(story, encounter):
	if IsStoryTemporary(story): return
	active = m25getset.GetMissionVar('_story_'+story+'_active_list', default=None)
	if active == None:
		return
	num = 0
	for a in active:
		if a == encounter:
			del active[num]
			camp = 1
			if IsStoryTemporary(story):
				camp = 0
			m25getset.SetMissionVar('_story_'+story+'_active_list', active, campaign=camp)
			return
		num = num + 1

def IsStoryTemporary(story):
	if GetStoryline()[story].has_key('temporary'):
		if GetStoryline()[story]['temporary'] == 1:
			return 1
	return 0		
	

#-------------------------------------------------------------------
#
# Setting story encounter results
#
#-------------------------------------------------------------------	
					
def Story_SetEncounterResult(story, encounter, result):
	camp = 1
	if IsStoryTemporary(story): camp = 0
	m25getset.SetMissionVar('_story_'+m25string.RemoveSpaces(story)+'_enc_'+m25string.RemoveSpaces(encounter)+'_res', result, campaign=camp)
	
def Story_GetEncounterResult(story, encounter, default=[]):
	return m25getset.GetMissionVar('_story_'+m25string.RemoveSpaces(story)+'_enc_'+m25string.RemoveSpaces(encounter)+'_res', default=default)
	

#-------------------------------------------------------------------
#
# Setting story ended
#
#-------------------------------------------------------------------	
					
def Story_SetStoryEnded(story):
	camp = 1
	if IsStoryTemporary(story): camp = 0
	m25getset.SetMissionVar('_story_'+m25string.RemoveSpaces(story)+'_ended', 1, campaign=camp)
	
def Story_GetStoryEnded(story, default=None):
	return m25getset.GetMissionVar('_story_'+m25string.RemoveSpaces(story)+'_ended', default=default)
			
	
#-------------------------------------------------------------------
#
#  Story Add/Remove
#
#-------------------------------------------------------------------

def Storyline_AddStory(story):
	temp = 1
	newstory = GetStoryline()
	for name in story.keys():
		if newstory.has_key(name):
			m25report.Report(STORY_CHANNEL, 'Story ' + name + ' already exists in global story dictionary and will not be added.')
			return
		newstory[name] = story[name]
		if not story[name].has_key('temporary'):
			temp = 0
	SetStoryline(newstory,nosave=1)
	#if not temp:
	#	SaveStoryline()

def Storyline_RemoveStory(name):
	if GetStoryline().has_key(name):
		del GetStoryline[name]
		#SaveStoryline()
	else:
		m25report.Report(STORY_CHANNEL, 'Cannot remove story ' + name + ' as it does not exist in dictionary.')

def Storyline_HasStory(name):
	if GetStoryline().has_key(name):
		return 1
	return 0

#-------------------------------------------------------------------
#
# Running Stories
#
#-------------------------------------------------------------------	


def DoStoryline():
	all_stories = GetStoryline()
	for story in all_stories.keys():
		if not m25getset.GetMissionVar('story_%s_running'%story, default=0):
			m25getset.SetMissionVar('story_%s_running'%story, 1)
			#first check to see if the story has already been started
			active = GetActiveEncounters(story)
			if active != None:
				m25report.Report(STORY_CHANNEL, 'Restoring active story ' + story)
				for enc in active:
					DoStorylineState(story,enc,already_active=1)
			#otherwise start the story from the beginning
			else:
				m25report.Report(STORY_CHANNEL, 'Starting new story ' + story)
				for enc in all_stories[story]['encounters'].keys():
					e = all_stories[story]['encounters'][enc]
					if e.has_key('trigger') and ('start' in e['trigger']) and CheckStoryTrigger(story,enc,start=1):
						DoStorylineState(story,enc)
						#m25event.SetTimer(randint(1,3)*0.1,'DoStorylineStateEvent',story,enc)				
					
def DoStorylineStateEvent(event):
	story = event.object
	state = event.string
	DoStorylineState(story, state, event.user)
	
def DoStorylineObjective(story, state):
	enc = GetStoryline()[story]['encounters'][state]
	if enc.has_key('action'):
		for a in enc['action']:
			if isEncounterFunction(a):
				val, first, remainder = m25string.SplitAroundPhrase(a,'objectives = ')
				if val:
					objectives = m25string.TransList(remainder)
					for e in objectives:
						#check for silent objectives (used in freeroam to give xp and prestige without creating actual new objectives)
						if (e[0] > -1) and (e[1] != ''):
							m25objective.Objective_AddObjective(objective=e[1], prestige=e[2], objectivetype=e[0])
					
							
def DoStorylineState(story, state,already_active=0):
	m25report.Report(STORY_CHANNEL, 'Trying encounter ' + story + ' : ' + state)
	#don't try to run non-existent states
	if not GetStoryline()[story]['encounters'].has_key(state): return
	#don't try to run if the story has ended
	if Story_GetStoryEnded(story): return
	if not already_active:
		SetEncounterActive(story,state)
	m25report.Report(STORY_CHANNEL, 'Starting encounter ' + story + ' : ' + state)
	if CheckStoryTrigger(story, state, start=1, briefing=1, intermission=1):
		if not CheckStoryPreconditions(story,state):
			m25report.Report(STORY_CHANNEL, 'pre conditions failed for %s.'%(state))
			#even if the preconditions fail, we still want the objective to be shown in a freeroam mod
			if m25gametype.GameType_IsFreeRoam():
				DoStorylineObjective(story, state)
			return
	else:
		m25report.Report(STORY_CHANNEL, 'trigger not satisfied for %s.'%(state))
		return
	enc = GetStoryline()[story]['encounters'][state]
	has_encounter = 0
	return_val = []
	if enc.has_key('action'):
		action = enc['action']
		for a in action:
			if isEncounterFunction(a):
				enc_id = GenerateStorylineEncounterID(story,state)
				if not StorySetupEncounter(enc_id,story,state,a):
					m25report.Report(STORY_CHANNEL, 'Encounter did not initialise ' + story + ' : ' + state)
					return_val = ['error']
					m25event.SetTimer(randint(10,20),'DoStorylineStateEvent',story,state,user=1)
					return
				else:
					has_encounter = 1
			else:
				exec 'ff.' + a
	if not has_encounter:
		m25getset.SetMissionVar('story_enc_'+`int(0)`,[story,state])
		StorylineEncounterReturn(0,return_val,[])
			
	
def CheckStoryPreconditions(story,state):
	if not Story_CheckGameType(story,state): return 0
	if not GetStoryline()[story]['encounters'][state].has_key('pre'):
		return 1  #if no conditions
	pre = GetStoryline()[story]['encounters'][state]['pre']
	for t in pre:
		if m25string.IsFunction(t):
			val = 0
			try:
				exec 'val = ff.' + t
				if not val:
					return 0
			except:
				m25report.Report(STORY_CHANNEL, 'ERROR in trigger for ' + story + ' : ' + state + ' -> ' + trigger)
	return 1	
	
def Story_CheckGameType(story,state):
	#only check when in the freeroam game mode
	if not m25gametype.GameType_IsFreeRoam(): return 1
	#automatically pass the test if there are no gametypes to check for
	if not GetStoryline()[story]['encounters'][state].has_key('required_map'):
		return 1	
	#try out the gametypes
	gametypes = GetStoryline()[story]['encounters'][state]['required_map']
	check = 0
	if gametypes == []: check = 1
	for gt in gametypes:
		val = 0
		val = m25gametype.GameType_CheckMap(gt)
		if val == 1: 
			check = 1
	if check:
		if GetStoryline()[story]['encounters'][state].has_key('timeofday'):
			timeofday = GetStoryline()[story]['encounters'][state]['timeofday']
			if timeofday == 'day': return m25gametype.GameType_isDay()
			else:
				if timeofday == 'night':
					return m25gametype.GameType_isNight()
		return 1
	return 0
			
def CheckStoryTrigger(story,state,start=0,briefing=0,intermission=0):		
	if not GetStoryline()[story]['encounters'][state].has_key('trigger'):
		return 1  #if no conditions
	trigger = GetStoryline()[story]['encounters'][state]['trigger']
	for t in trigger:
		if m25string.IsFunction(t):
			val = 0
			try:
				exec 'val = ff.' + t
				if not val:
					return 0
			except:
				m25report.Report(STORY_CHANNEL, 'ERROR in trigger for ' + story + ' : ' + state + ' -> ' + trigger)
		elif (string.lower(t) == 'start'):
			#only trigger encounters with 'start' in them if we are actually at the start of the given story
			if not start:
				return 0
		elif (string.lower(t) == 'briefing'):
			if not briefing:
				return 0
		elif (string.lower(t) == 'intermission'):
			if not intermission:
				return 0
		else:
			if not TryStoryTrigger(story,t):
				return 0
	return 1

		
def StorySetupEncounter(enc_id,story,state,a, retries=0):
	m25getset.SetMissionVar('story_enc_'+`int(enc_id)`,[story,state])
	encname = m25string.GetLeftString(a,'(')
	encparams = m25string.GetRightString(a,'(')[:-1]
	val = 0
	try:
		val = eval('ff.%s(%i,%s,return_function="StorylineEncounterReturn")'%(encname,enc_id,encparams))
	except:
		m25report.Report('ERROR', 'The encounter %s does not exist or is broken.  You may need to download the appropriate plug-in, or check script.log for errors.'%(encname))
		print 'ff.%s(%i,%s,return_function="StorylineEncounterReturn")'%(encname,enc_id,encparams)
		raise
	#exec 'val = ff.' + encname + '(' + `enc_id` + ',' + encparams + ',' + 'return_function="StorylineEncounterReturn")'
	if val == 0 and (retries > 1):
		m25report.Report(STORY_CHANNEL, 'Encounter did not init: retrying...')
		return StorySetupEncounter(enc_id,story,state,a, retries=retries-1)
	else:
		 if val == 0:
			m25report.Report(STORY_CHANNEL, 'Encounter did not initialize!')
			return 0
	return 1


def StorylineEncounterReturn(enc_id, return_code, return_list):
	#print enc_id, return_code, return_list
	story, encounter = m25getset.GetMissionVar('story_enc_'+`int(enc_id)`,default=[None,None])
	# if we are returning from a briefing, then simply end the story
	if m25cutscene.CS_GetInBriefing():
		js.UI_Briefing_Done()
		return
	if story == None:
		m25report.Report(STORY_CHANNEL, 'Encounter returned, but no story set.')
		return
	# set the encounter as no longer active
	if 'keep_active' not in return_list:
		StopEncounterActive(story,encounter)
	
	#IF WE ARE IN A BASE SCREEN, DON'T RUN ADDITIONAL ENCOUNTERS - IT IS TOO EASY
	# FOR THE USER (or me for that matter) TO MESS THIS UP, SO DON'T ALLOW IT
	if m25gametype.GameType_InBase():
		return
	
	enc = GetStoryline()[story]['encounters'][encounter]
	#mark the encounter as ended with the return_code
	Story_SetEncounterResult(story, encounter, return_code)

	#perform any functions in the 'completed' section of the state
	ExecEncComplete(story, encounter, enc_id, return_code, return_list)
	
	#start any states that are directly triggered by this transition
	if enc.has_key('next'):
		for nextstate in enc['next']:
			#print nextstate
			if m25string.IsFunction(nextstate):
		 		newstate = ''
		 		exec 'newstate = ff.' + nextstate
		 		#DoStorylineState(story, newstate)
		 		if newstate not in ['',[],None]:
		 			m25event.SetTimer(randint(1,3)*0.1,'DoStorylineStateEvent',story,newstate)
		 	elif type(nextstate) == types.ListType:
		 		if Story_CodesMatch(nextstate[0],return_code):
		 			if type(nextstate[1]) == types.ListType:
		 				for ns in nextstate[1]:
		 					m25event.SetTimer(randint(1,2), 'DoStorylineStateEvent', story, ns)
		 			else:
		 				m25event.SetTimer(randint(1,3)*0.1, 'DoStorylineStateEvent', story, nextstate[1])
		 	else:
		 		#DoStorylineState(story, nextstate)	
		 		m25event.SetTimer(randint(1,3)*0.1,'DoStorylineStateEvent',story,nextstate)	
	#now try to find other encounters that are unlocked as a result of this one completing
	DoStoryLineTransition(story, encounter, return_code, return_list)
	#check to see if it is all over
	if Story_AllEnded(): Story_EndMission()

def Story_CodesMatch(required_code, return_code):
	if type(required_code) == types.ListType:
		for c in required_code:
			if not (string.lower(c) in return_code): 
				return 0
		return 1
	else:
		return required_code in return_code
	
def ExecEncComplete(story, encounter, enc_id, return_code, return_list):
	if GetStoryline()[story]['encounters'][encounter].has_key('complete'):
		enc = GetStoryline()[story]['encounters'][encounter]['complete']
		for code in enc.keys():
			if code in return_code:
				for e in enc[code]:
					if e[0:3] == 'Enc':
						try:
							exec 'ff.' + m25string.GetLeftString(e,'(') + `enc_id` + ',' + `return_code` + ',' + `return_list` + ',' + m25string.GetRightString(e,'(')
						except:
							m25report.Report(STORY_CHANNEL, 'Bad completion function ' + story + ' : ' + enc + ' -> ' + e)
					else:
						try:
							exec 'ff.' + e
						except:
							m25report.Report(STORY_CHANNEL, 'Bad completion function ' + story + ' : ' + enc + ' -> ' + e)


def DoStoryLineTransition(story, encounter, return_code, return_list):
	for enc in GetStoryline()[story]['encounters'].keys():
		#Note: you cannot start an encounter this way if it is already active, or has completed
		# otherwise they would re-run every time another encounter completed.
		if not IsEncounterActive(story, enc) and not Story_GetEncounterResult(story, enc, default=None):
			if GetStoryline()[story]['encounters'][enc].has_key('trigger'):
				if CheckStoryTrigger(story,enc,start=0,briefing=0,intermission=0):
					m25event.SetTimer(1,'DoStorylineStateEvent',story,enc)


def TryStoryTrigger(story, trigger):
	if type(trigger) == types.ListType:
		return TryStoryTriggerReturn(story, trigger)
	val, left, right = m25string.SplitAroundPhrase(trigger, ' and ')
	if val: return TryStoryTrigger(story,left) and TryStoryTrigger(story, right)
	val, left, right = m25string.SplitAroundPhrase(trigger, ' or ')
	if val: return TryStoryTrigger(story,left) or TryStoryTrigger(story, right)
	if string.lower(string.lstrip(trigger[:4])) == 'not ':
		return not TryStoryTrigger(story, trigger[4:])
	val, enc, return_code = m25string.SplitAroundPhrase(trigger, ' not at ')
	if val: return not TryStoryTriggerReturn(story, [string.strip(enc), string.strip(return_code)])
	val, enc, return_code = m25string.SplitAroundPhrase(trigger, ' at ')
	if val: return TryStoryTriggerReturn(story, [string.strip(enc), string.strip(return_code)])
	try:
		val = eval('ff.%s(%s)'%(trigger,story))
		return val
	except:
		return 0
							
def TryStoryTriggerReturn(story, trigger):
	required_enc = trigger[0]
	required_value = trigger[1]
	return_code = []
	encounters = GetStoryline()[story]['encounters']
	if encounters.has_key(required_enc):
		return_code = Story_GetEncounterResult(story, required_enc, default=[])
	#print story, trigger, return_code
	if return_code and (string.lower(string.strip(required_value)) in return_code):
		return 1
	return 0

	
#-------------------------------------------------------------------
#
#  Choose random elements from a list
#
#-------------------------------------------------------------------

def Story_GenerateRandomList(num_elements, choices):
	newlist = []
	num = 1
	while num <= num_elements:
		c = choices[randint(0,len(choices)-1)]
		newlist.append(c)
		num = num + 1
	return newlist

def Story_RandomListElement(the_list):
	return the_list[randint(0,len(the_list)-1)]
	
def Story_RandomizeList(the_list):
	newlist = []
	templist = the_list[0:]
	while len(templist) > 0:
		newlist.append(templist.pop(randint(0,len(templist)-1)))
	return newlist
	
	
#-------------------------------------------------------------------
#
# Return a random enemy either from those chosen at the
# rumble room screen, or placed in freeroam_native and 
# freeroam_custom
#
#-------------------------------------------------------------------	


def Story_GetRandomUserEnemy(subtype=None):
	return m25randomenc.RandomEnc_GetRandomEnemy(subtype)
	
	
#-------------------------------------------------------------------
#
# Working with Campaign based Stories
#
#-------------------------------------------------------------------	

def Story_DoBriefing():
	found = 0
	all_stories = GetStoryline()
	for story in all_stories.keys():
		m25report.Report(STORY_CHANNEL, 'Starting briefing ' + story)
		for enc in all_stories[story]['encounters'].keys():
			e = all_stories[story]['encounters'][enc]
			if e.has_key('trigger') and (('briefing' in e['trigger']) or ('Briefing' in e['trigger'])):
				DoStorylineState(story,enc)
				#m25event.SetTimer(randint(1,3)*0.1,'DoStorylineStateEvent',story,enc)
				found = 1
	return found


def Story_DoIntermission():
	found = 0
	all_stories = GetStoryline()
	for story in all_stories.keys():
		m25report.Report(STORY_CHANNEL, 'Starting intermission ' + story)
		for enc in all_stories[story]['encounters'].keys():
			e = all_stories[story]['encounters'][enc]
			#if e.has_key('trigger'): print e['trigger']
			if e.has_key('trigger') and (('intermission' in e['trigger']) or ('Intermission' in e['trigger'])):
				DoStorylineState(story,enc)
				#m25event.SetTimer(randint(1,3)*0.1,'DoStorylineStateEvent',story,enc)
				found = 1
	return found
	
	
#-------------------------------------------------------------------
#
# See if all stories are done
#
#-------------------------------------------------------------------	

def Story_AllEnded():
	storyline = GetStoryline()
	for story in storyline.keys():
		if not Story_GetStoryEnded(story):
			return 0
	return 1

def Story_EndMission():
	if not m25gametype.GameType_IsFreeRoam():
		m25missionover.SetMissionOver()
		js.Mission_Win()
		
def Story_WinMission():
	m25missionover.SetMissionOver()
	Story_EndMission()
	
def Story_LoseMission():
	if not m25gametype.GameType_IsFreeRoam():
		m25missionover.SetMissionOver()
		js.Mission_Lose()

