#m25team.py

#####################################################################################
#
# Teams
# 
# Description: set up various teams and their reactions to each other (allied, enemy, ect.)
#
#####################################################################################		

#print "Loading m25team.py v.3.2.0 final; branch = m25 branch 1"

#NOTE:  Call  SetDefaultTeamAlliances() to initialize the teams
#	this is done in InitScripts()

#Use:  
# To setup the standard team after spawning a character: 
#	SetDefaultTeam(char)

# To create two teams of NPCs that will fight each other:
#	SetTeam([list of characters in one team],2)
#	SetTeam([list of characters in other team],3)
#	SetTeamsEnemy(2,3)
#	SetBrawler([list of all charaters on both teams])


import js
import m25report
from m25getset import *

#for use with the report module
TEAM_CHANNEL = 'TEAM'

# team constants
TEAM_NONCOMBATANT = 0
TEAM_GOODGUYS = 1
TEAM_BADGUYS = 2
TEAM_ALWAYSHOSTILE = -1
TEAM_ENEMIES = 'enemy'
TEAM_ALLIES = 'friend'
TEAM_NEUTRAL = 'neutral'
TEAM_ANY = 'any'

TEAM_REACTION_ENEMIES = 1
TEAM_REACTION_NEUTRAL = 0
TEAM_REACTION_ALLIES = 2

Global_RegisterAttr('ai_team_initialized', 0)
Global_RegisterAttr('ai_team', 0)


#-------------------------------------------------------------------------------------
#
# Team status
#
#-------------------------------------------------------------------------------------	

def SetTeam(object, team):
	if team == -1: raise
	m25report.Report(TEAM_CHANNEL,object + ' set to team ' + `team`)
	if type(object) is types.ListType:
		for obj in object:
			js.Object_SetAttr(obj,'ai_team',team)
			#SetObjectVar(obj, 'team', team)
			MarkTeamSet(obj)
	else:
		js.Object_SetAttr(object, 'ai_team',team)
		#SetObjectVar(object, 'team', team)
		MarkTeamSet(object)

def GetTeam(object,default=TEAM_NONCOMBATANT):
	return int(js.Object_GetAttr(object, 'ai_team'))
	#return int(GetObjectVar(object, 'team', default=default, integer=1))
	
def SetDefaultTeam(object):
	objclass = js.Object_GetClass(object)
	if js.Object_GetAttr(object,'physical'):
		if (objclass & OC_CONTROLLABLE) or (objclass & OC_POLICE):
			SetTeam(object,TEAM_GOODGUYS)
		elif (objclass & OC_MINION) or (objclass & OC_VILLIAN):
			SetTeam(object,TEAM_BADGUYS)
		else:
			SetTeam(object,TEAM_NONCOMBATANT)
	else:
		SetTeam(object,TEAM_NONCOMBATANT)
	MarkTeamSet(object)

def IsNonCombatant(object):
	team = GetTeam(object)
	if team == TEAM_NONCOMBATANT:
		return 1
	return 0

def IsTeamSet(object):
	return js.Object_GetAttr(object,'ai_team_initialized')
	#return GetObjectVar(object,'team_initialized',default=0,integer=1)

def MarkTeamSet(object):
	if not IsTeamSet(object):
		js.Object_SetAttr(object,'ai_team_initialized',1)
		#SetObjectVar(object,'team_initialized',1)	

#-------------------------------------------------------------------------------------
#
# Allies and Enemies
#
#-------------------------------------------------------------------------------------		
	
_team_reaction = {}

def SetDefaultTeamAlliances():
	SetTeamsEnemy(TEAM_GOODGUYS, TEAM_BADGUYS)
	SetTeamsNeutral(TEAM_GOODGUYS, TEAM_NONCOMBATANT)
	SetTeamsNeutral(TEAM_BADGUYS, TEAM_NONCOMBATANT)
	SetTeamsEnemy(TEAM_GOODGUYS, TEAM_ALWAYSHOSTILE)
	SetTeamsEnemy(TEAM_BADGUYS, TEAM_ALWAYSHOSTILE)		

def SetTeamsAllied(team1, team2):
	m25report.Report(TEAM_CHANNEL, 'Ally reaction: ' + `team1` + ' and ' + `team2`)
	SetTeamsReaction(team1,team2,TEAM_ALLIES)
	#SetMissionVar('reaction_' + `team1` + '_' + `team2`, TEAM_ALLIES)
	#SetMissionVar('reaction_' + `team2` + '_' + `team1`, TEAM_ALLIES)
	
def SetTeamsEnemy(team1, team2):
	m25report.Report(TEAM_CHANNEL, 'Enemy reaction: ' + `team1` + ' and ' + `team2`)
	SetTeamsReaction(team1,team2,TEAM_ENEMIES)
	#SetMissionVar('reaction_' + `team1` + '_' + `team2`, TEAM_ENEMIES)
	#SetMissionVar('reaction_' + `team2` + '_' + `team1`, TEAM_ENEMIES)
	
def SetTeamsNeutral(team1, team2):
	m25report.Report(TEAM_CHANNEL, 'Neutral reaction: ' + `team1` + ' and ' + `team2`)
	SetTeamsReaction(team1,team2,TEAM_NEUTRAL)
	#SetMissionVar('reaction_' + `team1` + '_' + `team2`, TEAM_NEUTRAL)
	#SetMissionVar('reaction_' + `team2` + '_' + `team1`, TEAM_NEUTRAL)	
	
def SetTeamsReaction(team1, team2, reaction):
	global _team_reaction
	#js.Mission_SetAttr('reaction_%i_%i'%(team1,team2), reaction)
	#js.Mission_SetAttr('reaction_%i_%i'%(team2,team1), reaction)
	SetMissionVar('reaction_' + `team1` + '_' + `team2`, reaction)
	_team_reaction['reaction_%i_%i'%(team1,team2)] = reaction
	SetMissionVar('reaction_' + `team2` + '_' + `team1`, reaction)	
	_team_reaction['reaction_%i_%i'%(team2,team1)] = reaction
	
def Team_TranslateReaction(reaction):
	if type(reaction) == types.StringType:
		if reaction == 'enemy': return TEAM_REACTION_ENEMIES
		elif reaction == 'friend': return TEAM_REACTION_ALLIES
		elif reaction == 'neutral': return TEAM_REACTION_NEUTRAL
		else: return -1
	else:
		return reaction		
		
def GetTeamsReaction(team1, team2):
	global _team_reaction
	if team1 == team2:
		return TEAM_ALLIES
	else:
		if _team_reaction.has_key('reaction_%i_%i'%(team1,team2)): return _team_reaction['reaction_%i_%i'%(team1,team2)]
		#try:
		#	return int(js.Mission_GetAttr('reaction_%i_%i%'%(team1,team2)))
		#except:
		#	return TEAM_ENEMIES
		val = GetMissionVar('reaction_' + `team1` + '_' + `team2`, default=TEAM_ENEMIES)
		_team_reaction['reaction_%i_%i'%(team1,team2)] = val
		return val

def Team_GetReaction(object1, object2):
	return GetTeamsReaction(GetTeam(object1),GetTeam(object2))		
		
def TeamsReactionOfType(object1, object2, reaction):
	if reaction == TEAM_ANY:
		return 1
	team1 = GetTeam(object1)
	team2 = GetTeam(object2)
	
	#special team -1 is enemy to everyone by default (except team 0 or 1) - used for temporary constructs such as polymorphed victims
	if team1 == -1 or team2 == -1 and (team1 != team2):
		if reaction == TEAM_ENEMIES: 
			if team1 != 0 and team2 != 0 and team1 != 1 and team2 != 1:
				return 1
		return 0
		
	value = GetTeamsReaction(team1, team2)
	if value == reaction:
		return 1
	return 0		
			
	
def AreEnemies(object1, object2):
	return TeamsReactionOfType(object1, object2, TEAM_ENEMIES)
	
def AreAllies(object1, object2):
	return TeamsReactionOfType(object1, object2, TEAM_ALLIES)

def AreNeutral(object1, object2):
	return TeamsReactionOfType(object1, object2, TEAM_NEUTRAL)	
