######################################################################
##
## MLOG Reader Module
## By pimthe and Stumpy, with additional work by M25 and TMX
##
######################################################################
print "importing MLOG Reader 1.0.23" # please update as needed

# This module:
#   - enables the output of additional info to ff.log provided by FF's MLOG functions
#   - continuously reads ff.log to extract said info
#   - allows determination of animation states
#   - allows registering changes of animation states as event sinks
#
# (Top-Level functions descriptions and definitions are below change log.)


#v. 0.9 First public release 2005-04-08

#v. 0.91 Update 2005-04-17:
#   - line number argument removed from MLOG_WatchFlying() and MLOG_WatchDamage()
#   - speed improvements (?) due to use of:
#     - somewhat more conservative/intelligent use of MissionObjVar module
#       (don't forget to use the latest version of said module, as its performances are much improved)
#     - seek() for main loop

#v. 0.92 Update 2005-04-18:
#   - Removed possibly active modules from previously loaded missions during the same game session.

#v. 0.93 Update 2005-05-18 by M25 for FFX 3.00 merged:
#   - Improved the flight detection code and flailing detection added.
#   - Made the flight detection always on, thus making MLOG_IsFlying() and MLOG_IsFlyingCancel() irrelevant.

#v. 0.94 Update 2005-06-24:
#   - By default, every line of ff.log is now emptied every 30 seconds in the hope of improving performances.
#     EDIT: removed, doesn't work! Well, it was a long shot.
#   - Changed tabs back to 4 spaces.
#   - MLOG_IsFlying(): redone as a simple shortcut to GetObjectVar, since the MLOG_WatchFlying() code was completely redone by M25
#     to be mandatory, breaking the old version; also added MLOG_IsFlailing() using the same principle.

#v. 0.95 Update 2005-06-25:
#   - Removed the loop emptying ff.log every 30 seconds, as it didn't work, after all:
#     when the game writes new info to the file (a fraction of second later), if the file isn't the same size as previously, it writes
#     as many null (?) characters to replace the old content.
#   - removed obsolete references to MLOG_IsFlying() in MLOG_RegisterDamage().

#v. 0.96 Update 2005-07-14 by M25 for FFX 3.03 merged:
#   - Added timestamp
#   - Added MLOG_Shutdown()

#v. 0.97 2005-07-17:
#   - Corrected a stupid, stupid, stupid error introduced in v. 0.94 which broke calls to the Animator module
#   - You now define running modules in MLOG_RUNNING_MODULES (see under the "MLOG reader loop" header),
#     instead of in the analysis functions themsselves. Note that the analysis functions themselves
#     still need to be defined directly in MLOG_Read() for performances.

#v. 0.98 2005-07-22:
#   - Did I mention stupid? Bug patch submitted by Stumpy (thanks!) removed by accident when merging to 0.96 final.
#     Stumpy's patch: persistent EVENT_TIMER (used by OnMLOG_Read()) is called way too often (an engine bug), resulting
#     in performance hit; changed to a non-persistent version called again in OnMLOG_Read().
#   - Also according to Stumpy's suggestion, since the Damage module is currently unused, it has been desactivated
#     to improve performances. To enable it again, just uncomment le line in MLOG_RUNNING_MODULES.

#v. 0.981, 0.99 2005-09-13:
#   (stumpy:)
#   - Reworked log reading to speed up I/O and cut down long lags when starting a mission when the game has
#     been running for a while and has a big log file.
#   - Changed MLOG_Shutdown() so that it doesn't try to rewrite the log.ff file on shut down.
#   - Sped up log parsing with calls to chunkstring module.  NB: The import of chunkstring is now mandatory!
#   - Added/Changed watch code for flying, flailing, and landing.  Each now calls an event handler.
#   - Added animation event sinks and register calls so that a callback function can be called automatically
#     when flying, flailing, or landing is detected, rather than polling the object attributes with timers.
#   - Did some cleanup to use newer version of missionobjvar module with default values.  NB: newer version
#     of missionobjvar now required.

#v. 0.991 2005-10-04:
#   (stumpy:)
#   - added line on MLOG_Init() to cancel timer calls to OnMLOG_Read()
#     so that they don't pile up if the player loads saved games.

#v. 0.992 2005-10-28:
#   (stumpy:)
#   - changed first reader to start at 500KB before end of ff.log 

#v. 0.993 2006-02-10:
#   (stumpy:)
#   - Added event sinks for getting up animations: regGetupStart()
#     and regGetupStop()

#v. 0.994 2006-02-11:
#   (stumpy:)
#   - Added an optional string paramater to event registration calls
#     to be passed to the callback functions.

#v. 0.995 2006-02-13:
#   (stumpy:)
#   - Changed flight detection back so jumping and levitating again 
#     register as flying.
#   (pimthe:)
#   - Patched a bug in regMLOGDamage if DamageType wasn't set
#   - Corrected some inconsistant indentation. Wow. Mind-boggling change, isn't it?

#v. 0.996 2006-08-22:
# (m25:)
#  - removed import of m25getset and m25timestamp as they were no longer
#    being used
#  - added idle to list of animations checked for

#v. 1.0 alpha 1 2006-03-05:
#   (pimthe:)
#   - Added Jumping detection as a test of Movement module instead of animator
#   - Abstracted MLOG_Watch...() queue a bit; is the performance hit vs. dynamic management tradeoff worth it?

#v. 1.0 alpha 2 2006-03-05:
#   (pimthe:)
#   - Changed Jumping detection to improve performance.
#   - Flailing now registers as jump end.
#   - Add jump events registration because it's easy when somebody else has already done 99.999% of the job :).
#   - Moved MLOG_Init() after high-level functions, since MLOG_Watch...() functions are now encapsulated in a variable.

#v. 1.0 alpha 3 2006-03-16:
#   (pimthe:)
#   - Since there is now more than one module running, changed MLOG_Watch...() to avoid multiple checks of the same line
#     by different functions unless necessary. For example, 'Movement' module call will probably always be mutually exclusive
#     while 'Damage' won't.
#   - Movement function accordingly changed to avoid redundant code and to return a value.
#   - Added MLOG_IsJumping().
#   - Added functions to add and remove MLOG_Watch...() functions on the fly.

#v. 1.0 alpha 4 2006-04-28:
#   (pimthe:)
#   - Removed all the dynamic queue management code (mostly reverting back to 0.995), as it seemed to impact performances too much.
#   - MLOG_GetChar() redone, as the existing version was apparently completely broken. 

#v. 1.0 alpha 5 2006-07-15:
#   (pimthe:)
#   - Renamed MLOG_RegDamage() and MLOG_RegDamageCancel() to regMLOGDamage() and cancelMLOGDamageCallbackSink() for consistency 
#     with regAnimation() and its variants. Keep in mind however that the parameter format is still different.
#   - Moved generic helpers listAdd(), listRemove() and invertDict() to generalutils.py

#v. 1.0 alpha 6 2006-08-20:
#   (pimthe:)
#   - Jumping detection / Movement module:
#     - moved regJumpStart(), etc. to use regMLOGMovement() instead of parasiting regAnimation()
#     - Disabled jumping detection and the Movement module by default because of the poor useful info/total data produced by
#       that module.
#     - Added MLOGMovement() to enable/disable the module manually.
#     - Added tracking of event registration, so the module is automatically shut down when no callbacks are left.
#   - Solved an issue where Damage registration would be reinitialized when begining a mission from a saved game.
#   - Added some notes on flying vs. jumping vs. flailing.

#v. 1.0 alpha 7 2006-08-23:
#   (pimthe:)
#   - Added notes on the chunkstring module's functions.

#v. 1.0 alpha 8 2006-08-31:
#   (pimthe:)
#   - Added MLOG_RUNNING_<MLOG MODULE NAME> to accelerate the detection of whether the module is active or not. Currently
#     used by Movement only.
#   - Added 'activation' to the list of potentially useful modules.
#   - Completely rewrote Damage code.

#v. 1.0 alpha 8 stumpy branch 1 2006-09-08:
#   - automatic LOG_RUNNING_MODULES, additional chunkstring uses, etc.

#v. 1.0 alpha 9 2006-09-04:
#   (pimthe:)
#   - Updated to use the new chunkstring's MLOG_GetModuleName(line).
#   - Removed old Chunkstring documentation.
#   - Merged with M25's 0.996 branch.

#v. 1.0 alpha 9 m25 branch 1 2006-09-08:
#   (m25)
#   - Added knockback detection, including additions to Watch_Getup and Watch_Flailing
#   - put the Global_RegisterAttr lines back in: how are these being given defaults otherwise?

#v. 1.0 alpha 10 2006-09-10:
#   - merged Stumpy's branch of 1.0a8 and M25's branch of 1.0a9.

#v. 1.0 alpha 11 2006-09-10:
#   (stumpy:)
#   - fix for knockback detection

#v. 1.0 alpha 12 2006-09-14:
#   (pimthe:)
#   - Added cancelMLOGDamageCallbackSinkForTarget(target,callback).
#   - Corrected bug in cancelAllMLOGDamageCallbackSinksForTarget((target).

#v. 1.0 alpha 13 2006-09-27:
#   (stumpy:)
#   - Added suite of object spawn/destroy callback registration, handlers, 
#     and cancellation functions.
#   - Made 'activation' module mandatory.
#   - Changed Watch*() functions to return 1 when given a line from their 
#     module, 0 otherwise, so other Watch*() functions don't have to run.
#   - missionobjvar is not imported as *, since it now depends on this 
#     module

#v. 1.0 alpha 14 2006-10-11:
#   (stumpy:)
#   - changed some imports around to avoid CTD issues.
#   - removed MLOG_GetChar() since it wasn't being used in MLOG_WatchJumping()
#   - returned invertDict() to remove dependency on generalutils (circular ref)
#   - added cancelDestroySinksForObject(), cancelDestroySinksForFunction()
#   - changed cancelObjectSpawnCallbackSink() to allow matching for the function
#     and the event string and/or the event option when canceling spawn sinks, 
#     instead of just matching the function.

#v. 1.0 alpha 15 2006-10-14:
#   (stumpy:)
#   - changed callback handlers to quit if no callbacks were registered (rather
#     than register an empty list)
#   - Added event sinks for saved-game load detection
#   - Moved calls to MLOG_Init() nearer to beginning of load in ffx.py
#   - Added 'MLOG_GameTime' mission attribute
#   - Added 'MLOG_MissionName' mission variable
#   - Added 'MLOG_SavedGameLoaded' mission attribute
#   - Added MLOG modules initialization to MLOG_Read() so that saved-game loads
#     start the module because MLOG_Init() is not called in those cases).

#v. 1.0 alpha 16 2006-10-17:
#   (stumpy:)
#   - fixed silly error with saved game callback handling

#v. 1.0 alpha 17 2006-11-30:
#   (stumpy:)
#   - (minor) added some error checking to diagnose exceptions occuring in 
#     damage watch code.

#v. 1.0.18 final 2007-03-12:
#   (ep:)
#   - Added missing import of generalutils
#     Re invertDict(): in 1.0 a14: "returned invertDict() to remove dependency on generalutils (circular ref)."
#     Circular reference shouldn't be an issue, as we're using local import only (i.e., no "from generalutils import *"; anyway,
#     generalutils doesn't import mlogreader at all.
#   - HandleMLOGDamageCallbacks2(): added try/except to catch bug with Illusionist found by TMX which I haven't been able to reproduce.

#v. 1.0.19 final 2008-03-24:
#   (tmx:)
#   - Added regDodgeStart() and regDodgeStop(), MLOG_WatchDodge(), and MLOG_IsDodging() to detect Dodging animations

#v. 1.0.20 (alpha) 2008-04-20:
#   (stumpy:)
#   - Some fixes to the dodge animation detection. Some testing likely still needed.
#v. 1.0.21 (alpha) 2008-05-28:
#   (stumpy:)
#   - Fixed removal of ai_knockback global attribute from 1.0.20

#v. 1.0.22 2008-07-20 FFX 3.3:
#   (ep 2007-10-15:)
#   - MLOG_WatchDamage():
#     - Removed the damageType dictionary lookup which actually prevented state change powers from registering (thanks to TMX for testing it)
#     - Fixed (?) an infrequent bug in MLOG_WatchDamage() when char object is destroyed very quickly after damage occurs
#   - Moved MLOG_ALLTYPES_INVERTED from regMLOGDamage() to a global.
#   (ep 2008-04-17:)
#   - MLOG_WatchDamage():
#     - No code change, only added a note on an error in MLOG_WatchDamage(), which is caused by ff.log itself.

#v. 1.0.23 2008-08-12 FFX 3.3:
#   (ep:)
#   - MLOG_WatchDamage():
#     - Replaced the try... except for invalid targets by Object_Exists()




######################################################################
####
#### Top-Level Functions
####
######################################################################


######################################################################
##
## Imports and Global Attributes
##
######################################################################

import os
import sys
import string
import cshelper
from js import *

import datfiles         # for the filename heuristics
import chunkstring      # for parsing the log lines
import missionobjvar    # for storing long variables, e.g. the callback data
import generalutils     #for invertdict()

# for testing
import time

Global_RegisterAttr('ai_flying', 0)
Global_RegisterAttr('ai_flailing', 0)
Global_RegisterAttr('ai_landing', 0)
Global_RegisterAttr('ai_gettingup', 0)
Global_RegisterAttr('ai_jumping', 0)
Global_RegisterAttr('ai_idle', 1)
Global_RegisterAttr('ai_dodging', 0)
Global_RegisterAttr('ai_knockback', 0)

######################################################################
##
## Initializing the Module
##
######################################################################

# ***** ADD YOUR MLOG MODULES TO THIS LIST *****
MLOG_RUNNING_MODULES = (
   'animator',
   'Damage',
   'activation',
   #'Movement',
   #'CEffectHandler',
   #'controller',
   #'melee',
   #'ranged',
   )

#status of optional modules; 'activation', 'animator' and 'Damage' are mandatory
# stumpy 2006-09-05: make this automatic so people only have to change MLOG_RUNNING_MODULES
MLOG_ALL_MODULES = \
('CEffectHandler', 'Camera', 'Damage', 'Dodge', 'LightMgr', 'Movement',
 'MovementGrid', 'NetImmerseIO', 'PathFinder', 'Physics', 'PhysicsBehavior',
 'PhysicsNiCollision', 'Sound', 'SoundCache', 'SoundExists', 'SoundHandles',
 'SoundIO', 'SoundLib', 'SoundPlay', 'SoundSchema', 'Speech', 'StreamHandles',
 'Timer', 'activation', 'animator', 'arrowmgr', 'beam', 'controller', 'decal',
 'eventmgr', 'fonts', 'fonts_ext', 'formation', 'heromgr', 'idle', 'imageload',
 'localise', 'melee', 'nifcache', 'personality', 'ranged', 'rpg', 'stun')
for module in MLOG_ALL_MODULES:
    exec ('MLOG_RUNNING_%s = %d' % (string.upper(module),module in MLOG_RUNNING_MODULES))



# Initialize the module.
def MLOG_Init(keepRunningModules = 0):
    print 'MLOG_Init(keepRunningModules=%d): starting up' % keepRunningModules
    global fflog
##    tstart = time.clock()
    Event_CancelSinks('OnMLOG_Read')    # stop accumulation from multiple save/reloads
    if keepRunningModules == 0:
        # Registered modules are persistent for the game duration, not just for a single mission
        # To prevent loading unnecessary modules, all potentially used modules should be stopped first
        for module in MLOG_ALL_MODULES:
            MLOGKill(module)
    EnableMLOG(1)
    for module in MLOG_RUNNING_MODULES:
        MLOG(module)
    fflog = open('ff.log', 'r')
    try:
        fflog.seek(-500000,2)  # back up from end of long files
    except IOError:           # file not that long
        pass
    wholelog = fflog.readlines()        # read in the whole ff.log file
    position = fflog.tell()
    wholelog.reverse()
    for logline in wholelog:            # find where *this* mission started
        position = position - len(logline) - 1
        if logline[17:31] == 'CMission::init':
            _currentMission = string.upper(string.split(string.split(logline,',',1)[0],'(')[1])
            print 'mlogreader.MLOG_Init: current mission = %s' % repr(_currentMission)
            missionobjvar.Mission_SetVar('MLOG_MissionName',_currentMission)
            break
    fflog.seek(position)
    Mission_SetAttr('MLOG_GameTime',0.0)
    try:    # don't set this if it's already been set
        Mission_GetAttr('MLOG_SavedGameLoaded')
    except:
        Mission_SetAttr("MLOG_SavedGameLoaded",0)
    Event_RegisterSink(EVENT_TIMER, 'OnMLOG_Read', '', '', 0.1, 0, 0)
##    print 'MLOG_Init: time %17.7f sec' % (time.clock()-tstart)

######################################################################
##
## Basic Animation State Determination
##
######################################################################

# Checks if a character's isFlying attribute has been set and returns the value.
# If isFlying hasn't been set, adds the character to the flight check routine.
# Changed 0.94:
# - redone as a simple shortcut to GetObjectVar, since the MLOG_WatchFlying() code was completely redone by M25
#   to be mandatory
# - For simplicity's sake, characters whose movement hasn't been detected yet return 0, not None or -1
# Changed 0.981:
# Simply returns the attribute value, 0 or 1, which is set/unset for every
# character, since the previous version didn't add functionality.


# Checks if a character's ai_flying attribute has been set and returns the value.
# IMPORTANT: Jumping characters (including using Superleaper, but not using Jump Travel, Low Jumper
# or Clumsy Jumper) will also register as flying, since both movements use the hover animation.
# It's your responsability to make sure that only characters with a working flyer attribute or
# the Buoyant secondary state are checked (this will also improve
# MLOGreader's performance) - that is, unless you WANT to detect jumping.;)
# This can be done by using:
# if ffx.hasAttribute(char, 'flier') or ffx.hasAttribute(char, 'levitate') or ffx.isBuoyant(char):
#    MLOG_IsFlying(char)

def MLOG_IsFlying(char):
    return Object_GetAttr(char, 'ai_flying')

# Checks if a character's ai_flailing attribute has been set and returns the value.
def MLOG_IsFlailing(char):
    return Object_GetAttr(char, 'ai_flailing')

# Checks if a character's ai_landing attribute has been set and returns the value.
def MLOG_IsLanding(char):
    return Object_GetAttr(char, 'ai_landing')

# Checks if a character's ai_gettingup attribute has been set and returns the value.
def MLOG_IsGettingUp(char):
    return Object_GetAttr(char, 'ai_gettingup')
    
# Checks if a character's ai_idle attribute has been set and returns the value.    
def MLOG_IsIdle(char):
    return Object_GetAttr(char, 'ai_idle')

# Checks if a character's ai_knockback attribute has been set and returns the value.    
def MLOG_IsInKnockBack(char):
    return Object_GetAttr(char, 'ai_knockback')

# Checks if a character's ai_dodging attribute has been set and returns the value.
def MLOG_IsDodging(char):
    return Object_GetAttr(char, 'ai_dodging')

######################################################################
##
## Detecting changes in the Flight, Flailing, Landing, and Getting Up
## state of a character.
##
## Set up animation detection events.
## 
######################################################################

# The idea here is that, since we are checking the log file pretty
# frequently anyway, when we have functions that need to run when a
# character starts or ends flight (or flailing), we register them as
# event sinks and then handle them as callbacks to those sinks. This
# should be *much* more efficient than doing this through polling
# (e.g. testing for a change in the flight attribute every so many
# seconds or fractions of a second.  In addition, this means that we
# get our callbacks run just as soon as we know that there is a change
# in the status, rather than up to the polling period after it
# changes.  The potential fly in the ointment is how much overhead the
# registration of the sinks will add, particularly with the mission or
# object long variables if there are lots of objects registered.

# Most of this is cribbed from the power change event code in datfiles.

# Top-level functions:  These just call regAnimation(), but they are
# specific and probably a bit easier to use.  The event structure sent
# to the callback functions from each of these will have the character
# detected as event.object and the optional string parameter passed as
# event.string, along with the optional user and float parameters.  An
# additional parameter event.animation is passed containing the name
# of the animation that triggered the callback by starting or ending
# the registered animation state.

# Detect when a character first takes flight. This detects a jump as flight.
def regFlightStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=FLIGHT_START_EVENT, persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character's flight ends, either by landing or because
# he enters a flail state.  
def regFlightStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=FLIGHT_STOP_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character first starts flailing. 
# IMPORTANT: Some attribute have no choice but to force the character to flail during his special movement:
# Use of Jump Travel, Low Jumper, Clumsy Jumper or Fastball will register as flailing.
def regFlailStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=FLAIL_START_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character stops flailing, defined by no longer being
# in a flail animation.
def regFlailStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=FLAIL_STOP_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character starts to land, either by crashing into the
# ground from a flail state or by ending flight or ending a jump.  
def regLandStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=LAND_START_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character finishes landing, defined by no longer being
# in a land animation.  
def regLandStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=LAND_STOP_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character starts to get up, which they do after
# crashing into the ground after a flail state.
def regGetupStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=GETUP_START_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character finishes getting up, defined by no longer
# being in a get_up_* animation.  Note that this is useful when
# creating a more natural-looking point for characters to resume
# activity after they have been flailing (e.g. knocked back).  That
# is, if you are waiting for a flail to end before triggering a move
# or power use, use this function to start your action when the flail
# is done and the character has stood up, not when the flail itself
# ends (where the character is still laying on the ground).
def regGetupStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char,event_type=GETUP_STOP_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character is idle (hovering or standing on the ground).                 
def regIdleStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=IDLE_START_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character is no longer idle - attacking, running, etc.
def regIdleStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=IDLE_STOP_EVENT,persistent=persistent,\
                callback=function,float=float,user=user,string=string)		


# Detect when a character has been knocked back
def regKnockBackStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=KNOCKBACK_START_EVENT,persistent=persistent,\
                callback=function,float=float,user=user,string=string)

# Detect when a character has finished getting up after being knocked back
def regKnockBackStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=KNOCKBACK_STOP_EVENT,persistent=persistent,\
                callback=function,float=float,user=user,string=string)

#The Acrobatic FFX attribute interrupts knockback, so it should call ResetKnockBack to trigger the callbacks properly.
def MLOG_ResetKnockBack(char):
    if Object_GetAttr(char,'ai_knockback'):
        Object_SetAttr(char, 'ai_knockback', 0)
        if filter(lambda cb: cb[0]==KNOCKBACK_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
            HandleAnimationCallbacks(char,'scripted interrupt',KNOCKBACK_STOP_EVENT)    # stumpy 2006-09-24: words[] not defined here

# Detect when a character is dodging (dodge_right or dodge_left).                 
def regDodgeStart(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=DODGE_START_EVENT,persistent=persistent,\
                 callback=function,float=float,user=user,string=string)

# Detect when a character is no longer dodging - attacking, running, etc.
def regDodgeStop(char, function, persistent=0, float=0, user=0, string=''):
    regAnimation(object=char, event_type=DODGE_STOP_EVENT,persistent=persistent,\
                callback=function,float=float,user=user,string=string)		


# Cancel an object's regAnimation() sinks for a particular callback function
def cancelAnimationCallbackSink(object,callback):
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_ANIM_CB')
    if callbacks:   # don't make a list if there were no callbacks
        keptcallbacks = []
        for cb in callbacks:
            if cb[1] != callback:
                keptcallbacks.append(cb)
        missionobjvar.Object_SetVar(object,'MLOG_ANIM_CB',keptcallbacks)
    
# Cancel all regAnimation() sinks for a particular object
def cancelAnimationCallbackSinks(object):
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_ANIM_CB')
    if callbacks:   # don't make a list if there were no callbacks
        missionobjvar.Object_SetVar(object,'MLOG_ANIM_CB',[])

######################################################################
##
## Detecting when objects are spawned and destroyed
##
## (stumpy: putting the top-level stuff here.)
##
######################################################################

# Detect the spawning of any non-projectile object
def regObjectSpawn(function, persistent = 0, float=0, user = 0, string = ''):
    regActivationSpawnSink(OBJECT_SPAWN_EVENT, function, persistent = persistent,
                           floatVal=float, user = user, string = string, option = None)

# Detect the spawning of any projectile object (anything with a
# 'projectile_standin' template).
def regProjectileSpawn(function, persistent = 0, float=0, user = 0, string = ''):
    regActivationSpawnSink(PROJECTILE_SPAWN_EVENT, function, persistent = persistent,
                           floatVal=float, user = user, string = string, option = None)

# Detect the spawning of any object with the given template.
def regTemplateSpawn(function, template, persistent = 0, float=0, user = 0, string = ''):
    regActivationSpawnSink(TEMPLATE_SPAWN_EVENT, function, persistent = persistent,
                           floatVal=float, user = user, string = string, option = template)

# Detect the spawning of any object with the given class value.
# Matches based on logical AND match of object class to given class.
def regClassSpawn(function, objectClass, persistent = 0, float=0, user = 0, string = ''):
    regActivationSpawnSink(CLASS_SPAWN_EVENT, function, persistent = persistent,
                           floatVal=float, user = user, string = string, option = objectClass)

# Detect the spawning of any object with the given class value.
def regCharacterSpawn(function, persistent = 0, float=0, user = 0, string = ''):
    regActivationSpawnSink(CHARACTER_SPAWN_EVENT, function, persistent = persistent,
                           floatVal=float, user = user, string = string, option = None)

# Detect when the given object is removed from the game.
def regObjectDestroy(object, function, persistent = 0, float=0, user = 0, string = ''):
    ##print 'regObjectDestroy: called on object=%s function=%s string=%s' % (repr(object),repr(function),repr(string))
    regActivationDestroySink(object, function, persistent, float, user, string)

# Cancel all non-projectile object spawn sinks that call the given function.
def cancelObjectSpawnSink(function):
    cancelObjectSpawnCallbackSink(OBJECT_SPAWN_EVENT,function)

# Cancel all projectile spawn sinks that call the given function.
def cancelProjectileSpawnSink(function):
    cancelObjectSpawnCallbackSink(PROJECTILE_SPAWN_EVENT,function)

# Cancel all template spawn sinks that call the given function.
def cancelTemplateSpawnSink(function):
    cancelObjectSpawnCallbackSink(TEMPLATE_SPAWN_EVENT,function)

# Cancel all class object spawn sinks that call the given function.
def cancelClassSpawnSink(function):
    cancelObjectSpawnCallbackSink(CLASS_SPAWN_EVENT,function)

# Cancel all character object spawn sinks that call the given function.
def cancelCharacterSpawnSink(function):
    cancelObjectSpawnCallbackSink(CHARACTER_SPAWN_EVENT,function)

# Cancel an object destroy sink for the given object and function.
def cancelObjectDestroySink(object, function):
    cancelObjectDestroyCallbackSink(object, function)

# Cancel all regActivationDestroySink() sinks for a given object.
def cancelDestroySinksForObject(object):
    cancelObjectDestroyCallbackSinksForObject(object)

# Cancel regActivationDestroySink() sinks for a given function.
def cancelDestroySinksForFunction(function):
    cancelObjectDestroyCallbackSinksForFunction(function)

# some testing
def TestSpawnSinks():
    regObjectSpawn('mlogreader.SpawnCallback',persistent=0,float=21.0,user=3,string='test regObjectSpawn')
    regProjectileSpawn('mlogreader.SpawnCallback',persistent=1,float=22.0,user=4,string='test regProjectileSpawn')
    regTemplateSpawn('mlogreader.SpawnCallback','building_debris_0', persistent=1,float=23.0,user=5,string='test regTemplateSpawn')
    regClassSpawn('mlogreader.SpawnCallback',OC_MINION,persistent=1,float=24.0,user=6,string='test regClassSpawn')
    regCharacterSpawn('mlogreader.SpawnCallback',persistent=1,float=25.0,user=7,string='test regCharacterSpawn')
    pass

def SpawnCallback(event):
    Mission_StatusText('%s event for %s' % (SPAWN_EVENT_NAMES[event.event],event.object))
    print 'SpawnCallback: %s event for %s with option=%s' % (SPAWN_EVENT_NAMES[event.event],event.object,repr(event.option))
    print 'SpawnCallback: string=<%s> float=<%f> data=<%d> user=%d' % (event.string,event.float,event.data,event.user)
    if ( event.event == CLASS_SPAWN_EVENT ) or ( event.event == PROJECTILE_SPAWN_EVENT ):
        regObjectDestroy(event.object,'mlogreader.SpawnCallback',persistent=0,float=-10.0,user=9,string='test regObjectDestroy')

def ShowCallbacks():
    for SE in SPAWN_EVENT_NAMES.keys():
        CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[SE] # where these events are stored
        callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
        print 'ShowCallbacks: %s = %s' % (CBVarName,callbacks)


######################################################################
##
## Basic Movement State Determination
##
######################################################################

# Checks if a character's ai_jumping attribute has been set and returns the value.
def MLOG_IsJumping(char):
    if not MLOG_RUNNING_MOVEMENT:
        print "MLOG_IsJumping() non-fatal error:"
        print "    'Movement' module is not enabled; jumping cannot be detected."
        print "    Use MLOG_Movement(1) to enable the module first."
        return 0
    return Object_GetAttr(char, 'ai_jumping')


######################################################################
## 
## Detecting changes in the jumping state of the character
## 
######################################################################

# Activate or desactivate the 'Movement' MLOG module used to check jumping, etc.
# enable should be 1 to enable or 0 to disable.
# Note that using the Movement module adds a lot of overhead to the game.
# Use it only if really necessary and disable it after use.
def MLOGMovement(enable = 1):
    global MLOG_RUNNING_MODULES
    global MLOG_RUNNING_MOVEMENT
    if enable and not ('Movement' in MLOG_RUNNING_MODULES):
        MLOG_RUNNING_MODULES.append('Movement')
        MLOG_RUNNING_MOVEMENT = 1
        MLOG('Movement')
        print "MLOG_Movement(): 'Movement' logging Activated."
    elif not enable and ('Movement' in MLOG_RUNNING_MODULES):
        MLOG_RUNNING_MODULES.remove('Movement')
        MLOG_RUNNING_MOVEMENT = 0
        MLOGKill('Movement')
        print "MLOG_Movement(): 'Movement' logging Disabled."
        jumpers = missionobjvar.Mission_GetVar('MLOGJumpers', [])
        for char in jumpers:
            if Object_Exists(char):
                Object_SetAttr(char, 'ai_jumping', 0)
        missionobjvar.Mission_SetVar('MLOGJumpers', [])
    #else:
        #print "MLOG_Movement(): No change required to 'Movement' logging status."


# Detect when a character starts jumping.
def regJumpStart(char, function, persistent=0, float=0, user=0, string=''):
    regMLOGMovement(object=char,event_type=JUMP_START_EVENT, persistent=persistent,callback=function,float=float,user=user,string=string)

# Detect when a character's jump ends, either by landing or because he enters a flail state.  
def regJumpStop(char, function, persistent=0, float=0, user=0, string=''):
    regMLOGMovement(object=char,event_type=JUMP_STOP_EVENT,persistent=persistent,callback=function,float=float,user=user,string=string)


# Cancel an object's regMLOGMovement() sinks for a particular callback function
def cancelMLOGMovementCallbackSink(object,callback):
    currentCallbacks = Mission_GetAttr('MLOGMovement_calls')
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_MOVEMENT_CB')
    if callbacks:   # don't make a list if there were no callbacks
        keptcallbacks = []
        for cb in callbacks:
            if cb[1] != callback:
                keptcallbacks.append(cb)
            else:
                currentCallbacks = currentCallbacks - 1
        missionobjvar.Object_SetVar(object,'MLOG_MOVEMENT_CB',keptcallbacks)
    Mission_SetAttr('MLOGMovement_calls', currentCallbacks)
    if currentCallbacks < 1:
        MLOGMovement(0)

# Cancel all regMLOGMovement() sinks for a particular object
def cancelMLOGMovementCallbackSinks(object):
    currentCallbacks = Mission_GetAttr('MLOGMovement_calls')
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_MOVEMENT_CB', [])
    if callbacks:   # don't make a list if there were no callbacks
        missionobjvar.Object_SetVar(object,'MLOG_MOVEMENT_CB',[])
        currentCallbacks = currentCallbacks - len(callbacks)
    Mission_SetAttr('MLOGMovement_calls', currentCallbacks)
    if currentCallbacks < 1:
        MLOGMovement(0)

######################################################################
## 
## Detecting Saved-Game Loads
##
## (no wrappers currently, see regSavedGameLoad(),
## cancelSavedGameSinks(), and cancelSavedGameCallbackSink() below.
## 
######################################################################


######################################################################
##
## Running and Stopping the MLOG reader loop
##
######################################################################

def OnMLOG_Read(event):
    Event_RegisterSink(EVENT_TIMER, 'OnMLOG_Read', '', '', 0.1, 0, 0)
    MLOG_Read()

# stumpy 2006-09-06: for the optional modules, let's put the 'if MLOG_RUNNING_*' 
# check here, so it only has to happen once for each module and we don't have 
# the function call overhead.
def MLOG_Read():
    global fflog
##    tstart = time.clock()
    GameTime = Mission_GetAttr('MLOG_GameTime')  # stumpy 2006-10-10: this will be useful
    try:
        lines = fflog.readlines()
    except NameError:   # likely load from a saved game
        print 'MLOG_Read: likely mid-mission load'
        Mission_SetAttr("MLOG_SavedGameLoaded",1)
        HandleSavedGameCallbacks(GameTime)
        fflog = open('ff.log', 'r')
        try:
            fflog.seek(-500000,2)  # back up from end of long files
        except IOError:           # file not that long
            pass
        wholelog = fflog.readlines()
        wholelog.reverse()
        loglinenum = 0
        for logline in wholelog:
            ##if logline[17:31] == 'CMission::init':
            if logline[17:43] == 'CPathNodeList::preprocess:':  # stumpy 2006-10-27: to skip past massive spawnage at mission load
                break
            loglinenum = loglinenum + 1
        lines = wholelog[:loglinenum]
        lines.reverse()
        # If we load a saved game when MLOG hasn't been running, MLOG doesn't get turned on
        EnableMLOG(1)
        for module in MLOG_RUNNING_MODULES:
            MLOG(module)
    Mission_SetAttr('MLOG_GameTime',GameTime+0.1)
    for thisLine in lines:
        # ***** YOUR MLOG ANALYSIS FUNCTIONS HERE *****
        #animation module:
        found = \
        MLOG_WatchDodge(thisLine) + \
        MLOG_WatchFlying(thisLine) + \
        MLOG_WatchFlailing(thisLine) + \
        MLOG_WatchLanding(thisLine) + \
        MLOG_WatchGetup(thisLine) + \
        MLOG_WatchIdle(thisLine)
        if not found:
            #Damage module:
            found = MLOG_WatchDamage(thisLine)
        if not found:
            #Activation module:
            found = MLOG_WatchSpawnAndDestroy(thisLine)
        if not found:
            #Movement module:
            if MLOG_RUNNING_MOVEMENT:
                found = MLOG_WatchJumping(thisLine)            
        # ***** YOUR MLOG ANALYSIS FUNCTIONS ABOVE *****
##    print 'MLOG_Read: time %17.7f sec' % (time.clock()-tstart)

def MLOG_Shutdown():
   for module in MLOG_RUNNING_MODULES:
       MLOGKill(module)
   EnableMLOG(0)


######################################################################
##
## Watch function return values
##
######################################################################

# The watch functions should return 0 when they find something from
# their module and 0 if they do not.  That is so that we can assume
# that if the functions watching for animation changes found something,
# we don't have to bother with the functions watching for signals from
# other modules.

######################################################################
##
## Watching for Animation State Changes
##
######################################################################

# The actual state detection code, used in MLOG_Read()

# MLOG animator module: BlendActivate word sequences indicating flight (not necessarily begining of)
# or end of flight or jump up/down.
# A line in ff.log is something like this:
# 02:16:44.89: FF: animator:(hero_1) BlendActivate run -> melee_idle
# Note that MLOG_WatchFlying() expect the first two flyChanges elements to mean "flying" and anything
# after that to mean "not flying".

# Check if a character is currently flying.
# IMPORTANT: Jumping characters (including using Superleaper, but not using Jump Travel, Low Jumper
# or Clumsy Jumper) will also register as flying, since both movements use the hover animation.
# It's your responsability to make sure (by reading the info from the save game, character.dat or
# .hero file) that only characters with a working flyer attribute are checked (this will also improve
# MLOGreader's performance). That is, unless you WANT to detect jumping.;)
# Changed 0.981:
# Why count jumping as flying?  I don't think we want the AIs to use
# their fly tactics on jumpers, right? I took out the hover check,
# since I don't think you can ever be flying without going through a
# 'fly' animation.
# Answer v0.995: without the hover animation, we won't detect
# levitation as flight and we won't detect flight where the
# character flies straight up (until he moves horizontally).

def MLOG_WatchFlying(line):
    words = chunkstring.ParseAnimations(line)
    if not words:
        return 0
    char = words[0]
    if not Object_Exists(char): 
        return 1
    already_flying = Object_GetAttr(char,'ai_flying')
    if ( words[1] == 'fly' ) or ( words[1] == 'hover' ):
        if not already_flying:
            Object_SetAttr(char,'ai_flying',1)
            ##print '%s\t%s now flying' % (line,char)
            if filter(lambda cb: cb[0]==FLIGHT_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],FLIGHT_START_EVENT)
    # stumpy 2006-09-06: M25 added the 'idle' check.  Can a char stop flying without going through a land or flail
    elif ( words[1][:4] == 'land' ) or ( words[1][:8] == 'flailing' ) or ( words[1][:5]=='melee' ):
        if already_flying:
            Object_SetAttr(char,'ai_flying',0)
            ##print '%s\t%s done flying' % (line,char)
            if filter(lambda cb: cb[0]==FLIGHT_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],FLIGHT_STOP_EVENT)
    return 1

def MLOG_WatchDodge(line):
	words = chunkstring.ParseAnimatorByKeyword(line,'startAnimating',1, 'end startAnimating')  or  chunkstring.ParseAnimatorByKeyword(line,'CAnimator::resumeAnimating: Activate Sequence',1)
	if not words: return 0
	#print line, words
	#char, anim = words
	if not Object_Exists(words[0]): return 1
	if ( words[1] in ['dodge_left','dodge_right']):
		Object_SetAttr(words[0],'ai_dodging',1)
		if filter(lambda cb: cb[0]==DODGE_START_EVENT, missionobjvar.Object_GetVar(words[0],'MLOG_ANIM_CB',[])):
			HandleAnimationCallbacks(words[0],words[1],DODGE_START_EVENT)
	else:
		Object_SetAttr(words[0],'ai_dodging',0)
		if filter(lambda cb: cb[0]==DODGE_STOP_EVENT, missionobjvar.Object_GetVar(words[0],'MLOG_ANIM_CB',[])):
			HandleAnimationCallbacks(words[0],words[1],DODGE_STOP_EVENT)
	return 1

# IMPORTANT: Some attribute have no choice but to force the character to flail during his special movement:
# Use of Jump Travel, Low Jumper, Clumsy Jumper or Fastball will register as flailing. These special states
# can be detected by using ffx.FFX_ObjectGetAttr(char, 'voluntary_flailing')
def MLOG_WatchFlailing(line):
    words = chunkstring.ParseAnimations(line)
    if not words:
        return 0
    char = words[0]
    if not Object_Exists(char): 
        return 1
    already_flailing = Object_GetAttr(char,'ai_flailing')
    if words[1][:8] == 'flailing':
        if not already_flailing:
            Object_SetAttr(char,'ai_flailing',1)
            ##print '%s\t%s now flailing' % (line,char)
            if filter(lambda cb: cb[0]==FLAIL_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],FLAIL_START_EVENT)
            #KNOCKBACK starts as well # stumpy 2006-09-12: this should be inside the previous conditional
            if not Object_GetAttr(char,'ai_knockback'):
                ##print '%s\t%s now in KB' % (line,char)
                Object_SetAttr(char, 'ai_knockback',1)
                if filter(lambda cb: cb[0]==KNOCKBACK_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                    HandleAnimationCallbacks(char,words[1],KNOCKBACK_START_EVENT)
    else:
        if already_flailing:
            Object_SetAttr(char,'ai_flailing',0)
            ##print '%s\t%s done flailing' % (line,char)
            if filter(lambda cb: cb[0]==FLAIL_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],FLAIL_STOP_EVENT)
    return 1

def MLOG_WatchLanding(line):
    words = chunkstring.ParseAnimations(line)
    if not words:
        return 0 
    char = words[0]
    if not Object_Exists(char): 
        return 1
    already_landing = Object_GetAttr(char,'ai_landing')
    if words[1][:4] == 'land':
        if not already_landing:
            Object_SetAttr(char,'ai_landing',1)
            ##print '%s\t%s now landing' % (line,char)
            if filter(lambda cb: cb[0]==LAND_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],LAND_START_EVENT)
    else:
        if already_landing:
            Object_SetAttr(char,'ai_landing',0)
            ##print '%s\t%s done landing' % (line,char)
            if filter(lambda cb: cb[0]==LAND_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],LAND_STOP_EVENT)
    return 1

def MLOG_WatchGetup(line):
    words = chunkstring.ParseAnimations(line)
    if not words:
        return 0
    char = words[0]
    if not Object_Exists(char): 
        return 1
    already_gettingup = Object_GetAttr(char,'ai_gettingup')
    if words[1][:7] == 'get_up_':
        if not already_gettingup:
            Object_SetAttr(char,'ai_gettingup',1)
            ##print '%s\t%s now getting up' % (line,char)
            if filter(lambda cb: cb[0]==GETUP_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],GETUP_START_EVENT)
    else:
        if already_gettingup:
            Object_SetAttr(char,'ai_gettingup',0)
            ##print '%s\t%s done getting up' % (line,char)
            if filter(lambda cb: cb[0]==GETUP_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                HandleAnimationCallbacks(char,words[1],GETUP_STOP_EVENT)  
            #KNOCKBACK ends # stumpy 2006-09-12: this really must be inside the previous conditional
            if Object_GetAttr(char,'ai_knockback'):
                ##print '%s\t%s now ending KB' % (line,char)
                Object_SetAttr(char, 'ai_knockback', 0)
                if filter(lambda cb: cb[0]==KNOCKBACK_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_ANIM_CB',[])):
                    HandleAnimationCallbacks(char,words[1],KNOCKBACK_STOP_EVENT)
    return 1


#m25 August 2006: 
# WatchIdle is used by the AI to determine when a character has gone idle and is ready to act again.     
# it looks for startAnimating <animation>
# or CAnimator::resumeAnimating: Activate Sequence <animation>
def MLOG_WatchIdle(line):
	words = chunkstring.ParseAnimatorByKeyword(line,'startAnimating',1, 'end startAnimating')  or  chunkstring.ParseAnimatorByKeyword(line,'CAnimator::resumeAnimating: Activate Sequence',1)
	if not words: return 0
	#print line, words
	#char, anim = words
	if not Object_Exists(words[0]): return 1
	if ( words[1] in ['melee_idle','hover','idle','dodge_right','dodge_left']):  #'dodge_right','dodge_left' ??
		Object_SetAttr(words[0],'ai_idle',1)
		if filter(lambda cb: cb[0]==IDLE_START_EVENT, missionobjvar.Object_GetVar(words[0],'MLOG_ANIM_CB',[])):
			HandleAnimationCallbacks(words[0],words[1],IDLE_START_EVENT)
	else:
		Object_SetAttr(words[0],'ai_idle',0)
		if filter(lambda cb: cb[0]==IDLE_STOP_EVENT, missionobjvar.Object_GetVar(words[0],'MLOG_ANIM_CB',[])):
			HandleAnimationCallbacks(words[0],words[1],IDLE_STOP_EVENT)
	return 1

######################################################################
##
## Watching for activation spawn or destroy events
##
######################################################################

def MLOG_WatchSpawnAndDestroy(line):
    # Look for a spawn
    words = chunkstring.ParseActivationForNewObject(line) 
    if words:
        object = words[0]
        if not Object_Exists(object): 
            return 1
        NIF = words[1]
        HandleActivationSpawnCallbacks(object,NIF)
        return 1
    # Look for a destroy
    words = chunkstring.ParseByModuleAndKeyword(line,'activation', 'CGameObjFactory::removeObject: removing', 1)
    if words:
        HandleActivationDestroyCallbacks(words[1])
        return 1
    # nothing found
    return 0
 

######################################################################
##
## Watching for Movement State Changes
##
######################################################################

# The actual state detection code, used in MLOG_Read()

# MLOG animator module: BlendActivate word sequences indicating flight (not necessarily begining of)
# or end of flight or jump up/down.
# A jumping start line in ff.log is something like this:
# 01:08:39.13: FF: Movement:(hero_0) jumping state starting
# landing from a jump:
# 00:27:52.97: FF: Movement:(hero_1) finish: JUMPING
# (04:31:44.80: FF: Movement:(hero_0) jump hit landing point can't be used, as it's not called when the system starts and cancel a jump immediately)
# flailing while jumping:
# 06:42:06.79: FF: Movement:(hero_1) starting movement command: RESET (from JUMPING)

# Check if a character is currently jumping.
# Must return a boolean value, as it's set as mutually exclusive with other fuctions of the same module
# stumpy 2006-09-06: changed to use chunkstring to do string parsing
def MLOG_WatchJumping(line):
    words = chunkstring.ParseByModuleAndKeyword(l,'Movement','jumping state starting',0) or \
            chunkstring.ParseByModuleAndKeyword(l,'Movement','finish: JUMPING',0)
    if not words:
        return 0
    char = words[0]
    if not Object_Exists(char):
        return 1
    already_jumping = Object_GetAttr(char,'ai_jumping')
    #jumping state starting
    if ( words[1][0] == 'j' ):
        if not already_jumping:
            Object_SetAttr(char,'ai_jumping',1)
            #Add the character to the list of jumpers so we can reset the attribute to 0 when using MLOGMovement(0) to avoid false positives once
            #if the module is restaored afterward.
            jumpers = missionobjvar.Mission_GetVar('MLOGJumpers', [])
            if not char in jumpers:
                jumpers.extend([char])
                missionobjvar.Mission_SetVar('MLOGJumpers', jumpers)
            #Mission_StatusText('%s now jumping' % char)
            #print '%s\t%s now jumping' % (line,char)
            if filter(lambda cb: cb[0]==JUMP_START_EVENT, missionobjvar.Object_GetVar(char,'MLOG_MOVEMENT_CB',[])):
                HandleMLOGMovementCallbacks(char,'jump_stand',JUMP_START_EVENT)
    #finish: JUMPING / starting movement command: RESET (from JUMPING)
    else:
        if already_jumping:
            Object_SetAttr(char,'ai_jumping',0)
            #Mission_StatusText('%s done jumping' % char)
            #print '%s\t%s done jumping' % (line,char)
            if filter(lambda cb: cb[0]==JUMP_STOP_EVENT, missionobjvar.Object_GetVar(char,'MLOG_MOVEMENT_CB',[])):
                HandleMLOGMovementCallbacks(char,'land_from_jump',JUMP_STOP_EVENT)
    return 1



# Object-specific events are those where the callback is registered to
# trigger on a particular game object.  The other type of event is
# nonobject-specific, meaning that the registration call has no idea
# which object will trigger the callback function.  Object-specific
# events are stored with the object and nonobject-specific events are
# stored with the callback function.

######################################################################
##
## Handling Animation Events
##
######################################################################

# Make a pseudo-event register sink for animation changes to make
# scripting them more familiar.  These are the event_type values to be
# passed to regAnimation().

# (I'm just making up these numbers.)
# Animation Events, these are object-specific sinks
# (Note that the SAVED_GAME_LOADED event is defined below as 100.)
FLIGHT_START_EVENT = 66
FLIGHT_STOP_EVENT = 67
FLAIL_START_EVENT = 70
FLAIL_STOP_EVENT = 71
LAND_START_EVENT = 72
LAND_STOP_EVENT = 73
GETUP_START_EVENT = 74
GETUP_STOP_EVENT = 75
IDLE_START_EVENT = 76
IDLE_STOP_EVENT = 77
KNOCKBACK_START_EVENT = 78
KNOCKBACK_STOP_EVENT = 79
DODGE_START_EVENT = 80
DODGE_STOP_EVENT = 81

# Lower-level stuff.

class PseudoAnimationEvent:
    def __init__(self, eventType=FLIGHT_START_EVENT, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0, animStr=''):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal
        self.animation = animStr

# Add a sink so that a script function is called when a given
# animation event type (one of those int constants above) is spotted
# during the MLOG_Watch* functions.
def regAnimation(object,event_type,callback,persistent=0,float=0,user=0, string=''):
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_ANIM_CB',[])
    callback = (event_type,callback,persistent,float,user,string)
    if not callback in callbacks:
        callbacks.append(callback)
    missionobjvar.Object_SetVar(object,'MLOG_ANIM_CB',callbacks)

# Cancel all regAnimation() sinks
def cancelAllAnimationSinks():
    missionobjvar.Object_SetVar(object,'MLOG_ANIM_CB',[])

# Execute the callback functions
def HandleAnimationCallbacks(object,anim,event_type):
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_ANIM_CB',[])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:   
        if ( event_type == cb[0] ):
            if not cb[2]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoAnimationEvent(event_type,objectStr=object, stringStr=cb[5], floatVal=cb[3], dataVal=cb[2], userVal=cb[4], animStr=anim)
            fn = cb[1]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Object_SetVar(object,'MLOG_ANIM_CB',keptcallbacks)

def eventeval(fn,event):
    if '.' in fn:   # module was specified
        try:
            eval(fn+'(event)')
        except:
            try:    # maybe we haven't imported it
                dotpos = len(fn)-2
                while fn[dotpos]!='.':  dotpos=dotpos-1
                module = fn[:dotpos]
                exec('import '+module)
                eval(fn+'(event)')
            except:
                print 'eventeval: problem evaluating string %s' % repr(fn+'(event)')
                raise
    else:   # no module specified: take some guesses then try them all
        success = 0
        tryfirst = ['ff.','ffx.','m25ai.','missionobjvar.','cshelper.',]
        for module in tryfirst: # try the preferred modules
            try:
                eval(module+fn+'(event)')
                success = 1
                break
            except:
                try:    # maybe we haven't imported it
                    exec('import '+module[:-1])
                    eval(module+fn+'(event)')
                    success = 1
                    break
                except:
                    pass
        if not success:
            for module in sys.modules.keys():   # try all the imported modules
                try:
                    eval(module+'.'+fn+'(event)')
                    success = 1
                    break
                except:
                    try:    # maybe we haven't imported it
                        exec('import '+module)
                        eval(module+'.'+fn+'(event)')
                        success = 1
                        break
                    except:
                        pass
        if not success:
            print 'eventeval: unable to evaluate string %s in any module.' \
                  % repr(fn+'(event)')
            raise


######################################################################
##
## Handling Events Upon Saved Game Detection 
##
######################################################################

# With only one event type, this isn't really needed, but I will stick
# with the model.
SAVED_GAME_LOADED_EVENT = 100

# Like the other event classes, generally.  The float field will be
# the game time, the object will be the full path of the saved game.
# The data field will be whether the event was persistent.
class PseudoSavedGameEvent:
    def __init__(self, eventType=SAVED_GAME_LOADED_EVENT, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal

# Add a sink so that a script function is called when we detect that a
# saved game has loaded during the MLOG_Read() function.
def regSavedGameLoad(callback,persistent=0,user=0,string=''):
    callbacks = missionobjvar.Mission_GetVar('SAVED_GAME_LOAD_CB',[])
    callback = (SAVED_GAME_LOADED_EVENT,callback,persistent,user,string)
    if not callback in callbacks:
        callbacks.append(callback)
    missionobjvar.Mission_SetVar('SAVED_GAME_LOAD_CB',callbacks)

# Cancel all regAnimation() sinks
def cancelSavedGameSinks():
    missionobjvar.Mission_SetVar('SAVED_GAME_LOAD_CB',[])

# Cancel all regSavedGameLoad() sinks for the given function.  Also
# allows matching with the registered event string.
def cancelSavedGameCallbackSink(function,string=None):
    callbacks = missionobjvar.Mission_GetVar('SAVED_GAME_LOAD_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    # handle combinations of event string and option matching
    if ( string==None ):    # function matching
        for cb in callbacks:
            if not ( cb[0] == function ):
                keptcallbacks.append(cb)
    else:    # function and string matching
        for cb in callbacks:
            if not ( (cb[1] == function) and (cb[4] == string) ):
                keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar('SAVED_GAME_LOAD_CB', keptcallbacks)


# If a saved game has been detected, call this function and it will
# figure out what the saved game was, update the missionobjvar cached
# variable data, and call other functions registered as saved game
# sinks.
def HandleSavedGameCallbacks(GameTime):
    FileName = datfiles.GetLoadedSavedGame(GameTime)
    # This could be a registered event itself, but since it cleans up
    # the missionobjvar cache (which includes the event lists), we
    # want to do it before anything else.
    missionobjvar.CleanVarCache(FileName)
    # Now handle any other events...
    callbacks = missionobjvar.Mission_GetVar('SAVED_GAME_LOAD_CB',[])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    event_type = SAVED_GAME_LOADED_EVENT    # not strictly necessary, since that's the only event type this handler handles...
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:   
        if ( event_type == cb[0] ):
            if not cb[2]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoSavedGameEvent(event_type,objectStr=FileName, stringStr=cb[4], floatVal=GameTime, dataVal=cb[2], userVal=cb[3])
            fn = cb[1]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Mission_SetVar('SAVED_GAME_LOAD_CB',keptcallbacks)

######################################################################
##
## Handling Activation (object spawn/destroy) Events
##
######################################################################

# Activation events, spawn events are nonobject-specific, destroy events are object-specific
SPAWN_EVENT_NAMES = {
    90 : 'OBJECT_SPAWN_EVENT',        # any non-projectile object spawn
    91 : 'PROJECTILE_SPAWN_EVENT',    # any projectile object spawn
    92 : 'TEMPLATE_SPAWN_EVENT',      # any object with the given template spawn
    93 : 'CLASS_SPAWN_EVENT',         # any object of the given class spawn
    94 : 'CHARACTER_SPAWN_EVENT',     # any character-class object spawn
    96 : 'OBJECT_DESTROY_EVENT',      # any object destroyed
    # NB: It's tough to have specific destroy events (like a
    # CHARACTER_DESTROY_EVENT) because we can't check the object's
    # class, template, etc. once it's destroyed. In fact: we can't even store DESTROY
    # events as object data at all since the object will be gone and we
    # can't grab the callback data from it then. The option of storing
    # the destroy data in a mission var works, but could result
    # in some large mission vars (lists of registered objects).  
    }
for SE in SPAWN_EVENT_NAMES.keys():
    exec ('%s = %d' % (SPAWN_EVENT_NAMES[SE],SE))

# Generally, the reg function can pass an arbitrary string, float, and
# user values.  object is returned by the handler (the
# spawned/destroyed object), data is the persistent setting and option
# is the filter option (either the object class or template string)
# for spawns.
class PseudoActivationEvent:
    def __init__(self, eventType=OBJECT_SPAWN_EVENT, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0, optionVal=None):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal
        self.option = optionVal

# Trigger callback for any new object spawned.  Object
# and NIF will be set by the callback handler.
def regActivationSpawnSink(event_type, function, persistent = 0, floatVal=0, user = 0, string = '', option = None):
    CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
    callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
    callback = (function, persistent, floatVal, user, string, option)
    if not callback in callbacks:
        callbacks.append(callback)
    missionobjvar.Mission_SetVar(CBVarName, callbacks)

# Trigger callback for any new non-projectile object spawned.  Object
# and NIF will be set by the callback handler.
def regActivationDestroySink(object, function, persistent = 0, floatVal=0, user = 0, string = ''):
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    callback = (object, function, persistent, floatVal, user, string)
    if not callback in callbacks:
        callbacks.append(callback)
    missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', callbacks)

# Cancel all regActivationSpawnSink() sinks of the specifed event_type
# for a given function.  Also allows matching with the registered
# event string and/or event option values.
def cancelObjectSpawnCallbackSink(event_type,function,string=None,option=None):
    CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
    callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    # handle combinations of event string and option matching
    if ( string==None ) and ( option==None ) :    # function matching
        for cb in callbacks:
            if not ( cb[0] == function ):
                keptcallbacks.append(cb)
    elif ( string!=None ) and ( option==None ) :    # function and string matching
        for cb in callbacks:
            if not ( (cb[0] == function) and (cb[4] == string) ):
                keptcallbacks.append(cb)
    elif ( string==None ) and ( option!=None ) :    # function and option matching
        for cb in callbacks:
            if not ( (cb[0] == function) and (cb[5] == option) ):
                keptcallbacks.append(cb)
    elif ( string!=None ) and ( option!=None ) :    # function, string, and option matching
        for cb in callbacks:
            if not ( (cb[0] == function) and (cb[4] == string) and (cb[5] == option) ):
                keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar(CBVarName, keptcallbacks)

# Cancel regActivationDestroySink() sinks for a given object and
# function.
def cancelObjectDestroyCallbackSink(object, function):
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    for cb in callbacks:
        if not ( (cb[0] == object) and (cb[1] == function) ):
            keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', keptcallbacks)

# Cancel all regActivationDestroySink() sinks for a given object.
def cancelObjectDestroyCallbackSinksForObject(object):
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    for cb in callbacks:
        if not ( cb[0] == object ):
            keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', keptcallbacks)

# Cancel regActivationDestroySink() sinks for a given function.
def cancelObjectDestroyCallbackSinksForFunction(function):
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    for cb in callbacks:
        if not ( cb[1] == function ):
            keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', keptcallbacks)

# Execute the nonobject-specific callback functions
def HandleActivationSpawnCallbacks(object,NIF):
    # the first two event types are exclusive
    if NIF[-23:] != '\\projectile_standin.nif': # non-projectile spawn
        event_type = 90     # OBJECT_SPAWN_EVENT
        CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
        callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
        changed = 0
        keptcallbacks = []  # the persistent ones
        for cb in callbacks:   
            if not cb[1]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(event_type,objectStr=object, stringStr=cb[4], floatVal=cb[2], dataVal=cb[1], userVal=cb[3], optionVal=cb[5])
            fn = cb[0]
            eventeval(fn,event)
        if changed:
            missionobjvar.Mission_SetVar(CBVarName,keptcallbacks)
    else:   # projectile spawn
        event_type = 91     # OBJECT_SPAWN_EVENT
        CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
        callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
        changed = 0
        keptcallbacks = []  # the persistent ones
        for cb in callbacks:   
            if not cb[1]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(event_type,objectStr=object, stringStr=cb[4], floatVal=cb[2], dataVal=cb[1], userVal=cb[3], optionVal=cb[5])
            fn = cb[0]
            eventeval(fn,event)
        if changed:
            missionobjvar.Mission_SetVar(CBVarName,keptcallbacks)

    event_type = 92     # TEMPLATE_SPAWN_EVENT
    CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
    callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:   
        if Object_GetTemplate(object) == cb[5]: # template match
            if not cb[1]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(event_type,objectStr=object, stringStr=cb[4], floatVal=cb[2], dataVal=cb[1], userVal=cb[3], optionVal=cb[5])
            fn = cb[0]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Mission_SetVar(CBVarName,keptcallbacks)

    objclass = Object_GetClass(object)
    
    event_type = 93     # CLASS_SPAWN_EVENT
    CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
    callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:

        if objclass & cb[5]: # class match
            if not cb[1]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(event_type,objectStr=object, stringStr=cb[4], floatVal=cb[2], dataVal=cb[1], userVal=cb[3], optionVal=cb[5])
            fn = cb[0]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Mission_SetVar(CBVarName,keptcallbacks)

    event_type = 94     # CHARACTER_SPAWN_EVENT
    if objclass & OC_CHARACTER:
        CBVarName = '%s_CB' % SPAWN_EVENT_NAMES[event_type] # where these events are stored
        callbacks = missionobjvar.Mission_GetVar(CBVarName, defaultVal = [])
        changed = 0
        keptcallbacks = []  # the persistent ones
        for cb in callbacks:   
            if not cb[1]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(event_type,objectStr=object, stringStr=cb[4], floatVal=cb[2], dataVal=cb[1], userVal=cb[3], optionVal=cb[5])
            fn = cb[0]
            eventeval(fn,event)
        if changed:
            missionobjvar.Mission_SetVar(CBVarName,keptcallbacks)

# Execute the object-specific callback functions.  We still handle
# persistent sinks, in case it comes in handy for tracking re-spawned
# objects with the same name.
def HandleActivationDestroyCallbacks(object):
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:
        if cb[0] == object:
            if not cb[2]:   # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoActivationEvent(OBJECT_DESTROY_EVENT,objectStr=object, stringStr=cb[5], floatVal=cb[3], dataVal=cb[2], userVal=cb[4])
            fn = cb[1]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', keptcallbacks)

######################################################################
##
## Handling Movement Events
##
######################################################################

JUMP_START_EVENT = 80
JUMP_STOP_EVENT = 81

# Lower-level stuff.

class PseudoMovementEvent:
    def __init__(self, eventType=JUMP_START_EVENT, objectStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0, animStr=''):
        self.event = eventType
        self.object = objectStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal
        self.animation = animStr

try:
    dummyvar = Mission_GetAttr('MLOGMovement_calls')
    del dummyvar
except:
    Mission_SetAttr('MLOGMovement_calls', 0)

# Add a sink so that a script function is called when a given
# movement event type (one of those int constants above) is spotted
# during the MLOG_Watch* functions.
def regMLOGMovement(object,event_type,callback,persistent=0,float=0,user=0, string=''):
    MLOGMovement(1)
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_MOVEMENT_CB',[])
    callbacks.extend([(event_type,callback,persistent,float,user,string)])
    missionobjvar.Object_SetVar(object,'MLOG_MOVEMENT_CB',callbacks)
    Mission_SetAttr('MLOGMovement_calls', Mission_GetAttr('MLOGMovement_calls') + 1)

# Cancel all regAnimation() sinks
def cancelAllMLOGMovementSinks():
    for object in Mission_GetDynamicObjects():
        cancelMLOGMovementCallbackSinks(object)

# Execute the callback functions
def HandleMLOGMovementCallbacks(object,anim,event_type):
    callbacks = missionobjvar.Object_GetVar(object,'MLOG_MOVEMENT_CB',[])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:   
        if ( event_type == cb[0] ):
            if not cb[2]:   # this was not a persistent sink
                changed = 1
                #reduce the sink count for the module
                currentCallbacks = Mission_GetAttr('MLOGMovement_calls') - 1
                Mission_SetAttr('MLOGMovement_calls', currentCallbacks)
                #if there is no sink left at all, shut down the module
                if currentCallbacks < 1:
                    MLOGMovement(0)
            else:
                keptcallbacks.append(cb)
            event = PseudoMovementEvent(event_type,objectStr=object, stringStr=cb[5], floatVal=cb[3], dataVal=cb[2], userVal=cb[4], animStr=anim)
            fn = cb[1]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        missionobjvar.Object_SetVar(object,'MLOG_MOVEMENT_CB',keptcallbacks)

######################################################################
##
## Damage detection
##
######################################################################

# Uses include:
# - setting global state swaps on a character by character basis, including direct, area and ranged
#   area attacks
# - getting exact knockback info
# - knowing when a character has been the target of attacks which don't change character state
# - adapting AI tactics to past damage and success rate against target template
# Other possible uses, requiring additional coding in the MLOG Reader module:
# - detecting the effects of exploding objects
# - in conjunction with the MLOG 'ranged' module and regPowerUse(), using script-based powers
#   on any target spot (character, object, building, ground) for state-altering missile and beam attacks.

# Notes/Issues:
# - Applied damage: Indirect damage from knockback, irradiation, Trigger_damage, etc. is currently
#   unsupported, as the damage source isn't available. An applied damage line in ff.log might look
#   like this:
#   00:53:48.45: FF: Damage:(_skthug_with_bat01)  (applied) dmg: 0, stun time: 0
#   Note that applied damage from thrown objects is especially unreliable, since the ff.log damage is
#   output not only when hitting a target, but also every time the player displays the damage preview
#   in the status bar when targeting a potential enemy or object!
# - IMPORTANT: Target characters with spaces in their name are currently not supported. Mod makers
#   beware when creating characters or hero start position markers!


# Dictionnary of all special types powers
# Values are set as strings rather than constants because not all attacks change character state;
# values (not keys) which don't are in lowercase here.
# To get the value of a character state, simply use eval(returnedValue)
# Notes:
# - Special power Teleport Swap doesn't register at all.
# - Any attempt to hypnotise a character from the same side (hero/police vs villain/minion) is in fact
#   converted by the game engine into a 0% success rate attempt to set the target's state to normal!?!
#TODO: MLOG_ALLTYPES isn't used anymore; convert to MLOG_ALLTYPES_INVERTED format and remove the later.
MLOG_ALLTYPES = {
   '(calc': 'physical',
   'frozen': 'PCSTATE_FROZEN',
   'stun': 'PCSTATE_STUNNED', # not a state attack per se
   'static': 'PCSTATE_STATIC',
   'blank': 'PCSTATE_BLANK',
   'enraged': 'PCSTATE_ENRAGED',
   'panicked': 'PCSTATE_PANICKED',
   'hypnotised': 'PCSTATE_HYPNOTISED',
##   'normal': 'PCSTATE_HYPNOTISED', #!
   'normal': 'PCSTATE_NORMAL', # not a state attack
   'DIMENSION_SHIFT': 'PCSTATE_DIMENSION_SHIFT',
   'interrogate': 'PCSTATE_INTEROGATE', # not a state attack
   'object': 'PCSTATE_OBJECT',
   'toad': 'PCSTATE_TOAD',
   'DOPPELGANGER': 'PCSTATE_DOPPELGANGER',
   'polymorphed': 'PCSTATE_POLYMORPHED',
   'electrified': 'OBJSTATE_ELECTRIFIED', # not a state attack
   'MOLECULAR_EXCITE': 'OBJSTATE_MOLECULAR_EXCITE',
   'ALTER_GENETIC_STRUCTURE': 'OBJSTATE_ALTERED_STRUCTURE',
   'RALLY': 'OBJSTATE_ATTACKBONUS', # also available through Object_GetAttr(char, 'attackBonus')
   'DENSITY_MAXIMIZE': 'OBJSTATE_DENSITY_CHANGE',
   'irradiated': 'SCSTATE_IRRADIATED',
   'blind': 'SCSTATE_BLIND',
   'hexed': 'SCSTATE_HEXED',
   'acid_burnt': 'SCSTATE_ACID_BURNT',
   'invisible': 'SCSTATE_INVISIBLE', # not a state attack
   'POWER_NULLIFICATION': 'SCSTATE_POWER_NULLIFICATION',
   'SPAWN': 'SCSTATE_SPAWN', # not found as a power; makes the character semi-transparent.
   'ENERGIZE': 'SCSTATE_ENERGIZED',
   'empathy': 'SCSTATE_EMPATHY', # not a state attack
   'ENERGIZE_AREA': 'SCSTATE_AREA_ENERGIZED',
   'displace_image': 'SCSTATE_DISPLACE_IMAGE', # not a state attack
   'speed increased': 'SCSTATE_SPEED_INCREASED',
   'speed decreased': 'SCSTATE_SPEED_DECREASED',
   'weakened': 'SCSTATE_WEAKENED',
   'BUOYANCY': 'SCSTATE_BUOYANT',
   'zombie': 'SCSTATE_ZOMBIE', # doesn't register as attack, as target is dead
   'DEEP_SENSOR': 'SCSTATE_DEEP_SENSORED',
   'afterburner': 'afterburner', # not a state attack
   'ALTERATION': 'alteration',
   'HP_BALANCE': 'balance hp',
   'burrow': 'burrow', # not a state attack; same secondary state number as invisible
   'CEREBRAL_NORMALISATION': 'cerebral normalisation',
   'clone self': 'clone self', # not a state attack
   'CLONE_TARGET': 'clone target',
   'DISRUPT_OBJECT': 'disrupt object',
   'EP_DRAW': 'energy draw',
   'EP_TRANSFER': 'energy transfer',
   'IGNITE': 'ignite',
   'image projector': 'image projector', # not a state attack
   'law <-> order': 'law <-> order', # not a state attack
   'LIFE_GIFT': 'life gift',
   'living fires': 'living fires', # not a state attack
   'morph': 'morph', # not found
   'NORMALISE': 'normalise',
   'POSSESSION': 'possession', # destroying the possessor makes possession permanent
   'powerup': 'powerup', # not a state attack
   'REMOVE_DEFENCE': 'remove defence',
   'RESTORATION': 'restoration',
   'sonic boom': 'sonic boom', # not a state attack
   'SPEEDING_BULLET': 'speeding bullet', # note that Speeding Bullet never indicates any damage or KB!
   'sprint': 'sprint', # not a state attack
   'teleport self': 'teleport self', # not a state attack
   'teleport swap': 'teleport swap', # not a state attack
   'tumble': 'tumble', # not a state attack
   'tunnel': 'tunnel', # not a state attack
   'uncloak': 'uncloak', # not a state attack; same secondary state number as invisible
   }

# Inverted key-value pairs version of MLOG_ALLTYPES
# Ep. 2007-10-15: moved from regMLOGDamage() to improve performances a bit if regMLOGDamage() is actually used
MLOG_ALLTYPES_INVERTED = generalutils.invertDict(MLOG_ALLTYPES)

# Entries from MLOG_ALLTYPES which register in MLOG 'Damage' module
# Values are set as strings rather than constants because not all attacks change character state;
# values which don't are in lowercase here.
# To get the value of a character state, simply use eval(returnedValue)
MLOG_DAMAGETYPES = (
   'physical', # returns stun and KB
   'PCSTATE_FROZEN',
   'PCSTATE_STATIC',
   'PCSTATE_BLANK',
   'PCSTATE_ENRAGED',
   'PCSTATE_PANICKED',
   'PCSTATE_HYPNOTISED',
   'PCSTATE_DIMENSION_SHIFT', # returns stun and KB
   'PCSTATE_OBJECT',
   'PCSTATE_TOAD',
   'PCSTATE_DOPPELGANGER', # returns stun and KB
   'PCSTATE_POLYMORPHED',
   'OBJSTATE_MOLECULAR_EXCITE', # returns stun and KB
   'OBJSTATE_ALTERED_STRUCTURE', # returns stun and KB
   'OBJSTATE_ATTACKBONUS', # returns stun and KB
   'OBJSTATE_DENSITY_CHANGE', # returns stun and KB
   'SCSTATE_IRRADIATED',
   'SCSTATE_BLIND',
   'SCSTATE_HEXED',
   'SCSTATE_ACID_BURNT',
   'SCSTATE_POWER_NULLIFICATION', # returns stun and KB
   'SCSTATE_ENERGIZED', # returns stun and KB
   'SCSTATE_AREA_ENERGIZED', # returns stun and KB
   'SCSTATE_SPEED_INCREASED',
   'SCSTATE_SPEED_DECREASED',
   'SCSTATE_WEAKENED',
   'SCSTATE_BUOYANT', # returns stun and KB
   'SCSTATE_DEEP_SENSORED', # returns stun and KB
   'alteration', # returns stun and KB
   'balance hp', # returns stun and KB
   'cerebral normalisation', # returns stun and KB
   'clone target', # returns stun and KB
   'disrupt object', # returns stun and KB
   'energy draw', # returns stun and KB
   'energy transfer', # returns stun and KB
   'ignite', # returns stun and KB
   'life gift', # returns stun and KB
   'normalise', # returns stun and KB
   'possession', # returns stun and KB
   'remove defence', # returns stun and KB
   'restoration', # returns stun and KB
   'speeding bullet' # returns stun and KB, but always with zero values, even for damage
   )


class PseudoDamageEvent:
    def __init__(self, eventType='', objectStr='', targetStr='', stringStr='', floatVal=0.0, dataVal=0.0, userVal=0.0, \
                 successBol=1, damageVal=0, stunVal=0, knockbackVal=0):
        self.event = eventType
        self.object = objectStr
        self.target = targetStr
        self.string = stringStr
        self.float = floatVal
        self.data = dataVal
        self.user = userVal
        self.success = successBol
        self.damage = damageVal
        self.stun = stunVal
        self.knockback =knockbackVal

#sample damage callback function
def OnDamageTest(event):
    print "OnDamageTest(event):"
    print "    event           %s" % event
    print "    event.event     %s" % repr(event.event)
    print "    event.object    %s" % repr(event.object)
    print "    event.target    %s" % repr(event.target)
    print "    event.string    %s" % repr(event.string)
    print "    event.float     %f" % event.float
    print "    event.data      %f" % event.data
    print "    event.user      %f" % event.user
    print "    event.success   %i" % event.success
    print "    event.damage    %i" % event.damage
    print "    event.stun      %i" % event.stun
    print "    event.knockback %i" % event.knockback
        
# Add a sink so that a script function is called when a given
# Damage event type (one of those int constants above) is spotted
# during the MLOG_Watch* functions.
#
# function: the function to be called. Ex": 'myFunction'. A sample function might be:
#   def OnResistGeneticDamage(event):
#       Object_SetPrimaryState(target, PCSTATE_TOAD, 15)
#   which, when registered as
#     regMLOGDamage('OnResistGeneticDamage', object=''hero_3, target='', event_type='OBJSTATE_ALTERED_STRUCTURE', success=0, persistent=1, float=0.0, user=0.0, string='')
#   with Microwave as 'hero_3', would transform into toads targets who successfully resist his Genetic Damage direct attack.
# char = '': the attacker to watch. If empty, any character.
# target = ''. the target of an attack to watch. If empty, any target... including non-characters.
# event_type = '': the type of damage. It must be one of MLOG_DAMAGETYPES. Sadly, physical damage subtypes
#   are not available.
# success = 1: if set on 1, function will be called only if attack is successful (damage or stun or
#   effective KB is done or resistance failed). If set to 0, it will be called only on failure.
#   If set to -1, function will be called no matter what.
#   - Note that 'speeding bullet' always returns success. You'll need to check its efficiency by additional
#     means, such as checking for flailing animation or target health before/after comparison.
#   - Also note that some attack types return effect duration, while some return damage, stun duration
#     and knockback force. The rule of thumb is that offensive primary and secondary state attack usually
#     only return duration, while other types usually return damage (which is actual damage only for
#     physical type, of course), stun and KB. See the comments in the MLOG_DAMAGETYPES tuple above
#     for each type.
# persistent = 0: If set to 0, function will only be called once; if set to 1, it will be called ad vitam
#   aeternam, until explicitely removed with either cancelMLOGDamageCallbackSink(), cancelMLOGDamageCallbackSinks(), 
#   cancelAllMLOGDamageSinks() or cancelAllMLOGDamageSinksForTarget().
# float = 0: User-defined float value.
# user = '': User-defined float value.
# string = '': User-defined string value.
#TODO: add PCSTATE_NORMAL as additional detection if success 0 or -1?
def regMLOGDamage(callback, object='', target='', event_type='', success=1, persistent=0, float=0, user=0, string=''):
    #finds the ff.log name of the damage in MLOG_ALLTYPES
    damageTypeKey = ''
    if event_type != '':
        if MLOG_ALLTYPES_INVERTED.has_key(event_type):
            damageTypeKey = MLOG_ALLTYPES_INVERTED[event_type]
        else:
            print "regMLOGDamage() fatal error:"
            print "    event_type = '%s' not found"%event_type
            return
    if object == '':
        callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB', [])
        callbacks.extend( [(damageTypeKey, callback, persistent, float, user, string, target, success)] )
        missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB', callbacks)
    else:
        callbacks = missionobjvar.Object_GetVar(object, 'MLOG_DAMAGE_CB', [])
        callbacks.extend( [(damageTypeKey, callback, persistent, float, user, string, target, success)] )
        missionobjvar.Object_SetVar(object, 'MLOG_DAMAGE_CB', callbacks)

    
# Cancel all regAnimation() sinks, including generic ones (i.e., object='')
def cancelAllMLOGDamageSinks():
    missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB', [])
    for object in Mission_GetDynamicObjects():
        cancelMLOGDamageCallbackSinks(object)

# Cancel an object's regMLOGDamage() sinks for a particular callback function
def cancelMLOGDamageCallbackSink(object,callback):
    if object == '':
        callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB')
        if callbacks:   # don't make a list if there were no callbacks
            keptcallbacks = []
            for cb in callbacks:
                if cb[1] != callback:
                    keptcallbacks.append(cb)
            missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB',keptcallbacks)
    else:
        callbacks = missionobjvar.Object_GetVar(object,'MLOG_DAMAGE_CB')
        if callbacks:   # don't make a list if there were no callbacks
            keptcallbacks = []
            for cb in callbacks:
                if cb[1] != callback:
                    keptcallbacks.append(cb)
            missionobjvar.Object_SetVar(object,'MLOG_DAMAGE_CB',keptcallbacks)

# Cancel all regMLOGDamage() sinks for a particular object
def cancelMLOGDamageCallbackSinks(object):
    if object == '':
        callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB', [])
        if callbacks:   # don't make a list if there were no callbacks
            missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB',[])
    else:
        callbacks = missionobjvar.Object_GetVar(object,'MLOG_DAMAGE_CB', [])
        if callbacks:   # don't make a list if there were no callbacks
            missionobjvar.Object_SetVar(object,'MLOG_DAMAGE_CB',[])

# Cancel a target's regMLOGDamage() sinks for a particular callback function
def cancelMLOGDamageCallbackSinkForTarget(target,callback):
    #generic sinks
    callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB')
    if callbacks:   # don't make a list if there were no callbacks
        keptcallbacks = []
        for cb in callbacks:
            if (cb[1] != callback) and (cb[-2] != target):
                keptcallbacks.append(cb)
        missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB', keptcallbacks)
    #specific object sinks
    for object in Mission_GetDynamicObjects():
        callbacks = missionobjvar.Object_GetVar(object, 'MLOG_DAMAGE_CB')
        if callbacks:   # don't make a list if there were no callbacks
            keptcallbacks = []
            for cb in callbacks:
                if (cb[1] != callback) and (cb[-2] != target):
                    keptcallbacks.append(cb)
            missionobjvar.Object_SetVar(object, 'MLOG_DAMAGE_CB', keptcallbacks)

# Cancel all regMLOGDamage() sinks for a particular target
def cancelAllMLOGDamageSinksForTarget(target):
    #generic sinks
    callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB')
    if callbacks:   # don't make a list if there were no callbacks
        keptcallbacks = []
        for cb in callbacks:
            if cb[-2] != target:
                keptcallbacks.append(cb)
        missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB', keptcallbacks)
    #specific object sinks
    for object in Mission_GetDynamicObjects():
        callbacks = missionobjvar.Object_GetVar(object, 'MLOG_DAMAGE_CB')
        if callbacks:   # don't make a list if there were no callbacks
            keptcallbacks = []
            for cb in callbacks:
                if cb[-2] != target:
                    keptcallbacks.append(cb)
            missionobjvar.Object_SetVar(object, 'MLOG_DAMAGE_CB', keptcallbacks)

            

# Execute the callback functions
def HandleMLOGDamageCallbacks(object, target, event_type, damage, stun, knockback, success):
    HandleMLOGDamageCallbacks2(object, target, event_type, damage, stun, knockback, success)
    HandleMLOGDamageCallbacks2('', target, event_type, damage, stun, knockback, success)

def HandleMLOGDamageCallbacks2(object, target, event_type, damage, stun, knockback, success):
    if object == '':
        callbacks = missionobjvar.Mission_GetVar('MLOG_DAMAGE_CB', [])
    else:
        try: #Ep. 2007-03-12: issue reported by TMX related to Illusionists; I haven't been able to reproduce it, so hopefully this try/except will solve it
            callbacks = missionobjvar.Object_GetVar(object, 'MLOG_DAMAGE_CB', [])
        except:
            callbacks = []
    if not callbacks:   # don't make a list if there were no callbacks
        return
    changed = 0
    keptcallbacks = []  # the persistent ones
    for cb in callbacks:
        if ( cb[0] in (event_type,  '') ) and \
           ( cb[6] in (target, '') ) and \
           ( cb[7] in (success, -1) ):
            if not cb[2]: # this was not a persistent sink
                changed = 1
            else:
                keptcallbacks.append(cb)
            event = PseudoDamageEvent(eventType=event_type, objectStr=object, targetStr=target, stringStr=cb[5], floatVal=cb[3], dataVal=cb[2], userVal=cb[4], \
                                      successBol=success, damageVal=damage, stunVal=stun, knockbackVal=knockback)
            fn = cb[1]
            eventeval(fn,event)
        else:
            keptcallbacks.append(cb)
    if changed:
        if object == '':
            missionobjvar.Mission_SetVar('MLOG_DAMAGE_CB', keptcallbacks)
        else:
            missionobjvar.Object_SetVar(object,'MLOG_DAMAGE_CB',keptcallbacks)



MLOG_DAMAGE_PREVIOUSLINE = ""

def MLOG_WatchDamage(line):
    if chunkstring.MLOG_GetModuleName(line) != 'Damage':
        return 0
    global MLOG_DAMAGE_PREVIOUSLINE
    words = string.split(line, " ")
    #Standard damage. Second line output by the Damage module.
    if words[4][0:9] == '(resolve)': #beware, as sometimes (resolve) and the state word are together, sometimes not, even for the same term!
        char = MLOG_DAMAGE_PREVIOUSLINE[2][8:-1]
        if not Object_Exists(char): #Ep. 2007-10-15: fixing (?) an infrequent bug when char object is destroyed very quickly
            return 0
        target = MLOG_DAMAGE_PREVIOUSLINE[4]
        damageType = MLOG_DAMAGE_PREVIOUSLINE[5]
        if damageType == 'speed':
            damageType = "%s %s"%(damageType, MLOG_DAMAGE_PREVIOUSLINE[6])
        success = 0
        damage = 0
        stun = 0
        knockback = 0
        #check for state or physical damage success in lines like:
        #04:15:43.67: FF: Damage:(_skthug_with_bat08)  (resolve) dmg: 16, stun time: 0, KB: 1966
        #or
        #03:43:31.04: FF: Damage:(hero_1)  (resolve) dmg: 0, stun time: 0, KB: 0
        #(note that this might have been successful if the first line was:
        #00:16:27.18: FF: Damage:(hero_1)  _skthug_with_grenade07 SPEEDING_BULLET res: -1.000000
        if (words[-2] == 'KB:'): 
            if  (words[-1][:-1] != '0'): #words[-1][:-1] because of the \r character
                knockback = int(words[-1]) #int is much faster than eval and slightly faster than float (in Python 2.4, at least)
##                try:
                    # stumpy 2006-10-20: I see tracebacks "SystemError: NULL result without error in call_object", so let's see what's going on
                    # Ep. 2008-05-17: Got one case of it, it's a bug in FF: the target character's name was prefixed with a junk "c" character. From ff.log:
                    # 25:25:21.16: FF: Damage:(_impobj_146)  chero_2 (calc pwr)dmg: 5, stun pct: 15, KB: 8000
                    # If we can reproduce the bug, we could see if the pattern is always the same and try removing the extra character in case of failure
                    # Ep. 2008-08-12: Got another case of it, but the target was destroyed - could probably happen with static objects, tanks, russian dolls, etc. From ff.log:
                    # 00:09:47.57: FF: Damage:(centurion_1)  tank_ger1 (calc pwr)dmg: 1, stun pct: 0, KB: 0
                    # 00:09:47.57: FF: Damage:(centurion_1)  (resolve) dmg: 0, stun time: 0, KB: 0
                if Object_Exists(target):
                    if knockback - Object_GetAttr(target, 'minForce') < 0:
                        knockback = 0
                elif Object_Exists(target[1:]): #if the target name is wrong due to the FF.log extra 1st character bug
                    if knockback - Object_GetAttr(target[1:], 'minForce') < 0:
                        knockback = 0
##                except:
##                    print "MLOG_WatchDamage: error in Object_GetAttr(%s, 'minForce'), Object_Exists(%s)=%d" % (repr(target),repr(target),Object_Exists(target))
##                    print '\twords=%s MLOG_DAMAGE_PREVIOUSLINE=%s' % (words,MLOG_DAMAGE_PREVIOUSLINE)
##                    raise              
            stun = int(words[-3][:-1])
            damage = int(words[-6][:-1])
            if (damageType == 'SPEEDING_BULLET'):
                success = 1
                damage = 0 #this wasn't actual damage           
            elif (knockback + stun + damage != 0): #marginally faster than doing an equality comparison, itself marginally faster than an is comparison
                success = 1 
        #check for state success in lines like:
        #02:27:57.86: FF: Damage:(hero_0)  (resolve) panicked: 15
        #02:27:57.87: FF: Damage:(hero_0)  (resolve)panicked: resisted
        #IMPORTANT: these state changes don't return KB or stun, so we can only check for the success of the state attack itself
        elif words[-1][:-1] != 'resisted': #if words[-1][:-1] == 'resisted' we don't do anything, as the variables are already set
            success = 1
        HandleMLOGDamageCallbacks(char, target, damageType, damage, stun, knockback, success) #Ep. 2007-10-15: removed the damageType dictionary lookup which actually prevented state change powers from registering
    #Applied damage, resulting from knockback. Note that, consequently, it doesn't return KB. Damage module only shows one line.
    ##03:44:16.18: FF: Damage:(hero_1)  (applied) dmg: 10, stun time: 0
    elif words[4][0:9] == '(applied)':
        target = words[3][8:-1]
        stun = int(words[-1][:-1])
        damage = int(words[-4][:-1])
        success = (stun + damage > 0)
        HandleMLOGDamageCallbacks('', target, 'physical', damage, stun, 0, success)
    #Wielded damage. Note that it doesn't return KB. Second line output by the Damage module.
    ##02:28:06.72: FF: Damage:(hero_1)  _skthug_with_bat02 (calc pwr)dmg: 12, stun : 12, KB: 12000
    ##02:28:06.72: FF: Damage:(_skthug_with_bat02)  (wield) dmg: 17, stun time: 0
    elif words[4][0:7] == '(wield)':
        char = MLOG_DAMAGE_PREVIOUSLINE[2][8:-1]
        target = MLOG_DAMAGE_PREVIOUSLINE[4]
        stun = int(words[-1][:-1])
        damage = int(words[-4][:-1])
        success = (stun + damage > 0)
        HandleMLOGDamageCallbacks(char, target, 'physical', damage, stun, 0, success)
    #Standard or wielded damage. First line output by the Damage module, usually resolved on the following line.
    #Exceptions: successful active and passive defences get only the first line; TEMP!!! pass through?
    else: 
        MLOG_DAMAGE_PREVIOUSLINE = words
    return 1


