# -------------------------------------------------------------------------------
# skXCamera2.py (v0.75beta)
# -------------------------------------------------------------------------------
# download: http://www.frpmods.com/files/kuertee/skXCamera2V0.75Beta.zip
# read-me: http://www.frpmods.com/files/kuertee/skXCamera2.txt

# what's new:
# 1. bug-fix: doesn't fail when a query to the character's target returns a
#	non-string data type.
# 2. tweak/bug-fix: invisibility of the hero-squad removed. it was too hard to
#	determine when they should be invisible. (e.g. normal ff when the special
#	character ---------- wasn't available.) the onus is now on the module
#	that triggers the camera function to turn the heroes (if any) invisible.
# 3. tweak/bug-fix: invisible characters should not be chosen as follow or track
#	targets.
# 4. new modes:
#	0 (default) = all custom-actions available.
#	1 = watch-mode. all map entities become visible. best in watch-mode
#		skirmishes.
#		available commands: tv and follow.
#	2 = player-mode. camera will limit itself to the player. best while
#		playing a campaign or player-skirmish.
#		available commands: cinematic and lookAts.
# 5. add custom actions (or change modes on-the-fly) with:
#		fSkXWatchMode () or fSkXPlayerMode ()
# 6. remove all custom actions with: fSkXCameraClearCustomActions ()
# 7. fSkXCameraTvAvailable () is now obsolete. it'll run fSkXWatchMode ()
#	when executed.
# 8. launch from cshelper doesn't add the "add look commands" onto the portraits
# 9. save-game bug-fix. previously, the camera instance doesn't exists upon load
#	of a saved game
# 10. import routine (e.g. from cshelper) has changed. see the instructions below.
# 11. added feature: monologue with device. this triggers when a character is
#	alone and is facing an object. e.g.: Liberty Bay's mission 2 prelude when
#	Doc Astounding contacts the police commisioner.

# -------------------------------------------------------------------------------
# pros and cons:
# -------------------------------------------------------------------------------
# ffvsttr camera:
# pro 1: you can pan and rotate by moving the mouse to the edge of the screen
# pro 2: in cinematic mode (by pressing z), the camera switches between heroes
# con 1: (a biggie) you can't look upwards
 
# skXCamera:
# pro 1: you can look at anything from your character's point-of-view (e.g. look
# 	up from your character's shoulder at a blitzkrieg poster)
# pro 2: in skXCinematic mode (by choosing cinematic mode in the custom actions
# 	menu), the camera automatically rotates to the closest enemy that he/she is
# 	facing
# con 1: (a biggie) you cannot pan and rotate the camera by moving the mouse to
# 	the edge of the screen
# con 2: (another biggie) in skXCinematic mode, the camera doesn't switch
# 	between heroes. you need to specifically use the custom action on the heroes
# 	to switch the camera to him/her. tried to get this working, but i don't
# 	think there is a solution:
# 	(http://www.freedomreborn.net/invision/inde...showtopic=37356)
# con 3: (a biggie as well) skXCamera doesn't like working in underground maps.
# 	i think there is a bug in the actual ffvsttr camera. it dies when its
# 	z-location goes too far below 0.

# conclusion:
# skXCamera cinematic mode is not really playable. it does give good visuals,
# 	but it is a definate killer in big fire-fights - especially when its on a
# 	melee-type character (because the camera may move down to ground level
# 	during a brawl).

# however, skXCamera will let you enjoy the visuals of the map better. cinematic
# 	mode on flying characters is a treat. and the "look at" function paints a
# 	prettier picture than the default camera.

# -------------------------------------------------------------------------------
# features:
# -------------------------------------------------------------------------------
# 1. auto-follow and auto-track: keep both targets on screen.
# 2. 5 skXCamera modes: battle, melee, shoulder, dialogue, cinematic
# 	a. battle and melee: both targets are viewed from the side. battle-mode is
# 		very much like ff's default camera which is viewed from above.
# 		melee-mode puts skXCamera right at the duelist's level.
# 	b. shoulder: the track-target is viewed from the follow-target's right
# 		shoulder.
# 	c. dialogue: when no lookAt commands are issued during a cut-scene step,
# 		this mode views the speaker from the previous speaker's right shoulder.
# 	d. cinematic: this is similar to ff's cinematic camera. however, skXCamera
# 		tracks the closest enemy and changes the viewing angle (modes) to keep
# 		the player and the enemy in view.

# 		however, the camera is locked on the character. cinematic view will not
# 		switch to newly selected heroes without specifically turning it on from
# 		their portrait.

# 		this feature is "as playable" as the default cinematic view - which is not
# 		very much. some may find enjoyment in this "toy", though. (good for
# 		screen-shots.)
# 	battle, melee, shoulder, and dialogue modes are automatically used.
# 	cinematic mode is set by choosing "cinematic" from the custom actions.
# 3. custom actions: to enable the menu, issue an "Add look commands" from a
# 	player's portrait. i.e.: right click on the currently selected hero, and
# 	click on "Add look commands".
# 	a. Add look commands: all objects except the ground get a "Look at" action.
# 		i.e. the cursor changes into a hand when it is over an object
# 	b. Look at: look at anything in the game from the hero's point-of-view
# 	c. Cinematic: enable cinematic mode.
# 	d. Stop looking/Stop cinematic: release the SkXCamera.
# 	e. Remove look commands: removes all "Look at" actions
#	these two below are available in watch-mode:
#	f. T.V.: randomly select follow and track targets
#	g. Follow: manually select a follow target for the camera.
# 4. SkXCamera-enabled AFF Superhero TV skirmish file is included.
# 5. SkXCamera-enabled M25AI Watch-mode skirmish file is included.

# -------------------------------------------------------------------------------
# installation:
# -------------------------------------------------------------------------------
# 1. extract into a folder called skXCamera2 in the ffvsttr\ folder. ensure that
# 	the files are in this folder and not in another skXCamera2\ folder.
# 2. use startSkXCamera2.bat to start the game.
# 2. in the rumble room, select the skXCamera-enabled versions of either AFF's
# 	hero t.v. or m25ai's watch mode. note that you need ffx3 for m25ai's
# 	watch-mode.
# 3. to see skXCamera working in cut-scenes, play the campaign. you can also
# 	load a previously saved game. however, you need to restart that mission. (i
# 	think that the save game file may not initialise new modules that were not
# 	incorporated in it when it was saved.)
# 4. if you have ffx3 installed, read-on:
# 	a. ffx3 installs four modules: cshelper.py, mp.py, prestige.py, and sk.py in
# 		both the main ffvsttr\ and ffvsttr\data\missions\scripts folders.
# 	b. customised version of these modules in the mod folder will never load if
# 		they exists in ffvsttr's system folders (i.e. system\,
# 		data\missions\scripts, etc.)
# 	c. without ffx3, skXCamera is imported from a modified cshelper located in its
# 		mod folder - skXCamera2\missions\scripts\. skXCamera is loaded from cshelper
# 		so that it is enabled in all missions (campaign or skirmish) - saves from
# 		having to edit all mission files.
# 	d. with ffx3 installed, skXCamera's modified cshelper will never be used. so
# 		you need to import skXCamera at the end of the cshelper in the main
# 		ffvsttr\ folder. (because that's the cshelper that is used.) if you don't
# 		have cshelper in the main ffvsttr\, edit the one in
# 		ffvsttr\data\missions\scripts\ folder. add these lines to the very bottom
# 		of cshelper.

#		def fSkXCamera2 (tevent = None):
#			try: test = skXCamera2
#			except:
#				import skXCamera2
#				skXCamera2.fSkXCameraInitialise (0)
#			regTimer ("fSkXCamera2", 1)
#		regTimer ("fSkXCamera2", 1)

# -------------------------------------------------------------------------------
# how to use in a different mod:
# -------------------------------------------------------------------------------
# 1. copy these files from Missions\Scripts\ to the mod's Missions\Scripts\ folder:
# 	a. skXCamera2.py
# 	b. skXAngles.py
# 	c. skXMapInfo.py
# 	d. datfiles.py (if the mod doesn't already have this)
# 	e. heightcheck.py (if the mod doesn't already have this)
# 	f. heightcheckdata.py (if the mod doesn't already have this)
# 	g. missionobjvar.py (if the mod doesn't already have this)
# 2. import the module with "import skXCamera2" or "from skXCamera2 import *"
# 3. initialise with "fSkXCameraInitialise (1)" to enable the look-at and
# 	cinematic custom actions or "fSkXCameraInitialise (0)" to not.
# 4. see below for cut-scene use.

# -------------------------------------------------------------------------------
# cut-scene developement (scripting commands):
# -------------------------------------------------------------------------------
# normal speak and lookAt commands still work. however, cut-scene steps with a
# 	speak command but no lookAt command will enable skXCamera dialogue mode.
# 	this mode will continoually focus the camera on the speaker from the previous
# 	speaker's right shoulder. its best to limit the use of lookAt commands to
# 	objects spoken about and let skXCamera focus on the speakers.

# fSkXCameraStart (follow target, track target, loop):
# 	starts skXCamera.
# 	follow target = object name or tuple position
# 	track target = object name or tuple position
# 	loop = 1 or 0. defaults to 0.
# 	if the follow and track targets are not specified, then skXCamera will
# 		randomly choose a follow target and will use enemies as track targets.
# 	if loop is set to 1, the camera will change targets at every interval.
# 	e.g. spawn a hero on top of a building and set the camera to follow him
# 		and track a moving car below: fSkXCameraStart ("hero_0", "car_0")
# 	e.g. spawn two heroes. one flies toward the other and lands. set the camera
# 		to follow the waiting hero and track the flying hero:
# 		fSkXCameraStart ("hero_1", "hero_2")

# fSkXCameraStop ():
# 	stop skXCamera.
# 	apart from this, the camera will automatically be released when:
# 	a. these commands are issued: lookAtObject, lookAtMarker, lookAtPos
# 	b. at the end of the cut-scene.

# -------------------------------------------------------------------------------
# heightcheck and create building data tool
# -------------------------------------------------------------------------------
# 1. launch this in the rumble room when a new building is added to the dat-files.
#	i believe that map-conversions adds objects to the dat-files.
# 2. open heightcheck.py to read more information on this.
# 3. these files are not required after the data file is created. they can be deleted
#	safely prior to mod-delivery.

#	Art\level layout\blank_city\*.*
#	Missions\empty_city\*.*

# 4. skXCamera relies on character height data to position and point the camera
#	properly. however, by default, characters with no height information is assumed
#	to be 40 units high. you may want to create this data. heightcheck.py has more
#	information.

# -------------------------------------------------------------------------------
# credits
# -------------------------------------------------------------------------------
# irrational: ffvsttr
# kuertee: main code, trigonometry functions
# epimethee: heightcheck
# m25ai: m25ai watch-mode
# abenavides: aff superhero tv
# stumpy: missionobjvar and datfiles (both used in heightcheck)

# -------------------------------------------------------------------------------
# change-log
# -------------------------------------------------------------------------------
# v0.74Beta:
# 1. The Tv option is (by default) not available to the player. This is because
#	the Tv option will show characters/objects not in the player's visible
#	range.
# 2. To make it available (for watch-mode games), issue the command fSkXCameraTvAvailable ()
#	after initialisation.
# 3. Uses MapInfo to determine whether the camera is blocked by a building. This
#	is a big improvement compared to the previous version. However, other
#	objects (e.g. newstands, trees) may still block the view.
# 4. Bug fixed: prevents the use of building-blocked code during "dialogue" mode in
#	cutscenes
# 5. Bug fixed: prevents initialisation on non-valid maps (e.g. the Irrational Logo at
#	the beginning loads cshelper.py).
# 6. Changed: installation 4.d. - code-block to copy and paste into cshelper.py only
#	if you have cshelper.py in both the main ffvsttr folder and/or its data folder.
# 7. Added: heightcheck tool - create building data. This is a Rumble room Game mode
#	that retrieves then saves building dimensions from the objects.dat. Run this
#	when new building objects are added to the dat-files. (Note: this is optional.
#	SkXCamera2 should still work without this data. It may, however, not detect
#	that the building is blocking your view.)
# 8. Added: blurb about heightcheck (before the credits). This may also need to be run
#	when new 3D models are added to the dat-files. (Again, this is optional.
#	however, the camera may not point at a character's face or sit on a
#	character's shoulder without their proper height information.)
#
# v0.71Beta:
# 1. cinematic camera is now from a much higher (better) angle. its playable,
# 	but still too clunky for use in more than one-one-one battles. also, the
# 	feature of the view-point resolving from a mid-point of nearby heroes has
# 	been removed.
# 2. trigonometry functions corrected. correct calculation of pitch and yaw.
#
# v0.7Beta:
# 1. complete re-write of dynamicCamera
# 2. cinematic camera
# 3. block checker: checks whether a building is in the way. (it's not 100% 
# 	because i don't think there's an actual ray-trace function. however, it's
# 	better than not having any.)





#-------------------------------------------------------------------------------
#module imports
#-------------------------------------------------------------------------------
from cshelper import *
from string import *

from skXAngles import *
import heightcheck
import skXMapInfo
#


gbCameraIsOn = None  # stumpy 2006-09-17: otherwise global not declared before first reference


#-------------------------------------------------------------------------------
#speakInternal and lookAt* replacements
#-------------------------------------------------------------------------------
def fSkXCameraspeakInternal(object, tag, fn = '', user = 0, isTemplate = 0, pan = 0.5, position = -1, telepathy = 2, forceNonMission = 0, interrupt = 0):
	if (csHelper.isTrivial()):
		sayDynamic(object, tag)
	else:
		if interrupt:
			AI_CancelAllSpeech(CHANNEL_SCRIPT)
		if not isCSPlaying ():
			AI_SayPortraitSpeech(object, tag, fn, user, isTemplate, pan, position, telepathy, -1, forceNonMission)
		else:
			fSkXCameraSpeak (object, tag, fn, user, isTemplate, pan, position, telepathy, -1, forceNonMission)
speakInternal = fSkXCameraspeakInternal
ff.speakInternal = fSkXCameraspeakInternal

def fSkXCameralookAtObject (object, zoom = 0, heading = 0, time = 0, fn = ''):
	if not js.Object_Exists(object):
		return
	global gbSpeakTimeOut
	global gbCameraIsOn
	global giCameraHeading
	global giMonologueStep
	giMonologueStep = 0
	if gbCameraIsOn:
		miOffset = fSkXResolveAngle (giCameraHeading) - heading
		heading = giCameraHeading - miOffset
	if not fSkXCameraLookAt (object, fn):
		gbSpeakTimeOut = 1
		regTimer ("fSkXCameraSpeakTimeOn", 5)
		if time == 0:
			js.Camera_LookAtObject(object, zoom, 0, heading, 0, js.CPM_SNAP, js.CA_MOVE, fn)
		else:
			js.Camera_LookAtObject(object, zoom, 0, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)
lookAtObject = fSkXCameralookAtObject
ff.lookAtObject = fSkXCameralookAtObject

def fSkXCameralookAtMarker (marker, zoom = 0, heading = 0, time = 0, fn = ''):
	global giMonologueStep
	giMonologueStep = 0
	if gbCameraIsOn:
		miOffset = fSkXResolveAngle (giCameraHeading) - heading
		heading = giCameraHeading - miOffset
	global gbSpeakTimeOut
	if not fSkXCameraLookAt (Get_MarkerPos (marker), fn):
		gbSpeakTimeOut = 1
		regTimer ("fSkXCameraSpeakTimeOn", 5)
		global giCameraHeading
		if time == 0:
			js.Camera_LookAtMarker(marker, zoom, 0, heading, time, js.CPM_SNAP, js.CA_MOVE, fn)
		else:
			js.Camera_LookAtMarker(marker, zoom, 0, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)
lookAtMarker = fSkXCameralookAtMarker
ff.lookAtMarker = fSkXCameralookAtMarker

def fSkXCameralookAtPos (pos, zoom = 0, pitch = 0, heading = 0, time = 0, fn = ''):
	global gbCameraIsOn
	global giMonologueStep
	giMonologueStep = 0
	if gbCameraIsOn:
		miOffset = fSkXResolveAngle (giCameraHeading) - heading
		heading = giCameraHeading - miOffset
	global gbSpeakTimeOut
	if not fSkXCameraLookAt (pos, fn):
		gbSpeakTimeOut = 1
		regTimer ("fSkXCameraSpeakTimeOn", 5)
		global giCameraHeading
		if time == 0:
			js.Camera_LookAtPos(pos, zoom, pitch, heading, 0, js.CPM_SNAP, js.CA_MOVE, fn)
		else:
			js.Camera_LookAtPos(pos, zoom, pitch, heading, time, js.CPM_SCROLLTO, js.CA_MOVE, fn)
lookAtPos = fSkXCameralookAtPos
ff.lookAtPos = fSkXCameralookAtPos

def fSkXCameraendCS():
	js.Camera_SetWidescreen(0)
	js.UI_Enable(1)

#	  js.Camera_Release()
	fSkXCameraStop ()
	AI_EnableDynamicSpeech(1)
endCS = fSkXCameraendCS
ff.endCS = fSkXCameraendCS

try:
	ff.cshelper.speakInternal = fSkXCameraspeakInternal
	ff.cshelper.lookAtObject = fSkXCameralookAtObject
	ff.cshelper.lookAtMarker = fSkXCameralookAtMarker
	ff.cshelper.lookAtPos = fSkXCameralookAtPos
	ff.cshelper.endCS = fSkXCameraendCS
except:
	pass

def fSkXCameraSpeak (object, tag, fn = '', user = 0, isTemplate = 0, pan = 0.5, position = -1, telepathy = 2, forceNonMission = 0, interrupt = 0):
	fSkXCameraInitialise ()
	fSkXCameraCustomActions ()
	global gbCameraIsOn
	global giCameraClock
	global gsLastSpeaker
	global gbIsSpeaking
	global gaListeners
	global gbSpeakTimeOut
	global gtLookAt
	global giMonologueStep
	#
	#get camera, speaker and listener data
	if not len (gtLookAt):
		mtCamera = Camera_GetLookAtInfo ()
		gtLookAt = (mtCamera [0], mtCamera [1], mtCamera [2])
	#
	#get speaker
	msSpeaker = object
	if not Object_Exists (msSpeaker):
		#make guess who is speaking by choosing an object with a suitable template
		maPossibles = []
		for msC in Mission_GetDynamicObjects ():
			if object [:5] == Object_GetTemplate (msC) [:5]:
				maPossibles.append (msC)
		if len (maPossibles):
			msSpeaker = fGetClosest (gtLookAt, maPossibles)
		if not Object_Exists (msSpeaker) and Object_Exists (gsLastSpeaker):
			msSpeaker = gsLastSpeaker
	if Object_Exists (msSpeaker):
		mtSpeaker = Get_ObjectPos (msSpeaker)
	else:
		mtSpeaker = gtLookAt
	#
	#get listener
	msListener = gsLastSpeaker
	if msListener == "":
		msListener = msSpeaker
	if isTuplePosition (msListener):
		mtListener = msListener
	elif Object_Exists (msListener):
		mtListener = Get_ObjectPos (msListener)
	else:
		msListener = msSpeaker
		mtListener = mtSpeaker
	miDistance = fSkXGetDistance (mtListener, mtSpeaker)
	if miDistance > DISTANCE_MEDIUM:
		msListener = msSpeaker
		mtListener = mtSpeaker
		miDistance = 0
	#
	#turn those not scheduled to animate to the speaker
	maListeners = gaListeners [:]
	if msSpeaker in maListeners:
		maListeners.remove (msSpeaker)
	if not len (maListeners) and msSpeaker == gsLastSpeaker: giMonologueStep = giMonologueStep + 1
	else: giMonologueStep = 0
	maAnimating = []
	if csHelper.isPlaying ():
		maCSCommands = csHelper.csList [csHelper.stepIndex]
		for msCSCommand in maCSCommands:
			for msC in maListeners:
				if find (msCSCommand, msC [:5]) > -1 and Object_Exists (msC):
					maAnimating.append (msC)
	for msC in maListeners:
		if msC not in maAnimating and Object_Exists (msC):
			Trigger_Move (msC, "turn", mtSpeaker)
	#
	#if not timed out or disance < medium, then do skXCamera
	if not gbSpeakTimeOut:
		gbSpeakTimeOut = 0
		gbIsSpeaking = 1
		gbCameraIsOn = 1
		fSkXCameraSetUp (tsFollow = msListener, tsTarget = msSpeaker, tbLoop = 0)
	#
	if not isTuplePosition (gsLastSpeaker) and Object_Exists (msSpeaker):
		gsLastSpeaker = msSpeaker
	if msSpeaker not in gaListeners:
		gaListeners.append (msSpeaker)
	AI_SayPortraitSpeech (object, tag, fn, user, isTemplate, pan, position, telepathy, -1, forceNonMission)

def fSkXCameraIsLookAtInCSStep (tiStep):
	mbLookAtInStep = 0
	if csHelper.isPlaying () and len (csHelper.csList [tiStep]) < csHelper.csLength:
		maCSCommands = csHelper.csList [tiStep]
		for msCSCommand in maCSCommands:
			if find (msCSCommand, "look") > -1:#find (msCSCommand, "lookAt") > -1:# or find (msCSCommand, "speak") > -1:
				mbLookAtInStep = 1
	return mbLookAtInStep

def fSkXCameraLookAt (object, fn = ""):
	fSkXCameraInitialise ()
	fSkXCameraCustomActions ()
	global gsLastSpeaker
	global gbCameraIsOn
	global gsCallBack
	global gtLookAt
	mbOK = 0
	msLookAt = object
	if isTuplePosition (msLookAt):
		mtLookAt = msLookAt
	else:
		mtLookAt = Get_ObjectPos (msLookAt)
	gtLookAt = mtLookAt
	#
	#check if CS step has look command
	mbLookAtInCSStep = 0
	mbLookAtInNextCSStep = 0
	if csHelper.isPlaying ():
		mbLookAtInCSStep = fSkXCameraIsLookAtInCSStep (csHelper.stepIndex)
		if csHelper.stepIndex < csHelper.csLength:
			mbLookAtInNextCSStep = fSkXCameraIsLookAtInCSStep (csHelper.stepIndex + 1)
		else:
			mbLookAtInNextCSStep = 0
	#
	if not mbLookAtInCSStep and (Object_Exists (gsLastSpeaker) or isTuplePosition (gsLastSpeaker)):
		if isTuplePosition (gsLastSpeaker):
			mtLookFrom = gsLastSpeaker
		else:
			mtLookFrom = Get_ObjectPos (gsLastSpeaker)
		miDistance = fSkXGetDistance (mtLookFrom, mtLookAt)
		if miDistance < DISTANCE_MEDIUM:
			mbOK = 1
			gbCameraIsOn = 1
			gsCallBack = fn
			fSkXCameraSetUp (tsFollow = gsLastSpeaker, tsTarget = msLookAt, tbLoop = 0, tsCallBack = fn)
	if not mbOK:
		gbCameraIsOn = 0
	return mbOK

def fSkXCameraSpeakTimeOn (tevent = ""):
	global gbSpeakTimeOut
	gbSpeakTimeOut = 0
ff.fSkXCameraSpeakTimeOn = fSkXCameraSpeakTimeOn
#



#
#-------------------------------------------------------------------------------
#initialisation
#-------------------------------------------------------------------------------
try:
	MODULE_FFX = ff.ffx
	Object_GetTemplate = ff.ffx.FFX_GetTemplate
except:
	MODULE_FFX = 0

try:
	MODULE_M25AI = ff.ffx.m25ai
	DISTANCE_VISIBLE = MODULE_M25AI.DISTANCE_VISIBLE
	DISTANCE_LONG = MODULE_M25AI.DISTANCE_LONG
	DISTANCE_MEDIUM = MODULE_M25AI.DISTANCE_MEDIUM
	DISTANCE_SHORT = MODULE_M25AI.DISTANCE_SHORT
	DISTANCE_MELEE = MODULE_M25AI.DISTANCE_MELEE
except:
	MODULE_M25AI = 0
	DISTANCE_VISIBLE = 400
	DISTANCE_LONG = 300
	DISTANCE_MEDIUM = 200
	DISTANCE_SHORT = 100
	DISTANCE_MELEE = 50

WATCH_INTERVAL = 20
SKX_CAMERA_TICK_INTERVAL = 1
STANDARD_HUMAN_HEIGHT = 40
ANGLE_SNAP = 15
DISTANCE_SNAP = 25
GET_ENEMY = None
MODE_WATCH = 1
MODE_PLAYER = 2

def fSkXCameraInitialise (tbEnableCustomActions = 0, tiMode = None):
	global gbSkXCameraIsInitialised
	try: mbTest = gbSkXCameraIsInitialised
	except:
		gbSkXCameraIsInitialised = 1
		if len (Mission_GetObjects ()):
			#skXCamera-specific
			global gbCameraIsOn
			global gsFollowTarget
			global gsFollowTarget2
			global gsTrackTarget
			global gsTrackTarget2
			global gbCameraIsTicking
			global giCameraHeading
			global giCameraZoom
			global gtCameraCam
			global gbPaused
			global giCameraClock
			global gsFromSide
			gbCameraIsOn = 0
			gsFollowTarget = ""
			gsFollowTarget2 = ""
			gsTrackTarget = ""
			gsTrackTarget2 = ""
			gbCameraIsTicking = 0
			giCameraHeading = 0
			giCameraZoom = 0
			gtCameraCam = None
			gbPaused = 0
			giCameraClock = 0
			gsFromSide = "right"
			skXMapInfo.GetMapInfo ()
	
			#skXspeakInternal, skXlookAt*
			global gsLastSpeaker
			global gbIsSpeaking
			global gsCallBack
			global gaListeners
			global gbSpeakTimeOut
			global gtLookAt
			global giMonologueStep
			gsLastSpeaker = ""
			gbIsSpeaking = 0
			gsCallBack = ""
			gaListeners = []
			gbSpeakTimeOut = 0
			gtLookAt = ()
			giMonologueStep = 0
			
			#custom actions
			global gbIsLooking
			global gsLookingAt
			global gbIsFollowing
			global gsFollowing
			global gbIsCinematic
			global gbIsTV
			global gsLastActioner
			global gaPCs
			gbIsLooking = 0
			gsLookingAt = ""
			gbIsFollowing = 0
			gsFollowing = ""
			gbIsCinematic = 0
			gbIsTV = 0
			gsLastActioner = ""
			gaPCs = []
			global gbMode
			gbMode = tiMode
			fSkXCameraClearCustomActions ()
			#if tbEnableCustomActions:
			#	fSkXCameraCustomActions ()

def isSkXCameraInitialised ():
	global gbSkXCameraIsInitialised
	try: initialised = gbSkXCameraIsInitialised
	except: initialised = None
	return initialised

#-------------------------------------------------------------------------------
#custom actions
#-------------------------------------------------------------------------------
def fSkXWatchMode (tevent = None):
	global gbMode
	gbMode = MODE_WATCH
	fSkXCameraClearCustomActions ()
	fSkXCameraCustomActions ()

def fSkXPlayerMode (tevent = None):
	global gbMode
	gbMode = MODE_PLAYER
	fSkXCameraClearCustomActions ()
	fSkXCameraCustomActions ()

def fSkXCameraTvAvailable (tb = None):
	fSkXWatchMode ()
ff.fSkXCameraTvAvailable = fSkXCameraTvAvailable

def fSkXCameraAddLookCustomActions (tsTarget = "", tsSource = ""):
	fSkXCameraClearCustomActions (tsSource)
	global gbMode
	if gbMode:
		for msO in Mission_GetObjects ():
			if msO != "_impobj_0" and msO != tsSource:
				if gbMode != MODE_PLAYER: Mission_CustomAction ("Follow", tsSource, msO, "fSkXCameraFollow", 1000000, 0)
				if gbMode != MODE_WATCH: Mission_CustomAction ("Look at", tsSource, msO, "fSkXCameraLook", 1000000, 0)
				Mission_CustomAction ("Stop looking/following", tsSource, msO, "fSkXCameraStopLooking", 1000000, 0)
		if gbMode != MODE_WATCH: Mission_CustomAction ("Cinematic", tsSource, tsSource, "fSkXCameraCinematic", 1000000, 0)
		if gbMode != MODE_PLAYER: Mission_CustomAction ("T.V.", tsSource, tsSource, "fSkXCameraTV", 1000000, 0)
		Mission_CustomAction ("Remove look commands", tsSource, tsSource, "fSkXCameraCustomActions", 1000000, 0)
ff.fSkXCameraAddLookCustomActions = fSkXCameraAddLookCustomActions

def fSkXCameraCustomActions (tsTarget = "", tsSource = ""):
	global gbMode
	if gbMode:
		fSkXCameraClearCustomActions (tsSource)
		global gbIsLooking
		global gbIsFollowing
		global gbIsCinematic
		if gbIsLooking:
			fSkXCameraStopLooking (tsTarget, tsSource, 0)
		if gbIsFollowing:
			fSkXCameraStopFollowing (tsTarget, tsSource, 0)
		if gbIsCinematic:
			fSkXCameraStopCinematic (tsTarget, tsSource, 0)
		if tsSource == "":
			for msHero in getAllHeroes ():
				Mission_CustomAction ("Add look commands", msHero, msHero, "fSkXCameraAddLookCustomActions", 1000000, 0)
		else:
			Mission_CustomAction ("Add look commands", tsSource, tsSource, "fSkXCameraAddLookCustomActions", 1000000, 0)
ff.fSkXCameraCustomActions = fSkXCameraCustomActions

def fSkXCameraClearCustomActions (ts = "", tbLeaveLooks = 0, tbLeaveFollowings = 0):
	Mission_RemoveCustomAction ("Add look commands", ts)
	if not tbLeaveLooks:
		Mission_RemoveCustomAction ("Look at", ts)
		Mission_RemoveCustomAction ("Stop looking/following", ts)
	if not tbLeaveFollowings:
		Mission_RemoveCustomAction ("Follow", ts)
		Mission_RemoveCustomAction ("Stop Following", ts)
	Mission_RemoveCustomAction ("Cinematic", ts)
	Mission_RemoveCustomAction ("Stop cinematic", ts)
	Mission_RemoveCustomAction ("T.V.", ts)
	Mission_RemoveCustomAction ("T.V. off", ts)
	Mission_RemoveCustomAction ("Remove look commands", ts)

def fSkXCameraTV (tsTarget = "", tsSource = ""):
	global gbIsTV
	global gsLastActioner
	fSkXCameraClearCustomActions (tsSource)
	Mission_CustomAction ("T.V. off", tsSource, tsSource, "fSkXCameraTVOff", 1000000, 0)
	if gsLastActioner != tsSource and gsLastActioner != "":
		fSkXCameraCustomActions ("", gsLastActioner)
	see ()
	fSkXCameraStart (tbLoop = 1)
	gbIsTV = 1
	gsLastActioner = tsSource
ff.fSkXCameraTV = fSkXCameraTV

def fSkXCameraTVOff (tsTarget = "", tsSource = "", tbRestoreCommands = 1):
	global gbIsTV
	global gsLastActioner
	if tbRestoreCommands:
		fSkXCameraAddLookCustomActions (tsTarget, tsSource)
	fSkXCameraStop ()
	gbIsTV = 0
	gsLastActioner = ""
ff.fSkXCameraTVOff = fSkXCameraTVOff

def fSkXCameraCinematic (tsTarget = "", tsSource = ""):
	global gbIsCinematic
	global gsLastActioner
	fSkXCameraClearCustomActions (tsSource)
	Mission_CustomAction ("Stop cinematic", tsSource, tsSource, "fSkXCameraStopCinematic", 1000000, 0)
	if gsLastActioner != tsSource and gsLastActioner != "":
		fSkXCameraCustomActions ("", gsLastActioner)
	see ()
	fSkXCameraStart (tsSource, tbLoop = 0)
	gbIsCinematic = 1
	gsLastActioner = tsSource
ff.fSkXCameraCinematic = fSkXCameraCinematic

def fSkXCameraStopCinematic (tsTarget = "", tsSource = "", tbRestoreCommands = 1):
	global gbIsCinematic
	global gsLastActioner
	if tbRestoreCommands:
		fSkXCameraAddLookCustomActions (tsTarget, tsSource)
	blind ()
	fSkXCameraStop ()
	gbIsCinematic = 0
	gsLastActioner = ""
ff.fSkXCameraStopCinematic = fSkXCameraStopCinematic

def fSkXCameraLook (tsTarget = "", tsSource = ""):
	global gbIsLooking
	global gsLookingAt
	global gsLastActioner
	fSkXCameraClearCustomActions (tsSource, tbLeaveLooks = 1)
	if not gbIsLooking:
		Mission_CustomAction ("Stop looking/following", tsSource, tsSource, "fSkXCameraStopLooking", 1000000, 0)
	if gsLastActioner != tsSource and gsLastActioner != "":
		fSkXCameraCustomActions ("", gsLastActioner)
	if gsLookingAt == tsTarget:
		fSkXCameraStopLooking (tsTarget, tsSource)
	else:
		fSkXCameraStart (tsSource, tsTarget, tbLoop = 0)
	gbIsLooking = 1
	gsLookingAt = tsTarget
	gsLastActioner = tsSource
ff.fSkXCameraLook = fSkXCameraLook

def fSkXCameraStopLooking (tsTarget = "", tsSource = "", tbRestoreCommands = 1):
	global gbIsLooking
	global gbIsFollowing
	if gbIsFollowing:
		fSkXCameraStopFollowing (tsTarget, tsSource, tbRestoreCommands)
	elif gbIsLooking:
		global gsLastActioner
		if tbRestoreCommands:
			fSkXCameraAddLookCustomActions (tsTarget, tsSource)
		fSkXCameraStop ()
		gbIsLooking = 0
		gsLastActioner = ""
ff.fSkXCameraStopLooking = fSkXCameraStopLooking

def fSkXCameraFollow (tsTarget = "", tsSource = ""):
	global gbIsFollowing
	global gsFollowing
	global gsLastActioner
	fSkXCameraClearCustomActions (tsSource, tbLeaveFollowings = 1)
	if not gbIsFollowing:
		Mission_CustomAction ("Stop looking/following", tsSource, tsSource, "fSkXCameraStopLooking", 1000000, 0)
	if gsLastActioner != tsSource and gsLastActioner != "":
		fSkXCameraCustomActions ("", gsLastActioner)
	if gsFollowing == tsTarget:
		fSkXCameraStopFollowing (tsTarget, tsSource)
	else:
		fSkXCameraStart (tsTarget, tbLoop = 1)
	gbIsFollowing = 1
	gsFollowing = tsTarget
	gsLastActioner = tsSource
ff.fSkXCameraFollow = fSkXCameraFollow

def fSkXCameraStopFollowing (tsTarget = "", tsSource = "", tbRestoreCommands = 1):
	global gbIsFollowing
	global gsLastActioner
	if tbRestoreCommands:
		fSkXCameraAddLookCustomActions (tsTarget, tsSource)
	fSkXCameraStop ()
	gbIsFollowing = 0
	gsLastActioner = ""
ff.fSkXCameraStopFollowing = fSkXCameraStopFollowing

#-------------------------------------------------------------------------------
#skXCamera2
#-------------------------------------------------------------------------------
def fSkXCameraStart (tsFollow = None, tsTrack = None, tbLoop = 0, tsCallBack = ""):
	global gbCameraIsOn
	global giCameraClock
	fSkXCameraInitialise ()
	gbCameraIsOn = 1
	giCameraClock = 0
	fSkXCameraSetUp ("", tsFollow, tsTrack, tbLoop, tsCallBack)
ff.fSkXCameraStart = fSkXCameraStart

def fSkXCameraStop (tsTarget = "", tsSource = "", tbRelease = 1):
	global gbCameraIsOn
	global giCameraClock
	global gbCameraIsTicking
	global gbIsSpeaking
	global gsLastSpeaker
	global gbSpeakTimeOut
	gbCameraIsOn = 0
	giCameraClock = 0
	gbCameraIsTicking = 0
	gbIsSpeaking = 0
	gsLastSpeaker = ""
	gbSpeakTimeOut = 0
	if tbRelease:
		Camera_Release ()
ff.fSkXCameraStop = fSkXCameraStop

def fSkXCameraSetUp (tevent = "", tsFollow = None, tsTarget = None, tbLoop = 1, tsCallBack = ""):
	fSkXCameraInitialise ()
	global gbIsFollowing
	global gbCameraIsOn
	global gbCameraIsTicking
	global gsFollowTarget
	global gsTrackTarget
	if gbCameraIsOn:
		if tsFollow != None or isTuplePosition (tsFollow):
			gsFollowTarget = tsFollow
		elif not gbIsFollowing:
			if gsFollowTarget != "" and Object_IsAlive (gsFollowTarget):
				miHealth = Object_GetAttr (gsFollowTarget, "health")
				miMaxHealth = Object_GetAttr (gsFollowTarget, "maxHealth")
				msEnemy = fGetClosestEnemy (gsFollowTarget)
				if msEnemy != "":
					miDistanceToEnemy = fSkXGetDistance (Get_ObjectPos (gsFollowTarget), Get_ObjectPos (msEnemy))
				else:
					miDistanceToEnemy = DISTANCE_LONG + 1
				if miHealth > miMaxHealth * 0.25 or miDistanceToEnemy > DISTANCE_LONG:
					maCs = fGetValidCs ()
					if gsFollowTarget in maCs and len (maCs) > 1:
						maCs.remove (gsFollowTarget)
					if len (maCs):
						msC = maCs [randint (0, len (maCs) - 1)]
						gsFollowTarget = msC
			else:
				maCs = fGetValidCs ()
				if len (maCs):
					msC = maCs [randint (0, len (maCs) - 1)]
					gsFollowTarget = msC
		if tsTarget or isTuplePosition (tsTarget):
			gsTrackTarget = tsTarget
		else:
			gsTrackTarget = ""
		if gbCameraIsOn:
			if gbCameraIsTicking == 0:
				fSkXCameraTick ()
			if tbLoop:
				if type (gsFollowTarget) == StringType and Object_IsAlive (gsFollowTarget):
					regTimer ("fSkXCameraSetUp", WATCH_INTERVAL)
				else:
					regTimer ("fSkXCameraSetUp", SKX_CAMERA_TICK_INTERVAL)
	return gbCameraIsOn
ff.fSkXCameraSetUp = fSkXCameraSetUp

def fSkXCameraTick (tevent = ""):
	fSkXCameraInitialise ()
	ff.RPG_KB_JITTER_SCALE = 0
	global gbCameraIsOn
	global gsFollowTarget
	global gsFollowTarget2
	global gsTrackTarget
	global gsTrackTarget2
	global gbCameraIsTicking
	global giCameraClock
	global giCameraHeading
	global giCameraZoom
	global gtCameraCam
	global gbPaused
	global gsFromSide
	#
	global gbIsSpeaking
	global gsCallBack
	global gbIsLooking
	global gbIsFollowing
	global gbIsCinematic
	global gbIsTV
	if gbCameraIsOn:
		msFollow = gsFollowTarget
		mtFollow = None
		mbOK = 0
		if msFollow:
			if isTuplePosition (msFollow) \
			or (type (msFollow) == StringType and Object_Exists (msFollow)):
				mbOK = 1

		if mbOK:#msFollow == tuple or exists
			if not isTuplePosition (msFollow):
				miOrientation = fSkXResolveAngle (fSkXGetDegreesFromRadians (Object_GetOrientation (msFollow) [1]) - 180)
				try: mtFollow = Get_ObjectPos (msFollow)#ff.MapInfo [msFollow]["projectedPos"]
				except: pass
				if not isTuplePosition (mtFollow): mtFollow = Get_ObjectPos (msFollow)
			else: mtFollow = msFollow
			mtFollowCam = (mtFollow [0], mtFollow [1], mtFollow [2])

			msTrack = gsTrackTarget
			mtTrack = None
			mbOK2 = 0
			if isTuplePosition (msTrack): mbOK2 = 1
			elif type (msTrack) == StringType and Object_IsAlive (msTrack): mbOK2 = 1
			elif type (msTrack) == StringType and not Object_IsAlive (msTrack):
				if not isTuplePosition (msFollow) and not gbIsSpeaking and not gbIsLooking:
					#if follow=character and mode!=cutscene, track=enemy of follow
					msTrack = gsTrackTarget2
					if gbIsCinematic or not Object_IsAlive (msTrack) or msFollow != gsFollowTarget2:
						if not GET_ENEMY: msTrack = fGetClosestEnemy (msFollow, DISTANCE_LONG)
						else: msTrack = GET_ENEMY (msFollow) or ""
					if Object_IsAlive (msTrack):
						mbOK2 = 1
					else:
						msTrack = ""
						mbOK2 = 0

			if mbOK2:#msTrack == tuple or alive
				if not isTuplePosition (msTrack):
					try: mtTrack = Get_ObjectPos (msTrack)#ff.MapInfo [msTrack]["projectedPos"]
					except: pass
					if not isTuplePosition (mtTrack): mtTrack = Get_ObjectPos (msTrack)
				else: mtTrack = msTrack
				mtTrackCam = (mtTrack [0], mtTrack [1], mtTrack [2])

				miDistance = fSkXGetDistance (mtFollow, mtTrack)
				if gbIsSpeaking:#cut-scene
					if msFollow == msTrack:#close-up of speaker's face
						msMode = "face"
						maCams = fSkXCameraGetCameraPos (msFollow, "face")
					elif miDistance > DISTANCE_MEDIUM:#listener too far from speaker,
						#so switch (because speak passes listener as follow and speaker as track)
						msMode = "dialogue"
						maCams = fSkXCameraGetCameraPos (msTrack, "shoulder", msFollow)
					else:#default shoulder-cam
						msMode = "shoulder (gbIsSpeaking)"
						maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
				elif gbIsLooking or miDistance > DISTANCE_VISIBLE:#default shoulder-cam
					if isTuplePosition (msFollow) or isTuplePosition (msTrack):
						msMode = "shoulder (from/to tuple)"
						maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
					#elif Object_Exists (msFollow) and not isCharacterTemplate (Object_GetTemplate (msFollow)): 
					#	msMode = "shoulder (non-character follow)"
					#	maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
					#elif Object_Exists (msTrack) and not isCharacterTemplate (Object_GetTemplate (msTrack)):
					#	msMode = "shoulder (non-character track)"
					#	maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
					elif Object_Exists (msFollow) and Object_Exists (msTrack):
						msMode = "shoulder (object to object)"
						maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
					else:
						msMode = "far ahead"
						maCams = fSkXCameraGetCameraPos (msFollow, "far ahead")
				elif miDistance > DISTANCE_LONG:#default shoulder-cam
					#msFollow at shoulder and msTrack at face
					msMode = "shoulder (medium distance)"
					maCams = fSkXCameraGetCameraPos (msFollow, "shoulder", msTrack)
				elif miDistance > DISTANCE_SHORT or abs (mtTrackCam [2] - mtFollowCam [2]) > DISTANCE_SHORT:#combat cam
					msMode = "combat"
					maCams = fSkXCameraGetCameraPos (msFollow, "combat", msTrack)
				else:#melee cam
					msMode = "melee"
					maCams = fSkXCameraGetCameraPos (msFollow, "melee", msTrack)
			else:#look ahead
				msMode = "far ahead (track not valid)"
				maCams = fSkXCameraGetCameraPos (msFollow, "far ahead")
			mtFollowCam = maCams [0]
			mtTrackCam = maCams [1]
			#if msMode == "combat" or msMode == "melee":
			#	#add "margin" beyond mtFollowCam and mtTrackCam
			#	miA = fSkXGetAngle (mtFollowCam, mtTrackCam)
			#	mtFollowCam = fSkXGetPointFromVector (mtFollowCam, fSkXResolveAngle (miA + 180), DISTANCE_MELEE)
			#	mtTrackCam = fSkXGetPointFromVector (mtTrackCam, miA, DISTANCE_MELEE)

			#cinematic, check for facing
			if gbIsCinematic and not isTuplePosition(msTrack):
				if Object_Exists (msTrack):
					if isCharacterTemplate (Object_GetTemplate (msTrack)):
						miHeading = fSkXGetAngle (mtFollow, mtTrack)
						mbLookAhead = 0
						try: mbLookAhead = ff.MapInfo [msFollow]["distanceTravelled"] > DISTANCE_SNAP
						except: pass
						miFacing = fSkXSubtractAngles (miOrientation, miHeading)
						if mbLookAhead and miFacing > 90:
							#not facing target, so look far ahead
							msMode = "far ahead due to target not in facing"
							maCams = fSkXCameraGetCameraPos (msFollow, "far ahead")
							mtFollowCam = maCams [0]
							mtTrackCam = maCams [1]

			#block checker
			if not gbIsLooking and not gbIsSpeaking and skXMapInfo.MapInfo_IsPointInsideBuilding (mtFollowCam):
				mtMid = fSkXGetPointInBetween (mtFollowCam, mtFollow)
				mtOutOfBuilding = skXMapInfo.MapInfo_EnsurePointIsOutsideBuilding (mtMid)
				mtFollowCam = (mtOutOfBuilding [0], mtOutOfBuilding [1], mtFollowCam [2])

			#resolve zoom
			miZAngle = fSkXGetAngle (mtFollowCam, mtTrackCam, "pitch")
			if miZAngle == 0:#looking up
				miZAngle = 90#make look at same height level
			#angle 0 = zoom -500
			#angle 90 = zoom -250
			#angle 180 = zoom 0
			miZAngle180Factor = miZAngle / 180.0
			miZoom = -(550 - miZAngle180Factor * 550)

			#resolve heading
			miHeading = fSkXResolveAngle (fSkXGetAngle (mtFollowCam, mtTrackCam))
			miDiff = fSkXSubtractAngles (fSkXResolveAngle (giCameraHeading), miHeading)
			if abs (int (fSkXResolveAngle (giCameraHeading + miDiff)) - int (miHeading)) < ANGLE_SNAP:
				miHeading = giCameraHeading + miDiff
			elif abs (int (fSkXResolveAngle (giCameraHeading - miDiff)) - int (miHeading)) < ANGLE_SNAP:
				miHeading = giCameraHeading - miDiff
			if miDiff < ANGLE_SNAP: miHeading = giCameraHeading

			if not gtCameraCam: gtCameraCam = mtFollowCam
			if gbCameraIsTicking and fSkXGetDistance (mtFollowCam, gtCameraCam) < DISTANCE_VISIBLE:
				Camera_LookAtPos (mtFollowCam, miZoom, 0, miHeading - 45, SKX_CAMERA_TICK_INTERVAL, CPM_SCROLLTO, CA_MOVE, "fSkXCameraCallBack")
			else:
				Camera_LookAtPos (mtFollowCam, miZoom, 0, miHeading - 45, 0, CPM_SNAP, CA_MOVE, "fSkXCameraCallBack")
			if gbIsCinematic and not gbPaused and msTrack != gsTrackTarget2 and Object_Exists (msTrack):
				if isCharacterTemplate (Object_GetTemplate (msTrack)):
					regTimer ("fSkXCameraPause", 0.5)
			giCameraHeading = miHeading
			giCameraZoom = miZoom
			gtCameraCam = mtFollowCam
			gsTrackTarget2 = msTrack
			gsFollowTarget2 = msFollow
		gbCameraIsTicking = 1
		giCameraClock = giCameraClock + 1
		regTimer ("fSkXCameraTick", SKX_CAMERA_TICK_INTERVAL)
	else:
		gbCameraIsTicking = 0
		giCameraClock = 0
ff.fSkXCameraTick = fSkXCameraTick

def fSkXCameraPause (tevent = ""):
	global gbPaused
	Mission_SetPaused (1)
	gbPaused = 1
	regTimer ("fSkXCameraUnpause", 1)
ff.fSkXCameraPause = fSkXCameraPause

def fSkXCameraUnpause (tevent = ""):
	global gbPaused
	Mission_SetPaused (0)
	gbPaused = 0
ff.fSkXCameraUnpause = fSkXCameraUnpause

def fSkXCameraCallBack (tevent = ""):
	global gsCallBack
	if gsCallBack != "":
		exec (gsCallBack + " (tevent)")
		gsCallBack = ""

def fSkXCameraGetCameraPos (tsC, tsPos = "face", tsCFacing = ""):
	global gbIsSpeaking
	global gbIsCinematic
	global gsFromSide
	global giMonologueStep
	if isTuplePosition (tsC):
		mt = tsC
	else:
		mt = Get_ObjectPos (tsC)
		miOrientation = fSkXResolveAngle (fSkXGetDegreesFromRadians (Object_GetOrientation (tsC) [1]) - 180)
		maBipedInfo = fSkXCameraGetBipedInfo (tsC)
		mt = (mt [0], mt [1], mt [2] + maBipedInfo [BIPED_FACE_H])
	mtFollow = mt#defaults to mt

	if tsPos != "far ahead" and tsCFacing != "":
		if isTuplePosition (tsCFacing):
			mt2 = tsCFacing
		else:
			mt2 = Get_ObjectPos (tsCFacing)
			maBipedInfo2 = fSkXCameraGetBipedInfo (tsCFacing)
			mt2 = (mt2 [0], mt2 [1], mt2 [2] + maBipedInfo2 [BIPED_FACE_H])
	elif not isTuplePosition (tsC):#no tsCFacing. mt2 is forward of mt.
		mt2 = fSkXGetPointFromVector (mt, miOrientation, DISTANCE_LONG)
	mtTrack = mt2#defaults to mt2

# 	mbMoving = 0
# 	try: mbMoving = ff.MapInfo [tsC]["distanceTravelled"]
# 	except: pass
	if gbIsCinematic and (tsPos == "shoulder" or tsPos == "far ahead"):# and not mbMoving
		mt2 = fSkXGetPointFromVector (mt, miOrientation, DISTANCE_SHORT)
		mtTrack = mt2
		miD = fSkXGetDistance (mt2, mt) * 2
		mtFollow = fSkXGetPointFromVector (mt2, [fSkXResolveAngle (miOrientation + 180), 90 - ANGLE_SNAP], [miD, miD], "3d")
	elif tsPos == "face" and not isTuplePosition (tsC):
		miDistance = maBipedInfo [BIPED_HEIGHT]
		mtFollow = fSkXGetPointFromVector (mt, miOrientation, miDistance)
		msNearest = skXMapInfo.MapInfo_GetObjects (mtFollow, "nearest")
		if not isCharacterTemplate (msNearest):
			#looking at something? facing too close an object
			if not giMonologueStep or giMonologueStep / 2 == giMonologueStep / 2.0:
				#start, show from side
				mtTrack = mtFollow#Get_ObjectPos (msNearest)#mtFollow
				mtFollow = fSkXGetPointFromVector (mt, fSkXResolveAngle (miOrientation + 135), miDistance / 2.0)
			elif giMonologueStep != giMonologueStep / 2.0:
				#odd step, show face
				mtTrack = mt
				mtFollow = fSkXGetPointFromVector (mt, miOrientation, miDistance / 2.0)
		else: mtTrack = mt
	elif tsPos == "shoulder" and not isTuplePosition (tsC):
		if gbIsSpeaking:#dialogue, camera closer
			miSideAdj = maBipedInfo [BIPED_SHOULDERS_W]
			miBackAdj = (maBipedInfo [BIPED_HEIGHT] - maBipedInfo [BIPED_PECTORALS_H])
			miUpAdj = (maBipedInfo [BIPED_HEIGHT] - maBipedInfo [BIPED_PECTORALS_H])
		else:#shoulder
			miSideAdj = maBipedInfo [BIPED_SHOULDERS_W] * 2
			miBackAdj = maBipedInfo [BIPED_HEIGHT]
			miUpAdj = maBipedInfo [BIPED_HEIGHT]
		if tsCFacing != "":#get opposite shoulder of tsCFacing
			miHeading = fSkXGetAngle (mt2, mt)
			mtTrack = fSkXGetPointFromVector (mt2, fSkXResolveAngle (miHeading + 90), miSideAdj)
		else:
			mtTrack = mt2
		miHeading = fSkXGetAngle (mt, mtTrack)
		mtFollow = fSkXGetPointFromVector (mt, fSkXResolveAngle (miHeading + 90), miSideAdj)
		miH = fSkXResolveAngle (miHeading + 180)
		miP = fSkXResolveAngle (fSkXGetAngle (mtFollow, mtTrack, "pitch") + 180)
		mtFollow = fSkXGetPointFromVector (mtFollow, [miH, miP], [miBackAdj, miUpAdj], "3d")
	elif tsPos == "far ahead" and not isTuplePosition (tsC):
		mtTrack = mt2
		miSideAdj = maBipedInfo [BIPED_SHOULDERS_W] * 2
		miBackAdj = maBipedInfo [BIPED_HEIGHT]
		miUpAdj = maBipedInfo [BIPED_HEIGHT]
		miHeading = fSkXGetAngle (mt, mtTrack)
		mtFollow = fSkXGetPointFromVector (mt, fSkXResolveAngle (miHeading + 90), miSideAdj)
		miH = fSkXResolveAngle (miHeading + 180)
		miP = fSkXGetAngle (mtFollow, mtTrack, "pitch")
		mtFollow = fSkXGetPointFromVector (mtFollow, [miH, miP], [miBackAdj, miUpAdj], "3d")
	elif tsPos == "combat" or tsPos == "melee":
		miHeading = fSkXGetAngle (mt, mt2)
		miDistance = fSkXGetDistance (mt, mt2)
		if miDistance < DISTANCE_MEDIUM:
			miDistance = DISTANCE_MEDIUM
		mtMid = fSkXGetPointInBetween (mt, mt2)
		if gsFromSide == "right":
			miSide = miDistance
		else:
			miSide = -miDistance
		mtFollow = fSkXGetPointFromVector (mtMid, fSkXResolveAngle (miHeading + 90), miSide)
		if tsPos == "combat":
			mtFollow = (mtFollow [0], mtFollow [1], max (mt [2], mt2 [2]) + DISTANCE_SHORT)
		mtTrack = mtMid
# 	if Object_Exists ("m"):
# 		Trigger_Move ("m", "walk", mtFollow, "", 0)
# 		Trigger_Move ("mm", "walk", mtTrack, "", 0)
# 	else:
# 		Object_SpawnAt ("mentor", mtFollow, "m")
# 		Object_SpawnAt ("minute_man", mtTrack, "mm")

	#return
	return [mtFollow, mtTrack]

BIPED_HEIGHT = 0
BIPED_FACE_H = 1
BIPED_SHOULDERS_W = 2
BIPED_PECTORALS_H = 3

def fSkXCameraGetBipedInfo (tsC):
	ma = [0, 0, 0, 0]#0=height, 1=face height, 2=shoulders width, 3=pects height
	maHeightInfo = heightcheck.Template_GetHeight (Object_GetTemplate (tsC))
	if maHeightInfo [1] == 0:
		miHeight = STANDARD_HUMAN_HEIGHT
	else:
		miHeight = maHeightInfo [0]
	#0.618 segment of divine proportion. http://goldennumber.net/
	miPectsFromTop = (miHeight * 0.618 * 0.618 * 0.618)
	ma [BIPED_HEIGHT] = miHeight
	ma [BIPED_FACE_H] = miHeight - miPectsFromTop * 0.5#half-way between pects and top
	ma [BIPED_SHOULDERS_W] = miPectsFromTop
	ma [BIPED_PECTORALS_H] = miHeight - miPectsFromTop
	return ma

#-------------------------------------------------------------------------------
#utility functions
#-------------------------------------------------------------------------------
def isNumber (ti):
	return type (ti) == FloatType or type (ti) == IntType or type (ti) == LongType

def isTuplePosition (tt):
	return type (tt) == TupleType and len (tt) == 3 and isNumber (tt [0]) and isNumber (tt [1]) and isNumber (tt [2])

def fSkXgetAllHeroes():
	objects = js.Mission_GetDynamicObjects()
	heroes = []
	for object in objects:
		#kuertee removed
# 		if isHero (object)
# 			heroes.append(object)
		#kuertee removal end
		#kuertee added
		if not MODULE_M25AI and isHero(object):
			heroes.append(object)
		elif MODULE_M25AI and MODULE_M25AI.GetTeam (object) == 1:#assumes team 1 is player-controlled	
			heroes.append(object)
		#kuertee additions end
	return heroes
getAllHeroes = fSkXgetAllHeroes

def fSkXgetAllBaddies(checkLiving = 1, template = 'none'):
	objects = js.Mission_GetDynamicObjects()
	baddies = []
	for object in objects:
		#kuertee removed
# 		if not isBaddie(object):
# 			continue
		#kuertee removal end
		#kuertee added
		if not MODULE_M25AI and not isBaddie (object):
			continue
		elif MODULE_M25AI and MODULE_M25AI.GetTeam (object) > 1:#assumes any team other than 1 is not player-controlled
			continue
		#kuertee additions end
		if not ((template == 'none') or (js.Object_GetTemplate(object) == template)):
			continue
		if (checkLiving and not Object_IsAlive(object)):
			continue
		baddies.append(object)
	return baddies
getAllBaddies = fSkXgetAllBaddies

def fGetClosestEnemyDeleteThis (ts, tiDistance = 1000000):
	global gaPCs
	msEnemy = ""
	try:
		maEnemies = ff.MapInfo_GetNearbyEnemies (ts)
		mi = 0
		while mi < len (maEnemies):
			msEnemy = maEnemies [mi]
			if msEnemy in gaPCs: maEnemies.remove (msEnemy)
			else: mi = mi + 1
		if len (maEnemies):
			msEnemy = ff.MapInfo_GetNearest (ts, DISTANCE_VISIBLE, taObjects = maEnemies)
			if not msEnemy: msEnemy = ""
	except:
		if MODULE_M25AI:
			msEnemy = MODULE_M25AI.GetCurrentTarget (ts, MODULE_M25AI.TARGET_ENEMY)
			if msEnemy == None: msEnemy = ""
		else:
			if (isHero (ts)):
				ma = getAllBaddies ()
			else:
				ma = getAllHeroes ()
			if len (ma):
				msEnemy = fGetClosest (ts, fGetValidCs (ma), tiDistance)
				if ts in msEnemy:
					msEnemy.remove (ts)
				if len (msEnemy):
					msEnemy = fGetClosest (ts, msEnemy)
				else:
					msEnemy = ""
			else:
				msEnemy = ""
	return msEnemy

def fGetClosestEnemy (ts, tiDistance = 1000000):
	global gaPCs
	msEnemy = ""
	if MODULE_M25AI:
		msEnemy = MODULE_M25AI.GetCurrentTarget (ts, MODULE_M25AI.TARGET_ENEMY)
		if msEnemy == None: msEnemy = ""
	else:
		try:
			maEnemies = ff.MapInfo_GetNearbyEnemies (ts)
			mi = 0
			while mi < len (maEnemies):
				msEnemy = maEnemies [mi]
				if msEnemy in gaPCs or not isBaddie (msEnemy): maEnemies.remove (msEnemy)
				else: mi = mi + 1
			if len (maEnemies):
				msEnemy = ff.MapInfo_GetNearest (ts, DISTANCE_VISIBLE, taObjects = maEnemies)
				if not msEnemy: msEnemy = ""
		except:
				if (isHero (ts)):
					ma = getAllBaddies ()
				else:
					ma = getAllHeroes ()
				if len (ma):
					msEnemy = fGetClosest (ts, fGetValidCs (ma), tiDistance)
					if ts in msEnemy:
						msEnemy.remove (ts)
					if len (msEnemy):
						msEnemy = fGetClosest (ts, msEnemy)
					else:
						msEnemy = ""
				else:
					msEnemy = ""
	return msEnemy

def fGetClosest (ttPosition, taObjects = [], tiDistance = None):
	if len (taObjects) == 0:
		taObjects = Mission_GetDynamicObjects ()
	if type (ttPosition) == StringType:
		ttPosition = Get_ObjectPos (ttPosition)
	maInArea = []
	msClosest = ''
	miClosest = 1000000
	mt = ()
	for msObject in taObjects:
		try:
			mt = Get_ObjectPos (msObject)
		except:
			try:
				mt = Get_MarkerPos (msObject)
			except:
				pass
		if len (mt) > 0:
			mi = fSkXGetDistance (ttPosition, mt)
			if mi < miClosest:
				miClosest = mi
				msClosest = msObject
			if tiDistance != None:
				if mi < tiDistance:
					maInArea.append (msObject)
	if tiDistance == None:
		return msClosest
	else:
		return maInArea

def fGetValidCs (taCs = []):
	global gaPCs
	if len (taCs) == 0:
		maCs = getAllHeroes ()
		maCs.extend (getAllBaddies ())
	else:
		maCs = taCs [:]
	miC = 0
	maCs = filter (isPhysical, maCs)
	maCs = filter (Object_IsAlive, maCs)
	maCs = filter (isVisible, maCs)
	#while (miC < len (maCs)):
	#	msC = maCs [miC]
	#	if not Object_GetAttr (msC, 'physical') or not Object_IsAlive (msC) or msC in gaPCs:
	#		maCs.remove (msC)
	#	else:
	#		miC = miC + 1
	return maCs

def isPhysical (ts):
	return Object_GetAttr (ts, "physical")

def isVisible (ts):
	return not isInvisible (ts)

def isInvisible (ts):
	if Object_Exists (ts):
		if Object_GetSecondaryStates (ts) & SCSTATE_INVISIBLE: return 1
		if Object_GetSecondaryStates (ts) & SCSTATE_DISPLACE_IMAGE: return 1
		if Object_GetPrimaryState (ts) == PCSTATE_DIMENSION_SHIFT: return 1

def isBuilding (ts):
	return (Object_GetClass (ts) & OC_BUILDING)

def fSkXDebugCamera (ts):
	global gsDebug
	try:
		mbTest = gsDebug
	except:
		gsDebug = ""
	if type (ts) != StringType:
		ts = `ts`
	ts = '----------------------------------------' + ts
	if ts != gsDebug:
		print (ts)
	gsDebug = ts
