# -------------------------------------------------------------------------------
# MapInfo.py (0.79beta, update by M25 for compatibility with FFX 3.2)
# -------------------------------------------------------------------------------

# whats new (0.79 beta - m25 branch 2):
# ----------
# m25: changed map caliper red_sun to ffx_tiggot as red_sun caused problems on berlin map
# m25: MapInfo_UpdateFighters now uses alliance information from m25team.py
# m25: added check for ne_corner, nw_corner, se_corner and sw_corner markers
#      to make it easier to get the extents on problem maps (again the 'part of berlin' map)


# whats new (0.79 beta - m25 branch 1):
# ----------
# m25 modified skXMapInfo so that it tries to import m25distance and m25team instead of ffx.m25ai.
# m25 made this change since skXMapInfo wasn't picking up team data correctly.
# m25 added check against js.OC_CHARACTER in MapInfo_UpdateFighters so that it picks up 
#   characters that have had their portrait removed by ffx

# # -------------------------------------------------------------------------------
# MapInfo.py (0.78beta)
# -------------------------------------------------------------------------------
# download: http://www.frpmods.com/files/kuertee/skXMapInfoV0.78Beta.zip
# read-me: http://www.frpmods.com/files/kuertee/skXMapInfo.txt

# whats new:
# ----------
# bug-fix in retrieving building information when it has no data in heightcheck.
# bug-fix to 0.76beta: building-clusters invalid during mission briefings. delayed
#	initialisation until there are valid buildings.
# MapInfo [name]["healthLoss"]: health loss % of maxHealth
# MapInfo [name]["health"]: track health so that healthLoss can be calculated
# updated scripting reference re: other MapInfo queries that were available previously
# bug-fix in MapInfo_GetTeamMembers ()
# update scripting reference: remove obsolete MapInfoFighters [name] data
# MapInfo [name]["nearestAlly"]: neareast ally
# MapInfo [name]["nearestAllyVisible"]: in the case that nearest ally is invisible
# MapInfo [name]["nearestAllyAttackable"]: in the case that nearest ally is not attackable
# MapInfo [name]["nearestEnemy"]: nearest enemy
# MapInfo [name]["nearestEnemyVisible"]: in the case that nearest enemy is invisible
# MapInfo [name]["nearestEnemyAttackable"]: in the case that nearest enemy is not attackable
# MapInfo [name]["nearestAllyElsewhere"]: neareast ally not in fight-cluster
# MapInfo [name]["nearestEnemyElsewhere"]: nearest enemy not in fight-cluster
# IS_INVISIBLE_FUNCTION: optional filter used in enemy lists generation
# IS_ATTACKABLE_FUNCTION: optional filter used in enemy lists generation
# updated scripting reference: wrappers to functions set via GET_TEAM_FUNCTION,
#	IS_INVISIBLE_FUNCTION and IS_ATTACKABLE_FUNCTION
# city blocks data (building-clusters): to speed-up the process of ensuring a
#	point is projected out of a building's area, instead of iterating through
#	the process for each building in the vicinity one projection is used from
#	the data in building-clusters.
# MapInfo_GetObjects parameters:
#	tiDistance: number or string: "nearest", "furthest". only one object is returned
#		if "nearest" or "furthest" is used.
#	tfFilter: function. only objects that satisfy this check is returned
#	with these two parameters, the need in looping through the list returned by
#		MapInfo_GetObjects (e.g. to get the nearest object) is minimised.
# fight-clusters and queries with MapInfo_GetNearbyEnemies and MapInfo_GetNearbyAllies
#	return only living characters
# E and W swapped. had it the other way around. duh... (seriously, my way of thinking
#	was turned upside down because of standard cartesian systems: computer screens use
#	quad 2 but the convention use quad 1 for trignometry calculations.)
# MapInfo ["teams"] = array of team names used by MapInfo. e.g.: Team1.0 instead of 1.0
# MapInfo ["teamIds"] = array of actual team ids. e.g.: from Object_GetAttr (X, "team"), 1.0.

# features:
# ---------
# fight-clusters data compatible with m25-ai team data

# to use:
# -------
# 1. from skXMapInfo import * or import skXMapInfo
# 2. initialise with GetMapInfo ()
# 3. to get a list of objects (including those that move):
#	MapInfo_GetObjects ((x,y,z), distance)
# 4. to update the MapInfoFighters data: MapInfo_UpdateFighters (). do this once
#	every "tick", so that other functions return up-to-date information.
# 5. to get a list of fighters: MapInfo_GetFighters ((x,y,z), distance)
# 7. to get a list of allies near a character:
#	MapInfo_GetNearbyAllies (character name)
# 8. to get a list of enemies near a character:
#	MapInfo_GetNearbyEnemies (character name)

# uses:
# -----
# 1. returns an approximate bounding box and mid points of the map
# 2. an alternative to Mission_GetObjects ()
# 3. an alternative to Mission_GetDynamicObjects ()

# why?
# ----
# Mission_GetObjects () returns all objects in the map. while
#	MapInfo_GetObjects () returns only those objects around a particular point.
#	ultimately, looping through MapInfo_GetObjects () impacts the cpu less in
#	exchange for a small amount of memory (to hold the MapInfo [] dictionary).

# i.e.: these functions, instead of iterating through all objects in a map for
#	nearest/furtherest checks, iterate through objects retrieved from a look-up
#	dictionary. e.g. in airfield, retrieving all objects within a 100-unit
#	radius from (0,0,0) iterates through 40 objects instead of all 200.

# similarly, queries to MapInfoFighters saves from looping through
#	Mission_GetDynamicObjects () several times in a "tick" just to retrieve a
#	list of allies/enemies around a certain point. update the list once at the
#	start of a "tick" with MapInfo_UpdateFighters (), and use
#	MapInfo_GetNearbyAllies () or MapInfo_GetNearbyEnemies () to instantly
#	retrieve data.

# fight-cluster data can be used to quickly retrieve all teams/characters
#	information in a fight-cluster. a fight-cluster does not only contain
#	characters within a certain distance from a point, but all other characters
#	that may be pulled into the fight because of their proximity to an ally who
#	may be within a certain distance from an enemy.

# unlike the object data, fight-clusters need to be updated regularly. so a call
#	to MapInfo_UpdateFighters () is required.

# how?
# ----
# when GetMapInfo () is called (preferably at the beginning of the game in
#	OnPostInit ()), the map is read by calling Mission_GetObjects ().

# from this list of objects, the map's approximate dimensions are retrieved.
#	then, invisible "red_suns" are spawned at the edges of this bounding box and
#	triggered to move outwards. once they stop, the actual dimensions of the
#	map are retrieved.

# a grid with measurements given in MAPINFO_XGRID AND MAPINFO_YGRID is used to
#	divide the map. the grid positions of all objects in the map are stored in
#	the MapInfo dictionary.

# when MapInfo_GetObjects () is called, it returns all objects within the given
#	distance by reading the MapInfo [] dictionary. the number of objects returned is
#	much smaller than from Mission_GetObjects ().

# dynamic generic objects (i.e. not characters nor projectiles) are tracked
#	automatically. once they stop moving, their grid information is updated.
#	note that tracking automatically starts when MapInfo_GetObjects () is
#	called. and tracking automatically stops when all these generic objects
#	stop moving.

# data in MapInfoFighters works similarly. at every call to
#	MapInfo_UpdateFighters (), the grid positions of all characters and all moving
#	objects (including projectiles and thrown objects) are stored in the
#	MapInfoFighters dictionary.

# from the MapInfoFighters dictionary, the fight-cluster information is built.

# applications?
# -------------
# 1. (implemented:) skXCamera2 constantly looks for buildings that may have
#	blocked the view.
# 2. (implemented:) fight clusters working in my preliminary code to skX2.
# 3. (implemented:) projectiles/thrown-objects avoidance working in skX2.
# 3. (not yet:) custom ai looking for objects to throw (or objects to
#	hide behind.)

# ------------------------------------------------------------------------------
# scripting reference
# ------------------------------------------------------------------------------
# these three functions may be customised to further filter-out valid enemies.

# GET_TEAM_FUNCTION = is used to separate the characters into their teams
# IS_INVISIBLE_FUNCTION: optional filter used in "nearestEnemyVisible" assignment
# IS_ATTACKABLE_FUNCTION: optional filter used in "nearestEnemyAttackable" assignment

# the best way to customise these functions are to assign it a function from
# 	the importing module. for example:

# a "too pathetic to be attacked" attribute prevents any attacks. the function that
#	determines the presence of this attribute (along with the standard
#	non-attackable checks) need to be set in IS_ATTACKABLE_FUNCTION.

# import skXMapInfo
# def OnPostInit (event = None):
# 	skXMapInfo.GET_TEAM_FUNCTION = getTeam
# 	skXMapInfo.IS_INVISIBLE_FUNCTION = isInvisible
# 	skXMapInfo.IS_ATTACKABLE_FUNCTION = isAttackable
# def getTeam (name):
#	return Object_GetAttr (name, "team")
# def isInvisible (name):
# 	return Object_GetSecondaryStates (name) & SCSTATE_INVISIBLE
# def isAttackable (name):
# 	return not tooPatheticToBeAttacked (name) and not frozen (name) and not isInvisible (name) and not isObjectified (name)
# def isObjectified (name):
# 	banishedStates = [PCSTATE_OBJECT, PCSTATE_TOAD]
# 	return Object_GetPrimaryState (name) in banishedStates
# def tooPatheticToBeAttacked (name):
#	return Object_GetAttr (name, "strength") == 0

# by default, GET_TEAM_FUNCTION uses isHero (name) to differentiate between teams
# 	by default, IS_INVISIBLE_FUNCTION and IS_ATTACKABLE_FUNCTION are as per below:
# def isInvisible (name):
# 	return Object_GetSecondaryStates (name) & SCSTATE_INVISIBLE
# def isAttackable (name):
# 	return Object_IsAlive (name) and not frozen (name) and not isInvisible (name) and not staticked (name) and not isObjectified (name)
# def isObjectified (name):
# 	banishedStates = [PCSTATE_OBJECT, PCSTATE_TOAD]
# 	return Object_GetPrimaryState (name) in banishedStates

# ------------------------------------------------------------------------------
# MapInfo
# ------------------------------------------------------------------------------
# MapInfo ["boundingBox"] = tuple (xmin, ymin, zmin, xmax, ymax, zmax)
# MapInfo ["centre"] = tuple (xcentre, ycentre, zcentre)
# MapInfo ["NE"] = tuple NE corner
#	["nw"] = tuple NW corner
#	["sw"] = tuple SW corner
#	["se"] = tuple SE corner
#	["ceiling"] = height of heighest object in map. i.e. definately not real
#		ceiling
#	["floor"] = height of lowest object in map. i.e. may not be real floor
#	["mapExtents"] = (min x, min y, minz, max x, max y, max z)
# MapInfo [name]["health"] = health. required to keep for healthLoss
# 	calculations.
# MapInfo [name]["healthLoss"] = health loss % of maxHealth
# MapInfo [name]["prestige"] = prestige. saves from Object_CalcPrestige ().
# MapInfo [name]["team"] = string of "team" + team name/number
# MapInfo [name]["grid"] = array of grid position
# MapInfo [name]["pos"] = tuple of map position
# MapInfo [name]["heading"] = integer of heading of movement
#	(not orientation/facing)
# MapInfo [name]["pitch"] = integer of pitch of movement
# MapInfo [name]["distanceTravelled"] = actual/3d distance travelled
# MapInfo [name]["averageSpeed"] = average speed
# MapInfo [name]["projectedPos"] = tuple of position projected from
#	heading, pitch, and distanceTravelled
# MapInfo [name]["cluster"] = fight-cluster a member of
# MapInfo [name]["nearestAlly"] = neareast ally
# MapInfo [name]["nearestAllyVisible"]: in the case that nearest ally is invisible
# MapInfo [name]["nearestAllyAttackable"]: in the case that nearest ally is not attackable
# MapInfo [name]["nearestEnemy"] = nearest enemy
# MapInfo [name]["nearestEnemyVisible"]: in the case that nearest enemy is invisible
# MapInfo [name]["nearestEnemyAttackable"]: in the case that nearest enemy is not attackable
# MapInfo [name]["nearestAllyElsewhere"] = neareast ally not in fight-cluster
# MapInfo [name]["nearestEnemyElsewhere"] = nearest enemy not in fight-cluster
# MapInfo ["characters"] = array of characters
# MapInfo ["teams"] = array of team names for MapInfo. e.g.: Team1.0 instead of 1.0
# MapInfo ["teamIds"] = array of actual team ids. e.g.: from Object_GetAttr (X, "team"), 1.0.
# MapInfo ["team" + team name/number] = array of character names of members
# MapInfo ["team" + team name/number + "Dead"] = array of character names of dead members

# MapInfo_GetObjects (tuple (x, y, z), distance, tfFilter=function)
# -----------------------------------------------------------------
# distance: number or the strings: "nearest" or "furthest". if "nearest"/"furthest"
#	is used, only one object is returned.
# tfFilter: function (note: not string). only objects that satisfy this check
#	is returned.
# note that not all objects in the map is queried. only objects around the point
#	at the given distance is used. when not given or "nearest" or "furthest" is
#	used, distance defaults to the constant DISTANCE_VISIBLE = 400.
# returns an array of all objects around the given point or a string of the
#	objects's name that is "nearest" or "furthest" from the point.

# MapInfo_GetMapPosition (array of markers, objects or tuples)
# ------------------------------------------------------------
# returns a point inside the map bounding box.
# the point returned is either a random point anywhere in the map or a point
#	randomly selected from the given list.

# MapInfo_GetPointFromCentre (integer heading, integer distance from edge, float factor to edge)
# ----------------------------------------------------------------------------------------------
# returns a point from the centre of the map at the heading	angle and a certain
#	distance from the edge of the map.

# MapInfo_Update ()
# -----------------
# updates MapInfo's grid information. not necessary because a query with
#	MapInfo_GetObjects () will update the information automatically.

# ------------------------------------------------------------------------------
# MapInfoFighters
# ------------------------------------------------------------------------------
# MapInfoFighters ["clusters"] = array of fight clusters in map
# MapInfoFighters [cluster name] = dictionary of
#	(for each team in cluster:)
#	team<team name>: array of characters in cluster
#	team<team name>Health: integer total health of team in cluster
#	team<team name>MaxHealth: integer total maxHealth of team in cluster
#	team<team name>Prestige: integer total prestige of team in cluster

# MapInfo_UpdateFighters ()
# -------------------------
# updates MapInfoFighters' grid and fight cluster information

# MapInfo_GetFighters (tuple (x,y,z), distance, tfFilter=function)
# ----------------------------------------------------------------
# distance: number or the strings: "nearest" or "furthest". if "nearest"/"furthest"
#	is used, only one object is returned.
# tfFilter: function (note: not string). only objects that satisfy this check
#	is returned. this defaults to Object_IsAlive.
# returns an array of all characters around the given point or a string of the
#	characters's name that is "nearest" or "furthest" from the point.

# MapInfo_GetNearbyAllies (string character name)
# -----------------------------------------------
# returns an array of allies near the character.

# MapInfo_GetNearbyEnemies (string character name)
# ------------------------------------------------
# returns an array of enemies near the character.

# ------------------------------------------------------------------------------
# building information
# ------------------------------------------------------------------------------
# the functions relating to building information, require:
#	1. heightcheck.py (v1.2)
#	2. heightcheckdata.py with building data

# MapInfo_GetBuildingCorners (string building name)
# -------------------------------------------------
# returns the positions of the corners of the building

# MapInfo_GetBuildingPointsOfNearestSide (string building name, tuple (x,y,z))
# MapInfo_GetBuildingPointsOfFurthestSide (string building name, tuple (x,y,z))
# -----------------------------------------------------------------------------
# returns points of nearest/furthest side of the building from the given point

# MapInfo_PutPointOutsideBuilding (tuple (x,y,z), string building name)
# ---------------------------------------------------------------------
# returns a point that is outside a building

# MapInfo_EnsurePointIsOutsideBuilding (tuple (x,y,z))
# ----------------------------------------------------
# returns a point that is definately outside buildings by recursively running
#	MapInfo_PutPointOutsideBuilding until the point is outside any building

# MapInfo_IsPointInsideBuilding (tuple (x,y,z))
# ---------------------------------------------
# returns true if the point is inside a building

# MapInfo_GetBuildingPointInsideOf (tuple (x,y,z))
# ------------------------------------------------
# returns the building the point is inside of

#-------
#credits
#-------
# irrational: ffvsttr
# kuertee: main code, idea (inspired from stumpy's Mission_GetMostObjects ())
# epimethee: heightcheck.py
# stumpy: for ff1's Mission_GetMostObjects (). the idea i had was basically
#	in the line of "Mission_GetObjectsAtADistanceFrom ()" but without having to
#	loop through Mission_GetObjects () every time. stumpy's code got me thinking.

#ff modules
from cshelper import *
import time
#custom modules
from skXAngles import *
try: import heightcheck
except: heightcheck = None

print 'Starting skXMapInfo.py  v 0.79 beta'

MAPINFO_XGRID = 250
MAPINFO_YGRID = 250
try:
	import m25distance
	import m25team
	#MODULE_M25AI = ff
	DISTANCE_VISIBLE = m25distance.DISTANCE_VISIBLE
	MAPINFO_FIGHT_CLUSTER_RADIUS = m25distance.DISTANCE_VISIBLE
	GET_TEAM_FUNCTION = m25team.GetTeam
	print 'skXMapInfo: m25ai available'
except:
	print 'skXMapInfo: m25ai NOT available'
	MODULE_M25AI = 0
	DISTANCE_VISIBLE = 400
	MAPINFO_FIGHT_CLUSTER_RADIUS = 400
	GET_TEAM_FUNCTION = isHero

IS_ATTACKABLE_FUNCTION = None
UPDATE_CHARACTER_FUNCTION = None
MAPINFO_CHECKOBJECTS_INTERVAL = 3
MEASUREMENT_SNAP = 25

def GetMapInfo (tevent = ""):
	global gbGetMapInfoIsInitialised
	try: mbTest = gbGetMapInfoIsInitialised
	except:
		gbGetMapInfoIsInitialised = 1
		if len (Mission_GetObjects ()):
			print "!","GetMapInfo"
			#map info
			global gbMapInfoIsLooping
			global gaMapInfoObjectsToCheck
			gbMapInfoIsLooping = 0
			gaMapInfoObjectsToCheck = []
			ff.MapInfo = {}
			MapInfo_Init ()
			MapInfo_GetBuildingClusters ()
	
			#fight clusters
			ff.MapInfoFighters = {}
			MapInfo_UpdateFighters ()
			
			#measure mapextent
			if not Marker_Exists("ne_corner"):
				MapInfo_MeasureTerrain ()

def MapInfo_MeasureTerrain (tevent = ""):
	for mi in range (0,4):
		msC = "mapinfocaliper" + `mi`
		if mi == 0:
			mtPos = (ff.MapInfo ["NEObjectBound"][0],ff.MapInfo ["centre"][1],0)
			mtTo = (ff.MapInfo ["NE"][0] + 1000000, mtPos [1], 0)
		elif mi == 1:
			mtPos = (ff.MapInfo ["centre"][0],ff.MapInfo ["SEObjectBound"][1],0)
			mtTo = (mtPos [0], ff.MapInfo ["SE"][1] - 1000000, 0)
		elif mi == 2:
			mtPos = (ff.MapInfo ["NWObjectBound"][0],ff.MapInfo ["centre"][1],0)
			mtTo = (ff.MapInfo ["NW"][0] - 1000000, mtPos [1], 0)
		elif mi == 3:
			mtPos = (ff.MapInfo ["centre"][0],ff.MapInfo ["NEObjectBound"][1],0)
			mtTo = (mtPos [0], ff.MapInfo ["NE"][1] + 1000000, 0)
		Object_SpawnAt ("ffx_tiggot", mtPos, msC)
		disable (msC)
		Object_SetAttr (msC, "physical", 0)
		Object_SetAttr(msC,'speed',1000)
		Object_SetSecondaryState (msC, SCSTATE_INVISIBLE, 1000000, PERMANENT_STATE)
		Trigger_Move (msC, "walk", mtTo)
	regTimer ("MapInfo_CheckCalipers", 1)
# 	MapInfo_CheckCalipers ()

def MapInfo_CheckCalipers (tevent = ""):
	global gdCalipers
	global caliperDestination
	try: skXTest = gdCalipers
	except: gdCalipers = {}
	miActive = 0
	for mi in range (0,4):
		msC = "mapinfocaliper" + `mi`
		if Object_Exists (msC):
			Object_SetSecondaryState (msC, SCSTATE_INVISIBLE, 1000000, PERMANENT_STATE)
			mtPos = Get_ObjectPos (msC)
			try:
				if msC in gdCalipers.keys ():
					mtOldPos = gdCalipers [msC]
					miD = fSkXGetDistance (mtPos, mtOldPos)
				gdCalipers [msC] = mtPos
				if not miD:
					#stopped, update map extents
					MapInfo_UpdateMapExtents (mtPos)
					Object_Destroy (msC)
				else:
					miActive = miActive + 1
			except:
				if msC not in ff.MapInfo.keys (): ff.MapInfo [msC] = {}
				ff.MapInfo [msC]["pos"] = mtPos
				miActive = miActive + 1
	if miActive:
		regTimer ("MapInfo_CheckCalipers", 1)
# 		MapInfo_CheckCalipers ()
ff.MapInfo_CheckCalipers = MapInfo_CheckCalipers

def MapInfo_Init (tevent = ""):
	#print "!","MapInfo_Init"
	ff.MapInfo ["xGrid"] = MAPINFO_XGRID
	ff.MapInfo ["yGrid"] = MAPINFO_YGRID
	ma = list (Mission_GetObjects ())
	if "_impobj_0" in ma:
		ma.remove ("_impobj_0")
	miXMin = 0
	miXMax = 0
	miYMin = 0
	miYMax = 0
	miZMin = 0
	miZMax = 0
	
	
	for marker in ['ne_corner','nw_corner','se_corner','sw_corner']:
		if Marker_Exists(marker):
			maPoint = Get_MarkerPos(marker)
			miXMin = min (maPoint [0], miXMin)
			miXMax = max (maPoint [0], miXMax)
			miYMin = min (maPoint [1], miYMin)
			miYMax = max (maPoint [1], miYMax)
			miZMin = min (maPoint [2], miZMin)
			miZMax = max (maPoint [2], miZMax)
		
	for msObject in ma:
		try:
			maPoint = Get_MarkerPos (msObject)
		except:
			maPoint = Get_ObjectPos (msObject)
		miXSection = int (maPoint [0] / ff.MapInfo ["xGrid"])
		miYSection = int (maPoint [1] / ff.MapInfo ["yGrid"])
		if ff.MapInfo.has_key (miXSection) == 0:
			ff.MapInfo [miXSection] = {}
		if ff.MapInfo [miXSection].has_key (miYSection) == 0:
			ff.MapInfo [miXSection] [miYSection] = []
		ff.MapInfo [miXSection] [miYSection].append (msObject)
		ff.MapInfo [msObject] = [miXSection, miYSection]
		#
		miXMin = min (maPoint [0], miXMin)
		miXMax = max (maPoint [0], miXMax)
		miYMin = min (maPoint [1], miYMin)
		miYMax = max (maPoint [1], miYMax)
		miZMin = min (maPoint [2], miZMin)
		miZMax = max (maPoint [2], miZMax)
	#
	miXMid = miXMin + (miXMax - miXMin) * 0.5
	miYMid = miYMin + (miYMax - miYMin) * 0.5
	miZMid = miZMin + (miZMax - miZMin) * 0.5
	ff.MapInfo ["boundingBox"] = (miXMin, miYMin, miZMin, miXMax, miYMax, miZMax)
	ff.MapInfo ["NWObjectBound"] = (miXMin, miYMax, miZMid)
	ff.MapInfo ["NEObjectBound"] = (miXMax, miYMax, miZMid)
	ff.MapInfo ["SEObjectBound"] = (miXMax, miYMin, miZMid)
	ff.MapInfo ["SWObjectBound"] = (miXMin, miYMin, miZMid)
	#
	ff.MapInfo ["mapExtents"] = ff.MapInfo ["boundingBox"]
	ff.MapInfo ["centre"] = (miXMid, miYMid, miZMid)
	ff.MapInfo ["NW"] = (miXMin, miYMax, miZMid)
	ff.MapInfo ["NE"] = (miXMax, miYMax, miZMid)
	ff.MapInfo ["SE"] = (miXMax, miYMin, miZMid)
	ff.MapInfo ["SW"] = (miXMin, miYMin, miZMid)
	ff.MapInfo ["floor"] = miZMin
	ff.MapInfo ["ceiling"] = miZMax

def MapInfo_UpdateMapExtents (tt):
	#fSkXDebugMapInfo ("MapInfo_UpdateMapExtents " + `tt`)
	miXMin = ff.MapInfo ["mapExtents"][0]
	miYMin = ff.MapInfo ["mapExtents"][1]
	miZMin = ff.MapInfo ["mapExtents"][2]
	miXMax = ff.MapInfo ["mapExtents"][3]
	miYMax = ff.MapInfo ["mapExtents"][4]
	miZMax = ff.MapInfo ["mapExtents"][5]
	miSide = fSkXGetNearestSideToPoint (tt, ff.MapInfo ["mapExtents"])
	if miSide == 1: miYMax = tt [1];#maxy
	elif miSide == 2: miXMax = tt [0]#maxx
	elif miSide == 3: miYMin = tt [1]#miny
	elif miSide == 4: miXMin = tt [0]#minx
	miXMid = miXMin + (miXMax - miXMin) * 0.5
	miYMid = miYMin + (miYMax - miYMin) * 0.5
	miZMid = miZMin + (miZMax - miZMin) * 0.5
	ff.MapInfo ["mapExtents"] = (miXMin, miYMin, miZMin, miXMax, miYMax, miZMax)
	ff.MapInfo ["centre"] = (miXMid, miYMid, miZMid)
	ff.MapInfo ["NW"] = (miXMin, miYMax, miZMid)
	ff.MapInfo ["NE"] = (miXMax, miYMax, miZMid)
	ff.MapInfo ["SE"] = (miXMax, miYMin, miZMid)
	ff.MapInfo ["SW"] = (miXMin, miYMin, miZMid)
	ff.MapInfo ["floor"] = miZMin
	ff.MapInfo ["ceiling"] = miZMax

def Object_GetGridPos (ts):
	maPoint = Get_ObjectPos (ts)
	miXSection = int (maPoint [0] / ff.MapInfo ["xGrid"])
	miYSection = int (maPoint [1] / ff.MapInfo ["yGrid"])
	return [miXSection, miYSection]

def Object_UpdateGridPos (ts):
	if ts in ff.MapInfo.keys ():
		maGrid = ff.MapInfo [ts]
		if type (maGrid) == DictionaryType:
			if ts in ff.MapInfo ["projectiles"]: ff.MapInfo ["projectiles"].remove (ts)
			maGrid = maGrid ["grid"]
		if not maGrid [0] in ff.MapInfo.keys (): ff.MapInfo [maGrid [0]] = {}
		if not maGrid [1] in ff.MapInfo [maGrid [0]].keys (): ff.MapInfo [maGrid [0]][maGrid [1]] = []
		if ts in ff.MapInfo [maGrid [0]] [maGrid [1]]:
			ff.MapInfo [maGrid [0]] [maGrid [1]].remove (ts)
			del ff.MapInfo [ts]
	if Object_Exists (ts):
		maGrid = Object_GetGridPos (ts)
		if ts not in ff.MapInfo [maGrid [0]] [maGrid [1]]:
			ff.MapInfo [maGrid [0]] [maGrid [1]].append (ts)
		ff.MapInfo [ts] = maGrid
# 		fSkXDebugMapInfo ("MapInfo: " + `ts` + " updated into grid " + `maGrid [0]` + ", " + `maGrid [1]`)

def MapInfo_GetObjects (tt, tiDistance = None, tdObjects = None, tbAllowToLoop = 1, tfFilter = None):
	#tiDistance = number: get all objects within this distance
	#tiDistance = nearest/closest/furthest/farthest: get one object
	GetMapInfo ()
	msExclude = None
	if type (tt) == StringType:
		msExclude = tt
		tt = Get_ObjectPos (tt)
	miGridXYAverage = (ff.MapInfo ["xGrid"] + ff.MapInfo ["yGrid"]) * 0.5
	if not isNumber (tiDistance):
		msDistanceCheck = tiDistance
		tiDistance = DISTANCE_VISIBLE
	else: msDistanceCheck = None
	if tdObjects == None:
		tdObjects = ff.MapInfo
	mt = tt [:]
	mts = [mt]
	miX0 = int (tt [0] / ff.MapInfo ["xGrid"])
	miY0 = int (tt [1] / ff.MapInfo ["yGrid"])
	miX1 = miX0
	miX2 = miX0
	miY1 = miY0
	miY2 = miY0
	if tiDistance > miGridXYAverage:
		mi = 1
		while mi * ff.MapInfo ["xGrid"] < tiDistance:
			mi = mi + 1
			miX1 = miX1 - 1
			miX2 = miX2 + 1
		mi = 1
		while mi * ff.MapInfo ["xGrid"] < tiDistance:
			mi = mi + 1
			miY1 = miY1 - 1
			miY2 = miY2 + 1
	else:
		if mt [0] < miX0 * ff.MapInfo ["xGrid"] + ff.MapInfo ["xGrid"] * 0.5:
			miX2 = miX0 + 1
		else:
			miX2 = miX0 - 1
		if mt [1] < miY0 * ff.MapInfo ["yGrid"] + ff.MapInfo ["yGrid"] * 0.5:
			miY2 = miY0 + 1
		else:
			miY2 = miY0 - 1
	miXMin = min (miX1, miX2)
	miXMax = max (miX1, miX2)
	miYMin = min (miY1, miY2)
	miYMax = max (miY1, miY2)
	maObjects = []
	msObject = None
	miCheckDistance = 0
	if msDistanceCheck:
		if string.find (msDistanceCheck, "near") > -1 or string.find (msDistanceCheck, "close") > -1:
			miCheckDistance = 1000000
		elif string.find (msDistanceCheck, "far") > -1 or string.find (msDistanceCheck, "fur") > -1:
			miCheckDistance = -1000000
	for miXSegment in range (miXMin, miXMax + 1):
		if tdObjects.has_key (miXSegment):
			for miYSegment in range (miYMin, miYMax + 1):
				if tdObjects [miXSegment].has_key (miYSegment):
					mi = 0
					while mi < len (tdObjects [miXSegment][miYSegment]):
						ms = tdObjects [miXSegment][miYSegment] [mi]
						if Object_Exists (ms) and ms != msExclude:
							mbOK = 0
							miD = fSkXGetDistance (tt, Get_ObjectPos (ms))
							if msDistanceCheck:
								if string.find (msDistanceCheck, "near") > -1 or string.find (msDistanceCheck, "close") > -1:
									if miD < miCheckDistance: mbOK = 1
								elif string.find (msDistanceCheck, "far") > -1 or string.find (msDistanceCheck, "fur") > -1:
									if miD > miCheckDistance: mbOK = 1
							else:
								if miD < tiDistance: mbOK = 1
							if mbOK:
								mbOK2 = 1
								if tfFilter:
									if not tfFilter (ms): mbOK2 = 0
								if mbOK2:
									if msDistanceCheck:
										msObject = ms
										miCheckDistance = miD
									else: maObjects.append (ms)
							mi = mi + 1
						elif not Object_Exists (ms):
							tdObjects [miXSegment][miYSegment].remove (ms)
						else: mi = mi + 1
	if tbAllowToLoop:
		MapInfo_CheckObjects (tbAllowToLoop = tbAllowToLoop)
	if msDistanceCheck: maObjects = msObject
	return maObjects

def MapInfo_GetBuildings (tt, tiDistance = None):
	maObjects = MapInfo_GetObjects (tt, tiDistance, tbAllowToLoop = 0, tfFilter = isBuilding)
# 	return filter (isBuilding, maObjects)
	return maObjects

def MapInfo_Update (tevent = ""):
	MapInfo_CheckObjects2 (tbAllowToLoop = 0)

def MapInfo_CheckObjects (tevent = "", tbAllowToLoop = 1):
	GetMapInfo ()
	global gbMapInfoIsLooping
	global gaMapInfoObjectsToCheck
	global gbMapInfoIsLooping
	if not gbMapInfoIsLooping:
		gbMapInfoIsLooping = 1
		MapInfo_CheckObjects2 (tbAllowToLoop = tbAllowToLoop)

def MapInfo_CheckObjects2 (tevent = "", tbAllowToLoop = 1):
	global gbMapInfoIsLooping
	maMGDO = Mission_GetDynamicObjects ()
	#check if objects that need checking has stopped
	mi = 0
	while mi < len (gaMapInfoObjectsToCheck):
		ms = gaMapInfoObjectsToCheck [mi]
		if ms not in maMGDO or not Object_Exists (ms):#object stopped moving, so update its grid info
			Object_UpdateGridPos (ms)
			gaMapInfoObjectsToCheck.remove (ms)
		else:
			mi = mi + 1
	#get new objects that moved
	for ms in maMGDO:
		msT = Object_GetTemplate (ms)
		if not isHero (ms) and not isCharacterTemplate (msT):
			if ms not in gaMapInfoObjectsToCheck:
				gaMapInfoObjectsToCheck.append (ms)
	if len (gaMapInfoObjectsToCheck) and tbAllowToLoop:
		regTimer ("MapInfo_CheckObjects2", MAPINFO_CHECKOBJECTS_INTERVAL)
	else:
		gbMapInfoIsLooping = 0
# 	fSkXDebugMapInfo ("MapInfo_CheckObjects2 " + `gaMapInfoObjectsToCheck`)
ff.MapInfo_CheckObjects2 = MapInfo_CheckObjects2

def MapInfo_GetMapPosition (taMarkers = []):
	GetMapInfo ()
	maMarkers = taMarkers [:]
	if type (maMarkers) != ListType:
		maMarkers = [taMarkers]
	if len (maMarkers) > 0:
		if isNumber (maMarkers [0]):
			#tuple position, i.e. (10, 10, 10)
			maMarkers = [(taMarkers)]
	taMarkers = maMarkers [:]

	mtPosition = ()
	if len (taMarkers) > 0:
		msMarker = taMarkers [randint (0, len (taMarkers) - 1)]
		if type (msMarker) == StringType:
			#get marker or object pos
			try:
				mtPosition = Get_MarkerPos (msMarker)
			except:
				mtPosition = Get_ObjectPos (msMarker)
		elif isTuplePosition (msMarker):
			#specific position
			mtPosition = msMarker

	if len (mtPosition) == 0:
		#random marker
		mtBox = ff.MapInfo ["boundingBox"]
		miX = randint (int (mtBox [0]), int (mtBox [3]))
		miY = randint (int (mtBox [1]), int (mtBox [4]))
		miZ = randint (int (mtBox [2]), int (mtBox [5]))
		mtPosition = (miX, miY, miZ)

	mtPosition = MapInfo_PutPointWithinMap (mtPosition)

	return mtPosition [:]
ff.fSkXGetMapPosition = MapInfo_GetMapPosition

def MapInfo_PutPointWithinMap (ttPos):
	GetMapInfo ()
	miX = ttPos [0]
	miY = ttPos [1]
	if miX < ff.MapInfo ["NW"][0]: miX = ff.MapInfo ["NW"][0] + abs (miX - ff.MapInfo ["NW"][0])
	elif miX > ff.MapInfo ["NE"][0]: miX = ff.MapInfo ["NE"][0] - abs (ff.MapInfo ["NE"][0] - miX)
	if miY > ff.MapInfo ["NE"][1]: miY = ff.MapInfo ["NE"][1] - abs (miY - ff.MapInfo ["NE"][1])
	elif miY < ff.MapInfo ["SE"][1]: miY = ff.MapInfo ["SE"][1] + abs (ff.MapInfo ["SE"][1] - miY)
	mtPos = (miX, miY, ttPos [2])
	return mtPos
ff.fSkXPutPointWithinMap = MapInfo_PutPointWithinMap

def MapInfo_WarpPointIfOutside (ttPos):
	GetMapInfo ()
	miX = ttPos [0]
	miY = ttPos [1]
	if miX <= ff.MapInfo ["NW"][0]: miX = ff.MapInfo ["NE"][0]# - abs (ff.MapInfo ["NE"][0] - MEASUREMENT_SNAP)
	elif miX >= ff.MapInfo ["NE"][0]: miX = ff.MapInfo ["NW"][0]# + abs (MEASUREMENT_SNAP - ff.MapInfo ["NW"][0])
	if miY >= ff.MapInfo ["NE"][1]: miY = ff.MapInfo ["SE"][1]# + abs (MEASUREMENT_SNAP - ff.MapInfo ["NE"][1])
	elif miY <= ff.MapInfo ["SE"][1]: miY = ff.MapInfo ["NE"][1]# - abs (ff.MapInfo ["SE"][1] - MEASUREMENT_SNAP)
	mtPos = (miX, miY, ttPos [2])
	return mtPos

def MapInfo_ClipPointIfOutside (ttPos):
	GetMapInfo ()
	miX = ttPos [0]
	miY = ttPos [1]
	if miX <= ff.MapInfo ["NW"][0]: miX = ff.MapInfo ["NW"][0]
	elif miX >= ff.MapInfo ["NE"][0]: miX = ff.MapInfo ["NE"][0]
	if miY >= ff.MapInfo ["NE"][1]: miY = ff.MapInfo ["NE"][1]
	elif miY <= ff.MapInfo ["SE"][1]: miY = ff.MapInfo ["SE"][1]
	mtPos = (miX, miY, ttPos [2])
	return mtPos

def MapInfo_GetPointFromProjectionOfToPoints (ttPos, ttPos2):
	GetMapInfo ()
	miH = fSkXGetAngle (ttPos, ttPos2)
	miP = fSkXGetAngle (ttPos, ttPos2, "pitch")
	miD = fSkXGetDistance (ttPos, ttPos2)
	miD2 = fSkXGetDistance (ttPos, ttPos2, "pitch")
	mtPos = fSkXGetPointFromVector (ttPos2, [miH, miP], [miD, miD2], "3d")
	return mtPos
MapInfo_GetPointFromProjection = MapInfo_GetPointFromProjectionOfToPoints

def MapInfo_GetPointFromCentre (tiAngle = None, tiDistanceFromEdge = 100, tiFactorToEdge = 1):
	return MapInfo_GetPointAtEdgeFromVector (ff.MapInfo ["centre"], tiAngle, tiDistanceFromEdge, tiFactorToEdge)
ff.fSkXGetMapPointFromCentre = MapInfo_GetPointFromCentre
ff.fSkXGetMapPointFromCenter = MapInfo_GetPointFromCentre

def MapInfo_GetPointAtEdgeFromVector (ttPos, tiAngle = None, tiDistanceFromEdge = 100, tiFactorToEdge = 1):
	GetMapInfo ()
	if tiAngle == None:
		tiAngle = randint (0, 359)
	mtPoint = ()
	miAngle = fSkXResolveAngle (tiAngle)
	mtPoint = fSkXGetPointFromVector (ttPos, miAngle, 1000000)
	miTLAngle = fSkXGetAngle (ttPos, ff.MapInfo ["NW"])
	miTRAngle = fSkXGetAngle (ttPos, ff.MapInfo ["NE"])
	miBRAngle = fSkXGetAngle (ttPos, ff.MapInfo ["SE"])
	miBLAngle = fSkXGetAngle (ttPos, ff.MapInfo ["SW"])
	if miAngle >= miTLAngle or miAngle < miTRAngle:
		#min y
		mtPointIntersect = fSkXGetIntersectionFromPoints (ttPos, mtPoint, ff.MapInfo ["NE"], ff.MapInfo ["NW"])
	elif miAngle >= miTRAngle and miAngle < miBRAngle:
		#max x
		mtPointIntersect = fSkXGetIntersectionFromPoints (ttPos, mtPoint, ff.MapInfo ["NW"], ff.MapInfo ["SW"])
	elif miAngle >= miBRAngle and miAngle < miBLAngle:
		#max y
		mtPointIntersect = fSkXGetIntersectionFromPoints (ttPos, mtPoint, ff.MapInfo ["SW"], ff.MapInfo ["SE"])
	else: #miAngle >= 225 and miAngle < 315
		#min x
		mtPointIntersect = fSkXGetIntersectionFromPoints (ttPos, mtPoint, ff.MapInfo ["SE"], ff.MapInfo ["NE"])
	miDistance = fSkXGetDistance (ttPos, mtPointIntersect)
	miDistance = miDistance - tiDistanceFromEdge
	miDistance = miDistance * tiFactorToEdge
	mtPoint = fSkXGetPointFromVector (ttPos, miAngle, miDistance)
	return mtPoint

#-------------------------------------------------------------------------------
#MapInfoFighters
#-------------------------------------------------------------------------------
def fInitMapInfoDynamicObject (tsC):
	msC = tsC
	ff.MapInfo [msC] = {}
	ff.MapInfo [msC]["health"] = 0#do not set yet because we need to calculate healthLoss
	ff.MapInfo [msC]["healthLoss"] = 0#do not set yet because we need health set
	ff.MapInfo [msC]["prestige"] = 0#Object_CalcPrestige (msC)
	ff.MapInfo [msC]["team"] = None#do not set yet because they may have changed allegiance
	ff.MapInfo [msC]["grid"] = []
	ff.MapInfo [msC]["pos"] = Get_ObjectPos (msC)
	ff.MapInfo [msC]["heading"] = 0
	ff.MapInfo [msC]["pitch"] = 0
	ff.MapInfo [msC]["distanceTravelled"] = 0
	ff.MapInfo [msC]["maxSpeed"] = 0
	ff.MapInfo [msC]["averageSpeed"] = 0
	ff.MapInfo [msC]["projectedPos"] = {}
	ff.MapInfo [msC]["cluster"] = None
	ff.MapInfo [msC]["nearestAlly"] = None
	ff.MapInfo [msC]["nearestAllyVisible"] = None
	ff.MapInfo [msC]["nearestAllyAttackable"] = None
	ff.MapInfo [msC]["nearestEnemy"] = None
	ff.MapInfo [msC]["nearestEnemyVisible"] = None
	ff.MapInfo [msC]["nearestEnemyAttackable"] = None
	ff.MapInfo [msC]["nearestAllyElsewhere"] = None
	ff.MapInfo [msC]["nearestEnemyElsewhere"] = None

def MapInfo_UpdateFighters (tevent = ""):
	miTimeStarted = time.time ()
	if "fightersLastUpdated" in ff.MapInfo.keys (): miTimeSinceLastUpdate = miTimeStarted - ff.MapInfo ["fightersLastUpdated"]
	GetMapInfo ()
# 	msTeamSplitterFunction = GET_TEAM_FUNCTION
	maCs = list (Mission_GetDynamicObjects ())
	mi = 0
	while mi < len (maCs):
		msC = maCs [mi]
		if string.find (msC, "mapinfocaliper") > -1: maCs.remove (msC)
		else: mi = mi + 1
	if "characters" not in ff.MapInfo.keys (): ff.MapInfo ["characters"] = []
	if "teams" not in ff.MapInfo.keys (): ff.MapInfo ["teams"] = []
	if "teamIds" not in ff.MapInfo.keys (): ff.MapInfo ["teamIds"] = []
	if "projectiles" not in ff.MapInfo.keys (): ff.MapInfo ["projectiles"] = []
	ff.MapInfoFighters = {}
	#put into the grid
# 	maRemoves = []
	check = js.OC_CHARACTER | js.OC_VILLIAN | js.OC_CONTROLLABLE | js.OC_MINION | js.OC_POLICE
	for msC in maCs:
		if (Object_GetClass(msC) & check) and m25team.GetTeam(msC): #or isHero (msC) or isBaddie (msC) or isCharacterTemplate (Object_GetTemplate (msC)) or Object_GetAttr (msC, "mock"):
			#get team
			muTeam = GET_TEAM_FUNCTION (msC)#keep original team data, just in case
			msTeam = muTeam
			if type (msTeam) != StringType: msTeam = `msTeam`
			msTeam = "team" + msTeam
			if Object_IsAlive (msC):
				#global character info
				if msC not in ff.MapInfo ["characters"]:
					fInitMapInfoDynamicObject (msC)
					ff.MapInfo [msC]["prestige"] = Object_CalcPrestige (msC)
					ff.MapInfo [msC]["pos"] = Get_ObjectPos (msC)
					ff.MapInfo ["characters"].append (msC)
				#team info
				msOldTeam = ff.MapInfo [msC]["team"]
				if msOldTeam != None and msOldTeam != msTeam:
					if msC in ff.MapInfo [msOldTeam]: ff.MapInfo [msOldTeam].remove (msC)
				ff.MapInfo [msC]["team"] = msTeam
				if msTeam not in ff.MapInfo ["teams"]:
					ff.MapInfo [msTeam] = []
					ff.MapInfo ["teams"].append (msTeam)
					ff.MapInfo ["teamIds"].append (muTeam)
				if msC not in ff.MapInfo [msTeam]: ff.MapInfo [msTeam].append (msC)
				if msTeam + "Dead" not in ff.MapInfo.keys (): ff.MapInfo [msTeam + "Dead"] = []
				if msC in ff.MapInfo [msTeam + "Dead"]: ff.MapInfo [msTeam + "Dead"].remove (msC)
			else:
				if msC in ff.MapInfo [msTeam]: ff.MapInfo [msTeam].remove (msC)
				if msC not in ff.MapInfo [msTeam + "Dead"]: ff.MapInfo [msTeam + "Dead"].append (msC)
		else:#all other dynamic objects may be projectiles
			if msC not in ff.MapInfo ["projectiles"]:
				fInitMapInfoDynamicObject (msC)
				ff.MapInfo [msC]["team"] = "!@#projectiles"
				ff.MapInfo [msC]["pos"] = Get_ObjectPos (msC)
				ff.MapInfo ["projectiles"].append (msC)
		#object info
		if msC in ff.MapInfo.keys ():
# 			if ff.MapInfo [msC]["health"]:
# 				ff.MapInfo [msC]["healthLoss"] = (ff.MapInfo [msC]["healthLoss"] + (ff.MapInfo [msC]["health"] - Object_GetAttr (msC, "health"))) * 0.5
# 			ff.MapInfo [msC]["health"] = Object_GetAttr (msC, "health")
			#grid info
			mtOldPos = ff.MapInfo [msC]["pos"]
			mtPos = Get_ObjectPos (msC)
# 			ff.MapInfo [msC]["heading"] = fSkXGetAngle (mtOldPos, mtPos)
# 			ff.MapInfo [msC]["pitch"] = fSkXGetAngle (mtOldPos, mtPos, "pitch")
# 			ff.MapInfo [msC]["projectedPos"] = MapInfo_GetPointFromProjection (mtOldPos, mtPos)
# 			miDTravelled = fSkXGetDistance (mtOldPos, mtPos)
# 			if miDTravelled:
# 				miSpeed = miDTravelled / miTimeSinceLastUpdate
# 				if miSpeed > ff.MapInfo [msC]["maxSpeed"]:
# 					ff.MapInfo [msC]["maxSpeed"] = miSpeed
# 				miAverageSpeed = (ff.MapInfo [msC]["averageSpeed"] + miSpeed) * 0.5
# 				ff.MapInfo [msC]["averageSpeed"] = miAverageSpeed
# 			ff.MapInfo [msC]["distanceTravelled"] = miDTravelled
			ff.MapInfo [msC]["pos"] = mtPos
			maGrid = Object_GetGridPos (msC)
			ff.MapInfo [msC]["grid"] = maGrid
			if maGrid [0] not in ff.MapInfoFighters.keys (): ff.MapInfoFighters [maGrid [0]] = {}
			if maGrid [1] not in ff.MapInfoFighters [maGrid [0]].keys (): ff.MapInfoFighters [maGrid [0]][maGrid [1]] = []
			ff.MapInfoFighters [maGrid [0]][maGrid [1]].append (msC)
			#clear cluster info
			ff.MapInfo [msC]["cluster"] = None
	#build fight clusters
	miCluster = 0
	ff.MapInfoFighters ["clusters"] = []
	for msC in maCs:
		if msC in ff.MapInfo.keys () and m25team.GetTeam(msC):
			ff.MapInfo [msC]["nearestAlly"] = [None, 1000000]
			ff.MapInfo [msC]["nearestAllyVisible"] = [None, 1000000]
			ff.MapInfo [msC]["nearestAllyAttackable"] = [None, 1000000]
			ff.MapInfo [msC]["nearestAllyElsewhere"] = [None, 1000000]
			ff.MapInfo [msC]["nearestEnemy"] = [None, 1000000]
			ff.MapInfo [msC]["nearestEnemyVisible"] = [None, 1000000]
			ff.MapInfo [msC]["nearestEnemyAttackable"] = [None, 1000000]
			ff.MapInfo [msC]["nearestEnemyElsewhere"] = [None, 1000000]
			msCTeam = ff.MapInfo [msC]["team"]
			mtPos = Get_ObjectPos (msC)
			maFs = MapInfo_GetFighters (mtPos, MAPINFO_FIGHT_CLUSTER_RADIUS)
			maFs = filter (Object_IsAlive, maFs)
			msCluster = None
			if ff.MapInfo [msC]["cluster"] == None:
				for msF in maFs:#get cluster from vicinity
					if ff.MapInfo [msF]["cluster"]:
						msCluster = ff.MapInfo [msF]["cluster"]
						break
				if not msCluster: msCluster = "cluster" + `miCluster`
			else:
				msCluster = ff.MapInfo [msC]["cluster"]
			if msCluster not in ff.MapInfoFighters ["clusters"]:
				ff.MapInfoFighters [msCluster] = {}
				ff.MapInfoFighters [msCluster]["teams"] = []
				ff.MapInfoFighters ["clusters"].append (msCluster)
				miCluster = miCluster + 1
			for msF in maCs:#maFs:#maCs:
				msTeam = ff.MapInfo [msF]["team"]
				reaction = m25team.Team_GetReaction(msC,msF)
				if msF in maFs:
					if not ff.MapInfo [msF]["cluster"]:
						msTeam2 = msTeam
			 			if type (msTeam2) != StringType: msTeam2 = `msTeam2`
						if msTeam2 not in ff.MapInfoFighters [msCluster]["teams"]:
							ff.MapInfoFighters [msCluster][msTeam2] = []
							ff.MapInfoFighters [msCluster][msTeam2 + "health"] = 0
							ff.MapInfoFighters [msCluster][msTeam2 + "maxHealth"] = 0
							ff.MapInfoFighters [msCluster][msTeam2 + "prestige"] = 0
							ff.MapInfoFighters [msCluster]["teams"].append (msTeam2)
						if msTeam2 != "!@#projectiles":
							ff.MapInfoFighters [msCluster][msTeam2 + "health"] = ff.MapInfoFighters [msCluster][msTeam2 + "health"] + Object_GetAttr (msF, "health")
							ff.MapInfoFighters [msCluster][msTeam2 + "maxHealth"] = ff.MapInfoFighters [msCluster][msTeam2 + "maxHealth"] + Object_GetAttr (msF, "maxHealth")
							ff.MapInfoFighters [msCluster][msTeam2 + "prestige"] = ff.MapInfoFighters [msCluster][msTeam2 + "prestige"] + ff.MapInfo [msF]["prestige"]
						ff.MapInfoFighters [msCluster][msTeam2].append (msF)
						ff.MapInfo [msF]["cluster"] = msCluster
					if msC != msF and msTeam != "!@#projectiles":
						miDistance = fSkXGetDistance (Get_ObjectPos (msC), Get_ObjectPos (msF))
						#print 'TEAMSSTUFF!!!! %s, %s'%(msCTeam, msTeam)
						if reaction=='friend':
							if miDistance < ff.MapInfo [msC]["nearestAlly"][1]: ff.MapInfo [msC]["nearestAlly"] = [msF, miDistance]
							if isVisible (msF):
								if miDistance < ff.MapInfo [msC]["nearestAllyVisible"][1]: ff.MapInfo [msC]["nearestAllyVisible"] = [msF, miDistance]
							if isAttackable (msF):
								if miDistance < ff.MapInfo [msC]["nearestAllyAttackable"][1]: ff.MapInfo [msC]["nearestAllyAttackable"] = [msF, miDistance]
						elif reaction=='enemy':
							if miDistance < ff.MapInfo [msC]["nearestEnemy"][1]: ff.MapInfo [msC]["nearestEnemy"] = [msF, miDistance]
							if isVisible (msF):
								if miDistance < ff.MapInfo [msC]["nearestEnemyVisible"][1]: ff.MapInfo [msC]["nearestEnemyVisible"] = [msF, miDistance]
							if isAttackable (msF):
								if miDistance < ff.MapInfo [msC]["nearestEnemyAttackable"][1]: ff.MapInfo [msC]["nearestEnemyAttackable"] = [msF, miDistance]
						else: pass
				if msF not in maFs:
					if msC != msF and Object_IsAlive (msF) and msTeam != "!@#projectiles":
						miDistance = fSkXGetDistance (Get_ObjectPos (msC), Get_ObjectPos (msF))
						if reaction=='friend':
							if miDistance < ff.MapInfo [msC]["nearestAllyElsewhere"][1]: ff.MapInfo [msC]["nearestAllyElsewhere"] = [msF, miDistance]
						else:
							if reaction=='enemy':
								if miDistance < ff.MapInfo [msC]["nearestEnemyElsewhere"][1]: ff.MapInfo [msC]["nearestEnemyElsewhere"] = [msF, miDistance]
			if UPDATE_CHARACTER_FUNCTION: UPDATE_CHARACTER_FUNCTION (msC)
	#clean-up
	#clusters
	mi = 0
	while mi < len (ff.MapInfoFighters ["clusters"]):
		msCluster = ff.MapInfoFighters ["clusters"][mi]
		if msCluster in ff.MapInfoFighters.keys ():
			if not len (ff.MapInfoFighters [msCluster]["teams"]):
				del ff.MapInfoFighters [msCluster]
		if msCluster not in ff.MapInfoFighters.keys (): ff.MapInfoFighters ["clusters"].remove (msCluster)
		else: mi = mi + 1
	#characters
	mi = 0
	while mi < len (ff.MapInfo ["characters"]):
		msC = ff.MapInfo ["characters"][mi]
		if not Object_Exists (msC):
			msTeam = ff.MapInfo [msC]["team"]
			if msC in ff.MapInfo [msTeam]: ff.MapInfo [msTeam].remove (msC)
			if msC in ff.MapInfo [msTeam + "Dead"]: ff.MapInfo [msTeam + "Dead"].remove (msC)
			ff.MapInfo ["characters"].remove (msC)
			del ff.MapInfo [msC]
		else: mi = mi + 1
	#projectiles
	mi = 0
	while mi < len (ff.MapInfo ["projectiles"]):
		msP = ff.MapInfo ["projectiles"][mi]
		if not Object_Exists (msP) or msP not in Mission_GetDynamicObjects ():
			if not Object_Exists (msP): del ff.MapInfo [msP]
			else: Object_UpdateGridPos (msP)
			if msP in ff.MapInfo ["projectiles"]: ff.MapInfo ["projectiles"].remove (msP)
		else: mi = mi + 1
	#teams
	mi = 0
	while mi < len (ff.MapInfo ["teams"]):
		msTeam = ff.MapInfo ["teams"][mi]
		if not len (ff.MapInfo [msTeam]) and not len (ff.MapInfo [msTeam + "Dead"]):
			del ff.MapInfo [msTeam]
			del ff.MapInfo [msTeam + "Dead"]
			muTeam = ff.MapInfo ["teamIds"][ff.MapInfo ["teams"].index (msTeam)]
			ff.MapInfo ["teamIds"].remove (muTeam)
			ff.MapInfo ["teams"].remove (msTeam)
		else: mi = mi + 1
	ff.MapInfo ["fightersLastUpdated"] = miTimeStarted
	return time.time () - miTimeStarted

def MapInfo_GetFighters (tt = (), tiDistance = None, tfFilter = Object_IsAlive):
	if type (tt) == StringType:
		tt = Get_ObjectPos (tt)
	if tiDistance == None: tiDistance = MAPINFO_FIGHT_CLUSTER_RADIUS
	return MapInfo_GetObjects (tt, tiDistance, ff.MapInfoFighters, tbAllowToLoop = 0, tfFilter = tfFilter)

def MapInfo_GetNearbyAllies (tsC):
	return MapInfo_GetTeamInClusterInfo (tsC)
ff.MapInfo_GetNearbyAllies = MapInfo_GetNearbyAllies

def MapInfo_GetNearbyAlliesHealthLevel (tsC):
	miH = MapInfo_GetTeamInClusterInfo (tsC, tsData = "health")
	miMH = MapInfo_GetTeamInClusterInfo (tsC, tsData = "maxHealth")
	if not miH or not miMH:
		miHL = 0
	else:
		miHL = miH / float (miMH) * 100
	return miHL

def MapInfo_GetNearbyAlliesPrestige (tsC):
	return MapInfo_GetTeamInClusterInfo (tsC, tsData = "prestige")

def MapInfo_GetNearbyEnemies (tsC):
	return MapInfo_GetTeamInClusterInfo (tsC, tsType = "enemies")
ff.MapInfo_GetNearbyEnemies = MapInfo_GetNearbyEnemies

def MapInfo_GetNearbyEnemiesHealthLevel (tsC):
	miH = MapInfo_GetTeamInClusterInfo (tsC, tsData = "health", tsType = "enemies")
	miMH = MapInfo_GetTeamInClusterInfo (tsC, tsData = "maxHealth", tsType = "enemies")
	if not miH or not miMH:
		miHL = 0
	else:
		miHL = miH / float (miMH) * 100
	return miHL

def MapInfo_GetNearbyEnemiesPrestige (tsC):
	return MapInfo_GetTeamInClusterInfo (tsC, tsData = "prestige", tsType = "enemies")

def MapInfo_GetNearbyProjectiles (tsC):
	return MapInfo_GetTeamInClusterInfo (tsC, tsTeam = "!@#projectiles")
ff.MapInfo_GetNearbyProjectiles = MapInfo_GetNearbyProjectiles

def MapInfo_GetTeamInClusterInfo (tsCluster, tsTeam = None, tsData = "members", tsType = "allies"):
	GetMapInfo ()
	if tsCluster in ff.MapInfo ["characters"]:
		if tsTeam == None: tsTeam = ff.MapInfo [tsCluster]["team"]
		tsCluster = ff.MapInfo [tsCluster]["cluster"]
	if tsCluster in ff.MapInfo ["projectiles"]:
		if tsTeam == None: tsTeam = ff.MapInfo [tsCluster]["team"]
		tsCluster = ff.MapInfo [tsCluster]["cluster"]
	muValueCollection = []
	if tsCluster in ff.MapInfoFighters.keys ():
		mdCluster = ff.MapInfoFighters [tsCluster]
		#init return data to type of expected value
		if tsData == "members":
			muValueCollection = []
		else:
			ms1stTeam = mdCluster ["teams"][0]
			if type (ms1stTeam) != StringType: ms1stTeamString = `ms1stTeam`
			else: ms1stTeamString = ms1stTeam
			mu1stValue = mdCluster [ms1stTeamString + tsData]
			if type (mu1stValue) == ListType:
				muValueCollection == []
			elif isNumber (mu1stValue):
				muValueCollection = 0
			else:#assume strings will be returned as array
				muValueCollection = []
		maTeamsInCluster = mdCluster ["teams"]
		for msTIC in maTeamsInCluster:
			if tsData == "members":
				msData = msTIC
			else:
				if type (msTIC) != StringType: msTICString = `msTIC`
				else: msTICString = msTIC
				msData = msTICString + tsData
			if tsTeam == None or (tsType == "allies" and msTIC == tsTeam) or (tsType != "allies" and msTIC != tsTeam and msTIC != "!@#projectiles"):
				muValue = mdCluster [msData]
				if type (muValue) == ListType:
					muValueCollection.extend (muValue)
				elif isNumber (mu1stValue):
					muValueCollection = muValueCollection + muValue
				elif type (muValue) == StringType:
					muValueCollection.append (muValue)
	return muValueCollection

def MapInfo_GetTeamMembers (tsTeam, tsLife = None):
	GetMapInfo ()
	if isNumber (tsTeam): tsTeam = "team" + `float (tsTeam)`
	elif Object_Exists (tsTeam): tsTeam = ff.MapInfo [tsTeam]["team"]
	maMembers = []
	if tsLife != "dead": maMembers.extend (ff.MapInfo [tsTeam])
	if tsLife != "alive": maMembers.extend (ff.MapInfo [tsTeam + "Dead"])
	return maMembers

def isAttackable (tsC):
	return Object_IsAlive (tsC) and not frozen (tsC) and not isInvisible (tsC) and not isObjectified (tsC)

def isInvisible (tsC):
	return Object_GetSecondaryStates (tsC) & SCSTATE_INVISIBLE

def isObjectified (tsC):
	maBanishedStates = [PCSTATE_OBJECT, PCSTATE_TOAD]
	return Object_GetPrimaryState (tsC) in maBanishedStates

def isVisible (tsC):
	return not isInvisible (tsC)

if heightcheck:
	#---------------------------------------------------------------------------
	#buildings
	#---------------------------------------------------------------------------
	def MapInfo_GetBuildingCorners (ts):
		#print "!","MapInfo_GetBuildingCorners",ts
		GetMapInfo ()
		mt = Get_ObjectPos (ts)
		mdCorners = {}
		md = heightcheck.BuildingTemplate_GetDimensions (ts)
		if "nw" in md.keys ():
			miO = fSkXResolveAngle (fSkXGetDegreesFromRadians (Object_GetOrientation (ts) [1]))
			miA = fSkXGetAngle ((0,0,0), md ["nw"])
			miD = fSkXGetDistance ((0,0,0), md ["nw"])
			mtNW = fSkXGetPointFromVector (mt, fSkXResolveAngle (miO + miA), miD)
			miA = fSkXGetAngle ((0,0,0), md ["se"])
			miD = fSkXGetDistance ((0,0,0), md ["se"])
			mtSE = fSkXGetPointFromVector (mt, fSkXResolveAngle (miO + miA), miD)
			mtNE = fSkXGetPointFromVector (mtNW, fSkXResolveAngle (miO + 90), md ["w"])
			mtSW = fSkXGetPointFromVector (mtSE, fSkXResolveAngle (miO - 90), md ["w"])
			#align with world
			maCorners = [mtNW, mtNE, mtSE, mtSW]
			mtNW2 = mtNW
			mtNE2 = mtNE
			mtSE2 = mtSE
			mtSW2 = mtSW
			for mi in range (0, len (maCorners)):
				miAngle = fSkXGetAngle (mt, maCorners [mi])
				maCorners [mi] = (int (maCorners [mi][0]), int (maCorners [mi][1]), int (maCorners [mi][2]))
				if miAngle >= 270:
					mtNW2 = maCorners [mi]
				elif miAngle >= 180:
					mtSW2 = maCorners [mi]
				elif miAngle >= 90:
					mtSE2 = maCorners [mi]
				else:
					mtNE2 = maCorners [mi]
			mdCorners ["nw"] = mtNW2
			mdCorners ["ne"] = mtNE2
			mdCorners ["se"] = mtSE2
			mdCorners ["sw"] = mtSW2
		return mdCorners

# def fSkXGetNearestSideToPoint (ttPos, ttBoundingBox):
# 	#bounding box = (minx, miny, minz, maxx, maxy, maxz)
	def MapInfo_GetBuildingPointsOfNearestSide (tsBuilding, ttPos):
		GetMapInfo ()
		if type (ttPos) == StringType:
			ttPos = Get_ObjectPos (ttPos)
		maPoints = []
		mdCorners = MapInfo_GetBuildingCorners (tsBuilding)
		maBox = list (mdCorners ["nw"])
		maBox.extend (list (mdCorners ["se"]))
		miSide = fSkXGetNearestSideToPoint (ttPos, maBox)
		if miSide == 1: maPoints = [mdCorners ["nw"], mdCorners ["ne"]]
		elif miSide == 2: maPoints = [mdCorners ["ne"], mdCorners ["se"]]
		elif miSide == 3: maPoints = [mdCorners ["sw"], mdCorners ["se"]]
		elif miSide == 4: maPoints = [mdCorners ["nw"], mdCorners ["sw"]]
		return maPoints
	MapInfo_GetBuildingPointsOfClosestSide = MapInfo_GetBuildingPointsOfNearestSide

	def MapInfo_GetBuildingPointsOfFurthestSide (tsBuilding, ttPos):
		GetMapInfo ()
		if type (ttPos) == StringType:
			ttPos = Get_ObjectPos (ttPos)
		maPoints = []
		mdCorners = MapInfo_GetBuildingCorners (tsBuilding)
		maBox = list (mdCorners ["nw"])
		maBox.extend (list (mdCorners ["se"]))
		miSide = fSkXGetNearestSideToPoint (ttPos, maBox)
		if miSide == 1: maPoints = [mdCorners ["sw"], mdCorners ["se"]]
		elif miSide == 2: maPoints = [mdCorners ["nw"], mdCorners ["sw"]]
		elif miSide == 3: maPoints = [mdCorners ["nw"], mdCorners ["ne"]]
		elif miSide == 4: maPoints = [mdCorners ["ne"], mdCorners ["se"]]
		return maPoints
	MapInfo_GetBuildingPointsOfFarthestSide = MapInfo_GetBuildingPointsOfFurthestSide

	def MapInfo_PutPointOutsideBuilding (ttPos, tsBuilding = None, taExcludeBuildings = None, tsReturn = "point"):
		GetMapInfo ()
		muReturn = None
		if type (ttPos) == StringType: ttPos = Get_ObjectPos (ttPos)
		if not tsBuilding:
			maBuildings = MapInfo_GetBuildings (ttPos)
			if taExcludeBuildings:
				if type (taExcludeBuildings) == StringType: taExcludeBuildings = [taExcludeBuildings]
				fTrimArrayOf (maBuildings, taExcludeBuildings)
			tsBuilding = MapInfo_GetNearest (ttPos, taObjects = maBuildings)
		if tsBuilding:
			mtBuilding = Get_ObjectPos (tsBuilding)
			mdBuildingInfo = heightcheck.BuildingTemplate_GetDimensions (tsBuilding)
			if "h" in mdBuildingInfo.keys ():
				miHeight = mdBuildingInfo ["h"]
				maCorners = MapInfo_GetBuildingPointsOfNearestSide (tsBuilding, ttPos)
				miA = fSkXGetAngle (mtBuilding, ttPos)
				mtPointAtSide = fSkXGetIntersectionFromPoints (mtBuilding, ttPos, maCorners [0], maCorners [1])
				miD = fSkXGetDistance (mtBuilding, mtPointAtSide)
				miD2 = fSkXGetDistance (mtBuilding, ttPos)
				if miD > miD2 and ttPos [2] >= mtBuilding [2] and ttPos [2] <= mtBuilding [2] + miHeight:
					if tsReturn == "point":
						muReturn = fSkXGetPointFromVector (mtBuilding, miA, miD + MEASUREMENT_SNAP)
					else:
						muReturn = tsBuilding
				else:
					if tsReturn == "point":
						muReturn = ttPos
					else:
						muReturn = None
		return muReturn

	def MapInfo_EnsurePointIsOutsideBuilding (ttPos, tbPushToOtherSide = 0, tbWarpToOppositeEdgeOfMap = 0):
		mtPos = MapInfo_PutPointWithinMap (ttPos)
		msB = MapInfo_GetObjects (mtPos, tiDistance = "nearest", tfFilter = isBuilding)
		if msB:
			msCluster = ff.MapInfo ["buildingClusters"][msB]
			mdCluster = ff.MapInfo ["buildingClusters"][msCluster]
			if not tbPushToOtherSide: maCorners = MapInfo_GetClusterPointsOfNearestSide (mtPos, msCluster)
			else:  maCorners = MapInfo_GetClusterPointsOfFurthestSide (mtPos, msCluster)
			mtMid = fSkXGetPointInBetween (maCorners [0], maCorners [1])
			miD = fSkXGetDistance (mdCluster ["pos"], mtPos)
			miDToMid = fSkXGetDistance (mdCluster ["pos"], mtMid)
			if miD < miDToMid:
				mtPointOnSide = fSkXGetIntersectionFromPoints (mdCluster ["pos"], ttPos, maCorners [0], maCorners [1])
				miA = fSkXGetAngle (mdCluster ["pos"], ttPos)
				mtPos = fSkXGetPointFromVector (mtPointOnSide, miA, MEASUREMENT_SNAP)
				if tbWarpToOppositeEdgeOfMap: mtPos = MapInfo_WarpPointIfOutside (mtPos)
				else: mtPos = MapInfo_PutPointWithinMap (mtPos)
		return mtPos

	def MapInfo_IsPointInsideBuilding (ttPos, tsBuilding = None, taExcludeBuildings = None):
		GetMapInfo ()
		return ttPos != MapInfo_PutPointOutsideBuilding (ttPos, tsBuilding, taExcludeBuildings = taExcludeBuildings)

	def MapInfo_GetBuildingPointInsideOf (ttPos, taExcludeBuildings = None):
		GetMapInfo ()
		return MapInfo_PutPointOutsideBuilding (ttPos, taExcludeBuildings = taExcludeBuildings, tsReturn = "building")

	def MapInfo_GetAdjacentBuildings (ts):
		#print "!","MapInfo_GetAdjacentBuildings",ts
		msB = ts
		mtB = Get_ObjectPos (msB)
		mdCorners = MapInfo_GetBuildingCorners (msB)
		maAdjacents = []
		if "n" in mdCorners.keys ():
			mdCorners ["n"] = (mtB [0], mdCorners ["nw"][1], mtB [2])
			mdCorners ["e"] = (mdCorners ["ne"][0], mtB [1], mtB [2])
			mdCorners ["s"] = (mtB [0], mdCorners ["sw"][1], mtB [2])
			mdCorners ["w"] = (mdCorners ["nw"][0], mtB [1], mtB [2])
			for msCorner in mdCorners.keys ():
				mtCorner = mdCorners [msCorner]
				miA = fSkXGetAngle (mtB, mtCorner)
				miD = fSkXGetDistance (mtB, mtCorner)
				mtCorner = fSkXGetPointFromVector (mtB, miA, miD + MEASUREMENT_SNAP)
				mtCorner =(mtCorner [0], mtCorner [1], mtB [2] + MEASUREMENT_SNAP)
				msAdj = MapInfo_GetBuildingPointInsideOf (mtCorner, taExcludeBuildings = msB)
				if msAdj and msAdj not in maAdjacents: maAdjacents.append (msAdj)
		return maAdjacents

	#---------------------------------------------------------------------------
	#buildings clusters
	#---------------------------------------------------------------------------
	def MapInfo_GetClusterPointsOfNearestSide (ttPos, tsCluster = None):
		GetMapInfo ()
		if not tsCluster: tsCluster = MapInfo_GetBuildingCluster (ttPos)
		maCorners = []
		if tsCluster:
			mdCluster = ff.MapInfo ["buildingClusters"][tsCluster]
			maBoundingBox = [mdCluster ["nw"][0], mdCluster ["nw"][1], mdCluster ["floor"], mdCluster ["se"][0], mdCluster ["se"][1], mdCluster ["ceiling"]]
			miSide = fSkXGetNearestSideToPoint (ttPos, maBoundingBox)
			if miSide == 1: maCorners = [mdCluster ["nw"], mdCluster ["ne"]]
			elif miSide == 2: maCorners = [mdCluster ["ne"], mdCluster ["se"]]
			elif miSide == 3: maCorners = [mdCluster ["sw"], mdCluster ["se"]]
			elif miSide == 4: maCorners = [mdCluster ["nw"], mdCluster ["sw"]]
		return maCorners

	def MapInfo_GetClusterPointsOfFurthestSide (ttPos, tsCluster = None):
		GetMapInfo ()
		if not tsCluster: tsCluster = MapInfo_GetBuildingCluster (ttPos)
		maCorners = []
		if tsCluster:
			mdCluster = ff.MapInfo ["buildingClusters"][tsCluster]
			maBoundingBox = [mdCluster ["nw"][0], mdCluster ["nw"][1], mdCluster ["floor"], mdCluster ["se"][0], mdCluster ["se"][1], mdCluster ["ceiling"]]
			miSide = fSkXGetNearestSideToPoint (ttPos, maBoundingBox)
			if miSide == 1: maCorners = [mdCluster ["sw"], mdCluster ["se"]]
			elif miSide == 2: maCorners = [mdCluster ["nw"], mdCluster ["sw"]]
			elif miSide == 3: maCorners = [mdCluster ["nw"], mdCluster ["ne"]]
			elif miSide == 4: maCorners = [mdCluster ["ne"], mdCluster ["se"]]
		return maCorners

	def MapInfo_GetBuildingCluster (ttPos):
		GetMapInfo ()
		msCluster = None
		msBuilding = MapInfo_GetBuildingPointInsideOf (ttPos)
		if msBuilding: msCluster = ff.MapInfo ["buildingClusters"][msBuilding]
		return msCluster

	def MapInfo_GetClusterCorners (tsCluster = None):
		GetMapInfo ()
		if isTuplePosition (tsCluster): tsCluster = MapInfo_GetBuildingCluster (tsCluster)
		elif Object_Exists (tsCluster): tsCluster = ff.MapInfo ["buildingClusters"][tsCluster]
		mdCluster = None
		if tsCluster: mdCluster = ff.MapInfo ["buildingClusters"][tsCluster]
		return mdCluster

	def MapInfo_GetBuildingClusters ():
		#print "!","MapInfo_GetBuildingClusters"
		maBuildings = filter (isBuilding, Mission_GetObjects ())
		mdBuildingClusters = {}
		#mdBuildingClusters ["clusters"] = []
		miCluster = 0
		for msB in maBuildings:
			mtB = Get_ObjectPos (msB)
			msCluster = None
			#get adjacent buildings
			maAdjacents = MapInfo_GetAdjacentBuildings (msB)
			#get clusters of those buildings
			maAdjClusters = []
			for msAdj in maAdjacents:
				if msAdj in mdBuildingClusters.keys ():
					msCluster2 = mdBuildingClusters [msAdj]
					if msCluster2 not in maAdjClusters: maAdjClusters.append (msCluster2)
			if len (maAdjClusters): msCluster = maAdjClusters [0]
			#if no cluster yet, then new
			if not msCluster:#new cluster
				msCluster = "cluster" + `miCluster`
				miCluster = miCluster + 1
			if "clusters" not in mdBuildingClusters.keys (): mdBuildingClusters ["clusters"] = []
			if msCluster not in mdBuildingClusters ["clusters"]: mdBuildingClusters ["clusters"].append (msCluster)
			if msCluster not in mdBuildingClusters.keys ():
				mdBuildingClusters [msCluster] = {}
				mdBuildingClusters [msCluster]["buildings"] = []
			if msB not in mdBuildingClusters [msCluster]["buildings"]: mdBuildingClusters [msCluster]["buildings"].append (msB)
			mdBuildingClusters [msB] = msCluster
			#put all adjacents into this cluster
			for msB2 in maAdjacents:
				if msB2 not in mdBuildingClusters [msCluster]["buildings"]: mdBuildingClusters [msCluster]["buildings"].append (msB2)
				mdBuildingClusters [msB2] = msCluster
			#merge adjacent clusters together
			if len (maAdjClusters): del maAdjClusters [0]
			for msCluster2 in maAdjClusters:
				mdBuildingClusters [msCluster]["buildings"].extend (mdBuildingClusters [msCluster2]["buildings"])
				for msB2 in mdBuildingClusters [msCluster2]["buildings"]:
					mdBuildingClusters [msB2] = msCluster
				del mdBuildingClusters [msCluster2]
				mdBuildingClusters ["clusters"].remove (msCluster2)
		#get bounding box of cluster
		if "clusters" in mdBuildingClusters.keys ():
			for msCluster in mdBuildingClusters ["clusters"]:
				miXMin = 1000000; miXMax = -1000000
				miYMin = 1000000; miYMax = -1000000
				miZMin = 1000000; miZMax = -1000000
				for msB in mdBuildingClusters [msCluster]["buildings"]:
					mdCorners = MapInfo_GetBuildingCorners (msB)
					if "nw" in mdCorners.keys ():
						miXMin = min (miXMin, mdCorners ["nw"][0])
						miXMax = max (miXMax, mdCorners ["ne"][0])
						miYMin = min (miYMin, mdCorners ["sw"][1])
						miYMax = max (miYMax, mdCorners ["nw"][1])
						miZMin = min (miZMin, mdCorners ["nw"][2])
						miZMax = max (miZMin, mdCorners ["nw"][2] + heightcheck.BuildingTemplate_GetDimensions (msB)["h"])
					else:
						miXMin = 0; miXMax = 0
						miYMin = 0; miYMax = 0
						miZMin = 0; miZMax = 0
				mdBuildingClusters [msCluster]["nw"] = (miXMin, miYMax, miZMin)
				mdBuildingClusters [msCluster]["ne"] = (miXMax, miYMax, miZMin)
				mdBuildingClusters [msCluster]["se"] = (miXMax, miYMin, miZMin)
				mdBuildingClusters [msCluster]["sw"] = (miXMin, miYMin, miZMin)
				mdBuildingClusters [msCluster]["pos"] = fSkXGetPointInBetween (mdBuildingClusters [msCluster]["nw"], mdBuildingClusters [msCluster]["se"])
				mdBuildingClusters [msCluster]["floor"] = miZMin
				mdBuildingClusters [msCluster]["ceiling"] = miZMax
		ff.MapInfo ["buildingClusters"] = mdBuildingClusters

#-------------------------------------------------------------------------------
#utilities
#-------------------------------------------------------------------------------
def isNumber (ti):
	return type (ti) == FloatType or type (ti) == IntType or type (ti) == LongType

def isTuplePosition (tt):
	return type (tt) == TupleType and len (tt) == 3 and isNumber (tt [0]) and isNumber (tt [1]) and isNumber (tt [2])

def isBuilding (ts):
	return Object_GetClass (ts) & OC_BUILDING

def isGenericObject (ts):
	msT = Object_GetTemplate (ts)
	if Template_Exists (msT):
		return Template_GetClassString (msT) == "GAME_OBJ_GENERIC"

def MapInfo_GetNearest (ttPos, tiDistance = 1000000, taObjects = None, tsFilter = None):
	if type (ttPos) == StringType:
		if Object_Exists (ttPos): ttPos = Get_ObjectPos (ttPos)
		elif Marker_Exists (ttPos): ttPos = Get_MarkerPos (ttPos)
	if taObjects == None:
		taObjects = MapInfo_GetObjects (ttPos, tiDistance, tbAllowToLoop = 0)
	if type (tsFilter) == StringType:
		try:
			msExec = "maObjects2 = filter (" + tsFilter + ", taObjects)"
			exec (msExec)
			taObjects = maObjects2
		except:
			pass
	elif type (tsFilter) == FunctionType:
		taObjects = filter (tsFilter, taObjects)
	miClosest = tiDistance
	msClosest = None
	for ms in taObjects:
		miDistance = 1000000
		if type (ms) == StringType:
			if Object_Exists (ms):
				miDistance = fSkXGetDistance (ttPos, Get_ObjectPos (ms))
		elif isTuplePosition (ms):
			miDistance = fSkXGetDistance (ttPos, ms)
		if miDistance < miClosest:
			miClosest = miDistance
			msClosest = ms
	return msClosest
MapInfo_GetClosest = MapInfo_GetNearest
ff.MapInfo_GetNearest = MapInfo_GetNearest
ff.MapInfo_GetClosest = MapInfo_GetNearest

def MapInfo_GetFurthest (ttPos, tiDistance = 0, taObjects = None, tsFilter = None):
	if type (ttPos) == StringType:
		if Object_Exists (ttPos): ttPos = Get_ObjectPos (ttPos)
		elif Marker_Exists (ttPos): ttPos = Get_MarkerPos (ttPos)
	if taObjects == None:
		taObjects = MapInfo_GetObjects (ttPos, tiDistance, tbAllowToLoop = 0)
	if type (tsFilter) == StringType:
		try:
			msExec = "maObjects2 = filter (" + tsFilter + ", taObjects)"
			exec (msExec)
			taObjects = maObjects2
		except:
			pass
	miFurthest = tiDistance
	msFurthest = None
	for ms in taObjects:
		miDistance = 0
		if type (ms) == StringType:
			if Object_Exists (ms):
				miDistance = fSkXGetDistance (ttPos, Get_ObjectPos (ms))
		elif isTuplePosition (ms):
			miDistance = fSkXGetDistance (ttPos, ms)
		if miDistance > miFurthest:
			miFurthest = miDistance
			msFurthest = ms
	return msFurthest
MapInfo_GetFarthest = MapInfo_GetFurthest
ff.MapInfo_GetFurthest = MapInfo_GetFurthest
ff.MapInfo_GetFarthest = MapInfo_GetFurthest

def fTrimArrayOf (ta, ta2, tbExact = 0):
	mi = 0
	while mi < len (ta):
		mu = ta [mi]
		mbRemove = 0
		if type (ta2) == ListType:
			if tbExact:
				if mu in ta2: mbRemove = 1
			else:
				for mu2 in ta2: fTrimArrayOf (ta, mu2)
		elif type (ta2) == StringType:
			if string.find (mu, ta2) > -1: mbRemove = 1
		elif isNumber (ta2):
			if mu == ta2: mbRemove = 1
		if mbRemove: ta.remove (mu)
		else: mi = mi + 1
	return ta

# def fSkXDebugMapInfo (ts):
# # 	return
# 	global gsSkXMapInfoDebug
# 	try:
# 		mbSkXTest = gsSkXMapInfoDebug
# 	except:
# 		gsSkXMapInfoDebug = ""
# 	if ts != gsSkXMapInfoDebug:
# 		if type (ts) != StringType:
# 			ts = `ts`
# 		ts = "----------------------------------------" + ts
# 		print ts
# 		gsSkXMapInfoDebug = ts

def fSkXDebugMapInfo (ts):
#	return
	if type (ts) == DictionaryType:
		fSkXDebugDictMapInfo (ts)
	else:
		global gsSkXDebug
		try: mbSkXTest = gsSkXDebug
		except: gsSkXDebug = ""; giSkXDebugTick = 0
		if ts != gsSkXDebug:
			if type (ts) != StringType:
				ts = `ts`
			ts = "----------------------------------------" + ts
			print ts
			gsSkXDebug = ts

def fSkXDebugDictMapInfo (td, tiLevel = 0, tbOnScreen = 0):
	maKeys = td.keys ()
	maKeys.sort ()
	maLines = []
	for msKey in maKeys:
		if type (td [msKey]) == DictionaryType:
			msLine = " "*(tiLevel*4) + `msKey` + ":" + "{"
			maLines.append (msLine)
			print msLine
			if tbOnScreen: Mission_StatusText (msLine)
			maLines.append (fSkXDebugDictMapInfo (td [msKey], tiLevel + 1), tbOnScreen)
		else:
			msLine = " "*(tiLevel*4) + `msKey` + ": " + `td [msKey]` + ","
			maLines.append (msLine)
			print msLine
			if tbOnScreen: Mission_StatusText (msLine)
	msLine = " "*(tiLevel*4) + "},"
	maLines.append (msLine)
	print msLine
	return maLines

#-------------------------------------------------------------------------------
#test
#-------------------------------------------------------------------------------
def fTestMapInfo_GetObjects (tt = (0, 0, 0), tiDistance = 300):
	mi = -1
	for ms in ff.MapInfo_GetObjects (tt, tiDistance):
		mi = mi + 1
		Object_SpawnAt ("alchemiss", Get_ObjectPos (ms), "pos" + `mi`)

maAlchemiss = []
def fSampleShowObjectsAt (tiX, tiY):
	for ms in maAlchemiss:
		Object_Destroy (ms)
		fSkXDebugMapInfo (ms)
	if MapInfo.has_key (tiX):
		ma = MapInfo [tiX]
		if ma.has_key (tiY):
			ma = ma [tiY]
			for ms in ma:
				msName = ms + "marker" + `len (maAlchemiss)`
				Object_SpawnAt ("alchemiss", Get_ObjectPos (ms), msName)
				maAlchemiss.append (msName)
