#####################################################################################
#
# Module: m25skirmishepicmode
# Author: m25, modified by Dr Mike
# 
# Description: Allow for team battles and buddy matches  within the FFvTTR Rumble Room
#
#
# Version 1
# Date: June 4, 2005
#
#####################################################################################

import sk
from sk import *
import m25ai
import cshelper
from m25ai import *
import m25skirmish
from m25skirmish import *
import ffx
from ffx import *


#-------------------------------------------------------------------------------------
#
# Callbacks
#
#-------------------------------------------------------------------------------------
	
def OnPostInit():
	#let the player see the entire level
	cshelper.see()
	
	# initialize the AI
	m25ai.InitScripts()
	
	#used to start up FFX
	ffx.FFX_InitSkirmish(1)
	
	# check for any characters that already exist on the map (put there by the map-maker)
	m25skirmish.SetupPreexisting()
	
	# put the heroes on the appropriate teams (or set them as neutral if watch mode)
	m25skirmish.SetupHeroTeam(watch=1)

	#initialise hero AI so they react when they get de-portraited
	heroes = cshelper.getAllHeroes()
	for h in heroes:
		m25ai.SetBrawler(h)
		AI_SetAlerted(h, 1)
		#AIEnable(h)
		m25ai.AIDisable(h)

	#allow the player to select heroes for the AI to take over		
	#AddHeroAIToggles()

	# give the teams their orders
	m25skirmish.SetupGroups()

	#set up the autotransfer whenever a character gets KO'd:
	for o in Mission_GetDynamicObjects():
		if GetTeam(o)==m25ai.TEAM_GOODGUYS:
			FFX_ObjectSetAttr(o,'epdeathcmd',1)
			regDeath(o,'OnAutoTransfer')
			print 'autotransfer for %s'%(o)

	#set up polling loop for the end of the battle
	regTimer('SkirmishModeCheckEnd',1)
	
	#set up polling loop for new heroes who spawn
	regTimer('AddTransfer',1)

def AddTransfer(event):
	regTimer('AddTransfer',1)
	for o in Mission_GetDynamicObjects():
		if GetTeam(o)==m25ai.TEAM_GOODGUYS:
			if FFX_HasPortrait(o)==0 and FFX_ObjectGetAttr(o,'eptranscmd')==0:
				Mission_CustomAction('CUSTOM_TRANSFER','',o,'OnTransfer',100,0)
				FFX_ObjectSetAttr(o,'eptranscmd',1)
			if FFX_ObjectGetAttr(o,'epdeathcmd')==0:
				regDeath(o,'OnAutoTransfer')
				FFX_ObjectSetAttr(o,'epdeathcmd',1)
			#add check for M25 group goal memberships
			if m25ai.CharGetGroupName(o)==None:
				print 'adding %s to the team'%(o)
				m25ai.AddToGroup('team1',o)
				m25ai.GroupCommand_GuardLeader('team1','')

			
def AddHeroAIToggles(default='off'):
	heroes = cshelper.getAllHeroes()
	for h in heroes:
		print 'ADDing ai toggle for ' + h
		if default=='off':
			Mission_CustomAction('CUSTOM_TURNONAI','',h,'ToggleHeroAI',100,1)
		else:
			Mission_CustomAction('CUSTOM_TURNOFFAI','',h,'ToggleHeroAI',100,1)

def ToggleHeroAI(target, char):
	if FFX_HasPortrait(target):
		if m25ai.IsAIDisabled(target):
			m25ai.AIEnable(target)
			Mission_CustomAction('CUSTOM_TURNOFFAI','',target,'ToggleHeroAI',100,1)
		else:
			m25ai.AIDisable(target)
			Mission_CustomAction('CUSTOM_TURNONAI','',target,'ToggleHeroAI',100,1)
			
def OnReceiveSelectedEnemies(enemies):
	sk.PreSetup()
	for enemy in enemies:
		name=m25skirmish.SkirmishSpawnEnemy(enemy)
		if Object_Exists(name):
			print '%s:%d'%(FFX_GetTemplate(name),GetTeam(name))
			if GetTeam(name)==m25ai.TEAM_GOODGUYS:
				Mission_CustomAction('CUSTOM_TRANSFER','',name,'OnTransfer',100,0)
				FFX_ObjectSetAttr(name,'epictranscmd',1)

def OnSelectionRequiredCheck():
	return 1 #allows players to choose their opponents

def OnNumEnemiesCheck():
	return m25skirmish.WATCH_MAXCHARS

def OnGameTypeCheck():
    return sk.GetStandardMarkerNames()

def OnTransfer(target,char):
    if GetTeam(target)!=m25ai.TEAM_GOODGUYS:
        return
    #control is lost from the calling character, or a dead team member if we can find one
    loser=char
    for o in cshelper.Mission_GetDynamicObjects():
        if FFX_HasPortrait(o) and Object_IsAlive(o)==0:
            loser=o
    FFX_DestroyPortrait(loser)
    Mission_CustomAction('CUSTOM_TRANSFER','',loser,'OnTransfer',100,0)
    FFX_ObjectSetAttr(loser,'eptranscmd',1)
    AIRemoveGoal(loser,'Goal_HuntEnemy(')
    AIRemoveGoal(loser,'Goal_StayWithTeam(')
    m25ai.AIEnable(loser)

    #create a portrait for the target and select them
    FFX_AddPortrait(target)
    Mission_RemoveCustomAction('CUSTOM_TRANSFER','',target)
    Mission_SelectHero(target)
    m25ai.AIDisable(target)
    m25ai.Group_PromoteLeader('team1')


def OnAutoTransfer(event):
    loser=event.object
    #set it up again (heroic remedies)
    regDeath(loser,'OnAutoTransfer')
    if FFX_HasPortrait(loser)==0:
        return
    dormant=''
    alive=''
    for obj in Mission_GetDynamicObjects():
        if obj!=loser and FFX_HasPortrait(obj)==0 and GetTeam(obj)==m25ai.TEAM_GOODGUYS:
            if Object_IsAlive(obj):
                alive=obj
            else:
                if Object_GetAttr(obj,'heroPoints')>0:
                    dormant=obj
    target=''
    if dormant!='':
        target=dormant
    if alive!='':
        target=alive
    print 'loser=%s dormant=%s alive=%s target=%s'%(loser,dormant,alive,target)
    if target:
        OnTransfer(target,loser)
